Package: HVT
Type: Package
Date: 2026-04-29
Title: Constructing Hierarchical Voronoi Tessellations and Overlay
        Heatmaps for Data Analysis
Version: 26.1.3
Authors@R: c(
  person("Zubin", "Dowlaty", email = "zubin.dowlaty@mu-sigma.com", role = c("aut", "cre")),
  person("Mu Sigma, Inc.", role = "cph"))
Description: Facilitates building topology preserving maps for data analysis.
License: Apache License 2.0
Encoding: UTF-8
Imports: MASS, grDevices, splancs, stats, dplyr, NbClust, purrr,
        magrittr, ggplot2, tidyr, scales, cluster, reshape2, FNN,
        Rtsne, umap, plyr, markovchain, methods, deldir, gridExtra
Depends: R (>= 4.0.0)
BugReports: https://github.com/Mu-Sigma/HVT/issues
URL: https://github.com/Mu-Sigma/HVT
RoxygenNote: 7.3.3
Suggests: knitr,rmarkdown,testthat,geozoo, plotly, rlang,
        DT,patchwork,sp,Hmisc,data.table,gtable,
        htmlwidgets,skimr,tibble,devtools,gifski,gganimate,
        tidyverse,DataExplorer,htmltools,corrplot,kableExtra,polyclip,conf.design,
        furrr, future, dendextend
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-29 20:21:30 UTC; vishwavani
Maintainer: Zubin Dowlaty <zubin.dowlaty@mu-sigma.com>
Repository: CRAN
Date/Publication: 2026-04-29 21:40:02 UTC
Collate: 'Add_boundary_points.R' 'Corrected_Tessellations.R'
        'Transform_Coordinates.R' 'ScaleMat.R' 'DelaunayInfo.R'
        'Delete_Outpoints.R' 'HVTMSM_support.R' 'HVTMSMoptimization.R'
        'OptimizationResults.R' 'VQ_codebookSplit.R' 'constants.R'
        'clusterPlot.R' 'clustHVT.R' 'diagPlot.R' 'diagSuggestion.R'
        'displayTable.R' 'edaPlots.R' 'getCellId.R' 'getCentroids.R'
        'getCentroids_for_opti.R' 'getOptimalCentroids.R'
        'getTransitionProbability.R' 'global.R' 'hvq.R' 'madPlot.R'
        'mcmc_plots.R' 'msm.R' 'utilities.R' 'plotAnimatedFlowmap.R'
        'plotExAnteRawSeries.R' 'plotHVT.R' 'plotModelDiagnostics.R'
        'plotMsmKN.R' 'plotNovelCells.R' 'plotQuantErrorHistogram.R'
        'plotStateTransition.R' 'plotZscore.R'
        'reconcileTransitionProbability.R' 'removeNovelty.R'
        'scoreHVT.R' 'scoreLayeredHVT.R' 'summary.R' 'trainHVT.R'
Author: Zubin Dowlaty [aut, cre],
  Mu Sigma, Inc. [cph]
Built: R 4.5.2; ; 2026-04-30 01:56:46 UTC; unix

Package: HVT
Type: Package
Date: 2025-02-03
Title: Constructing Hierarchical Voronoi Tessellations and Overlay
        Heatmaps for Data Analysis
Version: 25.2.2
Authors@R: c(
	person("Zubin", "Dowlaty", email = "zubin.dowlaty@mu-sigma.com", role = "aut"),
    	person("Mu Sigma, Inc.", email = "ird.experiencelab@mu-sigma.com", role = "cre"))
Description: Facilitates building topology preserving maps for data analysis.
License: Apache License 2.0
Encoding: UTF-8
Imports: MASS, grDevices, splancs, stats, dplyr, NbClust, purrr,
        magrittr, ggplot2, tidyr, scales, cluster, reshape2,
        FNN,Rtsne,umap, plyr, gganimate, markovchain, methods,deldir,
        gridExtra
Depends: R (>= 4.0.0)
BugReports: https://github.com/Mu-Sigma/HVT/issues
URL: https://github.com/Mu-Sigma/HVT
RoxygenNote: 7.3.2
Suggests: knitr,rmarkdown,testthat,geozoo,plotly,
        DT,patchwork,sp,Hmisc,data.table,gtable,
        htmlwidgets,installr,skimr,tibble,devtools,gifski,
        tidyverse,DataExplorer,htmltools,corrplot,kableExtra,polyclip,conf.design
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2025-02-05 09:15:45 UTC; vishwavani
Author: Zubin Dowlaty [aut],
  Mu Sigma, Inc. [cre]
Maintainer: "Mu Sigma, Inc." <ird.experiencelab@mu-sigma.com>
Repository: CRAN
Date/Publication: 2025-02-05 12:50:02 UTC
Built: R 4.3.3; ; 2025-02-15 14:42:53 UTC; unix

Package: HVT
Type: Package
Date: 2026-04-29
Title: Constructing Hierarchical Voronoi Tessellations and Overlay
        Heatmaps for Data Analysis
Version: 26.1.3
Authors@R: c(
  person("Zubin", "Dowlaty", email = "zubin.dowlaty@mu-sigma.com", role = c("aut", "cre")),
  person("Mu Sigma, Inc.", role = "cph"))
Description: Facilitates building topology preserving maps for data analysis.
License: Apache License 2.0
Encoding: UTF-8
Imports: MASS, grDevices, splancs, stats, dplyr, NbClust, purrr,
        magrittr, ggplot2, tidyr, scales, cluster, reshape2, FNN,
        Rtsne, umap, plyr, markovchain, methods, deldir, gridExtra
Depends: R (>= 4.0.0)
BugReports: https://github.com/Mu-Sigma/HVT/issues
URL: https://github.com/Mu-Sigma/HVT
RoxygenNote: 7.3.3
Suggests: knitr,rmarkdown,testthat,geozoo, plotly, rlang,
        DT,patchwork,sp,Hmisc,data.table,gtable,
        htmlwidgets,skimr,tibble,devtools,gifski,gganimate,
        tidyverse,DataExplorer,htmltools,corrplot,kableExtra,polyclip,conf.design,
        furrr, future, dendextend
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-29 20:21:30 UTC; vishwavani
Maintainer: Zubin Dowlaty <zubin.dowlaty@mu-sigma.com>
Repository: CRAN
Date/Publication: 2026-04-29 21:40:02 UTC
Collate: 'Add_boundary_points.R' 'Corrected_Tessellations.R'
        'Transform_Coordinates.R' 'ScaleMat.R' 'DelaunayInfo.R'
        'Delete_Outpoints.R' 'HVTMSM_support.R' 'HVTMSMoptimization.R'
        'OptimizationResults.R' 'VQ_codebookSplit.R' 'constants.R'
        'clusterPlot.R' 'clustHVT.R' 'diagPlot.R' 'diagSuggestion.R'
        'displayTable.R' 'edaPlots.R' 'getCellId.R' 'getCentroids.R'
        'getCentroids_for_opti.R' 'getOptimalCentroids.R'
        'getTransitionProbability.R' 'global.R' 'hvq.R' 'madPlot.R'
        'mcmc_plots.R' 'msm.R' 'utilities.R' 'plotAnimatedFlowmap.R'
        'plotExAnteRawSeries.R' 'plotHVT.R' 'plotModelDiagnostics.R'
        'plotMsmKN.R' 'plotNovelCells.R' 'plotQuantErrorHistogram.R'
        'plotStateTransition.R' 'plotZscore.R'
        'reconcileTransitionProbability.R' 'removeNovelty.R'
        'scoreHVT.R' 'scoreLayeredHVT.R' 'summary.R' 'trainHVT.R'
Author: Zubin Dowlaty [aut, cre],
  Mu Sigma, Inc. [cph]
Built: R 4.6.0; ; 2026-05-01 22:04:56 UTC; unix

# Generated by roxygen2: do not edit by hand

export(HVTMSMoptimization)
export(OptimizationResults)
export(clustHVT)
export(displayTable)
export(edaPlots)
export(getTransitionProbability)
export(msm)
export(plotAnimatedFlowmap)
export(plotExAnteRawSeries)
export(plotHVT)
export(plotModelDiagnostics)
export(plotMsmKN)
export(plotNovelCells)
export(plotQuantErrorHistogram)
export(plotStateTransition)
export(plotZscore)
export(reconcileTransitionProbability)
export(removeNovelty)
export(scoreHVT)
export(scoreLayeredHVT)
export(summary)
export(trainHVT)
import(dplyr)
import(ggplot2)
import(markovchain)
import(methods)
import(tidyr)
importFrom(dplyr,across)
importFrom(dplyr,between)
importFrom(dplyr,case_when)
importFrom(dplyr,everything)
importFrom(dplyr,mutate)
importFrom(dplyr,select)
importFrom(ggplot2,element_text)
importFrom(ggplot2,expansion)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_colour_manual)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_x_datetime)
importFrom(ggplot2,theme_minimal)
importFrom(ggplot2,unit)
importFrom(magrittr,"%>%")
importFrom(plyr,rbind.fill)
importFrom(stats,as.dendrogram)
importFrom(stats,cutree)
importFrom(stats,dist)
importFrom(stats,kmeans)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(tidyr,pivot_longer)
importFrom(utils,combn)
importFrom(utils,data)
importFrom(utils,globalVariables)
importFrom(utils,head)
importFrom(utils,tail)

# HVT: Collection of functions used to build hierarchical topology preserving maps

#### Zubin Dowlaty

##### Created Date: 2018-11-15
##### Modified Date: 2026-04-29

<div id="TOC">

*   [<span class="toc-section-number">1.</span> Abstract](#abstract)
*   [<span class="toc-section-number">2.</span> Vignettes](#vignettes)
*   [<span class="toc-section-number">3.</span> Package Version History](#version-history)
    *   [<span class="toc-section-number">3.12</span> HVT (v26.1.3) | What’s New](#hvt-(v26.1.3)-whats-new)
    *   [<span class="toc-section-number">3.11</span> HVT (v26.1.2)](#hvt-(v26.1.2))
    *   [<span class="toc-section-number">3.10</span> HVT (v26.1.1)](#hvt-(v26.1.1))
    *   [<span class="toc-section-number">3.9</span> HVT (v25.2.8)](#hvt-(v25.2.8))
    *   [<span class="toc-section-number">3.8</span> HVT (v25.2.7)](#hvt-(v25.2.7))
    *   [<span class="toc-section-number">3.7</span> HVT (v25.2.6)](#hvt-(v25.2.6))
    *   [<span class="toc-section-number">3.6</span> HVT (v25.2.5)](#hvt-(v25.2.5))
    *   [<span class="toc-section-number">3.5</span> HVT (v25.2.4)](#hvt-(v25.2.4))
    *   [<span class="toc-section-number">3.4</span> HVT (v24.9.1)](#hvt-(v24.9.1))
    *   [<span class="toc-section-number">3.3</span> HVT (v24.5.2)](#hvt-(v24.5.2))
    *   [<span class="toc-section-number">3.2</span> HVT (v23.11.02)](#hvt-(v23.11.02))
    *   [<span class="toc-section-number">3.1</span> HVT (v22.12.06)](#hvt-(v22.12.06))
*   [<span class="toc-section-number">4.</span> Installation of HVT (v26.1.3)](#installation-of-hvt-(v26.1.3))


</div>

<div id="abstract" class="section level1" number="1">

# <span class="header-section-number">1.</span> Abstract

The HVT package offers a suite of R functions designed to construct <a href="https://link.springer.com/chapter/10.1007/1-84628-118-0_7" target="_blank">topology preserving maps</a> for in-depth analysis of multivariate data. It is particularly well-suited for datasets with numerous records. The package organizes the typical workflow into several key stages:

1.  **Data Compression**: Long datasets are compressed using Hierarchical Vector Quantization (HVQ) to achieve the desired level of data reduction.

2.  **Data Projection**:  Compressed cells are projected into one and two dimensions using dimensionality reduction algorithms, producing <a href="https://en.wikipedia.org/wiki/Embedding" target="_blank">embeddings</a> that preserve the original topology. This allows for intuitive visualization of complex data structures.

3.  **Tessellation**: Voronoi tessellation partitions the projected space into distinct cells, supporting hierarchical visualizations. Heatmaps and interactive plots facilitate exploration and insights into the underlying data patterns.

4.  **Scoring**: Test dataset is evaluated against previously generated maps, enabling their placement within the existing structure. Sequential application across multiple maps is supported if required.

5. **Temporal Analysis and Visualization**: Functions in this stage examine time-series data to identify patterns, estimate transition probabilities, and visualize data flow over time.

6. **Dynamic Forecasting**:  Monte Carlo simulations of Markov chain provides forecasting capabilities for both ex-post and ex-ante scenarios with meticulously handling problematic states when found.

The HVT package allows creation of visually stunning tessellations, showcasing the power of topology preserving maps. Below is an image depicting a captivating tessellation of a torus, see
<a href="https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/HVT_vignette.html" target="_blank">**vignette**</a> for more details.

<p align="center">
  <img src="https://raw.githubusercontent.com/Mu-Sigma/HVT/master/vignettes/pngs/torus2.png" width="642px" height="440px" />
</p>
<p align="center"><em>Figure 1: The Voronoi tessellation for layer 1 and number of cells 500 with the heat map overlaid for variable 'z'.</em></p>



</div>

<div id="vignettes" class="section level1" number="2">

# <span class="header-section-number">2.</span> Vignettes

 

Following are the links to the vignettes for the HVT package:

| Version | Vignette Title | Description |
|----|----------|----------------|
| v18.05.17 | <a href="https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/HVT_vignette.html" target="_blank" rel="noopener noreferrer">HVT Vignette</a> | Contains the workflow of the functions used for vector quantization and construction of Hierarchical Voronoi Tessellations for data analysis. |
| v18.05.17 | <a href="https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/HVT_model_diagnostics_vignette.html" target="_blank" rel="noopener noreferrer">HVT Model Diagnostics Vignette</a> | Contains demonstrations of functions used to perform model diagnostics and validation for the trained HVT model. |
| v23.05.16 | <a href="https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/Scoring_Cells_with_Layers_using_scoreLayeredHVT.html" target="_blank" rel="noopener noreferrer">HVT Scoring Cells with Layers using scoreLayeredHVT</a> | Contains explanations of the functions used for scoring cells with layers based on a sequence of maps using scoreLayeredHVT. |
| v23.10.26 | <a href="https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/HVT_Temporal_Analysis.html" target="_blank" rel="noopener noreferrer">Temporal Analysis and Visualization: Leveraging Time Series Capabilities in HVT</a> | Contains implementations of the functions used for analyzing time series data and creating its state transition flow maps. |
| v24.05.16 | <a href="https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/LLM_Embeddings_in_HVT.html" target="_blank" rel="noopener noreferrer">Visualizing LLM Embeddings using HVT</a> | Contains implementation and analysis of hierarchical clustering using functions to evaluate and visualize token embeddings generated by OpenAI in 2D Space. |
| v24.08.14 | <a href="https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/Implementation_of_tsne_umap_in_trainHVT.html" target="_blank" rel="noopener noreferrer">Implementation of t-SNE and UMAP in trainHVT function</a> | Contains enhancements to the `trainHVT` function with advanced dimensionality reduction techniques such as t-SNE and UMAP, and includes a table of evaluation metrics to improve interpretability. |
| v25.03.01 | <a href="https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/Dynamic_Forecasting_macroeconomic_data.html" target="_blank" rel="noopener noreferrer">Dynamic Forecasting of Macroeconomic Time Series Dataset using HVT</a> | Contains enhancements to the HVT package for dynamic forecasting using Monte Carlo Simulations of Markov Chain (MSM) on macroeconomic time series dataset. |
| v25.08.25 | <a href="https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/Experimentation_of_hyperparameters_in_msm.html" target="_blank" rel="noopener noreferrer">Hyperparameter Experimentation for Champion Model Selection in MSM Dynamic Forecasting</a> | Contains enhancements to enable strategic selection of the champion model based on the lowest Mean Absolute Error by hyperparameter tuning in msm - dynamic forecasting. |

<div id="version-history" class="section level1" number="3">

# <span class="header-section-number">3.</span> Package Version History 

<div id="hvt-(v26.1.3)-whats-new" class="section level2" number="3.12">

## 3.12 HVT (v26.1.3) - What's New

29th April, 2026

In this version of the HVT package, several bug fixes and improvements have been made to enhance stability and performance.

**Bug Fixes:**

1. Fixed a crash in `msm` where comparisons against `n_nearest_neighbor` were performed without a null check, causing failures when the parameter was not supplied
2. Fixed missing initialization of intermediate variables before conditional blocks in `msm`, which caused undefined variable errors on early exit paths
3. Fixed `scoreHVT` not returning `centroidData` in its output, causing downstream null-access failures in `msm` and `clustHVT`
4. Fixed `clustHVT` crashing when `centroidData` was absent from scoring results; it now falls back to reconstructing centroid coordinates from `cellID_coordinates`

**Improvements:**

1. Extracted repeated polygon centroid, cell coordinate extraction, boundary coordinate construction, and cell label alignment logic into a shared `utilities.R` file, reducing duplication across `plotHVT`, `scoreHVT`, and `scoreLayeredHVT`.
2. Moved the hardcoded colour palette from `plotHVT` and `clusterPlot` into `constants.R` as a named constant.

</div>

<div id="hvt-(v26.1.2)" class="section level2" number="3.11">

## 3.11 HVT (v26.1.2)

5th February, 2026

*In this version of the HVT package, a minor issue related to the global initialization of a variable has been fixed.*
</div>

<div id="hvt-(v26.1.1)" class="section level2" number="3.1">

## 3.10 HVT (v26.1.1)

22nd January, 2026

In this version of the HVT package, the following enhancement and feature has been made:

**Enhancement**

1. **LLM Embeddings Vignette**: This vignette showcases the HVT workflow applied to numerical embedding data generated using an OpenAI model, featuring a more robust preprocessing pipeline and a revamped implementation aligned with recent AI developments.

**Feature**

1. **2D HVT Plots**: In this version, the aesthetics of compressed 2D HVT plots have been enhanced by adding new arguments to the `plotHVT` and `scoreHVT` functions. These updates allow users to toggle the display of cell centroids, adjust their sizes, and position cell IDs at the center, enabling clearer visualization of dense heatmaps with more legible centroids and cell labels.

</div>

<div id="hvt-(v25.2.8)" class="section level2" number="3.2">

## 3.9 HVT (v25.2.8)

16th December, 2025

In this version of the HVT package, the following new feature has been introduced:

1. **Ex-ante raw series forecasting**: A new feature enabling ex-ante raw series forecasting directly from transformed year-over-year (YoY) forecasts using a 12-month lookback approach. This allows the percentage changes from ex-ante transformed forecasts to be translated back into the raw series for each variable. This feature is introduced and explained in 'Dynamic Forecasting of Macroeconomic Time Series Dataset using HVT' vignette. 

</div>

<div id="hvt-(v25.2.7)" class="section level2" number="3.3">

## 3.8 HVT (v25.2.7) 

22nd October, 2025

In this version of the HVT package, the following new feature and vignette have been introduced:

**Feature**

1. **Experimentation of hyperparameters in `msm`**: This update introduces a new function called `HVTMSMoptimization` that runs grid search experiments across different hyperparameters (number of cells, clusters(k), nearest neighbors(nn)) by training and scoring HVT models, running MSM simulations for each combination and identify the champion model (lowest MAE across all results).

2. **Tabulation and Visualization**: Accessory functions to `HVTMSMoptimization` such as `OptimizationResults` and `plotMsmKN` has been added which helps to tabulate all the iterations and visualize the output via plotly object.

**Vignette** 

1. **Hyperparameter Experimentation for Champion Model Selection in MSM Dynamic Forecasting**: This vignette provides a comprehensive demonstration of using `HVTMSMoptimization`, covering the complete workflow from initial dataset handling, selection for train & test, executing hyperparameter tuning and identifying the champion model, implementing the champion model, and comparing MAE results.




</div>

<div id="hvt-(v25.2.6)" class="section level2" number="3.4">

## 3.7 HVT (v25.2.6)

14th October, 2025

*The issue with time-series animation plots from previous release has now been resolved with the latest gganimate update.*

</div>


<div id="hvt-(v25.2.5)" class="section level2" number="3.5">

## 3.6 HVT (v25.2.5)

04th July, 2025

*Dropping the time-series animation plots from the package since the latest version of gganimate doesn’t support them — a patched release will follow once the issue is resolved.*

</div>

<div id="hvt-(v25.2.4)" class="section level2" number="3.6">

## 3.5 HVT (v25.2.4) 

04th June, 2025

In this version of the HVT package, the following new features and vignette have been introduced:

**Features**

1. **Dynamic Forecasting of a Time Series Dataset**: This update introduces a new function called `msm` Monte Carlo Simulations of Markov Chain for dynamic forecasting of states in time series dataset. It supports both ex-post and ex-ante forecasting, offering valuable insights into future trends while resolving state transition challenges through clustering and nearest-neighbor methods to enhance simulation accuracy.

2. **Z score Plots**: This update introduces a new function called `plotZscore` that generates Z-score plots corresponding to the HVT cells for the given data, offering a visual representation of data distribution and highlighting potential outliers.

**Vignette** 

1. **Dynamic Forecasting of Macroeconomic Time Series Dataset using HVT**: This vignette illustrates the practical use of the new msm function on a macroeconomic dataset with 10 variables. It covers all steps, including data preparation, model training, scoring, and forecasting, while addressing challenges related to state transitions and evaluating performance using Mean Absolute Error (MAE).




</div>


<div id="hvt-(v24.9.1)" class="section level2" number="3.7">

## 3.4 HVT (v24.9.1) 

4th September, 2024

In this version of the HVT package, the following new features and vignettes have been introduced:

**Features**

1. **Implementation of t-SNE and UMAP in `trainHVT`**: This update incorporates dimensionality reduction methods like t-SNE and UMAP in the `trainHVT` function, complementing the existing Sammon's projection. It also enables the visualization of these techniques across all hierarchical levels within the HVT framework.

2. **Implementation of dimensionality reduction evaluation metrics**: This update introduces highly effective dimensionality reduction evaluation metrics as part of the output list of the `trainHVT` function. These metrics are organized into two levels: Level 1 (L1) and Level 2 (L2). The L1 metrics address key areas of dimensionality reduction which are mentioned below, by ensuring comprehensive evaluation and performance.

- Structure Preservation Metrics 
- Distance Preservation Metrics
- Human Centered Metrics
- Interpretive Quality Metrics
- Computational Efficiency Metrics


3. **Introduction of `clustHVT` function**: In this update, we introduced a new function called `clustHVT` specifically designed for Hierarchical clustering analysis. The function performs clustering of cells exclusively when the hierarchy level is set to 1, determining the optimal number of clusters by evaluating various indices. Based on user input, it conducts hierarchical clustering using AGNES with the default ward.D2 method. The output includes a dendrogram and an interactive 2D clustered HVT map that reveals cell context upon hovering. This function is not applicable when the hierarchy level is greater than 1.



**Vignettes** 

1. **Implementation of t-SNE and UMAP in `trainHVT` function**: This vignette showcases the integration of t-SNE and UMAP in the `trainHVT` function, offering a comprehensive guide on how to apply and visualize these dimensionality reduction techniques. It also covers the dimensionality reduction evaluation metrics and provides insights into their interpretation.

2. **Visualizing LLM Embeddings using HVT (Hierarchical Voronoi Tessellation)**: This vignette will outline the process of analyzing OpenAI-generated token embeddings using the HVT package, covering data compression, visualization, and hierarchical clustering, as well as comparing domain name assignments for clusters. It examines HVT's effectiveness in preserving contextual relationships between embeddings. Additionally, it provides a brief overview of the newly added `clustHVT` function and its parameters.



</div>


<div id="hvt-(v24.5.2)" class="section level2" number="3.8">

## 3.3 HVT (v24.5.2) 

2nd May, 2024

In this version of HVT package, the following new features have been introduced:

1. **Updated Nomenclature:** To make the function names more consistent and understandable/intuitive, we have renamed the functions throughout the package. Given below are the few instances.

* `HVT` to `trainHVT`
* `predictHVT` to `scoreHVT`
* `predictLayerHVT` to `scoreLayeredHVT`

2. **Restructured Functions:** The functions have been rearranged and grouped into new sections which are highlighted on the index page of package’s PDF documentation. Given below are the few instances.

* `trainHVT` function now resides within the `Training_or_Compression` section.
* `plotHVT` function now resides within the `Tessellation_and_Heatmap` section.
* `scoreHVT` function now resides within the `Scoring` section.

3. **Enhancements:** The pre-existed functions, `hvtHmap` and `exploded_hmap`, have been combined and incorporated into the `plotHVT` function. Additionally, `plotHVT` now includes the ability to perform 1D plotting.

4. **Temporal Analysis** 
- The new update focuses on the integration of time series capabilities into the HVT package by extending its foundational operations to time series data which is emphasized in this vignette.
- The new functionalities are introduced to analyze underlying patterns and trends within the data, providing insights into its evolution over time and also offering the capability to analyze the movement of the data by calculating its transitioning probability and creates elegant plots and GIFs.

Below are the new functions and its brief descriptions:

- `plotStateTransition`: Provides the time series flowmap plot.
- `getTransitionProbability`: Provides a list of transition probabilities.
- `reconcileTransitionProbability`: Provides plots and tables for comparing transition probabilities calculated manually and from markovchain function.
- `plotAnimatedFlowmap`: Creates flowmaps and animations for both self state and without self state scenarios.

</div>



<div id="hvt-(v23.11.02)" class="section level2" number="3.9">

## 3.2 HVT (v23.11.02) 

17th November, 2023

This version of HVT package offers functionality to score cells with layers based on a sequence of maps created using `scoreLayeredHVT`. Given below are the steps to created the successive set of maps.

1. **Map A** - The output of `trainHVT` function which is trained on parent data.

2. **Map B** - The output of `trainHVT` function which is trained on the 'data with novelty' created from `removeNovelty` function.

3. **Map C** - The output of `trainHVT` function which is trained on the 'data without novelty' created from `removeNovelty` function.

The `scoreLayeredHVT` function uses these three maps to score the test datapoints.

Let us try to understand the steps with the help of the diagram below

<p align="center">
  <img src="https://raw.githubusercontent.com/Mu-Sigma/HVT/master/vignettes/pngs/scoreLayeredHVT_function.png" width="672px" height="480px" />
</p>
<p align="center"><em>Figure 2: Data Segregation for scoring based on a sequence of maps using scoreLayeredHVT()</em></p>



</div>

<div id="hvt-(v22.12.06)" class="section level2" number="3.10">

## 3.1 HVT (v22.12.06) 

06th December, 2022

This version of HVT package offers features for both training an HVT model and eliminating outlier cells from the trained model.

1. **Training or Compression:** The initial step entails training the parent data using the `trainHVT` function, specifying the desired compression percentage and quantization error.

2. **Remove novelty cells:** Following the training process, outlier cells can be identified manually from the 2D hvt plot. These outlier cells can then be inputted into the `removeNovelty` function, which subsequently produces two datasets in its output: one containing 'data with novelty' and the other containing 'data without novelty'.


</div>


<div id="installation-of-hvt-(v26.1.3)" class="section level2" number="4">

# <span class="header-section-number">4.</span> Installation of HVT (v26.1.3)

<div class="sourceCode" id="cb1">

**CRAN Installation**

`install.packages("HVT")`

**Git Hub Installation**

`library(devtools)`

#increase the package download timelimit, if faced with error:  options(timeout = 1200) 

`devtools::install_github(repo = "Mu-Sigma/HVT")`

</div>

</div>

## ----setup, warning = FALSE, include = FALSE----------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  out.width = "auto",
  out.height = "480px",
  fig.width = 7,
  fig.height = 5,
  fig.align = "center",
  fig.retina = 1,
  dpi = 150,
  tangle = TRUE,
  eval=TRUE
)

## ----predictlayer_flow,echo=FALSE,warning=FALSE,fig.show='hold',message=FALSE,fig.cap='Figure 1:  Heatmap Visualization of a Torus with 500 Cells'----
knitr::include_graphics('./pngs/torus2.png')

## ----eval=TRUE, echo=FALSE----------------------------------------------------
library(knitr)
library(kableExtra)
# Create a data frame with hyperlinks
vignette_data <- data.frame(
  Version_Number = c("v18.05.17", "v18.05.17", "v23.05.16", "v23.10.26", "v24.05.16","v24.08.14", "v25.01.01"),
  Vignette_Title = c(
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/HVT_vignette.html' target='_blank'>HVT Vignette</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/HVT_model_diagnostics_vignette.html' target='_blank'>HVT Model Diagnostics Vignette</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/Scoring_Cells_with_Layers_using_scoreLayeredHVT.html' target='_blank'>HVT Scoring Cells with Layers using scoreLayeredHVT</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/HVT_Temporal_Analysis.html' target='_blank'>Temporal Analysis and Visualization: Leveraging Time Series Capabilities in HVT</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/LLM_Embeddings_in_HVT.html' target='_blank'>Visualizing LLM Embeddings using HVT</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/Implementation_of_tsne_umap_in_trainHVT.html' target='_blank'>Implementation of t-SNE and UMAP in trainHVT function</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/Dynamic_Forecasting_macroeconomic_data.html' target='_blank'>Dynamic Forecasting of Macroeconomic Time Series Dataset using HVT</a>"
),
  Description = c(
    "Contains descriptions of the functions used for vector quantization and construction of Hierarchical Voronoi Tessellations for data analysis.",
    "Contains descriptions of functions used to perform model diagnostics and validation for the HVT model.",
    "Contains descriptions of the functions used for scoring cells with layers based on a sequence of maps using scoreLayeredHVT.",
    "Contains descriptions of the functions used for analyzing time series data and its flow maps.",
    "Contains the implementation and analysis of hierarchical clustering using the `clustHVT` function to evaluate and visualize token embeddings generated by OpenAI.",
    "Contains enhancements to the `trainHVT` function with advanced dimensionality reduction techniques such as  t-SNE and UMAP, and includes a table of evaluation metrics to improve analysis, visualization, and interpretability.",
    "Contains enhancements to the HVT package for dynamic forecasting using Monte Carlo Simulations of Markov Chain (MSM) on macroeconomic time series dataset."
  ),
  stringsAsFactors = FALSE
)
 
# Generate the table
kable(vignette_data, format = "html", 
      col.names = c("Version", "Vignette Title", "Description"), 
      align = "l", escape = FALSE) %>% 
  kable_styling(full_width = TRUE) %>% column_spec(1, width = "15%") 

## ----mlayer_flow,echo=FALSE,warning=FALSE,fig.show='hold',message=FALSE,fig.cap='Figure 2: Data Segregation for scoring based on a sequence of maps using scoreLayeredHVT()'----
knitr::include_graphics('./pngs/scoreLayeredHVT_function.png')

## ----setup, warning = FALSE, include = FALSE----------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  out.width = "auto",
  out.height = "480px",
  fig.width = 7,
  fig.height = 5,
  fig.align = "center",
  fig.retina = 1,
  dpi = 150,
  eval=TRUE,
  tangle = FALSE
)

## ----predictlayer_flow,echo=FALSE,warning=FALSE,fig.show='hold',message=FALSE,fig.cap='Figure 1: The Voronoi tessellation for layer 1 and number of cells 500 with the heat map overlaid for variable z.'----
knitr::include_graphics('./pngs/torus2.png')

## ----eval=TRUE, echo=FALSE----------------------------------------------------
library(knitr)
library(kableExtra)
# Create a data frame with hyperlinks
vignette_data <- data.frame(
  Version_Number = c("v18.05.17", "v18.05.17", "v23.05.16", "v23.10.26", "v24.05.16","v24.08.14", "v25.03.01","v25.08.25"),
  Vignette_Title = c(
  "<a href='https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/HVT_vignette.html' target='_blank' rel='noopener noreferrer'>HVT Vignette</a>",
  "<a href='https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/HVT_model_diagnostics_vignette.html' target='_blank' rel='noopener noreferrer'>HVT Model Diagnostics Vignette</a>",
  "<a href='https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/Scoring_Cells_with_Layers_using_scoreLayeredHVT.html' target='_blank' rel='noopener noreferrer'>HVT Scoring Cells with Layers using scoreLayeredHVT</a>",
  "<a href='https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/HVT_Temporal_Analysis.html' target='_blank' rel='noopener noreferrer'>Temporal Analysis and Visualization: Leveraging Time Series Capabilities in HVT",
  "<a href='https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/LLM_Embeddings_in_HVT.html' target='_blank' rel='noopener noreferrer'>Visualizing LLM Embeddings using HVT</a>",
  "<a href='https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/Implementation_of_tsne_umap_in_trainHVT.html' target='_blank' rel='noopener noreferrer'>Implementation of t-SNE and UMAP in trainHVT function</a>",
  "<a href='https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/Dynamic_Forecasting_macroeconomic_data.html' target='_blank' rel='noopener noreferrer'>Dynamic Forecasting of Macroeconomic Time Series Dataset using HVT</a>",
  "<a href='https://raw.githack.com/Mu-Sigma/HVT/master/vignettes/Experimentation_of_hyperparameters_in_msm.html' target='_blank' rel='noopener noreferrer'>Hyperparameter Experimentation for Champion Model Selection in MSM Dynamic Forecasting</a>"
),
  Description = c(
    "Contains the workflow of the functions used for vector quantization and construction of Hierarchical Voronoi Tessellations for data analysis.",
    "Contains demonstrations of functions used to perform model diagnostics and validation for the trained HVT model.",
    "Contains explanations of the functions used for scoring cells with layers based on a sequence of maps using scoreLayeredHVT.",
    "Contains implementations of the functions used for analyzing time series data and creating its state transition flow maps.",
    "Contains implementation and analysis of hierarchical clustering using functions to evaluate and visualize token embeddings generated by OpenAI in 2D Space.",
    "Contains enhancements to the `trainHVT` function with advanced dimensionality reduction techniques such as t-SNE and UMAP, and includes a table of evaluation metrics to improve interpretability.",
    "Contains enhancements to the HVT package for dynamic forecasting using Monte Carlo Simulations of Markov Chain (MSM) on macroeconomic time series dataset.",
    "Contains enhancements to enable strategic selection of the champion model based on the lowest Mean Absolute Error by hyperparameter tuning in msm - dynamic forecasting."
  ),
  stringsAsFactors = FALSE
)
 
# Generate the table
kable(vignette_data, format = "html", 
      col.names = c("Version", "Vignette Title", "Description"), 
      align = "l", escape = FALSE) %>% 
  kable_styling(full_width = TRUE) %>% column_spec(1, width = "15%") 

## ----mlayer_flow,echo=FALSE,warning=FALSE,fig.show='hold',message=FALSE,fig.cap='Figure 2: Data Segregation for scoring based on a sequence of maps using scoreLayeredHVT()'----
knitr::include_graphics('./pngs/scoreLayeredHVT_function.png')

#' @name plotHVT 
#' @title Plot the hierarchical tessellations.
#' @description This is the main plotting function to construct hierarchical voronoi tessellations in 1D,2D or
#' Interactive surface plot.
#' @param hvt.results (1D/2DProj/2Dhvt/2Dheatmap/surface_plot) List. A list containing the output of \code{trainHVT} function
#' which has the details of the tessellations to be plotted.
#' @param plot.type Character. An option to indicate which type of plot should be generated. Accepted entries are 
#' '1D','2Dproj','2Dhvt','2Dheatmap'and 'surface_plot'. Default value is '2Dhvt'.
#' @param line.width (2Dhvt/2Dheatmap) Numeric Vector. A vector indicating the line widths of the
#' tessellation boundaries for each level.
#' @param color.vec (2Dhvt/2Dheatmap) Vector. A vector indicating the colors of the boundaries of
#' the tessellations at each level.
#' @param centroid.size (2Dhvt/2Dheatmap) Numeric Vector. A vector indicating the size of centroids
#' for each level.
#' @param centroid.color (2Dhvt/2Dheatmap) Numeric Vector. A vector indicating the color of centroids
#' for each level.
#' @param child.level (2Dheatmap/surface_plot) Numeric. Indicating the level for which the plot should
#' be displayed
#' @param hmap.cols (2Dheatmap/surface_plot) Numeric or Character. The column number or column name from
#' the dataset indicating the variables for which the heat map is to be plotted.
#' @param quant.error.hmap (2Dheatmap) Numeric. A number representing the quantization error threshold to be highlighted in the heatmap. 
#' When a value is provided, it will emphasize cells with quantization errors equal or less than the specified threshold,
#' indicating that these cells cannot be further subdivided in the next depth layer. The default value is NULL,
#' meaning all cells will be colored in the heatmap across various depths.
#' @param separation_width (surface_plot) Numeric. An integer indicating the width between hierarchical levels in surface plot
#' @param layer_opacity (surface_plot) Numeric. A vector indicating the opacity of each hierarchical levels in surface plot
#' @param dim_size  (surface_plot) Numeric. An integer controls the resolution or granularity of the 3D surface grid
#' @param cell_id (2Dhvt/2Dheatmap) Logical. A logical indicating whether the cell IDs should be displayed. Default is FALSE
#' @param cell_id_position (2Dhvt/2Dheatmap) Character. A character indicating the position of the cell IDs. Accepted entries are 'top' , 
#' 'bottom', 'left', 'right', and 'center'. Default is 'bottom'
#' @param cell_id_size (2Dhvt/2Dheatmap) Numeric. A numeric vector indicating the size of the cell IDs. Default is 2.6
#' @param centroids (2Dhvt/2Dheatmap) Logical. A logical indicating whether the centroid points should be displayed. Default value is TRUE.
#' @returns plot object containing the visualizations of reduced dimension(1D/2D) for the given dataset.
#' @author Shubhra Prakash <shubhra.prakash@@mu-sigma.com>, Sangeet Moy Das <sangeet.das@@mu-sigma.com>, Vishwavani <vishwavani@@mu-sigma.com>
#' @seealso \code{\link{trainHVT}} 
#' @keywords Tessellation_and_Heatmap
#' @importFrom magrittr %>%
#' @import ggplot2
#' @include utilities.R constants.R
#' @examples
#' data("EuStockMarkets")
#' hvt.results <- trainHVT(EuStockMarkets, n_cells = 60, depth = 1, quant.err = 0.1, 
#'                        distance_metric = "L1_Norm", error_metric = "max",
#'                        normalize = TRUE,quant_method="kmeans")
#'                        
#' #change the 'plot.type' argument to '2Dproj' or '2DHVT' to visualize respective plots.                      
#' plotHVT(hvt.results, plot.type='1D')
#' 
#' #change the 'plot.type' argument to 'surface_plot' to visualize the Interactive surface plot                   
#' plotHVT(hvt.results,child.level = 1, 
#' hmap.cols = "DAX", plot.type = '2Dheatmap')
#' @export plotHVT


plotHVT <- function(hvt.results, 
                    line.width = 0.5, 
                    color.vec =  'black', 
                    centroid.size = 0.6, 
                    centroid.color = "black", 
                    child.level = 1, 
                    hmap.cols,
                    separation_width = 7, 
                    layer_opacity = c(0.5, 0.75, 0.99), 
                    dim_size = 1000, 
                    plot.type = '2Dhvt',
                    quant.error.hmap = NULL,
                    cell_id = FALSE,
                    cell_id_position="bottom", 
                    cell_id_size = 2.6,
                    centroids = TRUE) {

  
  lev <- NULL
  n_cells <- max(stats::na.omit(hvt.results[[3]][["summary"]][["Cell.ID"]]))
  n_cells.hmap <- hvt.results[["model_info"]][["input_parameters"]][["n_cells"]]
  pch1 = 21
    
  if (is.null(plot.type)) {
    plot.type <- '2Dhvt'
  }
  
  if (plot.type == '1D') {
    generic_col=c("Segment.Level","Segment.Parent","Segment.Child","n","Quant.Error")
    hvq_k <- hvt.results[[6]]  
    temp_summary=hvq_k[["summary"]] %>% dplyr::select(!generic_col) %>% dplyr::mutate(id=row_number())
    cent_val= temp_summary %>% subset(.,stats::complete.cases(.)) 
    set.seed(123)
    sammon_1d_cord <- MASS::sammon(
      d = stats::dist(cent_val %>% dplyr::select(!id),method = "manhattan"),
      niter = 10 ^ 5,
      trace = FALSE,
      k=1
    )$points
    temp_df=data.frame(sammon_1d_cord,id=cent_val$id)%>%dplyr::arrange(sammon_1d_cord) %>% dplyr::mutate(Cell.ID=row_number()) %>% dplyr::select(!sammon_1d_cord)
    temp_summary = dplyr::left_join(temp_summary,temp_df,by="id") %>% select(!"id")
    hvq_k[["summary"]]$Cell.ID=temp_summary$Cell.ID
    
    x <-sammon_1d_cord
    y <- hvq_k[["summary"]][["Cell.ID"]]
    data_plot <- data.frame(x,y)
    
      if(length(y) <= 100) {dot_size <- 5} else if(length(y) <= 500) {dot_size <- 2.5} 
      else if(length(y) <= 1000) {dot_size <- 1.5} else {dot_size <- 1}
    
    plotly_obj <-  plotly::plot_ly(data_plot, x = ~y, y = ~x, type = 'scatter', mode = 'markers',
            marker = list(size = dot_size, color = 'blue', symbol = 'circle'),
            text = ~paste('Cell ID:', y, '<br>1D point:', round(x,4)), 
            hoverinfo = 'text') %>% 
            plotly::layout(title = 'Sammons 1D x Cell ID',
             xaxis = list(title = 'Cell ID', zeroline = FALSE),
             yaxis = list(title = '1D points',zeroline = FALSE))
    
     return(suppressMessages(plotly_obj))  
      
  } else if (plot.type == '2Dproj'){
   
     hvt_centroids_list <- hvt.results
      
    hvt_coordinates<- hvt_centroids_list[[2]][[1]][["1"]]
    centroids <- list()
    coordinates_value <- lapply(1:length(hvt_coordinates), function(x){
      centroids <-hvt_coordinates[[x]]
      coordinates <- centroids$pt
    })
    centroid_coordinates<<- do.call(rbind.data.frame, coordinates_value)  
    colnames(centroid_coordinates) <- c("x_coord","y_coord")
    centroid_coordinates$Row.No <- as.numeric(row.names(centroid_coordinates)) 
    centroid_coordinates <- centroid_coordinates %>% dplyr::select(Row.No,x_coord,y_coord)
    centroid_coordinates1 <- centroid_coordinates %>% data.frame() %>% round(4)

    gg_proj <- ggplot(centroid_coordinates1, ggplot2::aes(x_coord, y_coord)) +
      ggplot2::geom_point(color = "blue") +
      labs(title = "2D Projection plot of centroids",
           x = "X", y = "Y")
    
    
    return(suppressMessages(gg_proj))
    
     } else if (plot.type == '2Dhvt') {
      
    hvt_list <- hvt.results
    
    child.level <- min(child.level, max(hvt_list[[3]][["summary"]] %>% stats::na.omit() %>% dplyr::select("Segment.Level")))
    
    
    min_x <- 1e9
    min_y <- 1e9
    max_x <- -1e9
    max_y <- -1e9
    depthVal <- c()
    clusterVal <- c()
    childVal <- c()
    value <- c()
    x_pos <- c()
    y_pos <- c()
    x_cor <- c()
    y_cor <- c()
    depthPos <- c()
    clusterPos <- c()
    childPos <- c()
    levelCluster <- c()
    for (clusterNo in 1:length(hvt_list[[2]][[1]][[1]])) {
      bp_x <- hvt_list[[2]][[1]][[1]][[clusterNo]][["x"]]
      bp_y <- hvt_list[[2]][[1]][[1]][[clusterNo]][["y"]]
      
      
      if (min(bp_x) < min_x) {
        min_x <- min(bp_x)
      }
      if (max(bp_x) > max_x) {
        max_x <- max(bp_x)
      }
      if (min(bp_y) < min_y) {
        min_y <- min(bp_y)
      }
      if (max(bp_y) > max_y) {
        max_y <- max(bp_y)
      }
    }
    
    
    for (depth in 1:child.level) {
      for (clusterNo in 1:length(hvt_list[[2]][[depth]])) {
        for (childNo in 1:length(hvt_list[[2]][[depth]][[clusterNo]])) {
          current_cluster <- hvt_list[[2]][[depth]][[clusterNo]][[childNo]]
          
          x <- as.numeric(current_cluster[["x"]])
          y <- as.numeric(current_cluster[["y"]])
          x_cor <- c(x_cor, as.numeric(current_cluster[["pt"]][["x"]]))
          y_cor <- c(y_cor, as.numeric(current_cluster[["pt"]][["y"]]))
          depthVal <- c(depthVal, depth)
          clusterVal <- c(clusterVal, clusterNo)
          childVal <- c(childVal, childNo)
          depthPos <- c(depthPos, rep(depth, length(x)))
          clusterPos <- c(clusterPos, rep(clusterNo, length(x)))
          childPos <- c(childPos, rep(childNo, length(x)))
          x_pos <- c(x_pos, x)
          y_pos <- c(y_pos, y)
          levelCluster <- c(levelCluster, depth)
        }
      }
    }
    
    valuesDataframe <- data.frame(
      depth = depthVal,
      cluster = clusterVal,
      child = childVal
    )
    
    
    positionsDataframe <- data.frame(
      depth = depthPos,
      cluster = clusterPos,
      child = childPos,
      x = x_pos,
      y = y_pos
    )
    
    
    centroidDataframe <-
      data.frame(x = x_cor, y = y_cor, lev = levelCluster)
    
    datapoly <-
      merge(valuesDataframe,
            positionsDataframe,
            by = c("depth", "cluster", "child")
      )


      cellID_coordinates <- extract_cell_coordinates(hvt_list)

      
      centroidDataframe_2 <- merge(cellID_coordinates, centroidDataframe, by = c("x" ,"y"))
    
    # Calculate polygon geometric centroids for cell ID positioning
    polygon_centroids_list <- list()
    for (depth_val in 1:child.level) {
      depth_data <- datapoly[datapoly$depth == depth_val, ]
      if (nrow(depth_data) > 0) {
        polygon_centroids <- depth_data %>%
          dplyr::group_by(depth, cluster, child) %>%
          dplyr::summarise(
            x = calculate_polygon_centroid(x, y)$x,
            y = calculate_polygon_centroid(x, y)$y,
            .groups = 'drop'
          )
        polygon_centroids_list[[depth_val]] <- polygon_centroids
      }
    }
    polygon_centroids_df <- do.call(rbind, polygon_centroids_list)
    
    # Match polygon centroids with Cell.IDs for level 1
    if (child.level >= 1 && nrow(polygon_centroids_df) > 0) {
      level1_poly_centroids <- polygon_centroids_df[polygon_centroids_df$depth == 1, ]
      # Merge with Cell.IDs using cluster and child matching
      cellID_mapping <- hvt_list[[3]][["summary"]] %>%
        dplyr::filter(Segment.Level == 1) %>%
        dplyr::select(Segment.Parent, Segment.Child, Cell.ID) %>%
        dplyr::mutate(
          Segment.Parent = as.character(Segment.Parent),
          Segment.Child = as.numeric(Segment.Child)
        )
      
      cellID_polygon_centroids <- level1_poly_centroids %>%
        dplyr::mutate(
          cluster = as.character(cluster),
          child = as.numeric(child)
        ) %>%
        dplyr::left_join(cellID_mapping, 
                         by = c("cluster" = "Segment.Parent", "child" = "Segment.Child")) %>%
        dplyr::filter(!is.na(Cell.ID))
    } else {
      cellID_polygon_centroids <- data.frame(x = numeric(0), y = numeric(0), Cell.ID = numeric(0))
    }
    
    p <- ggplot2::ggplot()  
    for (i in child.level:1) {
      p <-
        p + ggplot2::geom_polygon(
          data = datapoly[which(datapoly$depth == i), ],
          ggplot2::aes(
            x = x,
            y = y,
            color = factor(depth),
            size = factor(depth),
            group = interaction(depth, cluster, child),
          ),
          fill = NA
        ) +
        ggplot2::scale_colour_manual(values = color.vec) +
        ggplot2::scale_size_manual(values = line.width, guide = "none") +
        ggplot2::labs(color = "Level")
    }
    
    if (cell_id == TRUE) {

    # Add centroid points if centroids = TRUE
    if (centroids == TRUE) {
      for (depth in 1:child.level) {
        depth_size <- centroid.size[child.level - depth + 1]
        centroid_color <- centroid.color[child.level - depth + 1]
        
        p <- p + ggplot2::geom_point(
          data = centroidDataframe[centroidDataframe["lev"] == depth, ],
          ggplot2::aes(x = x, y = y),
          size = depth_size,
          pch = pch1,
          fill = centroid_color,
          color = centroid_color
        ) 
      }
    }

    # Use polygon centroids if cell_id_position is "center", otherwise use point centroids
    if (cell_id_position == "center") {
      if (nrow(cellID_polygon_centroids) > 0) {
        subset_data <- cellID_polygon_centroids
        # Center alignment for polygon centroids
        hjust_val <- 0.5
        vjust_val <- 0.5
      } else {
        # Fallback to original method if polygon centroids not available
        subset_data <- subset(centroidDataframe_2, lev == 1)
        hjust_val <- 0.5
        vjust_val <- 0.5
      }
    } else {
      # Use original point centroids for other positions
      subset_data <- subset(centroidDataframe_2, lev == 1)
      
      alignments <- list(
        right = list(hjust = -0.5, vjust = 0.5),
        left = list(hjust = 1.5, vjust = 0.5),
        bottom = list(hjust = 0.5, vjust = 1.5),
        top = list(hjust = 0.5, vjust = -0.5)
      )
      
      # Get alignment values for the current position
      alignment <- rlang::`%||%`(alignments[[cell_id_position]], alignments$bottom)
      hjust_val <- alignment$hjust
      vjust_val <- alignment$vjust
    }
    
    # Add geom_text to the plot
    if (cell_id_position == "center") {
      p <- p + ggplot2::geom_text(
        data = subset_data,
        ggplot2::aes(x = x, y = y, label = Cell.ID),
        size = cell_id_size,
        color = "black",
        hjust = hjust_val,
        vjust = vjust_val)
    } else {
      p <- p + ggplot2::geom_text(
        data = subset_data,
        ggplot2::aes(x = x, y = y, label = Cell.ID),
        size = cell_id_size,
        color = "black",
        hjust = hjust_val,
        vjust = vjust_val,
        nudge_x = -0.02 * nchar(as.character(subset_data$Cell.ID)))
    }
     
    }else {
       # Add centroid points if centroids = TRUE
       if (centroids == TRUE) {
         for (depth in 1:child.level) {
           depth_size <- centroid.size[child.level - depth + 1]
           centroid_color <- centroid.color[child.level - depth + 1]
           
           p <- p + ggplot2::geom_point(
             data = centroidDataframe[centroidDataframe["lev"] == depth, ],
             ggplot2::aes(x = x, y = y),
             size = depth_size,
             pch = pch1,
             fill = centroid_color,
             color = centroid_color
           ) 
         }
       }
       
     }
  
 

    p <- p +
      ggplot2::scale_color_manual(
        name = "Level",
        values = color.vec
      ) +
      ggplot2::theme_bw() + ggplot2::theme(
        plot.background = ggplot2::element_blank(),
        plot.title = ggplot2::element_text(
          size = 20,
          hjust = 0.5,
          margin = ggplot2::margin(0, 0, 20, 0)
        ),
        panel.grid = ggplot2::element_blank(),
        panel.border = ggplot2::element_blank(),
        axis.ticks = ggplot2::element_blank(),
        axis.text = ggplot2::element_blank(),
        axis.title = ggplot2::element_blank(),
        panel.background = ggplot2::element_blank()
      ) + ggplot2::theme(plot.title = ggplot2::element_text(hjust = 0.5)) +
      ggplot2::scale_x_continuous(expand = c(0, 0)) +
      ggplot2::scale_y_continuous(expand = c(0, 0)) +
      if (!is.null(centroidDataframe$outlier_cell) && any(!is.na(centroidDataframe$outlier_cell))) {
        p <- p + ggplot2::geom_label(
          label = centroidDataframe$outlier_cell,
          nudge_x = 0.45, nudge_y = 0.1,
          label.padding = grid::unit(0.55, "lines"),
          label.size = 0.4,
          color = "white",
          fill = "#038225"
        )
      } +
      ggplot2::ggtitle(paste0("Tesellation Plot of ", n_cells, " cells"))+ ggplot2::theme(
        plot.title = ggplot2::element_text(hjust = 0, size = 10 ))
   

    return(suppressMessages(p))
    
    
  } else if (plot.type == '2Dheatmap') {
    hvt_list <- hvt.results
    
    child.level <- min(child.level, max(hvt_list[[3]][["summary"]] %>% stats::na.omit() %>% dplyr::select("Segment.Level")))
    summaryDF <- hvt_list[[3]][["summary"]]
    valuesDataframe <- data.frame(
      depth = 0,
      cluster = 0,
      child = 0,
      qe = 0,
      n_cluster = 0,
      value = 0,
      cellID = 0
    )
    positionsDataframe <- data.frame(
      depth = 0,
      cluster = 0,
      child = 0,
      x = 0,
      y = 0
    )
    
    centroidDataframe <- data.frame(x = 0, y = 0, lev = 0)
    
    for (depth in 1:child.level) {
      if (depth < 3) {
        for (clusterNo in names(hvt_list[[2]][[depth]])) {
          for (childNo in 1:length(hvt_list[[2]][[depth]][[clusterNo]])) {
            if (!is.null(hvt_list[[2]][[depth]][[clusterNo]])) {
              summaryFilteredDF <-
                summaryDF %>% dplyr::filter(
                  Segment.Level == depth,
                  Segment.Parent == clusterNo,
                  Segment.Child == childNo
                )
              
              current_cluster <- hvt_list[[2]][[depth]][[clusterNo]][[childNo]]
              
              
              val <- summaryFilteredDF[, hmap.cols]
              qe <- summaryFilteredDF[, "Quant.Error"]
              ncluster <- summaryFilteredDF[, "n"]
              x <- as.numeric(current_cluster[["x"]])
              y <- as.numeric(current_cluster[["y"]])
              if ("Cell.ID" %in% colnames(summaryFilteredDF)) {
                cell_ID <- summaryFilteredDF[, "Cell.ID"]
              } else {
                cell_ID <- 0
              }
              valuesDataframe <-
                rbind(
                  valuesDataframe,
                  data.frame(
                    depth = depth,
                    cluster = clusterNo,
                    child = childNo,
                    qe = qe,
                    n_cluster = ncluster,
                    value = val,
                    cellID = cell_ID
                  )
                )
              positionsDataframe <-
                rbind(
                  positionsDataframe,
                  data.frame(
                    depth = rep(depth, length(x)),
                    cluster = rep(clusterNo, length(x)),
                    child = rep(childNo, length(x)),
                    x = x,
                    y = y
                  )
                )
              
              centroidDataframe <-
                rbind(
                  centroidDataframe,
                  data.frame(
                    x = as.numeric(current_cluster[["pt"]][["x"]]),
                    y = as.numeric(current_cluster[["pt"]][["y"]]),
                    lev = depth
                  )
                )
            }
          }
        }
      } else {
        for (clusterNo in 1:n_cells.hmap^(child.level - 1)) {
          for (childNo in 1:length(hvt_list[[2]][[depth]][[as.character(clusterNo)]])) {
            if (!is.null(hvt_list[[2]][[depth]][[as.character(clusterNo)]])) {
              summaryFilteredDF <-
                summaryDF %>% dplyr::filter(
                  Segment.Level == depth,
                  Segment.Parent == clusterNo,
                  Segment.Child == childNo
                )
              
              current_cluster <- hvt_list[[2]][[depth]][[as.character(clusterNo)]][[childNo]]
              
              
              val <- summaryFilteredDF[, hmap.cols]
              qe <- summaryFilteredDF[, "Quant.Error"]
              
              x <- as.numeric(current_cluster[["x"]])
              y <- as.numeric(current_cluster[["y"]])
              if ("Cell.ID" %in% colnames(summaryFilteredDF)) {
                cell_ID <- summaryFilteredDF[, "Cell.ID"]
              } else {
                cell_ID <- 0
              }
              valuesDataframe <-
                rbind(
                  valuesDataframe,
                  data.frame(
                    depth = depth,
                    cluster = clusterNo,
                    child = childNo,
                    qe = qe,
                    n_cluster = ncluster,
                    value = val,
                    cellID = cell_ID
                  )
                )
              positionsDataframe <-
                rbind(
                  positionsDataframe,
                  data.frame(
                    depth = rep(depth, length(x)),
                    cluster = rep(clusterNo, length(x)),
                    child = rep(childNo, length(x)),
                    x = x,
                    y = y
                  )
                )
              
              centroidDataframe <-
                rbind(
                  centroidDataframe,
                  data.frame(
                    x = as.numeric(current_cluster[["pt"]][["x"]]),
                    y = as.numeric(current_cluster[["pt"]][["y"]]),
                    lev = depth
                  )
                )
            }
          }
        }
      }
    }
    valuesDataframe <- valuesDataframe[2:nrow(valuesDataframe), ]
    
    positionsDataframe <-
      positionsDataframe[2:nrow(positionsDataframe), ]
    
    centroidDataframe <-
      centroidDataframe[2:nrow(centroidDataframe), ]
    
    
    datapoly <-
      merge(valuesDataframe,
            positionsDataframe,
            by = c("depth", "cluster", "child")
      )

    # Calculate polygon geometric centroids for cell ID positioning
    polygon_centroids_list_hmap <- list()
    for (depth_val in 1:child.level) {
      depth_data <- datapoly[datapoly$depth == depth_val, ]
      if (nrow(depth_data) > 0) {
        polygon_centroids_hmap <- depth_data %>%
          dplyr::group_by(depth, cluster, child) %>%
          dplyr::summarise(
            x = calculate_polygon_centroid(x, y)$x,
            y = calculate_polygon_centroid(x, y)$y,
            .groups = 'drop'
          )
        polygon_centroids_list_hmap[[depth_val]] <- polygon_centroids_hmap
      }
    }
    polygon_centroids_df_hmap <- do.call(rbind, polygon_centroids_list_hmap)

    p <- ggplot2::ggplot()
    colour_scheme <- HVT_COLOR_SCHEME
    data <- datapoly
    if (child.level > 1) {
      for (i in 1:(child.level - 1)) {
        #index_tess <- which(data$depth == i & data$qe > quant.error.hmap & data$n_cluster > 3)
         index_tess <- which( data$qe > quant.error.hmap )
        
        if (length(index_tess) >0){
          data <- data[-index_tess, ] 
        } else {
            data = data
          }
         
      }
    }
    
    hvt_res1 <- hvt_list[[2]][[1]]$`1`
    hvt_res2 <- hvt_list[[3]]$summary$Cell.ID
    a <- 1: length(hvt_res1)
    b <- a[hvt_res2]
    b <-  as.vector(b)
    hvt_res2 <- stats::na.omit(b)
    
    coordinates_value1 <- lapply(1:length(hvt_res1), function(x) {
      centroids1 <- hvt_res1[[x]]
      coordinates1 <- centroids1$pt})
    cellID_coordinates <- do.call(rbind.data.frame, coordinates_value1)
    colnames(cellID_coordinates) <- c("x", "y")
    cellID_coordinates$Cell.ID <- hvt_res2
    centroidDataframe_2 <- merge(cellID_coordinates, centroidDataframe, by = c("x" ,"y"))
    
    # Match polygon centroids with Cell.IDs for level 1 (for heatmap)
    if (child.level >= 1 && nrow(polygon_centroids_df_hmap) > 0) {
      level1_poly_centroids_hmap <- polygon_centroids_df_hmap[polygon_centroids_df_hmap$depth == 1, ]
      # Merge with Cell.IDs using cluster and child matching
      cellID_mapping_hmap <- hvt_list[[3]][["summary"]] %>%
        dplyr::filter(Segment.Level == 1) %>%
        dplyr::select(Segment.Parent, Segment.Child, Cell.ID) %>%
        dplyr::mutate(
          Segment.Parent = as.character(Segment.Parent),
          Segment.Child = as.numeric(Segment.Child)
        )
      
      cellID_polygon_centroids_hmap <- level1_poly_centroids_hmap %>%
        dplyr::mutate(
          cluster = as.character(cluster),
          child = as.numeric(child)
        ) %>%
        dplyr::left_join(cellID_mapping_hmap, 
                         by = c("cluster" = "Segment.Parent", "child" = "Segment.Child")) %>%
        dplyr::filter(!is.na(Cell.ID))
    } else {
      cellID_polygon_centroids_hmap <- data.frame(x = numeric(0), y = numeric(0), Cell.ID = numeric(0))
    }
    
    p <- ggplot2::ggplot()
 
   
    
    # changing hoverText for torus demo
    if ("Cell.ID" %in% colnames(summaryFilteredDF)) {
      p <-
        p + ggplot2::geom_polygon(
          data = data,
          ggplot2::aes(
            x = x,
            y = y,
            group = interaction(depth, cluster, child),
            fill = value,
            text = paste(
              " Cell.ID:", cellID,
              "<br>", "Segment.Level:", depth,
              "<br>", "Segment.Parent:", cluster,
              "<br>", "Segment.Child:", child
            )
          )
        ) +
        ggplot2::scale_fill_gradientn(colours = colour_scheme, guide = guide_colorbar(order = 2)) +
        ggplot2::labs(fill = hmap.cols)
    } else {
      p <-
        p + ggplot2::geom_polygon(
          data = data,
          ggplot2::aes(
            x = x,
            y = y,
            group = interaction(depth, cluster, child),
            fill = value
          )
        ) +
        ggplot2::scale_fill_gradientn(colours = colour_scheme,guide = guide_colorbar(order = 2)) +
        ggplot2::labs(fill = hmap.cols)
    }
    
    
    for (i in child.level:1) {
      p <-
        p + ggplot2::geom_polygon(
          data = datapoly[which(datapoly$depth == i), ],
          ggplot2::aes(
            x = x,
            y = y,
            color = factor(depth),
            size = factor(depth),
            group = interaction(depth, cluster, child)
          ),
          fill = NA
        ) +
        ggplot2::scale_colour_manual(values = color.vec,guide = guide_legend(order = 1)) +
        ggplot2::scale_size_manual(values = line.width, guide = FALSE) +
        ggplot2::labs(color = "Level")
    }
    
    if (cell_id == TRUE) {

      # Add centroid points if centroids = TRUE
      if (centroids == TRUE) {
        for (depth in 1:child.level) {
          depth_size <- centroid.size[child.level - depth + 1]
          centroid_color <- centroid.color[child.level - depth + 1]
          
          p <- p + ggplot2::geom_point(
            data = centroidDataframe[centroidDataframe["lev"] == depth, ],
            ggplot2::aes(x = x, y = y),
            size = depth_size,
            fill = centroid_color,
            color = centroid_color
          ) 
        }
      }
      
      # Use polygon centroids if cell_id_position is "center", otherwise use point centroids
      if (cell_id_position == "center" && nrow(cellID_polygon_centroids_hmap) > 0) {
        subset_data <- cellID_polygon_centroids_hmap
      } else {
        subset_data <- subset(centroidDataframe_2, lev == 1)
      }
      alignment <- get_cell_label_alignment(cell_id_position, "ggplot")
      hjust_val <- alignment$hjust
      vjust_val <- alignment$vjust
      
      
      # Add geom_text to the plot
      if (cell_id_position == "center") {
        p <- p + ggplot2::geom_text(
          data = subset_data,
          ggplot2::aes(x = x, y = y, label = Cell.ID),
          size = cell_id_size,
          color = "black",
          hjust = hjust_val,
          vjust = vjust_val)
      } else {
        p <- p + ggplot2::geom_text(
          data = subset_data,
          ggplot2::aes(x = x, y = y, label = Cell.ID),
          size = cell_id_size,
          color = "black",
          hjust = hjust_val,
          vjust = vjust_val,
          nudge_x = -0.02 * nchar(as.character(subset_data$Cell.ID)))
      }
    }else {
      # Add centroid points if centroids = TRUE
      if (centroids == TRUE) {
        for (depth in 1:child.level) {
          depth_size <- centroid.size[child.level - depth + 1]
          centroid_color <- centroid.color[child.level - depth + 1]
          
          p <- p + ggplot2::geom_point(
            data = centroidDataframe[centroidDataframe["lev"] == depth, ],
            ggplot2::aes(x = x, y = y),
            size = depth_size,
            pch = pch1,
            fill = centroid_color,
            color = centroid_color
          ) 
        }
      }
      
    }
    
      

    
    # Continue with the rest of the plot modifications
    p <- p + ggplot2::theme(
      plot.background = ggplot2::element_blank(),
      plot.title = ggplot2::element_text(
        size = 20,
        hjust = 0.5,
        margin = ggplot2::margin(0, 0, 20, 0)
      ),
      panel.grid = ggplot2::element_blank(),
      panel.border = ggplot2::element_blank(),
      axis.ticks = ggplot2::element_blank(),
      axis.text = ggplot2::element_blank(),
      axis.title = ggplot2::element_blank(),
      panel.background = ggplot2::element_blank()
    ) +
      ggplot2::scale_x_continuous(expand = c(0, 0)) +
      ggplot2::scale_y_continuous(expand = c(0, 0)) +
      ggplot2::ggtitle(paste0("Heatmap of ", n_cells, " cells"))+ ggplot2::theme(
        plot.title = ggplot2::element_text(hjust = 0, size = 10 ))
   
    
   
    
    return(suppressMessages(p))
    
  } else if (plot.type == 'surface_plot') {
    
    child.level <- min(child.level, max(hvt.results[[3]][["summary"]]
                                     %>% stats::na.omit()
                                     %>% dplyr::select("Segment.Level")))
    summaryDF <- hvt.results[[3]][["summary"]]
    valuesDataframe <- data.frame(
      depth = 0,
      cluster = 0,
      child = 0,
      qe = 0,
      n_cluster = 0,
      value = 0,
      cellID = 0
    )
    positionsDataframe <- data.frame(
      depth = 0,
      cluster = 0,
      child = 0,
      x = 0,
      y = 0
    )
    
    centroidDataframe <- data.frame(x = 0, y = 0, lev = 0)
    
    for (depth in 1:child.level) {
      if (depth < 3) {
        for (clusterNo in names(hvt.results[[2]][[depth]])) {
          for (childNo in 1:length(hvt.results[[2]][[depth]][[clusterNo]])) {
            if (!is.null(hvt.results[[2]][[depth]][[clusterNo]])) {
              summaryFilteredDF <-
                summaryDF %>% dplyr::filter(
                  Segment.Level == depth,
                  Segment.Parent == clusterNo,
                  Segment.Child == childNo
                )
              
              current_cluster <- hvt.results[[2]][[depth]][[clusterNo]][[childNo]]
              
              
              val <- summaryFilteredDF[, hmap.cols]
              qe <- summaryFilteredDF[, "Quant.Error"]
              ncluster <- summaryFilteredDF[, "n"]
              x <- as.numeric(current_cluster[["x"]])
              y <- as.numeric(current_cluster[["y"]])
              if ("Cell.ID" %in% colnames(summaryFilteredDF)) {
                cell_ID <- summaryFilteredDF[, "Cell.ID"]
              } else {
                cell_ID <- 0
              }
              valuesDataframe <-
                rbind(
                  valuesDataframe,
                  data.frame(
                    depth = depth,
                    cluster = clusterNo,
                    child = childNo,
                    qe = qe,
                    n_cluster = ncluster,
                    value = val,
                    cellID = cell_ID
                  )
                )
              positionsDataframe <-
                rbind(
                  positionsDataframe,
                  data.frame(
                    depth = rep(depth, length(x)),
                    cluster = rep(clusterNo, length(x)),
                    child = rep(childNo, length(x)),
                    x = x,
                    y = y
                  )
                )
              
              centroidDataframe <-
                rbind(
                  centroidDataframe,
                  data.frame(
                    x = as.numeric(current_cluster[["pt"]][["x"]]),
                    y = as.numeric(current_cluster[["pt"]][["y"]]),
                    lev = depth
                  )
                )
            }
          }
        }
      } else {
        for (clusterNo in 1:n_cells.hmap^(child.level - 1)) {
          for (childNo in 1:length(hvt.results[[2]][[depth]][[as.character(clusterNo)]])) {
            if (!is.null(hvt.results[[2]][[depth]][[as.character(clusterNo)]])) {
              summaryFilteredDF <-
                summaryDF %>% dplyr::filter(
                  Segment.Level == depth,
                  Segment.Parent == clusterNo,
                  Segment.Child == childNo
                )
              
              current_cluster <- hvt.results[[2]][[depth]][[as.character(clusterNo)]][[childNo]]
              
              
              val <- summaryFilteredDF[, hmap.cols]
              qe <- summaryFilteredDF[, "Quant.Error"]
              
              x <- as.numeric(current_cluster[["x"]])
              y <- as.numeric(current_cluster[["y"]])
              if ("Cell.ID" %in% colnames(summaryFilteredDF)) {
                cell_ID <- summaryFilteredDF[, "Cell.ID"]
              } else {
                cell_ID <- 0
              }
              valuesDataframe <-
                rbind(
                  valuesDataframe,
                  data.frame(
                    depth = depth,
                    cluster = clusterNo,
                    child = childNo,
                    qe = qe,
                    n_cluster = ncluster,
                    value = val,
                    cellID = cell_ID
                  )
                )
              positionsDataframe <-
                rbind(
                  positionsDataframe,
                  data.frame(
                    depth = rep(depth, length(x)),
                    cluster = rep(clusterNo, length(x)),
                    child = rep(childNo, length(x)),
                    x = x,
                    y = y
                  )
                )
              
              centroidDataframe <-
                rbind(
                  centroidDataframe,
                  data.frame(
                    x = as.numeric(current_cluster[["pt"]][["x"]]),
                    y = as.numeric(current_cluster[["pt"]][["y"]]),
                    lev = depth
                  )
                )
            }
          }
        }
      }
    }
    valuesDataframe <- valuesDataframe[2:nrow(valuesDataframe), ]
    
    positionsDataframe <-
      positionsDataframe[2:nrow(positionsDataframe), ]
    
    centroidDataframe <-
      centroidDataframe[2:nrow(centroidDataframe), ]
    
    
    datapoly <-
      merge(valuesDataframe,
            positionsDataframe,
            by = c("depth", "cluster", "child")
      )
    
    
    
    Level_list <- lapply(1:child.level, function(x) {
      temp_df <- datapoly[datapoly$depth == x, ]
      min_x <- min(datapoly$x)
      max_x <- max(datapoly$x)
      min_y <- min(datapoly$y)
      max_y <- max(datapoly$y)
      temp_df <- temp_df %>% mutate(name = paste0("Segment", cluster, "-", child))
      returnList <- split(temp_df, f = temp_df$name)
      return(list(
        returnList,
        min_x,
        max_x,
        min_y,
        max_y
      ))
    })
    
    
    
    #####################################################################################
    depth_wise_surface <- function(finalList
                                  
    ) {
      column <- hmap.cols
      min_x <- unlist(finalList[[2]])
      max_x <- unlist(finalList[[3]])
      min_y <- unlist(finalList[[4]])
      max_y <- unlist(finalList[[5]])
      
      y <- rep(1:dim_size, times = dim_size)
      x <- rep(1:dim_size, each = dim_size)
      hvtVolcanoMatrix <- rep(0, each = dim_size^2)
      for (k in names(finalList[[1]])) {
        eval(parse(
          text = paste0(
            "polygon_x <- round(scales::rescale(finalList[[1]]$`",
            k,
            "`$x, to=c(0, ", dim_size, "), from  = c(min_x,max_x)))"
          )
        ))
        eval(parse(
          text = paste0(
            "polygon_y <- round(scales::rescale(finalList[[1]]$`",
            k,
            "`$y, to=c(0, ", dim_size, "), from  = c(min_y,max_y)))"
          )
        ))
        present <- sp::point.in.polygon(x, y, polygon_x, polygon_y)
        eval(parse(
          text = paste0(
            "quantError <- finalList[[1]]$`",
            k,
            "`$value[1]"
          )
        ))
        hvtVolcanoMatrix[which(present != 0)] <- quantError
      }
      
      hvtVolcanoMatrix <- matrix(hvtVolcanoMatrix, nrow = dim_size)
      return(hvtVolcanoMatrix)
    }
    
    #####################################################################################
    
    temp_hvtVolcanoMatrix <- lapply(Level_list, depth_wise_surface)
    number_of_layers <- length(temp_hvtVolcanoMatrix)
    p <- plotly::plot_ly(showscale = FALSE)
    
    temp <- lapply(1:number_of_layers, function(i) {
      hvtVolcanoMatrix <- temp_hvtVolcanoMatrix[[i]] - ((i - 1) * separation_width)
      p <<- p %>%
        plotly::add_surface(z = ~hvtVolcanoMatrix, opacity = layer_opacity[i], name = paste("Layer_", i), showlegend = TRUE)
    })
    
    p <- p %>%
      plotly::config(displaylogo = FALSE) %>%
      plotly::layout(
        zaxis = list(title = hmap.cols),
        title = paste(
          "Hierarchical Voronoi Tessellation Interactive surface plot with",
          hmap.cols,
          "Heatmap Overlay"
        ),
        scene = list(zaxis = list(title = hmap.cols))
      )
    
    
    
    return(suppressMessages(p))
    
  } else {
    stop("Invalid value for 'plot.type'. Expected '1D','2Dproj', '2Dhvt','2Dheatmap', or 'surface_plot'.")
  }
}

context("HVT package")

# ─────────────────────────────────────────────────────────────
# Shared setup
# ─────────────────────────────────────────────────────────────

USArrests_scaled <- scale(datasets::USArrests, center = TRUE, scale = TRUE)

eustock <- data.frame(
  t    = as.numeric(time(EuStockMarkets)),
  DAX  = EuStockMarkets[, "DAX"],
  SMI  = EuStockMarkets[, "SMI"],
  CAC  = EuStockMarkets[, "CAC"],
  FTSE = EuStockMarkets[, "FTSE"]
)

# ─────────────────────────────────────────────────────────────
# trainHVT
# ─────────────────────────────────────────────────────────────

test_that("trainHVT: L1_Norm + mean", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 10, depth = 1, quant.err = 0.2,
                               distance_metric = "L1_Norm", error_metric = "mean",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  expect_equal(length(hvt.results), 7)
  expect_equal(hvt.results[[3]]$summary[, "Quant.Error"],
               c(0.3609466, 0.2791825, 0.2990230, 0.3566383, 0.2383200,
                 0.3392648, 0.2637289, 0.0000000, 0.3795258, 0.2487440), tolerance = 1e-5)
  expect_equal(hvt.results[[3]]$compression_summary[["percentOfCellsBelowQuantizationErrorThreshold"]], 0.1)
})

test_that("trainHVT: L2_Norm + mean", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 10, depth = 1, quant.err = 0.2,
                               distance_metric = "L2_Norm", error_metric = "mean",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  expect_equal(length(hvt.results), 7)
  expect_equal(hvt.results[[3]]$summary[, "Quant.Error"],
               c(0.2027530, 0.1638378, 0.1638530, 0.2100104, 0.1350629,
                 0.1868272, 0.1477676, 0.0000000, 0.2131553, 0.1453146), tolerance = 1e-5)
  expect_equal(hvt.results[[3]]$compression_summary[["percentOfCellsBelowQuantizationErrorThreshold"]], 0.7)
})

test_that("trainHVT: L1_Norm + max", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 10, depth = 1, quant.err = 0.2,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  expect_equal(length(hvt.results), 7)
  expect_equal(hvt.results[[3]]$summary[, "Quant.Error"],
               c(0.5282025, 0.3198631, 0.4280770, 0.5259606, 0.2936061,
                 0.3972785, 0.2977948, 0.0000000, 0.4775840, 0.2994434), tolerance = 1e-5)
  expect_equal(hvt.results[[3]]$compression_summary[["percentOfCellsBelowQuantizationErrorThreshold"]], 0.1)
})

test_that("trainHVT: L2_Norm + max", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 10, depth = 1, quant.err = 0.2,
                               distance_metric = "L2_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  expect_equal(length(hvt.results), 7)
  expect_equal(hvt.results[[3]]$summary[, "Quant.Error"],
               c(0.3098097, 0.1954690, 0.2424398, 0.3029314, 0.1719962,
                 0.2006439, 0.1599492, 0.0000000, 0.2916686, 0.1742819), tolerance = 1e-5)
  expect_equal(hvt.results[[3]]$compression_summary[["percentOfCellsBelowQuantizationErrorThreshold"]], 0.5)
})

test_that("trainHVT: output structure is correct", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 5, depth = 1, quant.err = 0.2,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  expect_s3_class(hvt.results, "hvt.object")
  expect_true(all(c("summary", "compression_summary") %in% names(hvt.results[[3]])))
  expect_true(all(c("Segment.Level", "Segment.Parent", "Segment.Child", "n", "Quant.Error", "Cell.ID")
                  %in% colnames(hvt.results[[3]]$summary)))
  expect_equal(hvt.results[["model_info"]][["type"]], "hvt_model")
})

test_that("trainHVT: depth = 2 produces multi-level summary", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 3, depth = 2, quant.err = 0.2,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  expect_s3_class(hvt.results, "hvt.object")
  expect_true(max(hvt.results[[3]]$summary$Segment.Level, na.rm = TRUE) >= 1)
})

test_that("trainHVT: normalize = FALSE runs without error", {
  skip_on_cran()
  set.seed(420)
  expect_no_error(
    HVT::trainHVT(USArrests_scaled, n_cells = 5, depth = 1, quant.err = 0.2,
                  distance_metric = "L1_Norm", error_metric = "max",
                  normalize = FALSE, quant_method = "kmeans", dim_reduction_method = "sammon")
  )
})


# ─────────────────────────────────────────────────────────────
# scoreHVT
# ─────────────────────────────────────────────────────────────

test_that("scoreHVT: output structure is correct", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 10, depth = 1, quant.err = 0.2,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  scored <- HVT::scoreHVT(USArrests_scaled, hvt.results)
  expect_s3_class(scored, "hvt.object")
  expect_true(all(c("scoredPredictedData", "cellID_coordinates", "predictInput") %in% names(scored)))
})

test_that("scoreHVT: Cell.IDs are valid integers within trained cell range", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 10, depth = 1, quant.err = 0.2,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  scored <- HVT::scoreHVT(USArrests_scaled, hvt.results)
  cell_ids <- scored$scoredPredictedData$Cell.ID
  expect_true(all(!is.na(cell_ids)))
  expect_true(all(cell_ids >= 1 & cell_ids <= 10))
})

test_that("scoreHVT: scored rows match input rows", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 10, depth = 1, quant.err = 0.2,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  scored <- HVT::scoreHVT(USArrests_scaled, hvt.results)
  expect_equal(nrow(scored$scoredPredictedData), nrow(USArrests_scaled))
})

test_that("scoreHVT: rejects input with NA values", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 5, depth = 1, quant.err = 0.2,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  bad_data <- USArrests_scaled
  bad_data[1, 1] <- NA
  expect_error(HVT::scoreHVT(bad_data, hvt.results))
})

test_that("scoreHVT: rejects dataset missing training columns", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 5, depth = 1, quant.err = 0.2,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  bad_data <- USArrests_scaled[, 1:2]
  expect_error(HVT::scoreHVT(bad_data, hvt.results))
})

# ─────────────────────────────────────────────────────────────
# plotHVT
# ─────────────────────────────────────────────────────────────

test_that("plotHVT: returns a ggplot object for depth 1", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 5, depth = 1, quant.err = 0.2,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  p <- HVT::plotHVT(hvt.results, child.level = 1, hmap.cols = NULL, plot.type = "2Dhvt")
  expect_s3_class(p, "ggplot")
})

test_that("plotHVT: heatmap mode returns a ggplot object", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(USArrests_scaled, n_cells = 5, depth = 1, quant.err = 0.2,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  p <- HVT::plotHVT(hvt.results, child.level = 1, hmap.cols = "Murder", plot.type = "2Dhvt")
  expect_s3_class(p, "ggplot")
})

# ─────────────────────────────────────────────────────────────
# getTransitionProbability
# ─────────────────────────────────────────────────────────────

test_that("getTransitionProbability: output has required columns", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(eustock[, -1], n_cells = 10, depth = 1, quant.err = 0.1,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  scored     <- HVT::scoreHVT(eustock, hvt.results)
  state_data <- data.frame(cell_id = scored$scoredPredictedData$Cell.ID, t = eustock$t)
  tpm        <- HVT::getTransitionProbability(state_data, cellid_column = "cell_id", time_column = "t")
  expect_true(all(c("Current_State", "Next_State", "Transition_Probability") %in% colnames(tpm)))
})

test_that("getTransitionProbability: probabilities sum to 1 per state", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(eustock[, -1], n_cells = 10, depth = 1, quant.err = 0.1,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  scored     <- HVT::scoreHVT(eustock, hvt.results)
  state_data <- data.frame(cell_id = scored$scoredPredictedData$Cell.ID, t = eustock$t)
  tpm        <- HVT::getTransitionProbability(state_data, cellid_column = "cell_id", time_column = "t")
  row_sums   <- tapply(tpm$Transition_Probability, tpm$Current_State, sum)
  expect_true(all(abs(row_sums - 1) < 0.01))
})

# ─────────────────────────────────────────────────────────────
# msm
# ─────────────────────────────────────────────────────────────

test_that("msm: ex-post forecast returns expected output structure", {
  skip_on_cran()
  set.seed(420)
  hvt.results <- HVT::trainHVT(eustock[, -1], n_cells = 10, depth = 1, quant.err = 0.1,
                               distance_metric = "L1_Norm", error_metric = "max",
                               quant_method = "kmeans", dim_reduction_method = "sammon")
  scored     <- HVT::scoreHVT(eustock, hvt.results)
  state_data <- data.frame(Cell.ID = scored$scoredPredictedData$Cell.ID, t = eustock$t)
  tpm        <- HVT::getTransitionProbability(
    data.frame(cell_id = state_data$Cell.ID, t = state_data$t),
    cellid_column = "cell_id", time_column = "t"
  )
  train_states <- head(state_data, 1500)
  test_states  <- tail(state_data, nrow(state_data) - 1500)
  result <- HVT::msm(
    state_time_data             = train_states,
    forecast_type               = "ex-post",
    initial_state               = train_states$Cell.ID[nrow(train_states)],
    transition_probability_matrix = tpm,
    num_simulations             = 50,
    trainHVT_results            = hvt.results,
    scoreHVT_results            = scored,
    actual_data                 = NULL,
    raw_dataset                 = eustock,
    time_column                 = "t",
    handle_problematic_states   = FALSE
  )
  expect_true(is.list(result))
  expect_true("plots" %in% names(result))
})

#' @keywords internal


madPlot <- function(hvt.scoring) {
  requireNamespace("ggplot2")       #deldir function
  requireNamespace("scales")       #deldir function
  
  qe = hvt.scoring[["predictInput"]][["quant.err"]]
  df_scored = hvt.scoring[["scoredPredictedData"]]
  qe_val = df_scored$Quant.Error
  
  
  if (qe < 0.75) {
    x_breaks = seq(qe, 0.75, length.out = 10) %>% round(2)
    MAD_list = lapply(seq(qe, 0.75, by = 0.01), function(x) {
      anom_points = sum(qe_val > x)
      max_cells = length(qe_val)
      perct_anom = (anom_points / max_cells)
      
      return(c(x, anom_points, max_cells, perct_anom))
    })
  } else if (qe > 0.75) {
    x_breaks = seq(qe, (qe + 0.75), length.out = 10) %>% round(2)
    MAD_list = lapply(seq(qe, (qe + 0.75), by = 0.01), function(x) {
      anom_points = sum(qe_val > x)
      max_cells = length(qe_val)
      perct_anom = (anom_points / max_cells)
      
      return(c(x, anom_points, max_cells, perct_anom))
    })
    
    
    
  }
  
  
  comp_nclust_anom = MAD_list %>% purrr::reduce(rbind) %>% as.data.frame()
  names(comp_nclust_anom) = c(
    "Mean_Absolute_Deviation",
    "anom_points",
    "max_cells",
    "Percentage_of_Anomalous_Points"
  )
  
  ############################################
  
  mark_x = which(comp_nclust_anom$Percentage_of_Anomalous_Points < 0.01)
  mark_x = comp_nclust_anom[mark_x[1], ]
  
  
  
  mad_plot =
    ggplot()  +
    geom_point(
      data = comp_nclust_anom,
      mapping = aes(x = Mean_Absolute_Deviation,
                    y = Percentage_of_Anomalous_Points),
      shape = "circle",
      size = 1.5,
      colour = "midnightblue"
    ) +
    geom_hline(yintercept = 0.01,
               linetype = "dashed",
               colour = "brown2") +
    geom_vline(
      xintercept = mark_x$Mean_Absolute_Deviation,
      linetype = "dashed",
      colour = "brown2"
    ) +
  
    
    annotate(
      "text",
      x = mark_x$Mean_Absolute_Deviation + (0.03 * max(
        comp_nclust_anom$Mean_Absolute_Deviation
      )),
      y = mark_x$Percentage_of_Anomalous_Points + (0.1),
      label = mark_x$Mean_Absolute_Deviation,
      fontface = "bold"
    ) +
    annotate(
      "text",
      x = qe,
      y = 0.01 + (0.1),
      label = "1%",
      fontface = "bold"
    ) +
    
    ggplot2::theme_bw() +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 12),
      axis.text = element_text(size = 8),
      axis.title = element_text(size = 8)
    ) +
    labs(title = "Mean Absolute Deviation Plot") +
    scale_y_continuous(labels = scales::percent_format(), limits = c(0, 1)) +
    scale_x_continuous(breaks =x_breaks)
  
  return(mad_plot)
}

#' @name edaPlots
#' @title plots for data analysis
#' @description This is the main function that provides exploratory data analysis plots 
#' @param df Dataframe. A data frame object.
#' @param time_column Character. The name of the time column in the data frame.
#' Can be given only when the data is time series
#' @param output_type Character. The name of the output to be displayed. Options are 'summary',
#' 'histogram', 'boxplot', 'timeseries' & 'correlation'. Default value is summary.
#' @param n_cols Numeric. A value to indicate how many columns to be included in the output.
#' @param grey_bars List. A list of timestamps where each list contains two elements: start and end period,
#' which will be highlighted in gray in the time series plot. Default value is NULL.
#' @return Five objects which include time series plots, data distribution plots, 
#' box plots, correlation plot and a descriptive statistics table.
#' @author Vishwavani <vishwavani@@mu-sigma.com>
#' @keywords EDA
#' @examples
#' dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
#' DAX = as.numeric(EuStockMarkets[, "DAX"]),
#' SMI = as.numeric(EuStockMarkets[, "SMI"]),
#' CAC = as.numeric(EuStockMarkets[, "CAC"]),
#' FTSE = as.numeric(EuStockMarkets[, "FTSE"]))
#' edaPlots(dataset)
#' edaPlots(dataset, time_column = 'date', output_type = 'timeseries', n_cols = 4)
#' @export edaPlots


edaPlots <- function(df, time_column, output_type = "summary", n_cols = -1,grey_bars = NULL) {
    start <- end <-  NULL
  
  if(n_cols == -1){
    n_cols = ncol(df)
  }
  
  a <- ncol(df)
  if(n_cols == 0 || n_cols > a){
    stop(paste0("n_cols argument should be from 1 to ", a))  }
  
  df <- df[,1:n_cols]
  
  # Summary EDA Function
  summary_eda <- function(df) {
    variable  <-  `1st Quartile`<-median<- sd<-`3rd Quartile`<-  hist<- n_row<- n_missing<- NULL
    type <-complete_rate <-p0 <-p25 <-p50 <-p75<-p100<-NULL
    numeric_columns <- sapply(df, is.numeric)
    missing_counts <- sapply(seq_along(numeric_columns)[numeric_columns], function(i) sum(is.na(df[[i]])))
    missing_counts_vector <- unlist(missing_counts)
    
    result <- df %>%
      select(where(is.numeric)) %>%
      skimr::skim() %>%
      rename_with(~sub("^(skim_|numeric\\.)", "", .)) %>%
      select(-type, -n_missing, -complete_rate) %>%
      mutate(n_row = format(nrow(df), scientific = FALSE), n_missing = missing_counts_vector) %>%
      dplyr::rename(min = p0, `1st Quartile` = p25, median = p50, `3rd Quartile` = p75, max = p100) %>%
      dplyr::select(variable, min, `1st Quartile`, median, mean, sd, `3rd Quartile`, max, hist,n_row, n_missing) 
    
    
    calculate_dynamic_length_menu <- function(total_entries, base_step = 100) {
      max_option <- ceiling(total_entries / base_step) * base_step
      max_option <- max(max_option, 100)
      options <- seq(base_step, by = base_step, length.out = max_option / base_step)
      options <- c(10, options)
      return(options)
    }
    total_entries <- nrow(result)
    length_menu_options <- calculate_dynamic_length_menu(total_entries)
    eda_format <- DT::datatable(result, 
                                options = list(
                                  pageLength = 10,
                                  lengthMenu = length_menu_options,
                                  scrollX = TRUE,
                                  scrollY = TRUE
                                ),
                                rownames = FALSE
    ) %>%
      DT::formatRound(
        columns = c('min', 'max', 'mean', 'sd', 'median', '1st Quartile', '3rd Quartile'),
        digits = 2
      )
    
    
    return(eda_format)
  }
  
  # Histogram Function
  generateDistributionPlot <- function(data, column) {
    name  <-  NULL
    
    p1<- ggplot2::ggplot(data, ggplot2::aes(x = data[[column]])) +
      ggplot2::geom_histogram(ggplot2::aes(y = ..count..), fill = "midnightblue", size = 0.2, alpha=0.7) +
      ggplot2::xlab(column) + ggplot2::ylab("Count") +
      #ggplot2::labs(column) +
      ggplot2::theme_bw() + 
      ggplot2::theme(panel.border=ggplot2::element_rect(size=0.1),legend.position = c(0.8, 0.8))
    
    p2<-ggplot2::ggplot(data, ggplot2::aes(x = data[[column]])) +
      ggplot2::stat_function(ggplot2::aes(color = "Normal"), 
                             fun = stats::dnorm,args = list(mean = mean(data[[column]]),sd = stats::sd(data[[column]]))) +
      ggplot2::stat_density(ggplot2::aes(color = "Density"), geom = "line", linetype = "dashed")  +
      ggplot2::scale_colour_manual("", values = c("black", "orange")) +
      ggplot2::scale_linetype_manual("", values = c("Normal" = 2, "Density" = 1))  +
      ggplot2::guides(
        fill = ggplot2::guide_legend(keywidth = 1, keyheight = 1),
        linetype = ggplot2::guide_legend(keywidth = 3, keyheight = 1),
        colour = ggplot2::guide_legend(keywidth = 3, keyheight = 1)) + 
      ggplot2::ylab("Density") +
      ggplot2::xlab("Density") +
      ggplot2::theme(
        panel.background = ggplot2::element_blank(),
        panel.grid.minor = ggplot2::element_blank(),
        panel.grid.major = ggplot2::element_blank(),
        axis.text.y = ggplot2::element_text(colour="black", size=8),
        axis.text.x = ggplot2::element_text(size = 14),
        axis.ticks = ggplot2::element_line(colour = 'gray50'),
        axis.ticks.x = ggplot2::element_line(colour = "black"))
    
    g1 <- ggplot2::ggplotGrob(p1)
    g2 <- ggplot2::ggplotGrob(p2)
    
    pp <- c(subset(g1$layout, name == "panel"))
    g <- gtable::gtable_add_grob(g1, g2$grobs[[which(g2$layout$name == "panel")]], pp$t, 
                                 pp$l, pp$b, pp$l)
    
    # extract the y-axis of p2
    ia <- which(g2$layout$name == "axis-l")
    ga <- g2$grobs[[ia]]
    ax <- ga$children[[2]]
    
    # flip it horizontally
    ax$widths <- rev(ax$widths)
    ax$grobs <- rev(ax$grobs)
    
    # add the flipped y-axis to the right
    g <- gtable::gtable_add_cols(g, g2$widths[g2$layout[ia, ]$l], length(g$widths) - 1)
    g <- gtable::gtable_add_grob(g, ax, pp$t, length(g$widths) - 1, pp$b)
    distHistogram <- g
    return(distHistogram)
  }
  
  # Box Plot Function
  quantile_outlier_plots_fn <- function(data, outlier_check_var) {
    
    data <- data[sapply(data, is.numeric)]
    
    # Calculate thresholds for outlier detection
    lower_threshold <- stats::quantile(data[[outlier_check_var]], .25, na.rm = TRUE) - 
      1.5 * stats::IQR(data[[outlier_check_var]], na.rm = TRUE)
    upper_threshold <- stats::quantile(data[[outlier_check_var]], .75, na.rm = TRUE) + 
      1.5 * stats::IQR(data[[outlier_check_var]], na.rm = TRUE)
    
    # Identify outliers
    data$QuantileOutlier <- data[[outlier_check_var]] > upper_threshold | 
      data[[outlier_check_var]] < lower_threshold
    
    # Create boxplot
    quantile_outlier_plot <- ggplot2::ggplot(data, ggplot2::aes(x = "", y = data[[outlier_check_var]])) +
      ggplot2::geom_boxplot(fill = 'midnightblue', alpha = 0.7) + 
      ggplot2::theme_bw() + 
      ggplot2::theme(
        panel.border = ggplot2::element_rect(size = 0.1),
        panel.grid.minor.x = ggplot2::element_blank(),
        panel.grid.major.x = ggplot2::element_blank(),
        legend.position = "bottom"
      ) +
      ggplot2::ylab(outlier_check_var) + 
      ggplot2::xlab("")
    
    # Prepare output data
    data <- cbind(
      data[, !names(data) %in% c("QuantileOutlier")] %>% round(2),
      outlier = data[["QuantileOutlier"]]
    )
    
    # Return the results
    return(list(quantile_outlier_plot, data, lower_threshold, upper_threshold))
  }
  
  
  # Correlation Plot Function
  correlation_plot <- function(df, title = "Feature Correlation Analysis") {
    df <- df[sapply(df, is.numeric)]
    
    corrplot::corrplot(stats::cor(df, use = "complete.obs"), main = list(title , font = 1, cex = 1),method = "color", 
                       outline = TRUE, addgrid.col = "darkgray", addrect = 4,
                       rect.col = "black", rect.lwd = 5, cl.pos = "b", tl.col = "black", tl.cex = 0.7, 
                       cl.cex = 0.8,addCoef.col = "black", number.digits = 2, number.cex = 0.7, 
                       type = "lower",col = grDevices::colorRampPalette(c("maroon", "white", "#1D9CDB"))(200), mar=c(0,0,2,0))  
  }
  
  
  #####################timeseries plots
  if (output_type == "timeseries" && (!is.null(time_column)) ){
    
    df <- df[order(df[[time_column]]), ]
    
    generateTimeseriesPlot <- function(df, time_column, recession_periods = NULL) {
      if (!is.data.frame(df)) {
        stop("Input dataset must be a data frame")
      }
      
      if (!(time_column %in% names(df))) {
        stop("Specified time column does not exist in the dataset")
      }
      
      if (!inherits(df[[time_column]], "POSIXct") && !is.numeric(df[[time_column]])) {
        stop("Time column must be of class POSIXct or numeric.")
      }
      
      ordered_dataset <- df[order(df[[time_column]]), ]
      variable_names <- setdiff(names(df), time_column)
      
      if (!is.null(recession_periods)) {
        if (inherits(df[[time_column]], "POSIXct")) {
          # Convert recession periods to POSIXct for datetime time columns
          recession_df <- do.call(rbind, lapply(recession_periods, function(x) {
            if (length(x) == 2) {
              data.frame(start = as.POSIXct(x[1]), end = as.POSIXct(x[2]))
            } else {
              stop("Each recession period must have exactly two dates: start and end.")
            }
          }))
        } else if (is.numeric(df[[time_column]])) {
          # Ensure recession periods are numeric for numeric time columns
          recession_df <- do.call(rbind, lapply(recession_periods, function(x) {
            if (length(x) == 2) {
              data.frame(start = as.numeric(x[1]), end = as.numeric(x[2]))
            } else {
              stop("Each recession period must have exactly two numeric values: start and end.")
            }
          }))
        } else {
          stop("Recession periods must match the type of the time column.")
        }
      } else {
        recession_df <- NULL
      }
      
      # Generate plots
      plot_list <- lapply(variable_names, function(variable_name) {
        p <- ggplot(data = ordered_dataset, aes_string(x = time_column, y = variable_name)) +
          geom_line(color = "midnightblue") +
          labs(x = time_column, y = variable_name) +
          theme_minimal()
        
        # Add grey bars for recession periods (if applicable)
        if (!is.null(recession_df)) {
          p <- p +
            geom_rect(
              data = recession_df,
              aes(xmin = start, xmax = end, ymin = -Inf, ymax = Inf),
              inherit.aes = FALSE,
              fill = "grey",
              alpha = 0.5
            )
        }
        
        return(p)
      })
      
      gridExtra::grid.arrange(grobs = plot_list, ncol = 2, top = "Time series plots")
    }
    
    
  }
  if (output_type == "summary")  {
    eda_table <- summary_eda(df)
    return( eda_table)
  }
  
  if (output_type == "timeseries") {
    time_series_plot <- generateTimeseriesPlot(df, time_column, grey_bars)
    (time_series_plot)
  }
  
  if (output_type == "histogram") {
    eda_cols <- names(df)[sapply(df, is.numeric)]
    dist_list <- lapply(eda_cols, function(column) {
      histo <- generateDistributionPlot(df, column)
      plot(histo)
    }) 
  }
  
  if (output_type == "boxplot") {
    eda_cols <- names(df)[sapply(df, is.numeric)]
    box_plots <- lapply(eda_cols, function(column) {
      boxxo <- quantile_outlier_plots_fn(df, outlier_check_var = column)[[1]]
      plot(boxxo)
    })
  }
  
  if (output_type == "correlation") {
    correlation_plot(df,title = "Features Correlation Visualization")
    
  }
  
  
}

#' @name summary
#' @title Table for displaying summary
#' @description This is the main function for displaying summary from model training and scoring
#' @param data List. A listed object from trainHVT or scoreHVT
#' @param limit Numeric. A value to indicate how many rows to display.
#' @param scroll Logical. A value to indicate whether to display scroll bar or not. Default value is TRUE.
#' @return A consolidated table of summary for training, scoring and forecasting
#' @author Vishwavani <vishwavani@@mu-sigma.com>, Alimpan Dey <alimpan.dey@@mu-sigma.com>
#' @importFrom dplyr mutate 
#' @keywords Table_Formatting
#' @examples
#' data <- datasets::EuStockMarkets
#' dataset <- as.data.frame(data)
#' #model training
#' hvt.results <- trainHVT(dataset, n_cells = 60, depth = 1, quant.err = 0.1,
#'                       distance_metric = "L1_Norm", error_metric = "max",
#'                       normalize = TRUE, quant_method = "kmeans", dim_reduction_method = 'sammon')
#' summary(data =  hvt.results)
#' @export summary

summary <- function(data, limit = 20, scroll = TRUE) {
  
  
  Row_Number <- NULL
  
  if (inherits(data, "hvt.object")) {
    
    # Function to calculate scroll height based on the number of rows
    scrolLimit <- function(noOfRows) {
      if (noOfRows < 10) {
        swe <- paste(as.character(noOfRows * 50), "px")
      } else {
        swe <- "400px"
      }
      return(swe)
    }
    
    # Limit the number of rows displayed for each table
    apply_limit <- function(table, limit) {
      if (is.null(limit)) {
        limit <- if (nrow(table) > 20) 20 else nrow(table)
      }
      head(table, limit)
    }
    
    if (length(data) >= 3 && "compression_summary" %in% names(data[[3]])) {
      compression_summary_table <- data[[3]]$compression_summary %>% as.data.frame()
      compression_summary_table <- apply_limit(compression_summary_table, limit)
      
      table_1 <- compression_summary_table %>%
        dplyr::mutate(percentOfCellsBelowQuantizationErrorThreshold = round(compression_summary_table$percentOfCellsBelowQuantizationErrorThreshold, 2)) %>%
        dplyr::mutate(percentOfCellsBelowQuantizationErrorThreshold = dplyr::case_when(
          percentOfCellsBelowQuantizationErrorThreshold >= 0.8 ~ 
            kableExtra::cell_spec(percentOfCellsBelowQuantizationErrorThreshold, "html", color = "green"),
          TRUE ~ 
            kableExtra::cell_spec(percentOfCellsBelowQuantizationErrorThreshold, "html", color = "black")
        ))
      
      table_1 <- knitr::kable(table_1, "html", escape = FALSE, align = "c") %>%
        kableExtra::kable_styling(bootstrap_options = c("striped", "hover", "responsive"), font_size = 10.5) 

      return(table_1)
      
    } else if ("scoredPredictedData" %in% names(data)) {
      scorehvt_table <- data$scoredPredictedData %>% as.data.frame()
      scorehvt_table <- apply_limit(scorehvt_table, limit)
      
      value <- data[["model_info"]][["scored_model_summary"]][["mad.threshold"]]
      
      table_2 <- scorehvt_table %>%
        dplyr::mutate(
          Row_Number = dplyr::row_number(),
          Quant.Error = dplyr::case_when(
            Quant.Error > value ~ kableExtra::cell_spec(Quant.Error, "html", color = "red"),
            TRUE ~ kableExtra::cell_spec(Quant.Error, "html", color = "black")
          )
        ) %>%
        dplyr::select(Row_Number, everything())
      
      table_2 <- knitr::kable(table_2, "html", escape = FALSE, align = "c") %>%
        kableExtra::kable_styling(bootstrap_options = c("striped", "hover", "responsive")) %>%
        kableExtra::scroll_box(width = "100%", height = scrolLimit(nrow(scorehvt_table)))
      
      return(table_2)
      
    } else if ("actual_predictedTable" %in% names(data)) {
      scoreLayeredhvt_table <- data$actual_predictedTable %>% as.data.frame()
      scoreLayeredhvt_table <- apply_limit(scoreLayeredhvt_table, limit)
      
      scoreLayeredhvt_table <- scoreLayeredhvt_table %>%
        dplyr::mutate(Row_Number = dplyr::row_number()) %>%
        dplyr::select(Row_Number, everything())
      
      table_3 <- knitr::kable(scoreLayeredhvt_table, "html", escape = FALSE, align = "c") %>%
        kableExtra::kable_styling(bootstrap_options = c("striped", "hover", "responsive")) %>%
        kableExtra::scroll_box(width = "100%", height = scrolLimit(nrow(scoreLayeredhvt_table)))
      return(table_3)
    } else if(!is.null(data$problematic_states_list)){
      problematic_states_table <- data$problematic_states_list %>% as.data.frame()

      table_4 <- knitr::kable(problematic_states_table, "html", escape = FALSE, align = "c") %>%
        kableExtra::kable_styling(bootstrap_options = c("striped", "hover", "responsive")) 
       # kableExtra::scroll_box(width = "100%", height = scrolLimit(nrow(problematic_states_table)))
      return(table_4)
    } 
    
  } else {
    base::summary(data)
  }
}

#' @name plotExAnteRawSeries
#' @title Ex-ante raw series forecasting
#' @description Transforms ex-ante forecasts generated on a log-difference-12 scale back to the original raw data scale
#' @param ex_ante_results List. Output from `msm()` function with `forecast_type = "ex-ante"`
#' @param original_dataset Dataframe. The dataset imported with all features in its original scale of measures 
#' without normalization or log difference
#' @param transformed_dataset Dataframe. The dataset that is transformed for analysis including scaling and log difference
#' @param time_column Character. Name of the time column in the dataset.
#' @param mae_metric Character. Metric to highlight in plots ("mean", "median", or "mode"). Default is "median".
#' @return A list containing:
#' \item{reverted_forecasts}{List of data frames, one per variable, with columns: time, mean, median, mode (reverted values)}
#' \item{plots}{List of plotly objects, one per variable, showing historical data and reverted forecasts}
#' @author Vishwavani <vishwavani@@mu-sigma.com>
#' @keywords Timeseries_Analysis
#' @importFrom magrittr %>%
#' @importFrom stats na.omit
#' @importFrom dplyr select mutate across everything between case_when
#' @importFrom tidyr pivot_longer
#' @importFrom ggplot2 ggplot geom_line geom_point geom_vline scale_colour_manual scale_x_datetime 
#' @importFrom ggplot2 scale_x_continuous theme_minimal labs expansion element_text unit
#' @examples 
#' \dontrun{
#' # After running msm() for ex-ante forecasting
#' # Note: trainHVT_results and scoreHVT_results are needed for msm(), 
#' # but NOT for plotExAnteRawSeries() which uses the forecast values directly
#' ex_ante <- msm(state_time_data = temporal_data,
#'                forecast_type = "ex-ante",
#'                transition_probability_matrix = prob_trans_matx,
#'                initial_state = tail(temporal_data$Cell.ID, 1),
#'                n_ahead_ante = ex_ante_period,
#'                num_simulations = 500,
#'                scoreHVT_results = scoring,
#'                trainHVT_results = hvt.results,
#'                raw_dataset = entire_dataset,
#'                time_column = "t")
#' 
#' # Revert to raw scale - only needs ex_ante_results and original_dataset
#' raw_forecasts <- plotExAnteRawSeries(
#'   ex_ante_results = ex_ante,
#'   original_dataset = entire_dataset_original,  # Pre-transformation raw data
#'   transformed_dataset = entire_dataset,        # Post-transformation data 
#'   time_column = "t",
#'   mae_metric = "median"
#' )
#' 
#' # Access reverted forecasts
#' raw_forecasts$reverted_forecasts$CPI_Food
#' 
#' # Access plots
#' raw_forecasts$plots$CPI_Food
#' }
#' @export plotExAnteRawSeries

plotExAnteRawSeries <- function(ex_ante_results,
                            original_dataset,
                            transformed_dataset = NULL,
                            time_column,
                            mae_metric = "median") {
  
  # Global variables for CRAN warnings
  time <- value <- mean <- median <- mode <- NULL
  
  # ============================================================================
  # VALIDATION
  # ============================================================================
  
  # Check ex_ante_results structure
  if (is.null(ex_ante_results) || !is.list(ex_ante_results)) {
    stop("ERROR: ex_ante_results must be a list output from msm() function")
  }
  
  if (is.null(ex_ante_results$plots) || is.null(ex_ante_results$plots$variable_dfs)) {
    stop("ERROR: ex_ante_results must contain plots$variable_dfs from ex-ante forecasting")
  }
  
  # Check original_dataset
  if (is.null(original_dataset) || !is.data.frame(original_dataset)) {
    stop("ERROR: original_dataset must be a data frame")
  }
  
  if (!time_column %in% colnames(original_dataset)) {
    stop(paste0("ERROR: time_column '", time_column, "' not found in original_dataset"))
  }
  
  # Check mae_metric
  if (!mae_metric %in% c("mean", "median", "mode")) {
    stop("ERROR: mae_metric must be 'mean', 'median', or 'mode'")
  }
  
  # Check if we have enough original data (need at least 12 periods)
  if (nrow(original_dataset) < 12) {
    stop("ERROR: original_dataset must have at least 12 periods for log-diff-12 reversion")
  }
  
  # ============================================================================
  # EXTRACT FORECAST DATA
  # ============================================================================
  
  # Get variable forecasts from ex_ante_results
  # variable_dfs already contains transformed forecast values (mean, median, mode) for each variable
  variable_dfs <- ex_ante_results$plots$variable_dfs
  
  if (is.null(variable_dfs) || length(variable_dfs) == 0) {
    stop("ERROR: No variable forecasts found in ex_ante_results$plots$variable_dfs")
  }
  
  # Get variable names
  variable_names <- names(variable_dfs)
  
  # Check that all variables exist in original_dataset
  missing_vars <- setdiff(variable_names, colnames(original_dataset))
  if (length(missing_vars) > 0) {
    stop(paste0("ERROR: Variables not found in original_dataset: ", paste(missing_vars, collapse = ", ")))
  }
  
  # Validate that variable_dfs contains required columns (time, mean, median, mode)
  for (var_name in variable_names) {
    if (!all(c("time", "mean", "median", "mode") %in% colnames(variable_dfs[[var_name]]))) {
      stop(paste0("ERROR: variable_dfs[[", var_name, "]] must contain columns: time, mean, median, mode"))
    }
  }
  
  # ============================================================================
  # HELPER FUNCTIONS (from mcmc_plots.R)
  # ============================================================================
  
  # Standardize timestamp function
  standardize_timestamp <- function(data, time_col) {
    sample_time <- na.omit(data[[time_col]])[1]
    if(inherits(sample_time, "POSIXct")) return(data)
    
    formats <- c("%m/%d/%Y", "%m-%d-%Y", "%Y/%m/%d", "%Y-%m-%d", 
                 "%Y-%m-%d %H:%M:%S", "%d/%m/%Y")
    
    for(fmt in formats) {
      tryCatch({
        data[[time_col]] <- as.POSIXct(data[[time_col]], format = fmt)
        if(!all(is.na(data[[time_col]]))) break
      }, error = function(e) {})
    }
    return(data)
  }
  
  # Analyze timestamp characteristics
  analyze_timestamps <- function(data, time_col) {
    data <- standardize_timestamp(data, time_col)
    sorted_times <- sort(data[[time_col]])
    
    changes <- list(
      year = length(unique(format(sorted_times, "%Y"))),
      month = length(unique(format(sorted_times, "%Y-%m"))),
      day = length(unique(format(sorted_times, "%Y-%m-%d"))),
      hour = length(unique(format(sorted_times, "%Y-%m-%d %H"))),
      minute = length(unique(format(sorted_times, "%Y-%m-%d %H:%M"))),
      second = length(unique(format(sorted_times, "%Y-%m-%d %H:%M:%S")))
    )
    
    date_format <- if(changes$minute == changes$second) {
      if(changes$hour == changes$minute) {
        if(changes$day == changes$hour) {
          if(changes$month == changes$day) {
            if(changes$year == changes$month) "%Y" else "%Y-%m"
          } else "%Y-%m-%d"
        } else "%Y-%m-%d %H"
      } else "%Y-%m-%d %H:%M"
    } else "%Y-%m-%d %H:%M:%S"
    
    time_diffs <- diff(sorted_times)
    median_diff <- median(time_diffs)
    days_diff <- as.numeric(median_diff, units="days")
    
    interval_type <- dplyr::case_when(
      dplyr::between(days_diff, 57, 62) ~ "bimonthly",
      dplyr::between(days_diff, 27, 31) ~ "monthly",
      dplyr::between(days_diff, 88, 92) ~ "quarterly",
      dplyr::between(days_diff, 364, 366) ~ "yearly",
      dplyr::between(days_diff, 6, 8) ~ "weekly",
      dplyr::between(days_diff, 13, 15) ~ "biweekly",
      days_diff < 1 ~ "intraday",
      days_diff == 1 ~ "daily",
      TRUE ~ "irregular"
    )
    
    total_span <- as.numeric(difftime(max(sorted_times), min(sorted_times), units = "days"))
    
    return(list(
      interval_type = interval_type,
      date_format = date_format,
      total_span = total_span,
      standardized_data = data
    ))
  }
  
  # Get axis breaks
  get_axis_breaks <- function(time_analysis) {
    break_interval <- switch(time_analysis$interval_type,
                             "intraday" = if(grepl("%H:%M:%S", time_analysis$date_format)) "1 minute" 
                             else if(grepl("%H:%M", time_analysis$date_format)) "15 minutes" 
                             else "1 hour",
                             "daily" = "1 day", "weekly" = "1 week", "biweekly" = "2 weeks",
                             "monthly" = "1 month", "quarterly" = "3 months", "yearly" = "1 year",
                             "bimonthly" = "2 months",
                             {
                               months_span <- time_analysis$total_span / 30.44
                               if(months_span <= 1) "1 week"
                               else if(months_span <= 3) "2 weeks"
                               else if(months_span <= 6) "1 month"
                               else if(months_span <= 24) "1 month"
                               else if(months_span <= 60) "2 months"
                               else "1 year"
                             })
    
    return(list(break_interval = break_interval, date_format = time_analysis$date_format))
  }
  
  # ============================================================================
  # REVERSION ALGORITHM
  # ============================================================================
  
  # Standardize original dataset time column
  original_dataset <- standardize_timestamp(original_dataset, time_column)
  
  # Get last 12 original values for each variable
  n_orig <- nrow(original_dataset)
  last_12_indices <- (n_orig - 11):n_orig
  
  # Initialize reverted forecasts list
  reverted_forecasts <- list()
  
  # Process each variable
  for (var_name in variable_names) {
    
    # Get forecast data from variable_dfs (already contains transformed values)
    forecast_data <- variable_dfs[[var_name]]
    forecast_times <- forecast_data$time
    
    # Get transformed forecast values directly from variable_dfs (log-diff-12 space)
    forecast_mean <- as.numeric(forecast_data$mean)
    forecast_median <- as.numeric(forecast_data$median)
    forecast_mode <- as.numeric(forecast_data$mode)
    
    n_forecast <- length(forecast_mean)
    
    # Check for missing values
    if (any(is.na(forecast_mean)) || any(is.na(forecast_median)) || any(is.na(forecast_mode))) {
      warning(paste("Some forecasted values are NA for variable:", var_name))
    }
    
    # Get last 12 original values for this variable
    original_values <- original_dataset[[var_name]][last_12_indices]
    
    # Check for non-positive values
    if (any(original_values <= 0, na.rm = TRUE)) {
      stop(paste0("ERROR: Variable '", var_name, "' has non-positive values in last 12 periods. Log-diff-12 requires positive values."))
    }
    
    # Initialize reverted vectors
    reverted_mean <- numeric(n_forecast)
    reverted_median <- numeric(n_forecast)
    reverted_mode <- numeric(n_forecast)
    
    # Apply reversion formula
    for (i in 1:n_forecast) {
      if (i <= 12) {
        # For first 12 periods: use original value from 12 periods before
        base_index <- n_orig - 12 + i
        base_value <- round(original_dataset[[var_name]][base_index], 2)  # Round raw value to 2 decimals
        
        # Reversion: X_reverted = X_original × exp(Z_forecast)
        # Round exponential values to 4 decimals, then round final result to 2 decimals
        exp_mean <- round(exp(forecast_mean[i]), 4)
        exp_median <- round(exp(forecast_median[i]), 4)
        exp_mode <- round(exp(forecast_mode[i]), 4)
        
        reverted_mean[i] <- round(base_value * exp_mean, 2)
        reverted_median[i] <- round(base_value * exp_median, 2)
        reverted_mode[i] <- round(base_value * exp_mode, 2)
      } else {
        # For periods > 12: use previously reverted value from 12 periods back
        # Round exponential values to 4 decimals, then round final result to 2 decimals
        exp_mean <- round(exp(forecast_mean[i]), 4)
        exp_median <- round(exp(forecast_median[i]), 4)
        exp_mode <- round(exp(forecast_mode[i]), 4)
        
        reverted_mean[i] <- round(reverted_mean[i - 12] * exp_mean, 2)
        reverted_median[i] <- round(reverted_median[i - 12] * exp_median, 2)
        reverted_mode[i] <- round(reverted_mode[i - 12] * exp_mode, 2)
      }
    }
    
    # Create reverted forecast data frame
    reverted_forecasts[[var_name]] <- data.frame(
      time = forecast_times,
      mean = reverted_mean,    # Already rounded to 2 decimals
      median = reverted_median, # Already rounded to 2 decimals
      mode = reverted_mode,     # Already rounded to 2 decimals
      stringsAsFactors = FALSE
    )
  }
  
  # ============================================================================
  # PLOT GENERATION
  # ============================================================================
  
  # Define theme_plot
  theme_plot <- ggplot2::theme(
    plot.title = ggplot2::element_text(size = 16, face = "bold"),
    plot.subtitle = ggplot2::element_text(size = 12),
    axis.title = ggplot2::element_text(size = 12),
    axis.text = ggplot2::element_text(size = 10),
    legend.title = ggplot2::element_text(size = 14),
    legend.text = ggplot2::element_text(size = 12),
    legend.position = "right",
    legend.key.width = grid::unit(0.5, "cm")
  )
  
  # Initialize plots list
  plots <- list()
  
  # Generate plot for each variable
  for (var_name in variable_names) {
    
    variable_data <- reverted_forecasts[[var_name]]
    
    if (is.null(variable_data) || nrow(variable_data) == 0) {
      warning(paste("Skipping plot for variable:", var_name, "- no data"))
      next
    }
    
    # Standardize timestamps
    time_analysis <- analyze_timestamps(variable_data, "time")
    variable_data <- time_analysis$standardized_data
    
    # Get historical data (last 12 points)
    historical_data <- NULL
    if (var_name %in% colnames(original_dataset) && n_orig >= 12) {
      time_points_last12 <- original_dataset[[time_column]][last_12_indices]
      last_values <- original_dataset[[var_name]][last_12_indices]
      
      if (length(time_points_last12) == length(last_values) && length(last_values) > 0) {
        historical_data <- data.frame(
          time = time_points_last12,
          value = last_values,
          stringsAsFactors = FALSE
        )
        # Standardize historical timestamps
        historical_data <- standardize_timestamp(historical_data, "time")
      }
    }
    
    # Get the 13th point (first point after initial forecast)
    dyn_forecast_time <- NULL
    if (nrow(variable_data) >= 13) {
      dyn_forecast_time <- variable_data$time[13]
    }
    
    # Create base plot
    p1 <- ggplot2::ggplot()
    
    # Add vertical lines
    vline_data <- data.frame(
      time = as.numeric(variable_data$time[1]), 
      label = "Initial Point Forecast"
    )
    
    if (!is.null(dyn_forecast_time)) {
      vline_data_dyn <- data.frame(
        time = as.numeric(dyn_forecast_time),
        label = "Initial Point Dynamic Forecast"
      )
    }
    
    if (!is.null(historical_data)) {
      p1 <- p1 + 
        ggplot2::geom_vline(data = vline_data, 
                           ggplot2::aes(xintercept = time, color = label), 
                           linetype = "dashed", size = 0.5, show.legend = TRUE)
    }
    
    if (!is.null(dyn_forecast_time)) {
      p1 <- p1 + 
        ggplot2::geom_vline(data = vline_data_dyn, 
                           ggplot2::aes(xintercept = time, color = label), 
                           linetype = "dashed", size = 0.5, show.legend = TRUE)
    }
    
    # Add historical data
    if (!is.null(historical_data)) {
      p1 <- p1 +
        ggplot2::geom_line(data = historical_data, 
                          ggplot2::aes(x = time, y = value, color = "Historical"), 
                          size = 1.0) +
        ggplot2::geom_point(data = historical_data, 
                           ggplot2::aes(x = time, y = value, color = "Historical",
                                       text = paste("Time:", time, "<br>Historical", var_name, ":", round(value, 2))), 
                           size = 1.5)
    }
    
    # Add forecast lines and points
    p1 <- p1 +
      ggplot2::geom_line(data = variable_data, 
                        ggplot2::aes(x = time, y = mode, color = "Mode"), 
                        size = ifelse(mae_metric == "mode", 1.0, 0.4)) +
      ggplot2::geom_point(data = variable_data, 
                          ggplot2::aes(x = time, y = mode, color = "Mode", 
                                      text = paste("Time:", time, "<br>Mode", var_name, ":", round(mode, 2))), 
                          size = ifelse(mae_metric == "mode", 1.5, 1)) +
      
      ggplot2::geom_line(data = variable_data, 
                        ggplot2::aes(x = time, y = mean, color = "Mean"), 
                        size = ifelse(mae_metric == "mean", 1.0, 0.4)) +
      ggplot2::geom_point(data = variable_data, 
                         ggplot2::aes(x = time, y = mean, color = "Mean", 
                                     text = paste("Time:", time, "<br>Mean", var_name, ":", round(mean, 2))), 
                         size = ifelse(mae_metric == "mean", 1.5, 1)) +
      
      ggplot2::geom_line(data = variable_data, 
                        ggplot2::aes(x = time, y = median, color = "Median"), 
                        size = ifelse(mae_metric == "median", 1.0, 0.5)) +
      ggplot2::geom_point(data = variable_data, 
                         ggplot2::aes(x = time, y = median, color = "Median", 
                                     text = paste("Time:", time, "<br>Median", var_name, ":", round(median, 2))), 
                         size = ifelse(mae_metric == "median", 1.5, 1)) +
      
      ggplot2::scale_colour_manual(values = c("Median" = "red", 
                                              "Mean" = "darkgreen", 
                                              "Mode" = "#0901FF",
                                              "Historical" = "black",
                                              "Initial Point Forecast" = "gray",
                                              "Initial Point Dynamic Forecast" = "purple"
      ))
    
    # Combine all timestamps for axis breaks
    all_times <- variable_data$time
    if (!is.null(historical_data)) {
      all_times <- c(historical_data$time, all_times)
    }
    
    # Add appropriate axis scaling
    if (inherits(variable_data$time, "POSIXct")) {
      p1 <- p1 + 
        ggplot2::scale_x_datetime(
          breaks = all_times,
          date_labels = "%Y-%m",
          expand = ggplot2::expansion(mult = c(0.02, 0.02))
        ) +
        ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
    } else if (is.numeric(variable_data$time)) {
      p1 <- p1 + 
        ggplot2::scale_x_continuous(
          breaks = all_times,
          expand = ggplot2::expansion(mult = c(0.02, 0.02))
        ) +
        ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
    }
    
    # Add labels and theme
    p1 <- p1 +
      ggplot2::labs(
        x = "Timestamps",
        y = paste(var_name, "(raw units)"),
        title = paste("MSM: Ex-ante Predicted", var_name),
        color = " "
      ) +
      ggplot2::theme_minimal() +
      theme_plot
    
    # Convert to plotly
    p1_plotly <- plotly::ggplotly(p1, tooltip = "text") %>%
      plotly::layout(
        margin = list(r = 150, b = 50, t = 50),
        height = 400,
        yaxis = list(
          title = list(
            text = paste(var_name, "(raw units)"),
            standoff = 10
          )
        ),
        xaxis = list(
          title = list(
            text = "Timestamps",
            standoff = 10
          ),
          tickangle = -45
        ),
        legend = list(
          title = list(text = "")
        )
      )
    
    plots[[var_name]] <- p1_plotly
  }
  
  # ============================================================================
  # RETURN RESULTS
  # ============================================================================
  
  result <- list(
    reverted_forecasts = reverted_forecasts,
    plots = plots
  )
  
  class(result) <- "hvt.object"
  return(result)
}

#' @keywords internal

Delete_Outpoints <-
function(tile, check_dirsgs){
  check_dirsgs[, c("x1","x2","y1", "y2")]<-round(check_dirsgs[, c("x1","x2","y1", "y2")], 6)
  tile$x <- round(tile$x,6)
  tile$y <- round(tile$y,6)
  tile_vert_info <- list()
  for(tile_index in 1: length(tile$x)){
    #get the index of the vertices which are outside the parent tile; if a vertex is outside
    #the parent tile then the coordinate of that vertex will not be present in dirsgs
    if(length(intersect(which(check_dirsgs[, c("y1", "y2")] == tile$y[tile_index]),
                        which(check_dirsgs[, c("x1", "x2")] == tile$x[tile_index])))){
      tile_vert_info[[tile_index]] <- TRUE
    }else{
      tile_vert_info[[tile_index]] <- FALSE
    }
  }
  #delete the vertices which are outside the parent tile
  del_ind <- which(tile_vert_info == FALSE)
  if(length(del_ind)){
    tile$x <- tile$x[-del_ind]
    tile$y <- tile$y[-del_ind]
    tile$bp <- tile$bp[-del_ind]
  }
  return(tile)
}

#' @name plotMsmKN
#' @title plotting MAE across Number of cells
#' @description This function helps to visualize the `HVTMSMoptimization` function's
#' output. It creates a plotly object where the lowest MAE combination achieved for each cell 
#' will be plotted
#' @param optimization_results List. The output of `HVTMSMoptimization` function 
#' specific for a mae_metric
#' @return List containing visualization results for the selected MAE metric:
#' \item{[[plot]] }{A Plotly object of cell-level lowest MAE combination} 
#' \item{[[table]] }{A table of plotted data}
#' @author Vishwavani <vishwavani@@mu-sigma.com>, Nithya <nithya.sn@@mu-sigma.com>
#' @keywords Hyperparameter_Tuning
#' @include HVTMSMoptimization.R
#' @importFrom magrittr %>%
#' @export plotMsmKN

plotMsmKN <- function(optimization_results) {
  
  # Initialize NSE symbols to avoid R CMD check NOTES
  nclust <- mae <- k <- nn <- is_overall_best <- is_overall_worst <- hover_text <- NULL
  
  
  # Quick guard
  if (is.null(optimization_results)) {
    cat("No results provided.\n")
    return(list(plot = NULL, table = NULL))
  }
  
  # ---------- 1) Standardize a helper to pick columns safely ----------
  pick_col <- function(df, choices) {
    nm <- choices[choices %in% names(df)]
    if (length(nm) == 0) return(NULL)
    nm[1]
  }
  
  # ---------- 2) Prefer ALL RESULTS, fallback to nclust_best_results ----------
  src <- NULL
  if (!is.null(optimization_results$all_results) &&
      nrow(optimization_results$all_results) > 0) {
    src <- optimization_results$all_results
  } else if (!is.null(optimization_results$nclust_best_results) &&
             nrow(optimization_results$nclust_best_results) > 0) {
    src <- optimization_results$nclust_best_results
  } else {
    cat("No MSM optimization results to plot.\n")
    return(list(plot = NULL, table = NULL))
  }
  
  # Identify columns (support old/new schemas)
  nclust_col <- pick_col(src, c("Number_of_Cells","Number of Cells"))
  nn_col     <- pick_col(src, c("Number_of_nearest_neighbors","Number of nearest neighbors"))
  mae_col    <- pick_col(src, c("MAE","mae"))
  k_col      <- pick_col(src, c("k"))
  #metric_col <- pick_col(src, c("metric"))  # optional
  
  if (is.null(nclust_col) || is.null(nn_col) || is.null(mae_col) || is.null(k_col)) {
    stop("Required columns not found in results: need cells/k/nn/MAE.")
  }
  
  # ---------- 3) Build plot_data from ALL rows (ensures one point per tested cell) ----------
  # Coerce to numeric, drop problematic rows with NA in key fields
  src_clean <- src %>%
    dplyr::mutate(
      nclust = suppressWarnings(as.numeric(.data[[nclust_col]])),
      k      = suppressWarnings(as.numeric(.data[[k_col]])),
      nn     = suppressWarnings(as.numeric(.data[[nn_col]])),
      mae    = suppressWarnings(as.numeric(.data[[mae_col]]))
    ) %>%
    tidyr::drop_na(nclust, mae)  # keep only rows we can plot
  
  if (nrow(src_clean) == 0) {
    cat("No plottable rows (all NA after coercion).\n")
    return(list(plot = NULL, table = NULL))
  }
  
  # Choose one row per cell: the true best (min MAE) for that cell
  # If multiple rows tie, slice_min keeps one (with_ties = FALSE).
  plot_data <- src_clean %>%
    dplyr::group_by(nclust) %>%
    dplyr::slice_min(mae, n = 1, with_ties = FALSE) %>%
    dplyr::ungroup() %>%
    dplyr::arrange(nclust)
  
  # ---------- 4) Determine highest and lowest MAE ----------
  min_idx <- which.min(plot_data$mae)
  max_idx <- which.max(plot_data$mae)
  best_nclust <- plot_data$nclust[min_idx]
  worst_nclust <- plot_data$nclust[max_idx]
  
  plot_data <- plot_data %>%
    dplyr::mutate(
      is_overall_best = (nclust == best_nclust),
      is_overall_worst = (nclust == worst_nclust),
      hover_text = paste0(
        "Cells: ", nclust, "<br>",
        "k: ", k, "<br>",
        "nn: ", nn, "<br>",
        "MAE: ", sprintf("%.4f", mae)
      )
    )
  
  # ---------- 5) Build table ----------
  results_table <- plot_data %>%
    dplyr::transmute(
      `Number of Cells`   = nclust,
      k                   = k,
      `Nearest Neighbors` = nn,
      MAE                 = round(mae, 4),
      `Best Result`       = ifelse(is_overall_best, "\u2605", "")
    )
  
  # ---------- 6) Plot ----------
  n_points <- nrow(plot_data)
  point_size <- dplyr::case_when(
    n_points <= 20 ~ 3,
    n_points <= 50 ~ 2,
    n_points <= 100 ~ 1.5,
    TRUE ~ 1
  )
  
  p <- ggplot2::ggplot(plot_data, ggplot2::aes(x = nclust, y = mae)) +
    ggplot2::geom_line(color = "blue", alpha = 0.7, linewidth = 0.8) +
    ggplot2::geom_point(
      data = dplyr::filter(plot_data, !is_overall_best & !is_overall_worst),
      ggplot2::aes(text = hover_text),
      color = "blue",
      size = point_size,
      alpha = 0.8
    ) +
    ggplot2::geom_point(
      data = dplyr::filter(plot_data, is_overall_best),
      ggplot2::aes(text = hover_text),
      color = "#2e7d32",
      size = point_size * 1.5,
      alpha = 1
    ) +
    ggplot2::geom_point(
      data = dplyr::filter(plot_data, is_overall_worst),
      ggplot2::aes(text = hover_text),
      color = "red",
      size = point_size * 1.5,
      alpha = 1
    ) +
    ggplot2::scale_x_continuous(
      name = "Number of Cells",
      breaks = if (n_points <= 20) {
        plot_data$nclust
      } else if (n_points <= 50) {
        pretty(plot_data$nclust, n = 10)
      } else {
        pretty(plot_data$nclust, n = 8)
      },
      expand = ggplot2::expansion(mult = c(0.02, 0.02))
    ) +
    ggplot2::scale_y_continuous(
      name = "Mean Absolute Error (MAE)",
      limits = c(0, NA),
      expand = ggplot2::expansion(mult = c(0, 0.05)),
      labels = function(x) sprintf("%.3f", x)
    ) +
    ggplot2::theme_minimal() +
    ggplot2::theme(
      plot.title = ggplot2::element_text(size = 14, face = "bold", hjust = 0.5, margin = ggplot2::margin(b = 20)),
      axis.title = ggplot2::element_text(size = 12),
      axis.title.x = ggplot2::element_text(margin = ggplot2::margin(t = 15)),
      axis.title.y = ggplot2::element_text(margin = ggplot2::margin(r = 15)),
      axis.text = ggplot2::element_text(size = 10),
      axis.text.x = ggplot2::element_text(
        angle = if (n_points > 30) 45 else 0,
        hjust = if (n_points > 30) 1 else 0.5,
        margin = ggplot2::margin(t = 8)
      ),
      axis.text.y = ggplot2::element_text(margin = ggplot2::margin(r = 8)),
      panel.grid.minor = ggplot2::element_blank(),
      panel.grid.major.x = ggplot2::element_line(colour = "grey90", linewidth = 0.5),
      panel.grid.major.y = ggplot2::element_line(colour = "grey90", linewidth = 0.5),
      plot.margin = ggplot2::margin(30, 30, 30, 30)
    ) +
    ggplot2::ggtitle("MSM Optimization Results: MAE vs Number of Cells")
  
  interactive_plot <- plotly::ggplotly(p, tooltip = "text") %>%
    plotly::layout(
      autosize = TRUE,
      margin = list(l = 80, r = 50, t = 80, b = 100),
      hovermode = "closest",
      showlegend = FALSE,
      xaxis = list(title = list(standoff = 20), tickangle = if (n_points > 30) -45 else 0),
      yaxis = list(title = list(standoff = 30))
    ) %>%
    plotly::config(
      displayModeBar = TRUE,
      modeBarButtonsToRemove = c(
        "pan2d", "select2d", "lasso2d", "autoScale2d",
        "hoverClosestCartesian", "hoverCompareCartesian"
      ),
      displaylogo = FALSE,
      responsive = TRUE
    )
  
  best_row <- plot_data[which.min(plot_data$mae), ]
  
  list(
    plot = interactive_plot,
    table = results_table,
    best_nclust = best_row$nclust,
    best_k = best_row$k,
    best_nn = best_row$nn,
    best_mae = round(best_row$mae, 4)
  )
}

#' @name plotZscore
#' @title Plots of z scores
#' @description This is the main function to plot the z scores against cell ids.
#' @param data Data frame. A data frame of cell id and features. 
#' @param cell_range Vector. A numeric vector of cell id range for which the plot should be displayed. 
#' Default is NULL, which plots all the cells.
#' @param segment_size Integer.  A numeric value to indicate the size of the bars in the plot. 
#' Default is 2.
#' @param reference_lines Vector. A numeric vector of confidence interval values for the 
#' reference lines in the plot. Default is c(-1.65, 1.65).  
#' @return A grid of plots of z score against cell id of teh given features.
#' @author Vishwavani <vishwavani@@mu-sigma.com>
#' @keywords EDA
#' @examples 
#'data("EuStockMarkets")
#'dataset <- data.frame(t = as.numeric(time(EuStockMarkets)),
#'                      DAX = EuStockMarkets[, "DAX"],
#'                      SMI = EuStockMarkets[, "SMI"],
#'                      CAC = EuStockMarkets[, "CAC"],
#'                      FTSE = EuStockMarkets[, "FTSE"])
#'rownames(EuStockMarkets) <- dataset$t
#'hvt.results<- trainHVT(dataset[-1],n_cells = 60, depth = 1, quant.err = 0.1,
#'                       distance_metric = "L1_Norm", error_metric = "max",
#'                       normalize = TRUE,quant_method = "kmeans")
#'col_names <- c("Cell.ID","DAX","SMI","CAC","FTSE")
#'data <- dplyr::arrange(dplyr::select(hvt.results[[3]][["summary"]],col_names),Cell.ID)
#'data <- round(data, 2)
#'plotZscore(data)
#' @export plotZscore


plotZscore <- function(data, cell_range = NULL, segment_size = 2,
                       reference_lines = c(-1.65, 1.65)) {
  
  Value<-threshold_status <- NULL
  
  default_color = "gray50" 
  positive_color = "green3"
  negative_color = "red3"
  
  
  if (!"Cell.ID" %in% names(data)) {
    stop("Data must contain a 'Cell.ID' column")
  }
  

  # Function to round to nearest 0.2
  round_to_0.2 <- function(x) {
    round(x / 0.2) * 0.2
  }
  
  data_long <- data %>%
    pivot_longer(cols = !Cell.ID, 
                 names_to = "Variable", 
                 values_to = "Value") %>%
    mutate(
      threshold_status = case_when(
        Value >= 1.65 ~ "positive",
        Value <= -1.65 ~ "negative",
        TRUE ~ "within"
      ))
  
  if (!is.null(cell_range)) {
    data_long <- data_long %>%
      filter(Cell.ID >= cell_range[1], Cell.ID <= cell_range[2])
  }
  
  # Get all unique Cell IDs
  all_cell_ids <- sort(unique(data_long$Cell.ID))
  
  # Split data by Variable
  plot_data <- split(data_long, data_long$Variable)
  
  # Create plotting function
  create_plot <- function(df) {
    var_name <- unique(df$Variable)
    ggplot(df) +
      geom_vline(xintercept = reference_lines,
                 color = "black", linetype = "dashed", alpha = 0.5) +
      geom_segment(aes(x = 0, xend = Value, y = Cell.ID, yend = Cell.ID,
                       color = threshold_status),
                   size = segment_size) +
      scale_color_manual(
        values = c(
          "positive" = positive_color,
          "negative" = negative_color,
          "within" = default_color
        ),
        guide = "none") +
      scale_y_continuous(breaks = all_cell_ids,
                         labels = all_cell_ids,
                         expand = expansion(add = 1)) +
      labs(x = "Z-Score", y = "Cell ID", title = var_name) +
      scale_x_continuous(breaks = function(x) {
        unique(sort(c(x[1], -1.65, 0, 1.65, x[2])))
      }, labels = function(x) sprintf("%.2f", x)) +
      theme_bw() +
      theme(
        panel.grid.minor = element_blank(),
        panel.grid.major.y = element_line(color = "gray90"),
        axis.text.y = element_text(size = 8),
        axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.5, size = 8),
        plot.title = element_text(size = 12, face = "bold", hjust = 0.5),
        axis.title.x = element_text(size = 10, margin = margin(t = 10)),
        axis.title.y = element_text(size = 10, margin = margin(r = 10)),
        plot.margin = margin(t = 5, r = 20, b = 20, l = 20),
        panel.background = element_rect(fill = "white"),
        panel.border = element_rect(color = "black", fill = NA, size = 0.5),
        axis.line = element_line(color = "black"))
  }
  
  # Create all plots using purrr::map
  plots <- purrr::map(plot_data, create_plot)
  
  # Arrange plots in a grid with 3 columns
  arranged_plots <- do.call(gridExtra::grid.arrange, c(plots, ncol = 3))
  
  return(arranged_plots)
}

#' @name plotModelDiagnostics
#' @title Make the diagnostic plots for hierarchical voronoi tessellations
#' @description This is the main function that generates diagnostic plots for 
#' hierarchical voronoi tessellations models and scoring. 
#' @param model_obj List. A list obtained from the trainHVT
#' function or scoreHVT function
#' @returns For trainHVT, Minimum Intra-DataPoint Distance Plot, Minimum Intra-Centroid Distance Plot
#' Mean Absolute Deviation Plot, Distribution of Number of Observations in Cells, for Training Data and
#' Mean Absolute Deviation Plot for Validation Data are plotted.
#' For scoreHVT Mean Absolute Deviation Plot for Training Data and Validation Data are plotted
#' @author Shubhra Prakash <shubhra.prakash@@mu-sigma.com>
#' @seealso \code{\link{plotHVT}}
#' @keywords Diagnostics_or_Validation
#' @importFrom magrittr %>%
#' @import ggplot2 
#' @examples
#' data("EuStockMarkets")
#' hvt.results <- trainHVT(EuStockMarkets, n_cells = 10, depth = 1, quant.err = 0.1, 
#'                        distance_metric = "L1_Norm", error_metric = "max",
#'                        normalize = TRUE, quant_method="kmeans", diagnose = TRUE, 
#'                        hvt_validation = TRUE)
#' plotModelDiagnostics(hvt.results)
#' @export plotModelDiagnostics


plotModelDiagnostics <-
  function(model_obj) {
    if (!requireNamespace("patchwork", quietly = TRUE)) {
      stop("Package 'patchwork' is required but not installed.")
    } 


    if (model_obj[["model_info"]][["type"]] == "hvt_model") {
      if (is.na(model_obj[[4]])[1]) {
        stop("Diagnostic plots are not available. Re-run trainHVT with diagnose = TRUE.")
      }
      p1 <- model_obj[[4]]$datapoint_plot + ggplot2::ggtitle("Minimum Intra-DataPoint Distance Plot: Train Data")
      p2 <- model_obj[[4]]$cent_plot + ggplot2::ggtitle("Minimum Intra-Centroid Distance Plot: HVT Model | Train Data")
      p3 <- model_obj[[4]]$mad_plot_train + ggplot2::ggtitle("Mean Absolute Deviation Plot: Calibration: HVT Model | Train Data")
      p4 <- model_obj[[4]]$number_plot + ggplot2::ggtitle("Distribution of Number of Observations in Cells: HVT Model | Train Data")
      p5 <- model_obj[[4]]$singleton_piechart
      if (model_obj[["model_info"]][["input_parameters"]][["hvt_validation"]]) {
        p6 <- model_obj[[5]][["mad_plot"]] + ggplot2::ggtitle("Mean Absolute Deviation Plot:Validation")

        plotDiag <- (p3 / (p1 | p2) / (p4 | p5) / p6)
      } else {
        plotDiag <- (p3 / (p1 | p2) / (p4 | p5))
      }
    } else if (model_obj[["model_info"]][["type"]] == "hvt_prediction") {
      if (is.na(model_obj[["model_mad_plots"]])[1]) {
        stop("MAD plots are not available. Re-run trainHVT with diagnose = TRUE and hvt_validation = TRUE before scoring.")
      }
      mtrain <- model_obj[["model_mad_plots"]][["mtrain"]] + ggplot2::ggtitle("Mean Absolute Deviation Plot: Calibration on Train Data")
      mtest <- model_obj[["model_mad_plots"]][["mtest"]] + ggplot2::ggtitle("Mean Absolute Deviation Plot:Validation")
      mpred <- madPlot(model_obj) + ggplot2::ggtitle("Mean Absolute Deviation Plot:Test Data")
      plotDiag <- mtrain / mtest / mpred
    }
    return(plotDiag)
  }

#' @keywords internal

Transform_Coordinates <- function(sam_rectangle, par_rectangle, scale_mat) {
  xrange_sam <- range(sam_rectangle[, 1])
  yrange_sam <- range(sam_rectangle[, 2])
  xrange_par <- range(par_rectangle[, 1])
  yrange_par <- range(par_rectangle[, 2])
  xlength_sam <- xrange_sam[2] - xrange_sam[1]
  ylength_sam <- yrange_sam[2] - yrange_sam[1]
  xlength_par <- xrange_par[2] - xrange_par[1]
  ylength_par <- yrange_par[2] - yrange_par[1]
  
  transformed_coord_int <- t(sapply(1:4, function(i) {
    scale_mat %*% matrix(sam_rectangle[i, ], ncol = 1)
  }))
  
  translate_mat <- matrix(c(
    ((sum(xrange_par) / 2) - (sum(range(transformed_coord_int[, 1])) / 2)),
    ((sum(yrange_par) / 2) - (sum(range(transformed_coord_int[, 2])) / 2))
  ), ncol = 1)
  
  transformed_data <- c(scale_mat[1, 1], scale_mat[2, 2], translate_mat[1], translate_mat[2])
  
  return(transformed_data)
}

#' @name plotStateTransition
#' @title Creating State Transition Plot 
#' @description This is the main function to create a state transition plot from a data frame.
#' A state transition plot is a type of data visualization used to represent 
#' the changes or transitions in states over time for a given system. 
#' State refers to a particular condition or status of a cell at a specific point in time. 
#' Transition refers to the change of state for a cell from one condition to another over time. 
#' @param df Data frame. The Input data frame should contain two columns. 
#' Cell ID from scoreHVT function and time stamp of that dataset.
#' @param sample_size Numeric. An integer indicating the fraction of the data frame to visualize in the plot.
#' Default value is 0.2
#' @param line_plot Logical. A logical value indicating to create a line plot. Default value is NULL.
#' @param cellid_column Character. Name of the column containing cell IDs.
#' @param time_column Character. Name of the column containing time stamps.
#' @param v_intercept Numeric. A numeric value indicating the time stamp to draw a vertical line on the plot.
#' @param time_periods List. A list of vectors, each containing start and end times for highlighting time periods.
#' @return A plotly object representing the state transition plot for the given data frame.
#' @author PonAnuReka Seenivasan <ponanureka.s@@mu-sigma.com>, Vishwavani <vishwavani@@mu-sigma.com>
#' @keywords Timeseries_Analysis
#' @importFrom magrittr %>%
#' @examples
#' dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
#' DAX = EuStockMarkets[, "DAX"],
#' SMI = EuStockMarkets[, "SMI"],
#' CAC = EuStockMarkets[, "CAC"],
#' FTSE = EuStockMarkets[, "FTSE"])
#'
#' hvt.results<- trainHVT(dataset,n_cells = 60, depth = 1, quant.err = 0.1,
#'                        distance_metric = "L1_Norm", error_metric = "max",
#'                        normalize = TRUE,quant_method = "kmeans")
#' scoring <- scoreHVT(dataset, hvt.results)
#' cell_id <- scoring$scoredPredictedData$Cell.ID
#' time_stamp <- dataset$date
#' dataset <- data.frame(cell_id, time_stamp)
#' plotStateTransition(dataset, sample_size = 1, cellid_column = "cell_id",time_column = "time_stamp")
#' @export plotStateTransition

plotStateTransition <- function(df, sample_size = NULL, line_plot = NULL, 
                                cellid_column, time_column, v_intercept = NULL,
                                time_periods = NULL) { 
  ## For CRAN warnings, initializing empty vectors for these variables.
  Timestamp <- Frequency <- Next_State <- NULL
  
  # Rename column names for Time and Cell for consistency
  colnames(df)[colnames(df) == time_column] <- "Timestamp"
  colnames(df)[colnames(df) == cellid_column] <- "Cell.ID"
  
  # Validate time_periods parameter structure
  if (!is.null(time_periods)) {
    if (!is.list(time_periods) || !all(sapply(time_periods, length) == 2)) {
      stop("time_periods must be a list of vectors, each containing start and end times")
    }
    
    # Convert time_periods to match Timestamp data type
    if (inherits(df$Timestamp, "POSIXct")) {
      time_periods <- lapply(time_periods, function(x) {
        as.POSIXct(x, tz = attr(df$Timestamp, "tzone"))
      })
    } else if (is.numeric(df$Timestamp)) {
      time_periods <- lapply(time_periods, as.numeric)
    } else if  (inherits(df$Timestamp, "Date")) {
      time_periods <- lapply(time_periods, as.Date)
    }
  }
  
  # Ensure v_intercept is converted to the same data type as Timestamp
  if (!is.null(v_intercept)) {
    if (inherits(df$Timestamp, "POSIXct")) {
      v_intercept <- as.POSIXct(v_intercept, tz = attr(df$Timestamp, "tzone"))
    } else if (is.numeric(df$Timestamp)) {
      v_intercept <- as.numeric(v_intercept)
    } else if (inherits(df$Timestamp, "Date")) {
      v_intercept <- as.Date(v_intercept)
    } else {
      stop("Unsupported data type for Timestamp column.")
    }
  }
  
  # Set default values for sample_size and line_plot if they are NULL
  if (is.null(sample_size)) sample_size <- 0.2
  if (is.null(line_plot)) line_plot <- FALSE
  
  # Calculate the number of rows to sample and sample the data based on the specified sample_size
  sampling_percent <- round(sample_size * nrow(df))
  sampled_data <- df[(nrow(df) - sampling_percent + 1):nrow(df), ]
  
  # Group and count frequencies of cell IDs, then arrange by timestamp
  sampled_data <- sampled_data %>%
    dplyr::group_by(Cell.ID) %>%
    dplyr::mutate(Frequency = n()) %>%
    dplyr::arrange(Timestamp)
  
  max_cell_id <- max(sampled_data$Cell.ID, na.rm = TRUE)
  min_cell_id <- min(sampled_data$Cell.ID, na.rm = TRUE)
  
  range_size <- max_cell_id - min_cell_id
  dtick_val <- if (range_size <= 15) 2 else if (range_size <= 30) 5 else if (range_size <= 100) 10 else ceiling(range_size / 10)
  
  
  axis_settings <- list(
    xaxis = list(
      title = "Timestamp",
      range = range(sampled_data$Timestamp)
    ),
    yaxis = list(
      title = "Cell ID",
      range = c(1, max(sampled_data$Cell.ID) + 2),
      tickmode = "linear",
      dtick = dtick_val,
      tick0 = 0    )
  )
  
  # Create base plot with heatmap
  create_base_plot <- function(data, show_lines = FALSE) {
    p <- data %>%
      plotly::plot_ly(
        x = ~Timestamp, 
        y = ~Cell.ID, 
        z = ~Frequency,
        type = "heatmap", 
        hoverinfo = "text",  
        hovertext = ~sprintf(
          "Timestamp: %s<br>Cell ID: %d<br>Frequency: %d",
          Timestamp, Cell.ID, Frequency),
        showlegend = FALSE
      ) %>%
      plotly::colorbar(
        title = "Frequency",
        len = 0.5,
        thickness = 40,
        y = 0.8,
        yanchor = "middle"
      )
    
    # Add layout first
    p <- p %>%
      plotly::layout(
        autosize = TRUE,
        title = list(
          text = "Time series Flowmap",
          x = 0.03,
          y = 0.99,
          xanchor = "left"
        ),
        xaxis = axis_settings$xaxis,
        yaxis = axis_settings$yaxis,
        showlegend = FALSE,
        hovermode = "closest",
        hoverlabel = list(
          bgcolor = "black",
          font = list(color = "white") 
        )
        )
    
    # Add time period highlighting if specified
    if (!is.null(time_periods)) {
      # Create a list of shapes for all time periods
      shapes_list <- lapply(time_periods, function(period) {
        list(
          type = "rect",
          x0 = period[1],
          x1 = period[2],
          y0 = axis_settings$yaxis$range[1],
          y1 = axis_settings$yaxis$range[2],
          fillcolor = "black",
          opacity = 0.2,
          line = list(width = 0),
          layer = "below"
        )
      })
      
      # Add all shapes at once
      p <- p %>%
        plotly::layout(shapes = shapes_list)
    }
    
    # Add vertical line if specified
    if (!is.null(v_intercept)) {
      vline_df <- data.frame(
        x = c(v_intercept, v_intercept),
        y = c(min(data$Cell.ID), max(data$Cell.ID))
      )
      
      p <- p %>%
        plotly::add_trace(
          data = vline_df,
          x = ~x,
          y = ~y,
          type = "scatter",
          mode = "lines",
          line = list(
            color = "black",
            width = 2,
            dash = "4px,4px"
          ),
          showlegend = FALSE,
          hoverinfo = "none"
        )
      
      p <- p %>%
        plotly::layout(
          autosize = TRUE,
          annotations = list(
            list(
              x = 1.02,
              y = 0.55,
              text = "  --- End of\nTraining data",
              showarrow = FALSE,
              font = list(
                color = "black",
                size = 14
              ),
              xref = "paper",
              yref = "paper",
              xanchor = "left",
              yanchor = "top",
              align = "left"
            )
          )
        )
    }
    
    # Add state transition lines if requested
    if (show_lines) {
      state_transitions <- data %>%
        dplyr::select(Timestamp, Cell.ID, Frequency) %>%
        dplyr::mutate(Next_State = lead(Cell.ID))
      
      p <- p %>%
        plotly::add_trace(
          data = state_transitions,
          x = ~Timestamp,
          y = ~Cell.ID,
          type = "scatter",
          mode = "markers",
          line = list(color = "gray", width = 1),
          marker = list(color = "transparent", size = 1),
          showlegend = FALSE
        )
    }
    
    return(p)
  }
  
  # Return the appropriate plot
  if (sample_size <= 1) {
    if (line_plot == TRUE) {
      return(create_base_plot(sampled_data, show_lines = TRUE))
    } else if (line_plot == FALSE) {
      return(create_base_plot(sampled_data, show_lines = FALSE))
    } else {
      stop("Invalid line_plot parameter. Use TRUE or FALSE.")
    }
  } else {
    stop("Invalid sample_size parameter. Use values between 0.1 to 1.")
  }
}

#' @keywords internal

hvq <-
  function (x,
            min_compression_perc = NA,
            n_cells = NA,
            depth = 3,
            quant.err = 10,
            seed = 300,
            algorithm = "Hartigan-Wong",
            distance_metric = c("L1_Norm", "L2_Norm"),
            error_metric = c("mean", "max"),
            quant_method=c("kmeans","kmedoids")
  ) {    
    requireNamespace("dplyr")
    rescl <- list()
    resid <- list()
    resm <- list()
    maxqe <- list()
    meanqe <- list()
    resplt <- list()
    ztab3up <- list()
    ztab1 <- ztab2 <- ztabn <- NULL
    ztab11 <- ztab12 <- ztab13 <- NULL
    zdepth <- depth
    if(!is.na(n_cells)){
      quantinit <- rep(F, n_cells)
      ztab3upc<-matrix(0, nrow = ncol(x), n_cells)
      std<-matrix(0, nrow = ncol(x), n_cells)
    }
    # Ensure consistent seeding for kmeans calls
    set.seed(279)
  
    colSd <- function (x, na.rm=FALSE) apply(X=x, MARGIN=2, FUN=stats::sd, na.rm=na.rm)
 
    calculate_euclidean_distance_for_each_cluster <- function(centroid, data) {
      as.matrix(stats::dist(rbind(centroid, data)))[1, -1]
    }

    calculate_manhattan_distance_for_each_cluster <- function(centroid, data) {
      as.matrix(stats::dist(rbind(centroid, data), method = "manhattan"))[1, -1]
    }
    
    ## for distance metrics i.e; manhattan or eucleadian
    if(distance_metric == "L1_Norm"){
      function_to_calculate_distance_metric <- calculate_manhattan_distance_for_each_cluster
    } else if(distance_metric == "L2_Norm"){
      function_to_calculate_distance_metric <- calculate_euclidean_distance_for_each_cluster
    } else{
      stop('distance_metric must be L1_Norm (Manhattan), L2_Norm (Euclidean) or a custom distance function')
    }
    
    # for error metric i.e; mean or max
    if(error_metric %in% c("mean","max")){
      function_to_calculate_error_metric <- error_metric
    } else{
      stop('error_metric must be max,mean or custom function')
    }
    
   

    
    
    
    if(!is.na(min_compression_perc)){
      depth <- 1
      compress_percent <- 0
      n_cells_compress_perc = 2
      num_data_points <- rep(F, n_cells_compress_perc)
      n_cells_quant_err_list <- list()
      n_cells = NA

      while(n_cells_compress_perc < nrow(x) & compress_percent <= min_compression_perc){
      quantinit <- rep(F, n_cells_compress_perc)
      ztab3upc<-matrix(0, nrow = ncol(x), n_cells_compress_perc)
      std<-matrix(0, nrow = ncol(x), n_cells_compress_perc)
      # print(n_cells_compress_perc)
      # Ensure consistent seeding before kmeans
      set.seed(279)
      outkinit <- getCentroids(x=x,
                               kout = stats::kmeans(x, n_cells_compress_perc, iter.max=10^5, algorithm=algorithm),
                               n_cells=n_cells_compress_perc,
                               function_to_calculate_distance_metric=function_to_calculate_distance_metric,
                               function_to_calculate_error_metric=function_to_calculate_error_metric,
                               distance_metric=distance_metric,
                               error_metric = error_metric,
                               quant_method = quant_method)

      n_cells_quant_err_list <- unlist(outkinit$cent)
      noOfCells = sum(!is.na(n_cells_quant_err_list))
      noOfCellsBelowQuantizationError = sum(n_cells_quant_err_list < quant.err,na.rm = TRUE)
      compress_percent = round((noOfCellsBelowQuantizationError/noOfCells) * 100, 2) #percentOfCellsBelowQuantizationErrorThreshold
      n_cells_compress_perc <- n_cells_compress_perc + 1
    }
    n_cells_optimal <- n_cells_compress_perc - 1
    message(paste0("For the given dataset ",min_compression_perc,"% compression rate is achieved at n_cells : ", n_cells_optimal))
    }else{
      # Ensure consistent seeding before kmeans
      set.seed(279)
      outkinit <- getCentroids(x=x,
                             kout = stats::kmeans(x, n_cells, iter.max=10^5, algorithm=algorithm),
                             n_cells=n_cells,
                             function_to_calculate_distance_metric=function_to_calculate_distance_metric,
                             function_to_calculate_error_metric=function_to_calculate_error_metric,
                             distance_metric=distance_metric,
                             error_metric = error_metric,
                             quant_method = quant_method)
    
      n_cells_optimal <- n_cells
    }
    
   
    rescl[[1]] <- outkinit$val
    tet <- lapply(outkinit$val, row.names) 
    for (k in 1:length(tet)){
      tet[[k]] <- data.frame(tet[[k]], 1, 1, k)
    } 
    names(tet) <- paste(1:length(tet))
    resid[[1]] <- tet
    #centroids of each cluster
    resm[[1]] <- outkinit$cent
    maxqe[[1]] <- outkinit$maxQE
    meanqe[[1]] <- outkinit$meanQE
    #flag to check for quantization error
    resplt[[1]] <- unlist(outkinit$cent) > quant.err
    #initial values for next level k-means
    initclust <- outkinit$values  
    if (depth > 1) {
      i <- 1
      while (i < depth) {
        ijclust <- NULL
        #stores values of datapoints
        ijrescl <- list()
        #ids of the datapoints
        ijresid <- list()
        #to store the centroids
        ijresm <- list()
        # to store max qe as radius for anomalies
        ijmaxqe <- list()
        ijmeanqe <- list()
        #flag to check the quantization error
        ijresplt <- list()
        #number of datapoints in a cluster
        ijresnsize <- list()
        ijztab3up <- list()
        #flag which checks quantization error
        quantok <- unlist(resplt[[i]])
        j <- 1  
        ## Intervention 1
        while (j < (n_cells^i) + 1) {  
          #if datapoint exceeds quantization error and number of datapoints in that cluster is greater than n_cells
          if (quantok[j] & NROW(initclust[[j]]) > 3) {
            #k-means on the initclust to obtain the next level clustering(sub-clusters)
            z = data.frame(initclust[[j]])
            outk <- getOptimalCentroids(z, iter.max = 10^5, algorithm = algorithm, n_cells = n_cells, function_to_calculate_distance_metric = function_to_calculate_distance_metric, function_to_calculate_error_metric = function_to_calculate_error_metric, quant.err = quant.err, distance_metric = distance_metric, quant_method = quant_method)
            ijrescl[[j]] <- outk$val
            tet <- lapply(outk$val, row.names)
            tet[-which(sapply(tet, is.null))]
           
            for (k in 1: length(tet)){
              
              if(!is.null(tet[[k]])){
                tet[[k]] <- data.frame(tet[[k]],i + 1, j, k)  
              }
              else{
                tet[[k]] <- data.frame(tet[[k]]<-numeric(0))
              }
            }           
           
            ijresid[[j]] <- tet
            ijresm[[j]] <- outk$centers
            ijmaxqe[[j]] <- outk$maxQE
            ijmeanqe[[j]] <- outk$meanQE
            #size of a cluster
          
            ijresnsize[[j]] <- outk$nsize
            ijztab3up[[j]]<-matrix(0, nrow = ncol(x), n_cells)          
            ijztab3up[[j]] <- t(outk$centroid_val)
            #flag to check if the quantization error threshold is exceeded
            ijresplt[[j]] <- unlist(outk$centers) > quant.err
            #store the datapoints
            ijclust <- c(ijclust, outk$values)
          }else {
            ijrescl[[j]] <- rep(NA, n_cells)
            ijresid[[j]] <- NULL
            ijresm[[j]] <- rep(NA, n_cells)
            ijmaxqe[[j]] <- rep(NA, n_cells)
            ijmeanqe[[j]] <- rep(NA, n_cells)
            ijresnsize[[j]] <- rep(0, n_cells)
            ijztab3up[[j]] <- matrix(NA, ncol(x), n_cells)
            ijresplt[[j]] <- rep(F, n_cells)
            ijclust <- c(ijclust, rep(NA, n_cells))
          }
          n_cells_at_j <- length(tet)
          ztab1 <- c(ztab1, paste(i + 1, j, 1:n_cells_at_j, sep = ","))
          ztab11 <- c(ztab11, rep(i + 1, n_cells_at_j))
          ztab12 <- c(ztab12, rep(j, n_cells_at_j))
          ztab13 <- c(ztab13, 1: n_cells_at_j)
          j <- j + 1
        }
        rescl[[i + 1]] <- ijrescl
        resid[[i + 1]] <- ijresid
        resm[[i + 1]] <- ijresm
        maxqe[[i + 1]] <- ijmaxqe
        meanqe[[i + 1]] <- ijmeanqe
        resplt[[i + 1]] <- ijresplt
        ztab2 <- c(ztab2, unlist(ijresnsize))
        ztab3up[[i + 1]] <- data.frame(ijztab3up)
        ztabn <- c(ztabn, unlist(ijresm))
        initclust <- ijclust
        # flog.info("Output for Level %s is ready", i)
        i <- i + 1
        #check if the desired depth is reached
        if (!is.element(T, unlist(ijresplt))) {
          zdepth <- i
          i <- depth
        }
      }

      ztab <- data.frame(matrix(0, nrow = sum(n_cells^(1:zdepth)), 
                                ncol = (ncol(x) + 5)))
      #Segment Level
      ztab[1:n_cells, 1] <- rep(1, n_cells)
      #Segment Parent
      ztab[1:n_cells, 2] <- rep(1, n_cells)
      #Segment Child
      ztab[1:n_cells, 3] <- 1: n_cells
      #Size of the cluster
      ztab[1:n_cells, 4] <- unlist(outkinit$nsize)
      #Centroid/Quantization error of the cluster
      ztab[1:n_cells, 5] <- unlist(outkinit$cent)
      for(a in 1: n_cells){
        for(b in 1: ncol(x)){
          ztab3upc[b, a] <- as.matrix(mean(outkinit$val[[a]][, b], na.rm = TRUE))
          rownames(ztab3upc) <- colnames(x)
          # Calculating sd
          std[b, a] <- as.matrix(stats::sd(outkinit$val[[a]][, b], na.rm = TRUE))
          rownames(std) <- colnames(x)
        }
      }
      for (l in 1:length(ztab3up)){
        ztab3upc <- cbind(ztab3upc,ztab3up[[l]])
        std <- cbind(std, ztab3up[[l]])
      }    
      ztab[(n_cells + 1):sum(n_cells^(1:zdepth)), 1] <- ztab11
      ztab[(n_cells + 1):sum(n_cells^(1:zdepth)), 2] <- ztab12
      ztab[(n_cells + 1):sum(n_cells^(1:zdepth)), 3] <- ztab13
      ztab[(n_cells + 1):sum(n_cells^(1:zdepth)), 4] <- ztab2
      ztab[, 6: ncol(ztab)] <- t(ztab3upc)
      ztab[(n_cells + 1): sum(n_cells^(1: zdepth)), 5] <- ztabn
      names(ztab) <- c("Segment.Level", "Segment.Parent", "Segment.Child", 
                       "n", "Quant.Error", colnames(x))
    } else {
      # else loop if depth = 1
      ztab <- data.frame(matrix(0, nrow = n_cells_optimal, ncol = (ncol(x) +  5)))                                                          
      ztab[, 1] <- rep(1, n_cells_optimal)
      ztab[, 2] <- rep(1, n_cells_optimal)
      ztab[, 3] <- 1:n_cells_optimal
      ztab[, 4] <- unlist(outkinit$nsize)
      ztab[, 5] <- unlist(outkinit$cent)
      ztab_mean <- t(sapply(outkinit$val, colMeans, 
                            na.rm = TRUE))
      std <- t(sapply(outkinit$val, colSd, 
                      na.rm = TRUE))
      if(quant_method == "kmeans"){
        ztab[, 6:ncol(ztab)] <- ztab_mean # mean values
      } else {
        ztab[, 6:ncol(ztab)] <- outkinit[["sum_val"]] # Medoid/median values
      }
      names(ztab) <- c("Segment.Level", "Segment.Parent", "Segment.Child", 
                       "n", "Quant.Error", colnames(x))
    }
    if(depth > 1){
      meanCol = t(ztab3upc)
      std = t(std)
    }else{
      meanCol =  ztab_mean
      std = data.frame(std)
    }
    ## Calculate compress percentage
    compression_summary <- ztab %>% 
      dplyr::group_by(Segment.Level) %>% 
      dplyr::summarise( noOfCells = sum(!is.na(Quant.Error)), noOfCellsBelowQuantizationError = sum(Quant.Error < quant.err,na.rm = TRUE)) %>% dplyr::mutate(percentOfCellsBelowQuantizationErrorThreshold = (noOfCellsBelowQuantizationError/noOfCells))  
    colnames(compression_summary)[1] <- "segmentLevel" 
    if(any(is.nan(compression_summary$percentOfCellsBelowQuantizationErrorThreshold))){
      compression_summary <- compression_summary[stats::complete.cases(compression_summary),]}
    compression_summary$parameters=paste("n_cells:",n_cells_optimal,"quant.err:",quant.err,"distance_metric:",distance_metric[1],"error_metric:",error_metric[1],"quant_method:",quant_method[1])
    ridnames <- resid
    rclnames <- rescl
    return(list(clusters = initclust, nodes.clust = rescl, idnodes = resid, 
                error.quant = resm, max_QE = maxqe, plt.clust = resplt, summary = ztab, compression_summary = compression_summary,
                meanClust = meanCol, sdClust = std,mean_QE = meanqe, centroid_data = outkinit$dataframe_clusters))   
  }

#' @keywords internal
ScaleMat <-
function(samRectangle, parRectangle){
  
  xrange_sam <- range(samRectangle[, 1])
  yrange_sam <- range(samRectangle[, 2])
  
  xrange_par <- range(parRectangle[, 1])
  yrange_par <- range(parRectangle[, 2])
  
  xlength_sam <- xrange_sam[2] - xrange_sam[1]
  ylength_sam <- yrange_sam[2] - yrange_sam[1]
  
  xlength_par <- xrange_par[2] - xrange_par[1]
  ylength_par <- yrange_par[2] - yrange_par[1]
  
  x_scalefactor <- xlength_par / xlength_sam
  y_scalefactor <- ylength_par / ylength_sam
  
  scale_mat <- matrix(c(x_scalefactor, 0, 0, y_scalefactor), ncol = 2)
  return(scale_mat)
}

#' @importFrom magrittr %>%
#' @importFrom stats runif setNames
#' @keywords internal

# Calculate dataset statistics for scaling
calculate_dataset_stats <- function(numeric_dataset) {
  list(
    mean = numeric_dataset %>% dplyr::summarise(dplyr::across(dplyr::everything(), ~ round(mean(.), 4))),
    sd   = numeric_dataset %>% dplyr::summarise(dplyr::across(dplyr::everything(), ~ round(stats::sd(.),   4)))
  )
}

# Pre-compute HVT models sequentially to ensure consistency
precompute_hvt_models <- function(entire_dataset, ncell_range, hvt_params,
                                  verbose = FALSE, time_column) {
  
  # Pre-prepare dataset once
  numeric_dataset <- entire_dataset %>% dplyr::select(-dplyr::all_of(time_column))
  
  # Prepare HVT training dataset
  hvt_train_dataset <- numeric_dataset
  
  # Train models using lapply
  hvt_models <- lapply(ncell_range, function(nclust) {
    tryCatch({
      set.seed(279)
      hvt_args <- c(list(dataset = hvt_train_dataset, n_cells = nclust), hvt_params)  
      hvt_results <- do.call(trainHVT, hvt_args)
      if (is.null(hvt_results) || length(hvt_results) == 0) {
        stop("trainHVT returned NULL or empty results")
      }
      return(hvt_results)
    }, error = function(e) NULL)
  })
  
  # Name the list elements
  names(hvt_models) <- as.character(ncell_range)
  
  return(hvt_models)
}

# Process a single nclust value using pre-computed HVT model
process_single_nclust_with_precomputed_model <- function(nclust, hvt_model, numeric_dataset, entire_dataset, expost_forecasting,
                                  time_column, k_range, nn_range, num_simulations,
                                  time_values, n_ahead, mae_metric, 
                                  raw_dataset_stats, parallel, verbose) {
  
  tryCatch({
    if (verbose) cat("Processing nclust =", nclust, "\n")
    
    # Check if HVT model exists
    if (is.null(hvt_model)) {
      stop("Pre-computed HVT model for nclust = ", nclust, " is NULL")
    }
    
    # Score complete dataset using pre-computed model
    complete_dataset <- dplyr::bind_rows(entire_dataset, expost_forecasting)
    scoring <- score_hvt_model(complete_dataset, hvt_model)
    
    # Prepare temporal data
    temporal_data_all <- prepare_temporal_data(scoring, complete_dataset, time_column)
    
    # Create training-only temporal data for transition probabilities
    max_train_t <- max(entire_dataset[[time_column]], na.rm = TRUE)
    temporal_data_train <- temporal_data_all %>%
      dplyr::filter(.data[[time_column]] <= max_train_t)
    
    # Get transition probabilities using ONLY training data
    prob_trans_matx <- get_transition_probabilities(temporal_data_train, time_column)
    
    # Get initial state
    initial_state <- get_initial_state(temporal_data_train, time_column)
    
    # Prepare centroid data
    centroid_2d_points <- scoring$cellID_coordinates
    centroid_data_for_mae <- prepare_centroid_data_for_mae(
       hvt_model, numeric_dataset, scoring,
      complete_dataset, raw_dataset_stats, mae_metric, centroid_2d_points
    )
    
    # Detect problematic states
    scored_cells <- unique(scoring$scoredPredictedData$Cell.ID)
    problematic_detection <- detect_problematic_states(prob_trans_matx, scored_cells)
    problematic_states <- problematic_detection$all_problematic
    problematic_states <- intersect(problematic_states, scored_cells)
    
    # Filter to training cells only
    training_cells <- unique(temporal_data_train$Cell.ID)
    problematic_states <- problematic_states[problematic_states %in% training_cells]
    
    if (verbose && length(problematic_states) > 0) {
      cat("  Found", length(problematic_states), "problematic states\n")
    }
    
    # Initialize attempts storage
    all_attempts_by_metric <- initialize_attempts_storage(mae_metric)
    
    # Check if there are problematic states
    if (length(problematic_states) == 0) {
      # No problematic states - run baseline simulation only
      baseline_mae <- run_baseline_simulation(
        transition_matrix = prob_trans_matx,
        temporal_data_train = temporal_data_train,
        scoring = scoring,
        hvt_results = hvt_model,
        initial_state = initial_state,
        centroid_data_for_mae = centroid_data_for_mae,
        mae_metrics = mae_metric,
        raw_dataset_stats = raw_dataset_stats,
        time_column = time_column,
        entire_dataset = entire_dataset,
        expost_forecasting = expost_forecasting,
        num_simulations = num_simulations,
        time_values = time_values,
        n_ahead = n_ahead
      )
      
      # Store baseline results
      for (metric in mae_metric) {
        status_msg <- if (is.na(baseline_mae[[metric]]) || is.null(baseline_mae[[metric]])) {
          "ERROR: Data structure error in MSM baseline"
        } else {
          "Successful Iteration without Problematic state"
        }
        
        all_attempts_by_metric[[metric]][[1]] <- data.frame(
          `Number of Cells` = nclust, 
          k = "-", 
          `Number of nearest neighbors` = "-",
          mae = baseline_mae[[metric]], 
          status = status_msg, 
          stringsAsFactors = FALSE, 
          check.names = FALSE
        )
      }
      
      comprehensive_results <- list(
        simulation_results = list(),
        all_attempts = all_attempts_by_metric
      )
      
    } else {
      # Problematic states exist - run full k-nn combinations
      clustering_data <- centroid_2d_points %>% dplyr::select(-.data$Cell.ID)
      
      comprehensive_results <- process_all_k_nn_combinations(
        k_local_range = k_range,
        nn_range = nn_range,  
        problematic_states = problematic_states,
        centroid_2d_points = centroid_2d_points,
        clustering_data = clustering_data,
        hvt_results = hvt_model,
        scoring = scoring,
        nclust = nclust,
        mae_metric = mae_metric,
        all_attempts_by_metric = all_attempts_by_metric,
        prob_trans_matx = prob_trans_matx,
        temporal_data_complete = temporal_data_all,
        initial_state = initial_state,
        centroid_data_for_mae = centroid_data_for_mae,
        raw_dataset_stats = raw_dataset_stats,
        time_column = time_column,
        entire_dataset = entire_dataset,
        expost_forecasting = expost_forecasting,
        num_simulations = num_simulations,
        time_values = time_values,
        n_ahead = n_ahead,
        parallel = FALSE,
        verbose = verbose
      )
    }
    
    # Process and format results
    processed_results <- process_simulation_results(
      simulation_results = comprehensive_results$simulation_results,
      mae_metric = mae_metric,
      nclust = nclust,
      all_attempts_by_metric = comprehensive_results$all_attempts
    )
    
    return(list(
      successful_results = processed_results$successful_results,
      best_result = processed_results$best_result,
      all_attempts = processed_results$all_attempts,
      success = TRUE,
      nclust = nclust
    ))
  }, error = function(e) {
    if (verbose) cat("  ERROR with nclust", nclust, ":", e$message, "\n")
    
    all_config_failure <- list()
    for (metric in mae_metric) {
      all_config_failure[[metric]] <- data.frame(
        `Number of Cells` = nclust, k = "-", `Number of nearest neighbors` = "-",
        mae = NA, status = paste0("ERROR: High-level error: ", e$message),
        stringsAsFactors = FALSE, check.names = FALSE
      )
    }
    
    return(list(
      successful_results = NULL,
      best_result = NULL,
      all_attempts = all_config_failure,
      success = FALSE,
      error = e$message,
      nclust = nclust
    ))
  })
}

# Collect and format final optimization results

collect_final_results <- function(optimization_results, mae_metric, verbose) {
  
  # Efficient result categorization
  successful_results <- optimization_results[purrr::map_lgl(optimization_results, ~ .x$success %||% FALSE)]
  failed_results <- optimization_results[purrr::map_lgl(optimization_results, ~ !(.x$success %||% FALSE))]
  
  final_output <- list()
  
  if (length(optimization_results) > 0) {
    # Process each metric efficiently
    for (metric in mae_metric) {
      metric_key <- paste0(metric, "_mae")
      
      # Collect successful results with consistent data types (5 columns only)
      all_successful_metric <- purrr::map_dfr(optimization_results, function(x) {
        
        # Get k,nn simulation results
        sim_results <- if (!is.null(x$successful_results) && !is.null(x$successful_results[[metric]])) {
          result <- x$successful_results[[metric]]
          result$k <- as.character(result$k)
          result$`Number of nearest neighbors` <- as.character(result$`Number of nearest neighbors`)
          req <- c("Number of Cells","k","Number of nearest neighbors","mae","status")
          result <- result[, intersect(names(result), req), drop = FALSE]
          for (nm in setdiff(req, names(result))) {
            if (nrow(result) > 0) {
              result[[nm]] <- NA
            } else {
              result[[nm]] <- if (nm %in% c("Number of Cells")) integer(0) else if (nm %in% c("k", "Number of nearest neighbors", "status")) character(0) else numeric(0)
            }
          }
          result[req]
        } else {
          data.frame(
            `Number of Cells` = integer(0),
            k = character(0),
            `Number of nearest neighbors` = character(0),
            mae = numeric(0),
            status = character(0),
            check.names = FALSE
          )
        }
        
        # Get baseline results from all_attempts
        baseline_results <- if (!is.null(x$all_attempts) && !is.null(x$all_attempts[[metric]])) {
          tryCatch({
            if (is.list(x$all_attempts[[metric]]) && !is.data.frame(x$all_attempts[[metric]])) {
              attempts_list <- lapply(x$all_attempts[[metric]], function(df) {
                if (is.data.frame(df) && nrow(df) > 0) {
                  df$k <- as.character(df$k)
                  df$`Number of nearest neighbors` <- as.character(df$`Number of nearest neighbors`)
                }
                df
              })
              attempts_df <- do.call(rbind, attempts_list)
            } else {
              attempts_df <- x$all_attempts[[metric]]
            }
            
            if (is.data.frame(attempts_df) && "status" %in% colnames(attempts_df)) {
              success_rows <- attempts_df[grepl("^Successful Iteration", attempts_df$status), ]
              if (nrow(success_rows) > 0) {
                success_rows$k <- as.character(success_rows$k)
                success_rows$`Number of nearest neighbors` <- as.character(success_rows$`Number of nearest neighbors`)
              }
              req <- c("Number of Cells","k","Number of nearest neighbors","mae","status")
              success_rows <- success_rows[, intersect(names(success_rows), req), drop = FALSE]
              for (nm in setdiff(req, names(success_rows))) success_rows[[nm]] <- NA
              success_rows[req]
            } else {
              data.frame(
                `Number of Cells` = integer(0),
                k = character(0),
                `Number of nearest neighbors` = character(0),
                mae = numeric(0),
                status = character(0),
                check.names = FALSE
              )
            }
          }, error = function(e) {
            data.frame(
              `Number of Cells` = integer(0),
              k = character(0),
              `Number of nearest neighbors` = character(0),
              mae = numeric(0),
              status = character(0),
              check.names = FALSE
            )
          })
        } else {
          data.frame(
            `Number of Cells` = integer(0),
            k = character(0),
            `Number of nearest neighbors` = character(0),
            mae = numeric(0),
            status = character(0),
            check.names = FALSE
          )
        }
        
        # Ensure both sim_results and baseline_results have consistent column types
        if (nrow(sim_results) > 0) {
          sim_results$k <- as.character(sim_results$k)
          sim_results$`Number of nearest neighbors` <- as.character(sim_results$`Number of nearest neighbors`)
        }
        
        if (nrow(baseline_results) > 0) {
          baseline_results$k <- as.character(baseline_results$k) 
          baseline_results$`Number of nearest neighbors` <- as.character(baseline_results$`Number of nearest neighbors`)
        }
        
        # Safe combine with consistent structures and drop duplicates
        combined <- dplyr::bind_rows(sim_results, baseline_results)
        combined <- dplyr::distinct(
          combined,
          .data$`Number of Cells`, .data$k, .data$`Number of nearest neighbors`, .data$mae, .data$status,
          .keep_all = FALSE
        )
        combined
      })
      
      # Sort all_successful_metric by Number of Cells, k, and nn
      if (nrow(all_successful_metric) > 0) {
        # Convert to numeric for proper sorting, handling special cases
        all_successful_metric$k_sort <- suppressWarnings({
          ifelse(is.na(all_successful_metric$k) | all_successful_metric$k == "-", 999, as.numeric(all_successful_metric$k))
        })
        all_successful_metric$nn_sort <- suppressWarnings({
          ifelse(is.na(all_successful_metric$`Number of nearest neighbors`) | 
                 all_successful_metric$`Number of nearest neighbors` == "-", 999, 
                 as.numeric(all_successful_metric$`Number of nearest neighbors`))
        })
        # Replace any NAs from conversion with 999
        all_successful_metric$k_sort[is.na(all_successful_metric$k_sort)] <- 999
        all_successful_metric$nn_sort[is.na(all_successful_metric$nn_sort)] <- 999
        
        all_successful_metric <- all_successful_metric %>%
          dplyr::arrange(.data$`Number of Cells`, .data$k_sort, .data$nn_sort) %>%
          dplyr::select(-.data$k_sort, -.data$nn_sort)
      }
      
      # Collect best results per nclust
      best_results_metric <- purrr::map_dfr(successful_results, function(x) {
        if (!is.null(x$best_result) && !is.null(x$best_result[[metric]])) {
          result <- x$best_result[[metric]]
          result$k <- as.character(result$k)
          result$`Number of nearest neighbors` <- as.character(result$`Number of nearest neighbors`)
          result
        } else {
          data.frame()
        }
      })
      
      # Enhanced all attempts collection with error handling
      all_attempts_metric <- tryCatch({
        purrr::map_dfr(optimization_results, function(x) {
          if (!is.null(x$all_attempts) && !is.null(x$all_attempts[[metric]])) {
            if (length(x$all_attempts[[metric]]) > 0) {
              tryCatch({
                if (is.list(x$all_attempts[[metric]]) && !is.data.frame(x$all_attempts[[metric]])) {
                  attempts_df <- do.call(rbind, x$all_attempts[[metric]])
                } else {
                  attempts_df <- x$all_attempts[[metric]]
                }
                
                if (nrow(attempts_df) > 0) {
                  attempts_df$k <- as.character(attempts_df$k)
                  attempts_df$`Number of nearest neighbors` <- as.character(attempts_df$`Number of nearest neighbors`)
                }
                
                attempts_df
              }, error = function(e) {
                data.frame()
              })
            } else {
              data.frame()
            }
          } else {
            data.frame()
          }
        })
      }, error = function(e) {
        data.frame()
      })
      
      # Sort all results tables by Number of Cells, k, and nn
      if (nrow(all_attempts_metric) > 0) {
        # Convert to numeric for proper sorting, handling special cases
        all_attempts_metric$k_sort <- suppressWarnings({
          ifelse(is.na(all_attempts_metric$k) | all_attempts_metric$k == "-", 999, as.numeric(all_attempts_metric$k))
        })
        all_attempts_metric$nn_sort <- suppressWarnings({
          ifelse(is.na(all_attempts_metric$`Number of nearest neighbors`) | 
                 all_attempts_metric$`Number of nearest neighbors` == "-", 999, 
                 as.numeric(all_attempts_metric$`Number of nearest neighbors`))
        })
        # Replace any NAs from conversion with 999
        all_attempts_metric$k_sort[is.na(all_attempts_metric$k_sort)] <- 999
        all_attempts_metric$nn_sort[is.na(all_attempts_metric$nn_sort)] <- 999
        
        all_attempts_metric <- all_attempts_metric %>%
          dplyr::arrange(.data$`Number of Cells`, .data$k_sort, .data$nn_sort) %>%
          dplyr::select(-.data$k_sort, -.data$nn_sort)
      }
      
      if (nrow(best_results_metric) > 0) {
        # Convert to numeric for proper sorting, handling special cases
        best_results_metric$k_sort <- suppressWarnings({
          ifelse(is.na(best_results_metric$k) | best_results_metric$k == "-", 999, as.numeric(best_results_metric$k))
        })
        best_results_metric$nn_sort <- suppressWarnings({
          ifelse(is.na(best_results_metric$`Number of nearest neighbors`) | 
                 best_results_metric$`Number of nearest neighbors` == "-", 999, 
                 as.numeric(best_results_metric$`Number of nearest neighbors`))
        })
        # Replace any NAs from conversion with 999
        best_results_metric$k_sort[is.na(best_results_metric$k_sort)] <- 999
        best_results_metric$nn_sort[is.na(best_results_metric$nn_sort)] <- 999
        
        best_results_metric <- best_results_metric %>%
          dplyr::arrange(.data$`Number of Cells`, .data$k_sort, .data$nn_sort) %>%
          dplyr::select(-.data$k_sort, -.data$nn_sort)
        req <- c("Number of Cells","k","Number of nearest neighbors","mae","status")
        best_results_metric <- best_results_metric[, intersect(names(best_results_metric), req), drop = FALSE]
        for (nm in setdiff(req, names(best_results_metric))) best_results_metric[[nm]] <- NA
        best_results_metric <- best_results_metric[req]
      }
      
      # Find overall best
      overall_best_metric <- if (nrow(all_successful_metric) > 0) {
        tryCatch({
        best_row <- all_successful_metric %>% 
          dplyr::arrange(.data$mae) %>% 
          dplyr::slice(1)
          best_row
        }, error = function(e) {
          data.frame()
        })
      } else {
        data.frame()
      }
      
      # enforce 5-column shape everywhere
      req <- c("Number of Cells","k","Number of nearest neighbors","mae","status")
      shape <- function(df){
        if (is.null(df) || !nrow(df)) return(data.frame(`Number of Cells`=integer(0),k=character(0),`Number of nearest neighbors`=character(0),mae=numeric(0),status=character(0),check.names=FALSE))
        df <- df[, intersect(names(df), req), drop = FALSE]
        for (nm in setdiff(req, names(df))) df[[nm]] <- NA
        df[req]
      }
      final_output[[metric_key]] <- list(
        successful_results = shape(all_successful_metric),
        nclust_best_results = shape(best_results_metric),
        overall_best = shape(overall_best_metric),
        all_results = shape(all_attempts_metric)
      )
    }
  }
  
  # Print summary if verbose
  if (verbose) {
    cat("Optimization Summary:\n")
    cat("  Total configurations tested:", length(optimization_results), "\n")
    cat("  Successfully processed:", length(successful_results), "\n")
    cat("  Failed:", length(failed_results), "\n")
    
    for (metric in mae_metric) {
      metric_key <- paste0(metric, "_mae")
      all_results <- final_output[[metric_key]]$all_results
      if (nrow(all_results) > 0) {
        successful_count <- sum(grepl("^Successful Iteration", all_results$status), na.rm = TRUE)
        total_count <- nrow(all_results)
        success_rate <- round(100 * successful_count / total_count, 1)
        cat("  ", metric, "success rate:", success_rate, "% (", successful_count, "/", total_count, ")\n")
        
        if (nrow(final_output[[metric_key]]$overall_best) > 0) {
          best <- final_output[[metric_key]]$overall_best
          cat("  Best", metric, "MAE:", round(best$mae[1], 4), 
              "with", best$`Number of Cells`[1], "cells, k =", best$k[1], ", nn =", best$`Number of nearest neighbors`[1], "\n")
        }
      }
    }
  }
  
  return(final_output)
}



# ============================================================================
# HVT PROCESSING FUNCTIONS
score_hvt_model <- function(complete_dataset, hvt_results) {
  scoring <- scoreHVT(
    complete_dataset, hvt_results,
    analysis.plots = FALSE,
    names.column = complete_dataset[, 1]
  )
  
  if (is.null(scoring) || is.null(scoring$scoredPredictedData)) {
    stop("scoreHVT returned NULL or missing scoredPredictedData")
  }
  return(scoring)
}

# Prepare temporal data for MSM
prepare_temporal_data <- function(scoring, complete_dataset, time_column) {
  scored_data <- scoring$scoredPredictedData
  scored_data[[time_column]] <- as.POSIXct(complete_dataset[[time_column]])
  
  # Reorder columns to put time first
  time_col_pos <- which(names(scored_data) == time_column)
  if (length(time_col_pos) == 1) {
    other_cols <- names(scored_data)[-time_col_pos]
    scored_data <- scored_data[, c(time_column, other_cols), drop = FALSE]
  }
  
  temporal_data_all <- scored_data %>% dplyr::select(dplyr::all_of(c(time_column, "Cell.ID")))
  
  if (any(is.na(temporal_data_all[[time_column]]))) {
    stop("Failed to parse time column")
  }
  return(temporal_data_all)
}

# Get transition probabilities from training data ONLY
get_transition_probabilities <- function(temporal_data_train, time_column) {
  
  prob_trans_matx <- getTransitionProbability(
    df = temporal_data_train,  # ONLY training data
    cellid_column = "Cell.ID",
    time_column = time_column,
    type = "with_self_state"
  )
  
  if (is.null(prob_trans_matx) || nrow(prob_trans_matx) == 0) {
    stop("getTransitionProbability (train window) returned NULL or empty matrix")
  }
  
  return(prob_trans_matx)
}

# Get initial state for simulation
get_initial_state <- function(temporal_data_train, time_column) {
  last_t_train <- max(temporal_data_train[[time_column]], na.rm = TRUE)
  initial_state <- temporal_data_train$Cell.ID[temporal_data_train[[time_column]] == last_t_train][1]
  return(initial_state)
}



# Prepare centroid data for MAE calculation
prepare_centroid_data_for_mae <- function(hvt_results, numeric_dataset,
                                          scoring, complete_dataset, raw_dataset_stats,
                                          mae_metric, centroid_2d_points) {
  # Normal flow: align to training variables only
  centroid_data_for_mae <- hvt_results[[3]]$summary %>%
    dplyr::select(dplyr::any_of(colnames(numeric_dataset)))
  
  return(centroid_data_for_mae)
}
# Initialize attempts storage for each metric
initialize_attempts_storage <- function(mae_metric) {
  all_attempts_by_metric <- list()
  for (metric in mae_metric) {
    all_attempts_by_metric[[metric]] <- list()
  }
  return(all_attempts_by_metric)
}

# ============================================================================
# SIMULATION FUNCTIONS
# ============================================================================

# Run MSM simulation directly for optimization - MSM handles all validation
run_single_simulation <- function(k_val, n_nn, cluster_data, transition_matrix,
                                  temporal_data_train, scoring, hvt_results, initial_state,
                                  problematic_states, centroid_data_for_mae, mae_metrics,
                                  raw_dataset_stats, time_column, entire_dataset,
                                  expost_forecasting, num_simulations, time_values, n_ahead) {
  
  # Create temporal data properly for MSM
  scored_data <- scoring$scoredPredictedData
  complete_dataset <- rbind(entire_dataset, expost_forecasting)
  scored_data[[time_column]] <- as.POSIXct(complete_dataset[[time_column]])
  temporal_data_all <- scored_data %>% dplyr::select(dplyr::all_of(c(time_column, "Cell.ID")))
  
  tryCatch({
    # Set optimization flag for hvt_msm_opt
    options(hvt.msm.optimization = TRUE)
    on.exit(options(hvt.msm.optimization = FALSE), add = TRUE)
    
    # Let MSM handle ALL validation and clustering - no pre-validation
    msm_result <- msm(
      state_time_data = temporal_data_all,
      forecast_type = "ex-post",
      transition_probability_matrix = transition_matrix,
      initial_state = initial_state,
      num_simulations = num_simulations,
      scoreHVT_results = scoring,
      trainHVT_results = hvt_results,
      actual_data = expost_forecasting,
      raw_dataset = complete_dataset,
      k = k_val,
      handle_problematic_states = length(problematic_states) > 0,
      n_nearest_neighbor = n_nn,
      mae_metric = mae_metrics[1],
      show_simulation = FALSE,
      time_column = time_column,
      precomputed_problematic_states = problematic_states,
      plot_mode = "mae-only"
    )
    
    # Extract MAE from MSM results
    mae_results <- extract_mae_from_msm_results(msm_result, mae_metrics)
    
    # Cleanup heavy objects
    rm(msm_result, temporal_data_all, scored_data)
    gc(verbose = FALSE)
    
    # Return success with models
    if (!is.null(mae_results) && !is.na(mae_results[[mae_metrics[1]]])) {
      return(list(
        success = TRUE,
        k = k_val,
        nn = n_nn,
        mae_results = mae_results
        ))
    } else {
      return(list(success = FALSE, k = k_val, nn = n_nn, error = "MAE extraction failed"))
    }
    
  }, error = function(e) {
    # MSM failed (could be validation, clustering, or simulation failure)
    gc(verbose = FALSE)
    return(list(success = FALSE, k = k_val, nn = n_nn, error = as.character(e$message)))
  })
}

# Extract MAE values from MSM results
extract_mae_from_msm_results <- function(msm_result, mae_metrics) {
  mae_results <- list()
  
  # New compact structure from mae-only path
  if (!is.null(msm_result$mae_by_metric)) {
    for (metric in mae_metrics) {
      mm <- msm_result$mae_by_metric[[metric]]
      if (!is.null(mm)) {
        centroid_maes <- sapply(mm$centroid_mae_list, function(x) x[["mae"]])
        mae_results[[metric]] <- mean(centroid_maes, na.rm = TRUE)
      } else {
        mae_results[[metric]] <- NA
      }
    }
    return(mae_results)
  }
  
  # Backward-compatible path (plots-based)
  for (metric in mae_metrics) {
    if (!is.null(msm_result$plots) && length(msm_result$plots) >= 2) {
      centroid_maes <- sapply(msm_result$plots[[1]], function(x) x[["mae"]])
      mae_results[[metric]] <- mean(centroid_maes, na.rm = TRUE)
    } else {
      mae_results[[metric]] <- NA
    }
  }
  mae_results
}

# Run baseline simulation using MSM without problematic state handling
run_baseline_simulation <- function(transition_matrix, temporal_data_train, scoring, hvt_results,
                                    initial_state, centroid_data_for_mae, mae_metrics, raw_dataset_stats,
                                    time_column, entire_dataset, expost_forecasting, num_simulations,
                                    time_values, n_ahead) {
  
  # Create temporal data for MSM
  scored_data <- scoring$scoredPredictedData
  complete_dataset <- rbind(entire_dataset, expost_forecasting)
  scored_data[[time_column]] <- as.POSIXct(complete_dataset[[time_column]])
  temporal_data_all <- scored_data %>% 
    dplyr::select(dplyr::all_of(c(time_column, "Cell.ID")))
  
  tryCatch({
    # Set optimization flag for hvt_msm_opt
    options(hvt.msm.optimization = TRUE)
    on.exit(options(hvt.msm.optimization = FALSE), add = TRUE)
    
      # Baseline MSM call: no problematic state handling
      msm_result <- msm(
      state_time_data = temporal_data_all,
      forecast_type = "ex-post",
      transition_probability_matrix = transition_matrix,
      initial_state = initial_state,
      num_simulations = num_simulations,
      scoreHVT_results = scoring,
      trainHVT_results = hvt_results,
      actual_data = expost_forecasting,
      raw_dataset = complete_dataset,
      handle_problematic_states = FALSE,  # Baseline = no problematic state handling
      mae_metric = mae_metrics[1],
      show_simulation = FALSE,
      time_column = time_column,
      plot_mode = "mae-only"
    )
    
    # Extract MAE from MSM results
    mae_results <- extract_mae_from_msm_results(msm_result, mae_metrics)
    
    # Cleanup
    rm(msm_result, temporal_data_all, scored_data)
    gc(verbose = FALSE)
    
    return(mae_results)
    
  }, error = function(e) {
    # Return NA for all metrics if baseline fails
    mae_results <- list()
    for (metric in mae_metrics) {
      mae_results[[metric]] <- NA
    }
    return(mae_results)
  })
}

# Process all k-nn combinations with pre-filtering checkpoint for invalid combinations
process_all_k_nn_combinations <- function(k_local_range, nn_range, problematic_states,
                                          centroid_2d_points, clustering_data, hvt_results,
                                          scoring, nclust, mae_metric, all_attempts_by_metric,
                                          prob_trans_matx, temporal_data_complete, initial_state, 
                                          centroid_data_for_mae, raw_dataset_stats, time_column,
                                          entire_dataset, expost_forecasting, num_simulations,
                                          time_values, n_ahead, parallel,
                                          verbose = FALSE) {
  
  # Generate ALL possible k-nn combinations
  all_combinations <- expand.grid(k = k_local_range, nn = nn_range)
  
  if (verbose) {
    cat("  Total k-nn combinations:", nrow(all_combinations), "\n")
  }
  
  # CHECKPOINT: Pre-filter mathematically invalid combinations
  valid_combinations <- list()
  invalid_combinations <- list()
  
  for (i in seq_len(nrow(all_combinations))) {
    k_val <- all_combinations$k[i]
    nn_val <- all_combinations$nn[i]
    
    # Check 1: k must be less than or equal to number of cells
    if (k_val > nclust) {
      invalid_combinations[[length(invalid_combinations) + 1]] <- list(
        k = k_val,
        nn = nn_val,
        status = paste0("Mathematical bound reached - Invalid k (", k_val, ") for given n_cells (", nclust, ")")
      )
      next
    }
    
    # Check 2: nn must be <= (cells - k) [mathematical upper bound]
    max_possible_nn <- nclust - k_val
    if (nn_val > max_possible_nn) {
      invalid_combinations[[length(invalid_combinations) + 1]] <- list(
        k = k_val,
        nn = nn_val,
        status = paste0("Mathematical bound reached - Invalid nn (", nn_val, ") for given n_cells (", nclust, ") and k (", k_val, ")")
      )
      next
    }
    
    # Passed both checks - valid combination
    valid_combinations[[length(valid_combinations) + 1]] <- list(
      k = k_val,
      nn = nn_val,
      status = "ready_for_msm",
      valid = TRUE
    )
  }
  
  # Store invalid combinations in all_attempts_by_metric
  if (length(invalid_combinations) > 0) {
    for (combo in invalid_combinations) {
      for (metric in mae_metric) {
        idx <- length(all_attempts_by_metric[[metric]]) + 1
        all_attempts_by_metric[[metric]][[idx]] <- data.frame(
          `Number of Cells` = nclust,
          k = as.character(combo$k),
          `Number of nearest neighbors` = as.character(combo$nn),
          mae = NA,
          status = combo$status,  # Already contains "ERROR: " prefix
          stringsAsFactors = FALSE,
          check.names = FALSE
        )
      }
    }
  }
  
  if (verbose) {
    cat("  Valid combinations:", length(valid_combinations), "\n")
    cat("  Pre-filtered (invalid):", length(invalid_combinations), "\n")
  }
  
  # Run MSM simulations for VALID combinations only (invalid ones already filtered)
  simulation_results <- list()
  
  if (length(valid_combinations) > 0) {
    # Simulation function for valid combinations
    simulation_function <- function(combo) {
      # Call run_single_simulation which calls MSM
      result <- run_single_simulation(
        k_val = combo$k, n_nn = combo$nn, cluster_data = NULL,
        transition_matrix = prob_trans_matx, temporal_data_train = NULL,
        scoring = scoring, hvt_results = hvt_results, initial_state = initial_state,
        problematic_states = problematic_states, centroid_data_for_mae = NULL,
        mae_metrics = mae_metric, raw_dataset_stats = NULL,
        time_column = time_column, entire_dataset = entire_dataset,
        expost_forecasting = expost_forecasting, num_simulations = num_simulations,
        time_values = NULL, n_ahead = NULL
      )
      return(result)
    }
    
    # Execute ALL combinations
    if (parallel) {
      if (!requireNamespace("furrr", quietly = TRUE) || !requireNamespace("future", quietly = TRUE)) {
        stop("Packages 'furrr' and 'future' are required for parallel processing. ",
             "Please install them with: install.packages(c('furrr', 'future'))")
      }
      simulation_results <- furrr::future_map(valid_combinations, simulation_function,
                                       .options = furrr::furrr_options(seed = TRUE, packages = c("dplyr", "data.table")))
    } else {
      simulation_results <- purrr::map(valid_combinations, simulation_function)
    }
    
    if (verbose) {
      successful_count <- sum(sapply(simulation_results, function(x) x$success %||% FALSE))
      cat("  Completed:", length(simulation_results), "simulations,", successful_count, "successful\n")
    }
  }
  
  return(list(simulation_results = simulation_results, all_attempts = all_attempts_by_metric))
}

# Process simulation results - lightweight version
process_simulation_results <- function(simulation_results, mae_metric, nclust, all_attempts_by_metric) {
  
  if (length(simulation_results) == 0) {
    formatted_results <- list()
    best_results <- list()
    for (metric in mae_metric) {
      formatted_results[[metric]] <- data.frame()
      best_results[[metric]] <- data.frame()
    }
    return(list(
      successful_results = formatted_results,
      best_result = best_results,
      all_attempts = all_attempts_by_metric
    ))
  }
  
  formatted_results <- list()
  best_results <- list()
  successful_results_list <- list()
  
  # Initialize lists for each metric
  for (metric in mae_metric) {
    successful_results_list[[metric]] <- list()
  }
  
  # Process each simulation result
  for (result in simulation_results) {
    if (result$success && !is.null(result$mae_results)) {
      # Process successful results
      for (metric in mae_metric) {
        # Add to all attempts
        idx <- length(all_attempts_by_metric[[metric]]) + 1
        all_attempts_by_metric[[metric]][[idx]] <- data.frame(
          `Number of Cells` = nclust, 
          k = as.character(result$k), 
          `Number of nearest neighbors` = as.character(result$nn), 
          mae = round(result$mae_results[[metric]], 4), 
          status = "Successful Iteration by Handling Problematic state", 
          stringsAsFactors = FALSE, 
          check.names = FALSE
        )
        
        # Add to successful results WITHOUT models
        success_idx <- length(successful_results_list[[metric]]) + 1
        successful_results_list[[metric]][[success_idx]] <- data.frame(
          `Number of Cells` = nclust, 
          k = as.character(result$k), 
          `Number of nearest neighbors` = as.character(result$nn), 
          mae = round(result$mae_results[[metric]], 4),
          status = "Successful Iteration by Handling Problematic state",
          stringsAsFactors = FALSE,
          check.names = FALSE
        )
      }
    } else {
      # Process failed results
      for (metric in mae_metric) {
        idx <- length(all_attempts_by_metric[[metric]]) + 1
        status_msg <- if (!is.null(result$error)) {
          result$error  # Message comes as-is from MSM (already formatted)
        } else {
          "ERROR: MSM failed"
        }
        all_attempts_by_metric[[metric]][[idx]] <- data.frame(
          `Number of Cells` = nclust, 
          k = as.character(result$k), 
          `Number of nearest neighbors` = as.character(result$nn), 
          mae = NA, 
          status = status_msg, 
          stringsAsFactors = FALSE, 
          check.names = FALSE
        )
      }
    }
  }
  
  # Format successful results and find best combinations WITHOUT models
  for (metric in mae_metric) {
    if (length(successful_results_list[[metric]]) > 0) {
      # Extract configs and de-duplicate strictly on the 5 columns
      configs <- purrr::map_dfr(successful_results_list[[metric]], ~ .x)
      configs <- dplyr::distinct(configs, `Number of Cells`, k, `Number of nearest neighbors`, mae, status, .keep_all = FALSE)
      
      # Ensure numeric conversion for sorting
      configs$k_numeric <- as.numeric(configs$k)
      configs$nn_numeric <- as.numeric(configs$`Number of nearest neighbors`)
      configs$mae_numeric <- as.numeric(configs$mae)
      
      # Deduplicate and sort, keep only required 5 cols
      configs <- dplyr::distinct(configs, .data$`Number of Cells`, .data$k, .data$`Number of nearest neighbors`, .data$mae, .keep_all = TRUE)
      sort_order <- order(configs$`Number of Cells`, configs$k_numeric, configs$nn_numeric)
      formatted <- configs[sort_order, c("Number of Cells","k","Number of nearest neighbors","mae","status")]
      formatted_results[[metric]] <- formatted
      
      # Find best combination WITHOUT storing models
      best_idx <- which.min(configs$mae_numeric)
      best_combination <- configs[best_idx, ] %>%
        dplyr::select(-.data$k_numeric, -.data$nn_numeric, -.data$mae_numeric)
      best_results[[metric]] <- best_combination[, c("Number of Cells","k","Number of nearest neighbors","mae","status")]
      
    } else {
      formatted_results[[metric]] <- data.frame()
      best_results[[metric]] <- data.frame()
    }
  }
  
  # Sort all attempts
  for (metric in mae_metric) {
    if (length(all_attempts_by_metric[[metric]]) > 0) {
      combined_attempts <- do.call(rbind, all_attempts_by_metric[[metric]])
      
      # Sort by cells, k, nn
      combined_attempts$k_sort <- ifelse(combined_attempts$k == "-", 999, as.numeric(combined_attempts$k))
      combined_attempts$nn_sort <- ifelse(combined_attempts$`Number of nearest neighbors` == "-", 999, 
                                          as.numeric(combined_attempts$`Number of nearest neighbors`))
      
      combined_attempts <- combined_attempts %>%
        dplyr::arrange(.data$`Number of Cells`, .data$k_sort, .data$nn_sort) %>%
        dplyr::select(-.data$k_sort, -.data$nn_sort)
      
      # Convert back to list format
      all_attempts_by_metric[[metric]] <- split(combined_attempts, seq_len(nrow(combined_attempts)))
      all_attempts_by_metric[[metric]] <- lapply(all_attempts_by_metric[[metric]], as.data.frame)
    }
  }
  
  gc(verbose = FALSE)
  
  return(list(
    successful_results = formatted_results,
    best_result = best_results,
    all_attempts = all_attempts_by_metric
  ))
}

# ============================================================================
# PROBLEMATIC STATE DETECTION FUNCTIONS
# ============================================================================

# Detect all types of problematic states in a transition matrix
detect_problematic_states <- function(prob_trans_matx, scored_cells) {
  # Missing states: states in scored_cells but not in transition matrix
  transition_states <- unique(prob_trans_matx$Current_State)
  missing_states <- setdiff(scored_cells, transition_states)
  
  # Self-transition states (100% self-transition)
  self_transition_states <- prob_trans_matx %>%
    dplyr::group_by(.data$Current_State) %>%
    dplyr::summarize(
      is_self_only = dplyr::n() == 1 && all(.data$Current_State == .data$Next_State),
      .groups = 'drop'
    ) %>%
    dplyr::filter(.data$is_self_only == TRUE) %>%
    dplyr::pull(.data$Current_State)
  
  # Cyclic states detection
  cyclic_states <- find_cyclic_states(prob_trans_matx)
  
  # Combine all problematic states
  all_problematic <- unique(c(missing_states, self_transition_states, cyclic_states))
  
  return(list(
    missing_states = missing_states,
    self_transition_states = self_transition_states,
    cyclic_states = cyclic_states,
    all_problematic = all_problematic
  ))
}

# Detect cyclic states in a transition matrix
find_cyclic_states <- function(trans_table) {
  all_states <- unique(c(trans_table$Current_State, trans_table$Next_State))
  state_info <- trans_table %>%
    dplyr::group_by(.data$Current_State) %>%
    dplyr::summarize(next_states = list(unique(.data$Next_State)), .groups = 'drop')
  adjacency <- setNames(lapply(state_info$next_states, unlist), as.character(state_info$Current_State))
  two_way_cycles <- c(); one_way_traps <- c(); processed_pairs <- character(0)
  for (state_a in all_states) {
    if (!as.character(state_a) %in% names(adjacency)) next
    next_states_a <- adjacency[[as.character(state_a)]]
    other_states_a <- next_states_a[next_states_a != state_a]
    for (state_b in other_states_a) {
      pair_key <- paste(sort(c(state_a, state_b)), collapse = "-")
      if (pair_key %in% processed_pairs) next
      processed_pairs <- c(processed_pairs, pair_key)
      if (!as.character(state_b) %in% names(adjacency)) next
      next_states_b <- adjacency[[as.character(state_b)]]
      states_in_system <- c(state_a, state_b)
      reaches_outside_a <- setdiff(next_states_a, states_in_system)
      reaches_outside_b <- setdiff(next_states_b, states_in_system)
      if (length(reaches_outside_a) == 0 && length(reaches_outside_b) == 0) {
        a_can_reach_b <- state_b %in% next_states_a
        b_can_reach_a <- state_a %in% next_states_b
        if (a_can_reach_b && b_can_reach_a) two_way_cycles <- c(two_way_cycles, state_a, state_b)
        else if (a_can_reach_b && !b_can_reach_a) {
          if (length(next_states_b) == 1 && next_states_b[1] == state_b) one_way_traps <- c(one_way_traps, state_a, state_b)
        } else if (!a_can_reach_b && b_can_reach_a) {
          if (length(next_states_a) == 1 && next_states_a[1] == state_a) one_way_traps <- c(one_way_traps, state_a, state_b)
        }
      }
    }
  }
  unique(c(two_way_cycles, one_way_traps))
}

# ============================================================================
# MISSING MSM HELPER FUNCTIONS
# ============================================================================

# Validate MSM clustering for problematic states
validate_msm_clustering <- function(cluster_data, problematic_states, n_nearest_neighbor,
                                    centroid_2d_points, verbose = TRUE) {
  if (length(problematic_states) == 0) {
    return(list(is_valid = TRUE, message = "No problematic states found"))
  }
  validation_issues <- c()
  for (prob_state in problematic_states) {
    prob_cluster <- cluster_data$cluster[cluster_data$Cell.ID == prob_state]
    cluster_members <- cluster_data$Cell.ID[cluster_data$cluster == prob_cluster]
    valid_neighbors <- setdiff(cluster_members, c(prob_state, problematic_states))
    if (verbose) {
      cat("Problematic state", prob_state, "in cluster", prob_cluster,
          "- Valid neighbors:", length(valid_neighbors), "\n")
    }
    if (length(valid_neighbors) == 0) {
      validation_issues <- c(validation_issues,
                             paste0("State ", prob_state, " has no valid neighbors in cluster ", prob_cluster))
    } else if (length(valid_neighbors) < n_nearest_neighbor) {
      validation_issues <- c(validation_issues,
                             paste0("State ", prob_state, " has only ", length(valid_neighbors),
                                    " valid neighbors, but ", n_nearest_neighbor, " requested"))
    }
  }
  if (length(validation_issues) > 0) {
    return(list(
      is_valid = FALSE,
      message = paste(validation_issues, collapse = "; ")
    ))
  }
  list(is_valid = TRUE, message = "MSM clustering validation passed")
}

# Choose a nearest non-problematic neighbor for a problematic state
find_nearest_neighbor_enhanced <- function(problematic_state,
                                           centroid_2d_points,
                                           cluster_data,
                                           problematic_states,
                                           n_nearest_neighbor,
                                           scored_cells,
                                           initial_state) {
  if (is.null(cluster_data)) {
    valid_states <- setdiff(scored_cells, problematic_states)
    if (length(valid_states) > 0) return(sample(valid_states, 1))
    warning("No valid fallback states available, returning initial state")
    return(initial_state)
  }
  current_cluster <- cluster_data$cluster[cluster_data$Cell.ID == problematic_state]
  coords <- centroid_2d_points[centroid_2d_points$Cell.ID %in%
                                 cluster_data$Cell.ID[cluster_data$cluster == current_cluster], ]
  current_coords <- coords[coords$Cell.ID == problematic_state, c("x", "y")]
  distances <- sqrt((coords$x - current_coords$x)^2 + (coords$y - current_coords$y)^2)
  valid_neighbors <- coords$Cell.ID[!coords$Cell.ID %in% c(problematic_state, problematic_states)]
  if (length(valid_neighbors) == 0) {
    warning(paste0("No valid neighbors for problematic state ", problematic_state,
                   " in cluster ", current_cluster, ". Using global fallback."))
    all_coords <- centroid_2d_points
    all_distances <- sqrt((all_coords$x - current_coords$x)^2 +
                            (all_coords$y - current_coords$y)^2)
    global_valid <- setdiff(all_coords$Cell.ID, c(problematic_state, problematic_states))
    if (length(global_valid) > 0) {
      global_neighbor_distances <- all_distances[all_coords$Cell.ID %in% global_valid]
      closest_global <- global_valid[which.min(global_neighbor_distances)]
      return(closest_global)
    } else {
      warning("No valid neighbors globally available, returning initial state")
      return(initial_state)
    }
  }
  neighbor_distances <- distances[coords$Cell.ID %in% valid_neighbors]
  num_neighbors <- min(length(valid_neighbors), n_nearest_neighbor)
  nearest_n <- valid_neighbors[order(neighbor_distances)][1:num_neighbors]
  neighbor_probs <- 1/neighbor_distances[order(neighbor_distances)][1:num_neighbors]
  neighbor_probs <- neighbor_probs/sum(neighbor_probs)
  random_shock <- round(runif(min=0, max=1, n=1), 4)
  cumulative_probs <- cumsum(neighbor_probs)
  nearest_n[which.max(random_shock <= cumulative_probs)]
}

# Single simulation step for MSM
simulate_step_enhanced <- function(i, current_state, random_val,
                                   transition_probability_matrix,
                                   centroid_2d_points,
                                   cluster_data,
                                   problematic_states,
                                   n_nearest_neighbor,
                                   scored_cells,
                                   initial_state) {
  random_shock <- if (is.null(random_val)) round(runif(min=0, max=1, n=1), 4) else random_val
  if (i == 1) return(initial_state)
  
  # Handle problematic states only if handle_problematic_states = TRUE (indicated by cluster_data being non-NULL)
  # If cluster_data is NULL, just follow the transition matrix normally, even for problematic states
  if (!is.null(cluster_data) && current_state %in% problematic_states) {
    return(find_nearest_neighbor_enhanced(current_state, centroid_2d_points, cluster_data,
                                          problematic_states, n_nearest_neighbor, scored_cells, initial_state))
  }
  transition <- subset(transition_probability_matrix, Current_State == current_state)
  if (nrow(transition) == 0) {
    if (current_state %in% scored_cells) return(current_state)
    else return(find_nearest_neighbor_enhanced(current_state, centroid_2d_points, cluster_data,
                                               problematic_states, n_nearest_neighbor, scored_cells, initial_state))
  }
  transition <- transition[order(-transition$Transition_Probability), ]
  transition$cumulative_prob <- cumsum(transition$Transition_Probability)
  next_state <- transition$Next_State[which.max(random_shock <= transition$cumulative_prob)]
  next_state <- as.numeric(next_state)
  
  # Replace problematic states only if handle_problematic_states = TRUE (cluster_data is non-NULL)
  # If handle_problematic_states = FALSE, follow transition matrix even if it leads to problematic states
  if (!is.null(cluster_data) && next_state %in% problematic_states) {
    return(find_nearest_neighbor_enhanced(next_state, centroid_2d_points, cluster_data,
                                          problematic_states, n_nearest_neighbor, scored_cells, initial_state))
  }
  # Always replace invalid states (not in scored_cells)
  if (!next_state %in% scored_cells) {
    return(find_nearest_neighbor_enhanced(next_state, centroid_2d_points, cluster_data,
                                          problematic_states, n_nearest_neighbor, scored_cells, initial_state))
  }
  next_state
}

# Simulate a full MSM sequence
simulate_sequence_enhanced <- function(sim_index, n_ahead,
                                       initial_state, transition_probability_matrix,
                                       centroid_2d_points, cluster_data, problematic_states,
                                       n_nearest_neighbor, scored_cells) {
  current_state <- initial_state
  simulated_values <- sapply(1:n_ahead, function(i) {
    random_val <- NULL
    current_state <<- simulate_step_enhanced(i, current_state, random_val,
                                            transition_probability_matrix,
                                            centroid_2d_points, cluster_data,
                                            problematic_states, n_nearest_neighbor,
                                            scored_cells, initial_state)
    current_state
  })
  as.numeric(simulated_values)
}

# Adjust invalid initial state using neighbors
adjust_initial_state <- function(initial_state, transition_cells, problematic_states,
                                 scored_cells, centroid_2d_points, cluster_data,
                                 n_nearest_neighbor) {
  if (!(initial_state %in% transition_cells) || initial_state %in% problematic_states) {
    return(find_nearest_neighbor_enhanced(initial_state, centroid_2d_points, cluster_data,
                                          problematic_states, n_nearest_neighbor,
                                          scored_cells, initial_state))
  }
  initial_state
}

#' @keywords internal
#' @include constants.R

clusterPlot <- function(dataset, hvt.results, domains.column, highlight_cells = NULL,
                        cell_id = TRUE,
                        cell_id_position = "center",
                        centroids = TRUE,
                        cell_id_size = 4,
                        centroid.size = 1,
                        centroid.color = "black") {
  suppressWarnings({
    
    # ============================================================================
    # CONSTANTS
    # ============================================================================
    LINE_SIZE_NORMAL <- 0.5
    LINE_SIZE_HIGHLIGHT <- 1.5
    WRAP_LINE_LENGTH <- 100
    MAX_TEXT_CHARS <- 500
    CENTROID_BASE_SIZE <- 1.5
    CELL_ID_LABEL_YSHIFT <- -10
    CELL_ID_LABEL_SIZE <- 10
    
    
    # ============================================================================
    # HELPER FUNCTIONS
    # ============================================================================
    
    # Text wrapping function
    wrap_and_limit_text <- function(text, line_length = WRAP_LINE_LENGTH, 
                                     max_chars = MAX_TEXT_CHARS) {
      if (nchar(text) > max_chars) {
        text <- substr(text, 1, max_chars - 3)
        text <- paste0(text, "...")
      }
      wrapped <- sapply(seq(1, nchar(text), line_length), function(i) {
        substr(text, i, min(i + line_length - 1, nchar(text)))
      })
      paste(wrapped, collapse = "<br>")
    }
    
    # Add hover text to dataframe
    add_hover_text <- function(dataframe) {
      if (nrow(dataframe) == 0) {
        dataframe$hoverText <- NULL
        return(dataframe)
      }
      
      dataframe$hoverText <- apply(dataframe, 1, function(row) {
        cell_id <- row["Cell.ID"]
        full_names <- row["names.column"]
        formatted_names <- wrap_and_limit_text(full_names)
        paste("Cell.ID:", cell_id, "\nObservations:", formatted_names)
      })
      
      return(dataframe)
    }
    
    # Add polygon layer to plot
    add_polygon_layer <- function(plot, data, fill_colors, border_color, 
                                   line_size, hovertext_col) {
      if (nrow(data) == 0) {
        return(plot)
      }
      
      plot + ggplot2::geom_polygon(
        data = data,
        mapping = ggplot2::aes(
          x = x,
          y = y,
          group = interaction(depth, cluster, child),
          fill = fill_colors,
          text = hovertext_col
        ),
        colour = border_color,
        size = line_size,
        tooltip = "text",
        show.legend = TRUE
      )
    }
    
    # Extract color palette with sufficient colors
    get_color_palette <- function(n_needed, base_palette) {
      if (n_needed <= length(base_palette)) {
        return(base_palette[1:n_needed])
      }
      
      # Use colorRampPalette for smooth color generation
      color_func <- grDevices::colorRampPalette(base_palette)
      return(color_func(n_needed))
    }
    
    # Clean legend labels
    clean_legend_label <- function(label) {
      gsub("^\\(1,|\\)$", "", label)
    }
    
    # Safely modify plotly trace properties
    modify_plotly_trace <- function(trace) {
      # Clean legend name
      if (!is.null(trace$name)) {
        trace$name <- clean_legend_label(trace$name)
      }
      
      # Modify marker properties
      if (!is.null(trace$marker)) {
        trace$marker$line$width <- 0
        if (!is.null(trace$showlegend) && trace$showlegend) {
          trace$marker$line$color <- "rgba(0,0,0,0)"
        }
      }
      
      return(trace)
    }
    
    # Extract cluster information from nested list
    extract_cluster_info <- function(hvt_list, depth, cluster_no, child_no, hvt_res2) {
      current_cluster <- hvt_list[[2]][[depth]][[cluster_no]][[child_no]]
      
      list(
        x = as.numeric(current_cluster[["x"]]),
        y = as.numeric(current_cluster[["y"]]),
        x_cor = as.numeric(current_cluster[["pt"]][["x"]]),
        y_cor = as.numeric(current_cluster[["pt"]][["y"]]),
        cell_id = hvt_res2[child_no],
        depth = depth,
        cluster = cluster_no,
        child = child_no
      )
    }
    
    # ============================================================================
    # DATA EXTRACTION AND PROCESSING
    # ============================================================================
    
    # Extract HVT results
    hvt_list <- hvt.results
    hvt_res1 <- hvt_list[[2]][[1]]$`1`
    hvt_res2 <- hvt_list[[3]]$summary$Cell.ID
    a <- seq_along(hvt_res1)
    b <- a[hvt_res2]
    b <- as.vector(b)
    hvt_res2 <- stats::na.omit(b)
    
    # Extract coordinates
    coordinates_value1 <- lapply(seq_along(hvt_res1), function(x) {
      hvt_res1[[x]]$pt
    })
    cellID_coordinates <- do.call(rbind.data.frame, coordinates_value1)
    colnames(cellID_coordinates) <- c("x", "y")
    cellID_coordinates$Cell.ID <- hvt_res2
    cellID_coordinates <- cellID_coordinates %>% arrange(Cell.ID)


    maxDepth <- 1
    
    # ============================================================================
    # CALCULATE BOUNDARIES (for potential future use)
    # ============================================================================
    
    # Note: boundaries calculation kept for potential future boundary checks
    # Currently not used in plotting as ggplot handles scaling automatically
    
    # ============================================================================
    # EXTRACT CLUSTER INFORMATION
    # ============================================================================
    
    # Pre-calculate total number of iterations for vector pre-allocation
    total_cells <- sum(sapply(seq_len(maxDepth), function(depth) {
      sum(sapply(seq_along(hvt_list[[2]][[depth]]), function(cluster_no) {
        length(hvt_list[[2]][[depth]][[cluster_no]])
      }))
    }))
    
    # Initialize lists for collecting data
    cluster_info_list <- vector("list", total_cells)
    idx <- 1
    
    # Extract all cluster information
    for (depth in seq_len(maxDepth)) {
      for (cluster_no in seq_along(hvt_list[[2]][[depth]])) {
        for (child_no in seq_along(hvt_list[[2]][[depth]][[cluster_no]])) {
          cluster_info_list[[idx]] <- extract_cluster_info(
            hvt_list, depth, cluster_no, child_no, hvt_res2
          )
          idx <- idx + 1
        }
      }
    }
    
    # Convert collected data to vectors
    cell_info <- list(
      depthVal = sapply(cluster_info_list, `[[`, "depth"),
      clusterVal = sapply(cluster_info_list, `[[`, "cluster"),
      childVal = sapply(cluster_info_list, `[[`, "child"),
      cell_ids = sapply(cluster_info_list, `[[`, "cell_id"),
      x_cor = sapply(cluster_info_list, `[[`, "x_cor"),
      y_cor = sapply(cluster_info_list, `[[`, "y_cor")
    )
    
    # Create position data
    position_info <- lapply(cluster_info_list, function(info) {
      n_points <- length(info$x)
      data.frame(
        depth = rep(info$depth, n_points),
        cluster = rep(info$cluster, n_points),
        child = rep(info$child, n_points),
        x = info$x,
        y = info$y,
        stringsAsFactors = FALSE
      )
    })
    positionsDataframe <- do.call(rbind, position_info)
    
    # ============================================================================
    # CREATE DATAFRAMES
    # ============================================================================
    
    valuesDataframe <- data.frame(
      depth = cell_info$depthVal,
      cluster = cell_info$clusterVal,
      child = cell_info$childVal,
      cellid = cell_info$cell_ids,
      stringsAsFactors = FALSE
    )
    
    centroidDataframe <- data.frame(
      x = cell_info$x_cor,
      y = cell_info$y_cor,
      lev = cell_info$depthVal,
      stringsAsFactors = FALSE
    )
    
    # Merge dataframes
    datapoly <- merge(valuesDataframe, positionsDataframe, 
                      by = c("depth", "cluster", "child"))
    colnames(datapoly) <- c("depth", "cluster", "child", "Cell.ID", "x", "y")
    datapoly <- merge(datapoly, dataset, by = c("Cell.ID"))
    
    # Process centroid dataframes
    centroidDataframe <- centroidDataframe %>% 
      cbind(cellID_coordinates$Cell.ID)
    names(centroidDataframe) <- c("x", "y", "lev", "Cell.ID")
    
    dataset_new <- dataset %>% dplyr::select('Cell.ID', 'names.column')
    centroidDataframe <- merge(centroidDataframe, dataset_new, by = 'Cell.ID')
    
    centroidDataframe_2 <- merge(cellID_coordinates, centroidDataframe[, c("x", "y", "lev")], 
                                 by = c("x", "y"))
    centroidDataframe_2 <- centroidDataframe_2 %>% 
      cbind(centroidDataframe$names.column)
    colnames(centroidDataframe_2) <- c("x", "y", "Cell.ID", "lev", "names.column")
    
    # Helper function to calculate geometric centroid of a polygon
    calculate_polygon_centroid <- function(x_coords, y_coords) {
      n <- length(x_coords)
      if (n < 3) {
        return(list(x = mean(x_coords), y = mean(y_coords)))
      }
      if (x_coords[1] != x_coords[n] || y_coords[1] != y_coords[n]) {
        x_coords <- c(x_coords, x_coords[1])
        y_coords <- c(y_coords, y_coords[1])
        n <- n + 1
      }
      signed_area <- 0
      cx <- 0
      cy <- 0
      for (i in 1:(n - 1)) {
        cross_product <- x_coords[i] * y_coords[i + 1] - x_coords[i + 1] * y_coords[i]
        signed_area <- signed_area + cross_product
        cx <- cx + (x_coords[i] + x_coords[i + 1]) * cross_product
        cy <- cy + (y_coords[i] + y_coords[i + 1]) * cross_product
      }
      signed_area <- signed_area / 2
      if (abs(signed_area) < 1e-10) {
        return(list(x = mean(x_coords[1:(n-1)]), y = mean(y_coords[1:(n-1)])))
      }
      centroid_x <- cx / (6 * signed_area)
      centroid_y <- cy / (6 * signed_area)
      return(list(x = centroid_x, y = centroid_y))
    }
    
    # Calculate polygon centroids for cell ID positioning (when cell_id_position == "center")
    polygon_centroids_df <- data.frame(Cell.ID = numeric(0), x = numeric(0), y = numeric(0))
    if (cell_id == TRUE && cell_id_position == "center") {
      polygon_centroids_list <- list()
      for (depth_val in 1:maxDepth) {
        depth_data <- datapoly[datapoly$depth == depth_val, ]
        if (nrow(depth_data) > 0) {
          polygon_centroids <- depth_data %>%
            dplyr::group_by(depth, cluster, child) %>%
            dplyr::summarise(
              x = calculate_polygon_centroid(x, y)$x,
              y = calculate_polygon_centroid(x, y)$y,
              .groups = 'drop'
            )
          polygon_centroids_list[[depth_val]] <- polygon_centroids
        }
      }
      if (length(polygon_centroids_list) > 0) {
        polygon_centroids_df_base <- do.call(rbind, polygon_centroids_list)
        # Match with Cell.IDs for level 1
        if (maxDepth >= 1 && nrow(polygon_centroids_df_base) > 0) {
          level1_poly_centroids <- polygon_centroids_df_base[polygon_centroids_df_base$depth == 1, ]
          cellID_mapping <- hvt_list[[3]][["summary"]] %>%
            dplyr::filter(Segment.Level == 1) %>%
            dplyr::select(Segment.Parent, Segment.Child, Cell.ID) %>%
            dplyr::mutate(
              Segment.Parent = as.character(Segment.Parent),
              Segment.Child = as.numeric(Segment.Child)
            )
          polygon_centroids_df <- level1_poly_centroids %>%
            dplyr::mutate(
              cluster = as.character(cluster),
              child = as.numeric(child)
            ) %>%
            dplyr::left_join(cellID_mapping, 
                             by = c("cluster" = "Segment.Parent", "child" = "Segment.Child")) %>%
            dplyr::filter(!is.na(Cell.ID)) %>%
            dplyr::select(Cell.ID, x, y)
        }
      }
    }
        
    # ============================================================================
    # ADD HOVER TEXT
    # ============================================================================
    
    datapoly <- add_hover_text(datapoly)
    centroidDataframe <- add_hover_text(centroidDataframe)
    
    # ============================================================================
    # PREPARE COLORS
    # ============================================================================
    
    column_color <- as.factor(datapoly[, domains.column])
    n_colors_needed <- length(unique(column_color))
    domain_colors <- get_color_palette(n_colors_needed, DOMAIN_COLORS)
    
    # ============================================================================
    # CREATE PLOT
    # ============================================================================
    
    domains_plot <- ggplot2::ggplot()
    
    # Plot polygons for each depth level
    for (i in maxDepth:1) {
      # Pre-compute highlighted status
      is_highlighted <- datapoly[datapoly$depth == i, "Cell.ID"] %in% highlight_cells
      
      # Split data
      depth_data <- datapoly[datapoly$depth == i, ]
      data_normal <- depth_data[!is_highlighted, ]
      data_highlight <- depth_data[is_highlighted, ]
      
      # Get corresponding color subsets
      colors_normal <- column_color[datapoly$depth == i][!is_highlighted]
      colors_highlight <- column_color[datapoly$depth == i][is_highlighted]
      
      # Add polygon layers
      domains_plot <- add_polygon_layer(
        domains_plot, data_normal, colors_normal,
        "black", LINE_SIZE_NORMAL, data_normal$hoverText
      )
      
      domains_plot <- add_polygon_layer(
        domains_plot, data_highlight, colors_highlight,
        "white", LINE_SIZE_HIGHLIGHT, data_highlight$hoverText
      )
    }
    
    # Add centroids for each depth (conditionally based on centroids parameter)
    if (centroids == TRUE) {
      for (depth in seq_len(maxDepth)) {
        depth_size <- centroid.size[depth]
        depth_color <- centroid.color[depth]
        domains_plot <- domains_plot + ggplot2::geom_point(
          data = centroidDataframe[centroidDataframe["lev"] == depth, ],
          ggplot2::aes(x = x, y = y, text = hoverText),
          size = depth_size,
          pch = 21,
          fill = depth_color,
          color = depth_color,
          tooltip = "text"
        )
      }
    }
    
    # ============================================================================
    # STYLE PLOT
    # ============================================================================
    
    domains_plot <- domains_plot +
      ggplot2::scale_fill_manual(
        name = "Domains",
        values = domain_colors,
        breaks = levels(column_color)
      ) +
      ggplot2::theme_bw() + 
      ggplot2::theme(
        plot.background = ggplot2::element_blank(),
        plot.title = ggplot2::element_text(
          size = 20,
          hjust = 0.5,
          margin = ggplot2::margin(0, 0, 20, 0)
        ),
        panel.grid = ggplot2::element_blank(),
        panel.border = ggplot2::element_blank(),
        axis.ticks = ggplot2::element_blank(),
        axis.text = ggplot2::element_blank(),
        axis.title = ggplot2::element_blank(),
        panel.background = ggplot2::element_blank()
      ) +
      ggplot2::scale_x_continuous(expand = c(0, 0)) +
      ggplot2::scale_y_continuous(expand = c(0, 0))
    
    # ============================================================================
    # CONVERT TO PLOTLY AND FINALIZE
    # ============================================================================
    
    domainsPlot <- plotly::ggplotly(domains_plot, tooltip = "text")
    
    # Ensure centroids are filled circles in plotly
    if (centroids == TRUE) {
      n_traces <- length(domainsPlot$x$data)
      for (i in seq_len(n_traces)) {
        trace <- domainsPlot$x$data[[i]]
        if (!is.null(trace$mode) && grepl("markers", trace$mode)) {
          if (is.null(domainsPlot$x$data[[i]]$marker)) {
            domainsPlot$x$data[[i]]$marker <- list()
          }
          # Get the depth from the trace (if available) or use first depth color
          depth_idx <- min(i, length(centroid.color))
          domainsPlot$x$data[[i]]$marker$fillcolor <- centroid.color[depth_idx]
          if (is.null(domainsPlot$x$data[[i]]$marker$line)) {
            domainsPlot$x$data[[i]]$marker$line <- list()
          }
          domainsPlot$x$data[[i]]$marker$line$color <- centroid.color[depth_idx]
          domainsPlot$x$data[[i]]$marker$line$width <- 0.5
        }
      }
    }
    
    # Add cell ID annotations (conditionally based on cell_id parameter)
    if (cell_id == TRUE) {
      # Prepare annotation data based on cell_id_position
      if (cell_id_position == "center" && nrow(polygon_centroids_df) > 0) {
        annotation_data <- polygon_centroids_df
      } else {
        annotation_data <- subset(centroidDataframe_2, lev == 1)
      }
      
      # Calculate position offsets based on cell_id_position
      if (cell_id_position == "center") {
        xshift <- 0
        yshift <- 0
        xanchor <- "center"
        yanchor <- "middle"
      } else {
        alignments <- list(
          right = list(xshift = 5, yshift = 0, xanchor = "left", yanchor = "middle"),
          left = list(xshift = -5, yshift = 0, xanchor = "right", yanchor = "middle"),
          bottom = list(xshift = 0, yshift = -10, xanchor = "center", yanchor = "top"),
          top = list(xshift = 0, yshift = 10, xanchor = "center", yanchor = "bottom")
        )
        alignment <- rlang::`%||%`(alignments[[cell_id_position]], alignments$bottom)
        xshift <- alignment$xshift
        yshift <- alignment$yshift
        xanchor <- alignment$xanchor
        yanchor <- alignment$yanchor
      }
      
      # Convert cell_id_size from ggplot2 size to plotly font size
      plotly_font_size <- max(8, cell_id_size * 2.845)
      
      if (nrow(annotation_data) > 0) {
        domainsPlot <- domainsPlot %>%
          plotly::add_annotations(
            x = annotation_data$x,
            y = annotation_data$y,
            text = as.character(annotation_data$Cell.ID),
            showarrow = FALSE,
            font = list(size = plotly_font_size, color = "black"),
            xshift = xshift,
            yshift = yshift,
            xanchor = xanchor,
            yanchor = yanchor,
            xref = "x",
            yref = "y"
          )
      }
    }
    
    # Modify all traces in a single loop
    domainsPlot$x$data <- lapply(domainsPlot$x$data, modify_plotly_trace)
    
    # Add final layout modifications
    domainsPlot <- domainsPlot %>%
      plotly::layout(
        hoverlabel = list(bgcolor = "rgba(255,255,0,0.2)"),
        legend = list(
          title = list(text = "Domains"),
          itemsizing = "constant",
          itemdoubleclick = FALSE,
          itemclick = "toggleothers",
          traceorder = "reversed"
        )
      ) %>%
      plotly::config(displayModeBar = TRUE)
    
    return(domainsPlot)
  })
}

#' @name msm
#' @title Performing Monte Carlo Simulations of Markov Chain
#' @description This is the main function to perform Monte Carlo simulations of Markov Chain on
#' the dynamic forecasting of HVT States of a time series dataset. It includes both ex-post and ex-ante analysis 
#' offering valuable insights into future trends while resolving state transition 
#' challenges through clustering and nearest-neighbor methods to enhance simulation accuracy. 
#' @param state_time_data Dataframe. A dataframe containing state transitions over time(cell id and timestamp)
#' @param forecast_type Character. A character to indicate the type of forecasting.  
#' Accepted values are "ex-post" or "ex-ante".
#' @param initial_state Numeric. An integer indicatiog the state at t0. 
#' @param n_ahead_ante Numeric. A vector of n ahead points to be predicted further in ex-ante analyzes. 
#' @param transition_probability_matrix Dataframe. A dataframe of transition probabilities/ output of 
#' `getTransitionProbability` function
#' @param num_simulations Integer. A number indicating the total number of simulations to run.
#' Default is 100.
#' @param trainHVT_results List.`trainHVT` function output 
#' @param scoreHVT_results List. `scoreHVT` function output
#' @param actual_data Dataframe. A dataframe for ex-post prediction period with the actual raw data values
#' @param raw_dataset DataFrame. A dataframe of input raw dataset from the mean and standard deviation will 
#' be calculated to scale up the predicted values
#' @param k Integer. A number of optimal clusters when handling problematic states. 
#' Default is 5.
#' @param handle_problematic_states Logical. To indicate whether to handle problematic states or not.
#' Default is FALSE.
#' @param n_nearest_neighbor Integer. A number of nearest neighbors to consider when handling problematic states.
#' Default is 1.
#' @param show_simulation Logical. To indicate whether to show the simulation lines in plots or not. Default is TRUE.
#' @param time_column Character. The name of the time column in the input dataframe.
#' @param mae_metric Character. A character to indicate which metric to calculate Mean Absolute Error.
#' Accepted entries are "mean", "median", or "mode". Default is "median".
#' @param time_column Character. The name of the column containing time data. Used for aligning and plotting the results.
#' @param precomputed_problematic_states Vector. An internal parameter used in HVTMSMoptimization call 
#' to pass the problematic states detected
#' @param plot_mode Character. A character to indicate what should be included in output list 
#' "mae-only" values (used in HVTMSMoptimization) and "all" (used in direct call, include plots and mae)
#' @return A list object that contains the forecasting plots and MAE values.
#' \item{[[1]]}{Simulation plots and MAE values for state and centroids plot} 
#' \item{[[2]]}{Summary Table, Dendrogram plot and Clustered Heatmap when handle_problematic_states is TRUE} 
#' @author Vishwavani <vishwavani@@mu-sigma.com>, Nithya <nithya.sn@@mu-sigma.com>
#' @keywords Timeseries_Analysis
#' @importFrom magrittr %>%
#' @importFrom stats sd cutree
#' @include HVTMSM_support.R
#' @examples 
#' dataset <- data.frame(t = as.numeric(time(EuStockMarkets)),
#' DAX = EuStockMarkets[, "DAX"],
#' SMI = EuStockMarkets[, "SMI"],
#' CAC = EuStockMarkets[, "CAC"],
#' FTSE = EuStockMarkets[, "FTSE"])
#' hvt.results<- trainHVT(dataset[,-1],n_cells = 60, depth = 1, quant.err = 0.1,
#'                       distance_metric = "L1_Norm", error_metric = "max",
#'                       normalize = TRUE,quant_method = "kmeans")
#' scoring <- scoreHVT(dataset, hvt.results)
#' cell_id <- scoring$scoredPredictedData$Cell.ID
#' time_stamp <- dataset$t
#' temporal_data <- data.frame(cell_id, time_stamp)
#' table <- getTransitionProbability(temporal_data, 
#' cellid_column = "cell_id",time_column = "time_stamp")
#' colnames(temporal_data) <- c("Cell.ID","t")
#' ex_post_forecasting <- dataset[1830:1860,]
#' ex_post <- msm(state_time_data = temporal_data,
#'               forecast_type = "ex-post",
#'               transition_probability_matrix = table,
#'               initial_state = 2,
#'               num_simulations = 2,
#'               scoreHVT_results = scoring,
#'               trainHVT_results = hvt.results,
#'               actual_data = ex_post_forecasting,
#'               raw_dataset = dataset,
#'               mae_metric = "median",
#'               handle_problematic_states = FALSE,
#'              show_simulation = FALSE,
#'              time_column = 't')
#' @export msm


msm <- function(state_time_data,
                forecast_type = "ex-post",
                initial_state,
                n_ahead_ante = 10,
                transition_probability_matrix,
                num_simulations = 500,
                trainHVT_results,
                scoreHVT_results,
                actual_data = NULL,
                raw_dataset,
                k = NULL,
                handle_problematic_states = TRUE,
                n_nearest_neighbor = NULL,
                show_simulation = TRUE,
                mae_metric = "median",
                time_column,
                precomputed_problematic_states = NULL,
                plot_mode = c("all","mae-only")) {


  # Global variables for CRAN warnings
  time <- simulation <- median <- sd <- studentized_residuals <- value <- Cell.ID <- cluster <- nearest_neighbor <- NULL

  # Detect if called directly (not from optimization)
  called_directly <- !isTRUE(getOption("hvt.msm.optimization", FALSE))
  
  plot_mode <- match.arg(plot_mode)
  suppressWarnings(suppressMessages(requireNamespace('NbClust')))
  
  # ============================================================================
  # CONSOLIDATED VALIDATION
  # ============================================================================
  
  validate_inputs <- function() {
    errors <- c()
    warnings <- c()
    
    # Basic validation
    if(!"Cell.ID" %in% colnames(state_time_data)) 
      errors <- c(errors, "ERROR: Missing 'Cell.ID' column in state_time_data")
    if(!time_column %in% colnames(state_time_data)) 
      errors <- c(errors, paste0("ERROR: Missing '", time_column, "' column in state_time_data"))
    if (is.null(transition_probability_matrix) || nrow(transition_probability_matrix) == 0)
      errors <- c(errors, "ERROR: transition_probability_matrix cannot be empty")
    if (!mae_metric %in% c("median", "mean", "mode"))
      errors <- c(errors, "ERROR: mae_metric must be 'mean', 'median', or 'mode'")

    # Ex-post validation
    can_plot_states_actual <- FALSE
    if (forecast_type == "ex-post") {
      if (is.null(actual_data)) {
        warnings <- c(warnings, "No actual_data provided for ex-post - centroids will be plotted without actual comparison")
      } else {
        actual_times <- actual_data[[time_column]]
        available_times <- state_time_data[[time_column]]
        if (length(intersect(actual_times, available_times)) > 0) {
          can_plot_states_actual <- TRUE
        } else {
          stop("ERROR: No temporal state data available for actual data period - states comparison will be skipped")
        }
      }
    }
    
    # Transition matrix validation
    scored_cells_temp <- unique(scoreHVT_results$cellID_coordinates$Cell.ID)
    if(!is.null(transition_probability_matrix)) {
      transition_cells_temp <- unique(c(transition_probability_matrix$Current_State,
                                        transition_probability_matrix$Next_State))
      missing_cells <- setdiff(scored_cells_temp, transition_cells_temp)
      if(length(missing_cells) > 0) {
        warnings <- c(warnings, paste("Some scored cells missing from transition matrix:",
                                      length(missing_cells), "cells affected"))
      }
    }
    
    # Handle problematic states validation
    if (handle_problematic_states) {
      if (k > length(scored_cells_temp)) {
        errors <- c(errors, paste0("Invalid k vs cells (k=", k, ", cells=", length(scored_cells_temp), ")"))
      }
      max_possible_nn <- length(scored_cells_temp) - k
      if (!is.null(n_nearest_neighbor) && n_nearest_neighbor > max_possible_nn) {
        errors <- c(errors, paste0("Insufficient NN - requested nn=", n_nearest_neighbor,
                                  ", available nn=", max_possible_nn))
      }
    }
    
    if(length(warnings) > 0 && called_directly) {
      message("Validation Warnings:")
      for(w in warnings) message("  - ", w)
    }
    if(length(errors) > 0) stop(paste(paste(errors, collapse = "\n")))
    
    return(list(warnings = warnings, can_plot_states_actual = can_plot_states_actual))
  }
  
  # ============================================================================
  # SIMPLIFIED CLUSTERING
  # ============================================================================

  perform_clustering <- function(clustering_data, k, centroid_2d_points) {
    tryCatch({
      # Try clustHVT first
      if (exists("clustHVT", envir = .GlobalEnv, inherits = FALSE)) {
        clustFun <- get("clustHVT", envir = .GlobalEnv, inherits = FALSE)
      } else {
        clustFun <- clustHVT
      }
      
      clust.results <- clustFun(
        data = clustering_data,
        trainHVT_results = trainHVT_results,
        scoreHVT_results = scoreHVT_results,
        clustering_method = "ward.D2",
        indices = NULL,
        clusters_k = as.integer(k),
        type = "default",
        domains.column = NULL)
      
      clusters <- cutree(clust.results$hc, k = k)
      cluster_data <- data.frame(
        Cell.ID = centroid_2d_points$Cell.ID,
        cluster = clusters,
        stringsAsFactors = FALSE
      )
      

      return(list(cluster_data = cluster_data, clust.results = clust.results))
      
    }, error = function(e) {
      # Fallback: direct hierarchical clustering
      hc_fb <- stats::hclust(stats::dist(as.matrix(clustering_data)), method = "ward.D2")
      n_items <- nrow(clustering_data)
      if (k < 1 || k > n_items) {
        stop(sprintf("ERROR: Clustering error: k=%d out of range [1..%d]", k, n_items))
      }
      clusters_fb <- stats::cutree(hc_fb, k = k)
      
      cluster_data <- data.frame(
        Cell.ID = centroid_2d_points$Cell.ID,
        cluster = clusters_fb,
        stringsAsFactors = FALSE
      )
      
      clust.results <- list(hc = hc_fb, clusters = clusters_fb)
      return(list(cluster_data = cluster_data, clust.results = clust.results))
    })
  }
  # ============================================================================
  # MAIN EXECUTION
  # ============================================================================
  
  # Run validation
  validation_result <- validate_inputs()
  can_plot_states_actual <- validation_result$can_plot_states_actual
  
  # Prepare data
  centroid_2d_points <- scoreHVT_results$cellID_coordinates
  centroid_data <- trainHVT_results[[3]]$summary
  
  common_cols <- intersect(colnames(raw_dataset), colnames(centroid_data))
  clustering_data <- (scoreHVT_results$cellID_coordinates) %>% dplyr::select(-Cell.ID)
  centroid_data <- centroid_data[, common_cols]
  scored_cells <- unique(scoreHVT_results$cellID_coordinates$Cell.ID)
  transition_cells <- unique(c(transition_probability_matrix$Current_State,
                               transition_probability_matrix$Next_State))

  has_full_transition_coverage <- all(scored_cells %in% transition_cells)
  
  # Determine n_ahead
  if (forecast_type == "ex-ante") {
    time_vals <- state_time_data[[time_column]]
    time_step <- diff(time_vals)[1]
    horizon_len <- if (length(n_ahead_ante) == 1) n_ahead_ante else length(n_ahead_ante)
    n_ahead_ante <- seq(
      from = max(time_vals) + time_step,
      length.out = horizon_len,
      by = time_step
    )
  }

  if (forecast_type == "ex-post") {
    if(!is.null(actual_data)) {
      n_ahead <- nrow(actual_data)
    } else {
      n_ahead <- min(nrow(state_time_data), 10)
      if (called_directly) message("No actual_data provided, using temporal data length: ", n_ahead)
    }
  } else if (forecast_type == "ex-ante") {
    n_ahead <- length(n_ahead_ante)
  } else stop("forecast_type must be either 'ex-post' or 'ex-ante'")
  
  # Initialize variables
  stp_list <- NULL; cluster_heatmap <- NULL; clust.results <- NULL; cluster_data <- NULL
  
  # Detect problematic states
  problematic_detection <- detect_problematic_states(transition_probability_matrix, scored_cells)
  
  # Handle precomputed problematic states
  detected_problematic <- problematic_detection$all_problematic
  if (!is.null(precomputed_problematic_states)) {
    problematic_states <- intersect(unique(union(precomputed_problematic_states, detected_problematic)), scored_cells)
  } else {
    problematic_states <- intersect(detected_problematic, scored_cells)
  }


  # Handle problematic states with clustering
  if (handle_problematic_states && length(problematic_states) > 0) {
    cluster_data <- clust.results <- stp_list <- cluster_heatmap <- all_nearest_neighbors <- NULL
    tryCatch({
      #if (called_directly) message("Starting problematic states handling...")
      
      # Perform clustering
      clustering_result <- perform_clustering(clustering_data, k, centroid_2d_points)
      cluster_data <- clustering_result$cluster_data
      clust.results <- clustering_result$clust.results
      
      # Post-clustering validation: Check neighbor availability for problematic states
      if (!is.null(cluster_data) && length(problematic_states) > 0) {
        per_state_available <- sapply(problematic_states, function(ps) {
          cl_id <- cluster_data$cluster[cluster_data$Cell.ID == ps]
          members <- cluster_data$Cell.ID[cluster_data$cluster == cl_id]
          valid_neighbors <- setdiff(members, c(ps, problematic_states))
          length(valid_neighbors)
        })
        min_available <- if (length(per_state_available)) min(per_state_available) else 0
        
        # Category 1: Isolated problematic states (zero neighbors)
        if (min_available == 0) {
          stop(paste0("Isolated problematic states - At least one problematic state has no valid neighbors in its cluster",
                      "\nSuggestions:\n",
                      "1. Try a different k value (current: ", k, ")\n",
                      "2. Check if your data has sufficient non-problematic states"))
        }
        
        # Category 2: Limited availability of neighbors
        if (!is.null(n_nearest_neighbor) && n_nearest_neighbor > min_available) {
          stop(paste0("Limited Availability of neighbors (Req = ", n_nearest_neighbor, 
                      ", Pre = ", min_available, ")",
                      "\nSuggestions:\n",
                      "1. Reduce n_nearest_neighbor to ", min_available, " or less (current: ", n_nearest_neighbor, ")\n",
                      "2. Try a different k value (current: ", k, ") to create larger clusters"))
        }
      }
      
      # Add names.column if available
      if (!is.null(scoreHVT_results$centroidData) &&
          length(scoreHVT_results$centroidData$names.column) == nrow(cluster_data)) {
        cluster_data$names.column <- scoreHVT_results$centroidData$names.column
      }
      
      # Check for singleton clusters
      cluster_counts <- table(cluster_data$cluster)
      if(any(cluster_counts == 1)) {
        warning(paste0("Singleton clusters present with k = ", k,
                       ". Consider trying a different value of k."))
      }
      
      # Create neighbor mapping
      neighbor_mapping <- cluster_data %>%
        dplyr::group_by(cluster) %>%
        dplyr::group_split() %>%
        purrr::map_dfr(function(cluster_group) {
          prob_states_in_cluster <- intersect(cluster_group$Cell.ID, problematic_states)
          if (length(prob_states_in_cluster) == 0) return(NULL)
          coords <- centroid_2d_points[centroid_2d_points$Cell.ID %in% cluster_group$Cell.ID, c("Cell.ID", "x", "y")]
          rownames(coords) <- coords$Cell.ID
          coords_num <- coords[, c("x", "y")]
          dist_matrix <- as.matrix(dist(coords_num, method = "euclidean"))
          rownames(dist_matrix) <- coords$Cell.ID
          colnames(dist_matrix) <- coords$Cell.ID
          purrr::map_dfr(prob_states_in_cluster, function(prob_state) {
            distances <- dist_matrix[as.character(prob_state), ]
            valid_neighbors <- setdiff(names(sort(distances)), as.character(c(prob_state, problematic_states)))
            transition_problem <- dplyr::case_when(
              prob_state %in% problematic_detection$missing_states ~ "Absence Transitions",
              prob_state %in% problematic_detection$self_transition_states ~ "Self-State Only Transitions",
              prob_state %in% problematic_detection$cyclic_states ~ "Cyclic Transitions",
              TRUE ~ NA_character_
            )
            if (length(valid_neighbors) > 0) {
              num_neighbors <- min(length(valid_neighbors), n_nearest_neighbor)
              nearest_n <- as.numeric(valid_neighbors[1:num_neighbors])
              tibble::tibble(
                problematic_state = prob_state,
                `State Transition Problem` = transition_problem,
                nearest_neighbor = list(nearest_n)
              )
            } else NULL
          })
        })
      
      if(nrow(neighbor_mapping) > 0) {
        stp_list <- neighbor_mapping %>% dplyr::mutate(nearest_neighbor = sapply(nearest_neighbor, function(x) paste(x, collapse = ",")))
        all_nearest_neighbors <- neighbor_mapping %>% dplyr::pull(nearest_neighbor) %>% unlist() %>% unique()
        cluster_heatmap <- clusterPlot(
          dataset = data.frame(
            Cell.ID = cluster_data$Cell.ID,
            Cluster = as.factor(cluster_data$cluster),
            names.column = cluster_data$names.column
          ),
          hvt.results = trainHVT_results,
          domains.column = "Cluster",
          highlight_cells = c(problematic_states, all_nearest_neighbors)
        )
      }
      
    }, error = function(e) {
      warning(paste("Enhanced clustering failed:", e$message))
      # cluster_data <- NULL; stp_list <- NULL; cluster_heatmap <- NULL; clust.results <- NULL  # Commented out as not used
    })
  }
  
    
  # Final validation for problematic states handling
  if (handle_problematic_states && length(problematic_states) > 0) {
    if (is.null(cluster_data)) stop("ERROR: Clustering failed but problematic states exist. Cannot proceed with MSM.")
    # Neighbor availability already validated post-clustering
  }

  
  # Run simulations
  simulation_results <- sapply(seq_len(num_simulations), function(sim_index) {
    simulate_sequence_enhanced(sim_index = sim_index,
                               n_ahead = n_ahead,
                               initial_state = initial_state,
                               transition_probability_matrix = transition_probability_matrix,
                               centroid_2d_points = centroid_2d_points,
                               cluster_data = cluster_data,
                               problematic_states = problematic_states,
                               n_nearest_neighbor = n_nearest_neighbor,
                               scored_cells = scored_cells)
  })
  simulation_results <- as.data.frame(simulation_results)
  colnames(simulation_results) <- paste0("Sim_", seq_len(num_simulations))


    
  # Add time column
  time_values <- if(forecast_type == "ex-post") {
    if(!is.null(actual_data)) actual_data[[time_column]] else head(state_time_data[[time_column]], n_ahead)
  } else {
    n_ahead_ante[1:nrow(simulation_results)]
  }
  simulation_results$time <- time_values
  


  # Calculate summary statistics
  simulation_results <- simulation_results %>%
    dplyr::select(time, dplyr::everything()) %>%
    dplyr::rowwise() %>%
    dplyr::mutate(
      mean   = round(mean(dplyr::c_across(dplyr::starts_with("Sim_")))),
      median = round(stats::median(dplyr::c_across(dplyr::starts_with("Sim_")))),
      mode   = {
        tbl <- table(dplyr::c_across(dplyr::starts_with("Sim_")))
        as.numeric(names(tbl)[which.max(tbl)])
      }
    ) %>%
    dplyr::ungroup()
  
  # Fast MAE-only return
  if (plot_mode == "mae-only") {
    if (forecast_type != "ex-post" || is.null(actual_data))
      stop("mae-only mode currently expects ex-post with actual_data.")
    
    test_data_states <- state_time_data[state_time_data[[time_column]] %in% simulation_results$time, ]
    summary_data <- simulation_results[, c("time", "mean", "median", "mode")]
    
    # Calculate MAEs
    raw_dataset_wo_time <- raw_dataset %>% dplyr::select(-(time_column))
    mean_raw <- raw_dataset_wo_time %>% dplyr::summarise(dplyr::across(dplyr::everything(), ~ mean(., na.rm = TRUE)))
    sd_raw   <- raw_dataset_wo_time %>% dplyr::summarise(dplyr::across(dplyr::everything(), ~ stats::sd(., na.rm = TRUE)))
    
    name_columns <- colnames(centroid_data)
    centroid_map_df <- cbind(centroid_data, Cell.ID = centroid_2d_points$Cell.ID)
    
    actual_raw_dfs <- lapply(name_columns, function(col_name) {
      df <- actual_data[, c(time_column, col_name)]
      names(df) <- c("time", paste0("actual_", col_name))
      df
    })
    names(actual_raw_dfs) <- name_columns
    
    compute_mae_for_metric <- function(metric_name) {
      pred_states_metric <- summary_data[[metric_name]]
      states_mae_metric <- mean(abs(test_data_states$Cell.ID[seq_along(pred_states_metric)] - pred_states_metric), na.rm = TRUE)
      centroid_mae_list_metric <- lapply(name_columns, function(var) {
        map_vec <- setNames(centroid_map_df[[var]], centroid_map_df$Cell.ID)
        pred_numeric <- as.numeric(map_vec[as.character(pred_states_metric)])
        if (isTRUE(trainHVT_results[["model_info"]][["input_parameters"]][["normalize"]])) {
          pred_numeric <- pred_numeric * sd_raw[[1, var]] + mean_raw[[1, var]]
        }
        act_vec <- actual_raw_dfs[[var]][[paste0("actual_", var)]][seq_along(pred_numeric)]
        this_mae <- mean(abs(act_vec - pred_numeric), na.rm = TRUE)
        list(mae = as.numeric(this_mae))
      })
      list(centroid_mae_list = centroid_mae_list_metric, states_mae = as.numeric(states_mae_metric))
    }

    mae_by_metric <- list(
      mean   = compute_mae_for_metric("mean"),
      median = compute_mae_for_metric("median"),
      mode   = compute_mae_for_metric("mode")
    )

    out <- list(mae_by_metric = mae_by_metric)
    class(out) <- "hvt.object"
    return(out)
  }
  
  # Generate plots
  if (called_directly || plot_mode != "mae-only") {
    plots <- mcmc_plots(
      simulation_results = simulation_results,
      centroid_data = centroid_data,
      centroid_2d_points = centroid_2d_points,
      actual_data = actual_data,
      state_time_data = state_time_data,
      forecast_type = forecast_type,
      n_ahead_ante = n_ahead_ante,
      type = "MSM",
      raw_dataset = raw_dataset,
      show_simulation = show_simulation,
      mae_metric = mae_metric,
      time_column = time_column,
      trainHVT_results = trainHVT_results
    )
  } else {
    plots <- list()
  }
  
  if (handle_problematic_states && length(problematic_states) > 0) {
    clust.results.2 <- NULL
    clust.results.2 <- clustHVT(
      data = clustering_data,
      trainHVT_results = trainHVT_results,
      scoreHVT_results = scoreHVT_results,
      clustering_method = "ward.D2",
      indices = NULL,
      clusters_k = as.integer(k),
      type = "default",
      domains.column = NULL,
      only_dendro = TRUE,
      highlight_labels =c(problematic_states, all_nearest_neighbors))
    
  }
    
    
    
  # Return results
  if(handle_problematic_states && length(problematic_states) > 0) {
    output_list <- list(
      plots = plots,
      dendrogram = clust.results.2$dendrogram,
      problematic_states_list = stp_list,
      cluster_heatmap = cluster_heatmap,
      problematic_states = problematic_states,
      data_coverage = list(
        scored_cells = scored_cells,
        transition_cells = transition_cells,
        can_plot_states_actual = can_plot_states_actual,
        has_full_transition_coverage = has_full_transition_coverage
      )
    )
    class(output_list) <- "hvt.object"
    return(output_list)
  } else {
    output_list <- list(
      plots = plots,
      simulation_results = simulation_results,
      data_coverage = list(
        scored_cells = scored_cells,
        transition_cells = transition_cells,
        can_plot_states_actual = can_plot_states_actual,
        has_full_transition_coverage = has_full_transition_coverage
      )
    )
    class(output_list) <- "hvt.object"
    return(output_list)
  }
}

#' @importFrom utils combn
requireNamespace("deldir")

# ======================================================================================
# Global variables
# ======================================================================================

# quant.err <- 80
errorCode <- 0

# ______________________________________________________________________________________
# Splitting the codebooks
# ______________________________________________________________________________________

split_codevec <- function(codeVec, epsilon) {
  # errorValue=errorValue/100
  codeVec <- plyr::llply(codeVec, .fun = function(X) {
    return(list((1 + epsilon) * X, (1 - epsilon) * X))
  })
  return(codeVec)
}

# _____________________________________________________________________________________
# Checks if the MAPE for a given cluster and its codebook is lesser/greater
# than the specified treshold
# _____________________________________________________________________________________

findMape <- function(cluster, centroid) {
  mape <- 1 / nrow(cluster) * sum(apply(cluster, MARGIN = 1, FUN = function(X) {
    return(abs(sum((X - centroid), na.rm = TRUE) / length(X)))
  }))
  return(mape)
}

mapeTest <- function(cluster, centroid, errorValue) {
  if (is.null(dim(cluster))) {
    return(TRUE)
  }

  mape <- findMape(cluster, centroid)
  # print(paste("Quantization error:", mape))

  if (mape < errorValue) {
    return(TRUE)
  } else {
    return(FALSE)
  }
}

# _____________________________________________________________________________________
# allocate members to their centroids
# _____________________________________________________________________________________
findMembers <- function(dataset, centroid) {
  # compute the distance of every point in the data from a centroid
  # err <- lapply(centroid, FUN = function(X){ return(apply(dataset, MARGIN=1, FUN= function(Y)sqrt(sum((Y-X)^2))))})
  err <- lapply(centroid, FUN = function(x) sqrt(colSums((t(dataset) - x)^2)))
  err <- do.call(rbind, err)

  # Allocate the membership of a datum to its closest centroid
  clust.list <- apply(err, MARGIN = 2, FUN = function(X) {
    return(which(X == min(X), arr.ind = TRUE))
  })

  return(clust.list)
}

# _____________________________________________________________________________________
# Function to assign cluster membership of data points
# _____________________________________________________________________________________

findSeque <- function(dataset, clusters) {
  tryCatch(
    {
      rownames(dataset) <- NULL
      clusters <- lapply(clusters, unname)
      clusters <- lapply(clusters, as.matrix)
      clusters <- lapply(clusters, function(X) {
        colnames(X) <- colnames(dataset)
        return(X)
      })
      oneRowIndices <- which(lapply(lapply(clusters, dim), function(X) {
        return(X[2])
      }) == 1)
      clusters[oneRowIndices] <- lapply(clusters[oneRowIndices], t)
      pointPos <- lapply(clusters, FUN = function(X) rownames(plyr::match_df(dataset, as.data.frame(X))))
      pointPos <- lapply(pointPos, as.numeric)
      sequeDf <- as.data.frame(1:nrow(dataset))
      lapply(1:length(pointPos), FUN = function(X) {
        sequeDf[pointPos[[X]], ] <<- X
      })
      clust.list <- sequeDf[, 1]
    },
    error = function(err) {
      warning(err)
    },
    finally = {}
  )

  return(clust.list)
}

# _____________________________________________________________________________________
# allocate members to their centroids
# _____________________________________________________________________________________

allocateMembers <- function(dataset, centroid) {
  clust.list <- findMembers(dataset, centroid)

  return(lapply(unique(clust.list), FUN = function(X) which(clust.list == X)))
}

# _____________________________________________________________________________________
# Nearest neighbour condition - LBGVQ function
# _____________________________________________________________________________________

lbgVQ <- function(dataset, quant.err, epsilon) {
  # mean of the entire data set
  cnew <- list(apply(dataset, MARGIN = 2, FUN = function(X) mean(X, na.rm = TRUE)))
  cFin <- c()
  finalClusters <- c()

  # Set the quantization error
  mape <- 1 / nrow(dataset) * sum(apply(dataset, MARGIN = 1, FUN = function(X) {
    return(abs(sum((X - unlist(cnew)), na.rm = TRUE)) / length(X))
  }))
  if (mape != 0) {
    errorValue <- mape * quant.err
  } else {
    errorValue <- quant.err
  }

  # clust.list is the indices of the members of a cluster
  clust.list <- list(1:nrow(dataset))

  while (T) {
    # cluster is the actual data points in the cluster
    cluster <- lapply(clust.list, FUN = function(Y) {
      dataset[Y, ]
    })

    # codesToSplit is the codebooks which have q error greater than the treshold
    tryCatch(
      {
        codesToSplit <- cnew[which(!unlist(lapply(1:length(cluster), FUN = function(i) {
          mapeTest(cluster[[i]], cnew[[i]], errorValue)
        })))]
      },
      error = function(err) {
        warning(err)
      },
      finally = {}
    )

    # split the codebooks
    tryCatch(
      {
        cSplit <- unlist(split_codevec(codesToSplit, epsilon), recursive = FALSE)
      },
      error = function(err) {
        warning(err)
      },
      finally = {}
    )

    # remove the final codebook from the list whose q error is lesser than the treshold
    doneFlags <- !(cnew %in% codesToSplit)
    cFin <- c(cFin, cnew[doneFlags])
    if (length(which(doneFlags)) != 0) {
      # Update the finalClusters
      finalClusters <- c(finalClusters, cluster[which(!cnew %in% codesToSplit)])
      cluster[which(!cnew %in% codesToSplit)] <- NULL


      # Remove the clustered data from the original dataset
      dataset <- dataset[-unlist(clust.list[which(doneFlags)]), ]

      # Start afresh
      cnew <- list(apply(dataset, MARGIN = 2, FUN = mean))
      clust.list <- list(1:nrow(dataset))

      if (length(codesToSplit) == 0) {
        break
      }
      next
    }

    # update the clust.list with the new split codebooks and re-compute the mean codebook
    tryCatch(
      {
        clust.list <- allocateMembers(dataset, cSplit)
      },
      error = function(err) {
        warning(err)
      },
      finally = {}
    )

    cnew <- lapply(clust.list,
      FUN = function(X) {
        if (length(X) == 1) {
          return(dataset[X, ])
        } else {
          return(apply(dataset[unlist(X), ], MARGIN = 2, FUN = mean))
        }
      }
    )
  }

  cFin <- lapply(cFin, FUN = as.numeric)

  return(list(clusters = finalClusters, codebooks = cFin, errorValue = errorValue))
}

# _____________________________________________________________________________________
# Functions to merge the clusters that satisfy the quantization treshold
# _____________________________________________________________________________________

euDistance <- function(index, codebooks) {
  return((codebooks[[index[1]]] - codebooks[[index[2]]])^2)
}

# _____________________________________________________________________________________
# Function checks if the clusters should be merged based on the quantization error
# _____________________________________________________________________________________

shouldMergeClusters <- function(X, clusters, errorValue) {
  # merge the two clusters and compute the new mean
  newCluster <- do.call(rbind, list(clusters[[X[1]]], clusters[[X[2]]]))
  meanCb <- apply(newCluster, MARGIN = 2, FUN = mean)

  # check for the quantization error
  tryCatch(
    {
      mapeFlag <- mapeTest(newCluster, meanCb, errorValue)
    },
    error = function(err) {
      warning(err)
    },
    finally = {}
  )

  return(mapeFlag)
}


mergeClusters <- function(codebooks, clusters, errorValue) {
  # find the index of the centroids that are closest (X)
  pairs <- t(combn(length(codebooks), 2))

  tryCatch(
    {
      euD <- sqrt(colSums(apply(pairs, MARGIN = 1, FUN = function(X) {
        euDistance(X, codebooks)
      })))
    },
    error = function(err) {
      warning(err)
    },
    finally = {}
  )

  X <- pairs[which(euD == min(euD)), ]


  if (shouldMergeClusters(X, clusters, errorValue)) {
    # merge the two clusters and compute the new mean
    newCluster <- do.call(rbind, list(clusters[[X[1]]], clusters[[X[2]]]))
    meanCb <- list(apply(newCluster, MARGIN = 2, FUN = mean))

    # Update the codebooks and the clusters
    clusters[[X[1]]] <- newCluster
    clusters[[X[2]]] <- NULL
    codebooks[[X[1]]] <- meanCb[[1]]
    codebooks[[X[2]]] <- NULL

    # recurse to find the next merge-able codebook pair
    mergeClusters(codebooks, clusters, errorValue)
  } else {
    return(list(clusters = clusters, codebooks = codebooks, errorValue = errorValue))
  }
}

# _______________________________________________________________________________________
# Main Function to quantize
# ______________________________________________________________________________________

vectorQuantize <- function(dataset, quant.err, epsilon) {
  tryCatch(
    {
      lbgvq <- lbgVQ(dataset, quant.err, epsilon)
    },
    error = function(err) {
      warning(err)
    },
    finally = {}
  )

  if (length(lbgvq$codebooks) > 1) {
    tryCatch(
      {
        model <- mergeClusters(lbgvq$codebooks, lbgvq$clusters, lbgvq$errorValue)
      },
      error = function(err) {
        warning(err)
      },
      finally = {}
    )


    return(model)
  } else {
    return(lbgvq)
  }
}

# ==================================================================
# Function to Trilaterate
# Input : a distance matrix or a 'dist' object
# Output : a set of trilaterated points
# ==================================================================

# trilaterate <- function(d){
#   if(class(d)=='dist'){
#     d <- as.matrix(d)
#   }

trilaterate <- function(d) {
  # if(is(d, "dist")){
  if (inherits(d, "dist")) {
    d <- as.matrix(d)
  }


  # p1, p2, p3 are the initial triangulation points
  p1 <- c(0, 0)
  p2 <- c(0, d[1, 2])

  if (dim(d)[1] == 3) { # If there are only 3 points to trilaterate

    p3y <- (d[1, 3]^2 + d[1, 2]^2 - d[2, 3]^2) / (2 * d[1, 2])
    p3x <- sqrt(d[1, 3]^2 - p3y^2)
    p3 <- c(p3x, p3y)
    pdash <- rbind(p1, p2, p3)
    colnames(pdash) <- c("x", "y")
    return(pdash)
  } else if (dim(d)[1] > 3) { # If there are more than 3 points to trilaterate

    p3y <- (d[1, 3]^2 + d[1, 2]^2 - d[2, 3]^2) / (2 * d[1, 2])
    p3x <- sqrt(d[1, 3]^2 - p3y^2)
    p3 <- c(p3x, p3y)

    # iterator
    i <- 4:dim(d)[1]
    # compute the remaining points
    pdash <- lapply(i, FUN = function(ind) {
      y <- (d[1, ind]^2 + d[1, 2]^2 - d[2, ind]^2) / (2 * d[1, 2])
      x <- (d[1, ind]^2 + d[1, 3]^2 - d[3, ind]^2 - 2 * y * p3[2]) / (2 * p3[1])
      return(c(x, y))
    })

    pdash <- do.call(rbind, pdash)
    pdash <- rbind(p1, p2, p3, pdash)
    colnames(pdash) <- c("x", "y")
    return(pdash)
  } else { # If there are one or 2 points to trilaterate
    pdash <- rbind(p1, p2)
    colnames(pdash) <- c("x", "y")
    return(pdash)
  }
}



# ===========================================================================
# Main function to call
# ===========================================================================



#' @keywords internal
VQ_codebookSplit <- function(dataset, quant.err = 0.5, epsilon = NULL) {
  model <- list()

  if (is.null(epsilon)) {
    epsilon <- quant.err
  }

  if (!is.null(dataset)) {
    # Run VQ
    tryCatch(
      {
        system.time(model <- vectorQuantize(dataset, quant.err, epsilon))
      },
      error = function(err) {
        warning(err)
      },
      finally = {}
    )
  } else {
    # flog.error("Data is missing", name=logNS)
  }

  if (exists("model")) {
    # Reduce the dimensionality of the codebooks to 2
    codebooks <- do.call(rbind, model$codebooks)
    distances <- stats::dist(codebooks)
    if (length(distances) != 0) {
      tryCatch({
        twoDCodebooks <- trilaterate(distances)
      }, error = function(err) {
        warning(err)
      }, finally = {})
    } else {
      twoDCodebooks <- model$codebooks[[1]][1:2]
    }
    mapes <- lapply(1:length(model$clusters), FUN = function(i) {
      findMape(model$clusters[[i]], model$codebooks[[i]])
    })

    ztab <- data.frame(
      "Segment.Level" = rep(1, length(model$clusters)),
      "segment.Parent" = rep(1, length(model$clusters)),
      "Segment.Child" = 1:length(model$clusters),
      "n" = unlist(lapply(model$clusters, nrow)),
      "Quant.Error" = unlist(mapes)
    )
    ztab <- cbind(ztab, do.call(rbind, model$codebooks))

    idnodes <- lapply(model$clusters, rownames)
    idnodes <- list(lapply(1:length(idnodes), FUN = function(i) {
      data.frame(
        "tet..k.." = idnodes[[i]],
        "X1" = rep(1, length(idnodes[[i]])),
        "X1.1" = rep(1, length(idnodes[[i]])),
        "k" = i
      )
    }))

    return(list(
      "clusters" = model$clusters,
      "nodes.clust" = list(model$clusters),
      "idnodes" = idnodes,
      "error.quant" = list(mapes),
      plt.clust = list(mapes < model$errorValue),
      "summary" = ztab
    ))
  } else {
    return("error")
  }
}

#' @name trainHVT
#' @title Constructing Hierarchical Voronoi Tessellations
#' @description This is the main function to construct hierarchical voronoi tessellations.
#' This is done using hierarchical vector quantization(hvq). The data is represented in 2D
#' coordinates and the tessellations are plotted using these coordinates as
#' centroids. For subsequent levels, transformation is performed on the 2D
#' coordinates to get all the points within its parent tile. Tessellations are
#' plotted using these transformed points as centroids.
#' @param dataset Data frame. A data frame, with numeric columns (features) will be used for training the model.
#' @param min_compression_perc Numeric. An integer, indicating the minimum compression percentage to be achieved for the dataset. 
#' It indicates the desired level of reduction in dataset size compared to its original size.
#' @param n_cells Numeric. An integer, indicating the number of cells per hierarchy (level).
#' @param depth Numeric. An integer, indicating the number of levels. A depth of 1 means no hierarchy (single level), 
#' while higher values indicate multiple levels (hierarchy).
#' @param quant.err Numeric. A number indicating the quantization error threshold.
#' A cell will only breakdown into further cells if the quantization error of the cell is 
#' above the defined quantization error threshold.
#' @param normalize Logical. A logical value indicating if the dataset should be normalized. When set to TRUE, 
#' scales the values of all features to have a mean of 0 and a standard deviation of 1 (Z-score).
#' @param distance_metric Character. The distance metric can be L1_Norm(Manhattan) or L2_Norm(Eucledian). L1_Norm is selected by default.
#' The distance metric is used to calculate the distance between an n dimensional point and centroid.
#' @param error_metric Character. The error metric can be mean or max. max is selected by default. 
#' max will return the max of m values and mean will take mean of m values where
#' each value is a distance between a point and centroid of the cell.
#' @param quant_method Character. The quantization method can be kmeans or kmedoids. Kmeans uses means (centroids) as cluster centers
#'  while Kmedoids uses actual data points (medoids) as cluster centers. kmeans is selected by default.
#' @param scale_summary List. A list with user-defined mean and standard deviation values for all the features in the dataset. 
#' Pass the scale summary when normalize is set to FALSE.
#' @param diagnose Logical. A logical value indicating whether user wants to perform diagnostics on the model. 
#' Default value is FALSE.
#' @param hvt_validation Logical. A logical value indicating whether user wants to holdout a validation set and find 
#' mean absolute deviation of the validation points from the centroid. Default value is FALSE.
#' @param train_validation_split_ratio  Numeric. A numeric value indicating train validation split ratio. 
#' This argument is only used when hvt_validation has been set to TRUE. Default value for the argument is 0.8.
#' @param dim_reduction_method Character.The dim_reduction_method can be one of "tsne", "umap", "sammon".
#' @param tsne_perplexity Numeric.The tsne_perplexity is only used when dim_reduction_method is set to "tsne".
#' Default value is 30 and common values are between between 30 and 50.
#' @param tsne_theta Numeric.The tsne_theta is only used when dim_reduction_method is set to "tsne".
#' Default value is 0.5 and common values are between 0.2 and 0.5.
#' @param tsne_eta Numeric.The tsne_eta are used only when dim_reduction method is set to "tsne".
#' Default value is 200.
#' @param tsne_verbose Logical. A logical value which indicates the t-SNE algorithm to print 
#' detailed information about its progress to the console.
#' @param tsne_max_iter Numeric.The tsne_max_iter is used only when dim_reduction_method is set to "tsne".
#' Default value is 1000.More iterations can improve results but increase computation time.
#' @param umap_n_neighbors Integer.The umap_n_neighbors is used only when dim_reduction_method is set to "umap".
#' Default value is 15.Controls the balance between local and global structure in data.
#' @param umap_n_components Integer.The umap_n_components is used only when dim_reduction_method is set to "umap".
#' Default value is 2.Indicates the number of dimensions for embedding.
#' @param umap_min_dist Numeric.The umap_map_dist is used only when dim_reduction_method is set to "umap".
#' Default value is 0.1.Controls how tightly UMAP packs points together.
#' @return A Nested list that contains the hierarchical tessellation information. This
#' list has to be given as input argument to plot the tessellations.
#' \item{[[1]] }{A list containing information related to plotting tessellations. 
#' This information will include coordinates, boundaries, and other details necessary for visualizing the tessellations} 
#' \item{[[2]] }{A list containing information related to Sammon’s projection coordinates of the data points
#' in the reduced-dimensional space.}
#' \item{[[3]] }{A list containing detailed information about the hierarchical vector quantized data along with 
#' a summary section containing no of points, Quantization Error and the centroids for each cell.}
#' \item{[[4]] }{A list that contains all the diagnostics information of the model when diagnose is set to TRUE. 
#' Otherwise NA.}
#' \item{[[5]] }{A list that contains all the information required to generates a Mean Absolute Deviation (MAD) plot, 
#' if hvt_validation is set to TRUE. Otherwise NA}
#' \item{[[6]]}{A list containing detailed information about the hierarchical vector quantized data along with a
#'  summary section containing no of points, Quantization Error and the centroids for each cell which is the output of `hvq`}
#' \item{[[7]]}{model info: A list that contains model-generated timestamp, input parameters passed to the model ,
#' the validation results and the dimensionality reduction evaluation metrics table.}
#' @author Shubhra Prakash <shubhra.prakash@@mu-sigma.com>, Sangeet Moy Das <sangeet.das@@mu-sigma.com>, Shantanu Vaidya <shantanu.vaidya@@mu-sigma.com>,Bidesh Ghosh <bidesh.gosh@@mu-sigma.com>,Alimpan Dey <alimpan.dey@@mu-sigma.com>
#' @seealso \code{\link{plotHVT}}
#' @keywords Training_or_Compression
#' @importFrom magrittr %>%
#' @importFrom stats dist kmeans
#' @examples
#' data("EuStockMarkets")
#' hvt.results <- trainHVT(EuStockMarkets, n_cells = 60, depth = 1, quant.err = 0.1, 
#'                        distance_metric = "L1_Norm", error_metric = "max",
#'                        normalize = TRUE,quant_method="kmeans")
#' @include hvq.R getCellId.R madPlot.R displayTable.R
#' @include Add_boundary_points.R  Corrected_Tessellations.R  DelaunayInfo.R  Delete_Outpoints.R diagPlot.R
#' @export trainHVT


trainHVT <-
  function(dataset,
           min_compression_perc = NA,
           n_cells = NA,
           depth = 1,
           quant.err = 0.2,
           normalize = FALSE,
           distance_metric = "L1_Norm",
           error_metric =  "max",
           quant_method="kmeans",
           scale_summary = NA,
           diagnose=FALSE,
           hvt_validation=FALSE,
           train_validation_split_ratio=0.8,
           dim_reduction_method = "sammon",
           tsne_theta = 0.2,
           tsne_eta = 200,
           tsne_perplexity = 30,
           tsne_verbose = TRUE,
           tsne_max_iter = 500,
           umap_n_neighbors = 60,
           umap_n_components = 2, 
           umap_min_dist = 0.1
  ) {
    
    set.seed(279)
    requireNamespace("deldir")       #deldir function
    requireNamespace("Hmisc")        #ceil function
    requireNamespace("grDevices")    #chull function
    requireNamespace("splancs")      #csr function
    requireNamespace("sp")           #point.in.polygon function
    requireNamespace("conf.design")  #factorize functionScreenshot from 2022-10-26 16-15-44
  
    
    if (any(is.na(dataset))) {
      stop("Input dataframe contains missing (NA) values. Please handle missing data before using this function (e.g., via na.omit(), na.fill(), or imputation).")
       } 
    
    
    if(quant_method=="kmedoids"){message(' K-Medoids: Run time for vector quantization using K-Medoids is very high for large number of clusters.')}
    dataset <- as.data.frame(dataset)
    dataset=data.frame(
      lapply(dataset, as.numeric),
      check.names = FALSE,
      row.names = rownames(dataset)
    )
    data_structure <- dim(dataset)
    
    ## Diagnose Function
    if(hvt_validation){
      if(train_validation_split_ratio >1 | train_validation_split_ratio <0){stop("The train_validation_split_ratio should be in range 0-1 ")}
      message(paste0("Check MAD parameter has been set to TRUE, the train data will be randomly split by the ratio of  ",train_validation_split_ratio*100,":",(1-train_validation_split_ratio)*100))
      
      set.seed(273)
      smp_size <- floor(train_validation_split_ratio * nrow(dataset))
      train_ind <- sample(seq_len(nrow(dataset)), size = smp_size)
      
      validation_data <- dataset[-train_ind, ]
      dataset <- dataset[train_ind, ]
      
    }
    
    
    if (normalize) {
      scaledata <- scale(dataset, scale = TRUE, center = TRUE)
      rownames(scaledata) <- rownames(dataset)
      
      mean_data <- attr(scaledata, "scaled:center")
      std_data <- attr(scaledata, "scaled:scale")
      
      scale_summary <- list(mean_data = mean_data, std_data = std_data)
      
    } else {
      scaledata <- as.matrix(dataset)
      rownames(scaledata) <- rownames(dataset)
      if(!all(is.na(scale_summary))){
        scale_summary = scale_summary
      }else{
        scale_summary = NA
      }
      
    }
    
    polinfo <- hvqdata <- list()
    hvq_k <- hvq(
      scaledata,
      min_compression_perc = min_compression_perc,
      n_cells = n_cells,
      depth = depth,
      quant.err = quant.err,
      distance_metric = distance_metric,
      error_metric = error_metric,
      quant_method=quant_method
    )
    
    if(!is.na(min_compression_perc)){
      n_cells <- hvq_k$compression_summary$noOfCells
      depth <- 1
    }
    
    hvqoutput <- hvq_k$summary
    
    gdata <- hvqoutput  #assign the output of hvq file to gdata
    gdata <- hvqoutput[stats::complete.cases(hvqoutput), ]
    hvqdata <- gdata

    #to store hvqdata according to each level
    tessdata <- input.tessdata <- list()
    #stores sammon co-ordinates and segregated according to each level
    points2d <- list()
    #sammon datapoints to be stored in hierarchy
    rawdeldata <- list()
    #rawdeldata is transformed to be within its parent tile and stored in new_rawdeldata
    new_rawdeldata <- list()
    #contains the vertices of the parent polygon
    pol_info <- polygon_info <- list()
    #number of levels
    nlevel <- length(unique(gdata[, "Segment.Level"]))
    #verify if the transformed points are correct
    transpoints <- list()
    
    # New columns added to the dataset by the hvq function
    newcols <-  ncol(hvqoutput) - ncol(dataset)
    # Variable to store information on mapping
    par_map <- list()
    
    
    projection.scale <-  10
    for (i in 1:nlevel) {
      tessdata[[i]] <- gdata[which(gdata[, "Segment.Level"]==i),]
     
    }
    
   
    counter <- c(1:nlevel)
    
    
    dim_reduction <- function(x){
    requireNamespace("Rtsne")        
    requireNamespace("umap")
    requireNamespace("FNN")
    requireNamespace("MASS")
      data <- unique(tessdata[[x]][,(newcols +1):ncol(hvqoutput)])
      
      if (dim_reduction_method == "sammon"){
        return( projection.scale * MASS::sammon(stats::dist(data),niter = 10^5)$points)
      }else if (dim_reduction_method == "tsne"){
        set.seed(232)
        tsne_data <- (projection.scale * Rtsne::Rtsne(data,
                                                      theta = tsne_theta,
                                                      eta = tsne_eta,
                                                      perplexity = tsne_perplexity,
                                                      verbose = tsne_verbose,
                                                      max_iter = tsne_max_iter)$Y)
        rownames(tsne_data) <- rownames(tessdata[[x]])
        return(tsne_data)
      }else if (dim_reduction_method == "umap"){
        set.seed(232)
        return(projection.scale * umap::umap(data,
                                             n_neighbors = umap_n_neighbors,
                                             n_components = umap_n_components, 
                                             min_dist = umap_min_dist)$layout)
      }
    }
    start_time <- Sys.time()
    points2d <- lapply(counter,dim_reduction)
    end_time <-Sys.time()
    
    time_for_projection <- as.numeric(end_time - start_time)
    
    ## Function for Trustworthiness
    calculate_trustworthiness <- function(X, X_embedded, k = 5) {
      n <- nrow(X)
      # Compute the pairwise distance matrices
      dist_original <- as.matrix(dist(X))
      dist_embedded <- as.matrix(dist(X_embedded))
      # Compute ranks
      rank_original <- t(apply(dist_original, 1, rank))
      rank_embedded <- t(apply(dist_embedded, 1, rank))
      # Determine which points are in the top k in the embedded space but not in the original space
      U_k_matrix <- (rank_embedded <= k) & !(rank_original <= k)
      # Calculate the rank differences
      rank_differences <- (rank_original - k) * U_k_matrix
      # Sum all rank differences
      trustworthiness_sum <- sum(rank_differences[rank_differences > 0])
      # Normalize the trustworthiness score
      trustworthiness <- 1 - (2 / (n * k * (2 * n - 3 * k - 1)) * trustworthiness_sum)
      return(trustworthiness)
    }
    
    ## Function For Continuity
    calculate_continuity <- function(X, X_embedded, k = 5) {
      n <- nrow(X)
      # Compute the pairwise distance matrices
      dist_original <- as.matrix(dist(X))
      dist_embedded <- as.matrix(dist(X_embedded))
      # Compute ranks
      rank_original <- t(apply(dist_original, 1, rank))
      rank_embedded <- t(apply(dist_embedded, 1, rank))
      # Determine which points are in the top k in the original space but not in the embedded space
      U_k_matrix <- (rank_original <= k) & !(rank_embedded <= k)
      # Calculate the rank differences
      rank_differences <- (rank_embedded - k) * U_k_matrix
      # Sum all rank differences
      continuity_sum <- sum(rank_differences[rank_differences > 0])
      # Normalize the continuity score
      continuity <- 1 - (2 / (n * k * (2 * n - 3 * k - 1)) * continuity_sum)
      return(continuity)
    }
    
    ## RMSE
    calculate_normalized_rmse <- function(embeddingss) {
      
      # Ensure the original data is in matrix form
      if (is.list(embeddingss[["original"]])) {
        original_data <- do.call(cbind, embeddingss[["original"]])  # Convert list to matrix
      } else {
        original_data <- as.matrix(embeddingss[["original"]])  # Convert data frame to matrix if needed
      }
      
      # Replace the 'original' entry with the matrix
      embeddingss[["original"]] <- original_data
      
      # Check the number of points in the original and reduced data
      if (nrow(embeddingss[["original"]]) != nrow(embeddingss[["points_df"]])) {
        stop("The number of points in the original data and the reduced 2D data do not match.")
      }
      
      # Compute pairwise distances in the original space (3D)
      original_distances <- as.matrix(dist(embeddingss[["original"]], method = "euclidean"))
      
      # Compute pairwise distances in the reduced space (2D)
      reduced_distances <- as.matrix(dist(embeddingss[["points_df"]], method = "euclidean"))
      
      # Check if the distance matrices are conformable
      if (!all(dim(original_distances) == dim(reduced_distances))) {
        stop("The distance matrices are not conformable for RMSE calculation.")
      }
      
      # return(rmse_normalized)
      difference_matrix <- original_distances - reduced_distances
      
      # Square the differences
      squared_differences <- difference_matrix^2
      
      # Calculate the mean of the squared differences
      mean_squared_error <- mean(squared_differences)
      
      # Calculate the RMSE by taking the square root of the mean squared error
      rmse <- sqrt(mean_squared_error)
      
      return(rmse)
    }
    
    ###### Silhouette Score
    calculate_average_silhouette <- function(data, centers = 2, seed = 123) {
      
      set.seed(seed)
      
      # Apply k-means clustering
      kmeans_result <- kmeans(data, centers = centers)
      
      # Vectorized function to compute Silhouette score manually
      calculate_silhouette_vectorized <- function(data, clusters) {
        dist_matrix <- as.matrix(dist(data))
        
        # Compute the intra-cluster distance (a_i)
        a_i <- sapply(1:length(clusters), function(i) {
          cluster_points <- which(clusters == clusters[i])
          if (length(cluster_points) > 1) {
            mean(dist_matrix[i, cluster_points])
          } else {
            0
          }
        })
        
        # Compute the nearest-cluster distance (b_i)
        b_i <- sapply(1:length(clusters), function(i) {
          other_clusters <- unique(clusters)[unique(clusters) != clusters[i]]
          min(sapply(other_clusters, function(cluster) {
            cluster_points <- which(clusters == cluster)
            mean(dist_matrix[i, cluster_points])
          }))
        })
        
        # Calculate Silhouette scores
        sil_scores <- (b_i - a_i) / pmax(a_i, b_i)
        return(sil_scores)
      }
      
      # Calculate Silhouette scores
      silhouette_scores <- calculate_silhouette_vectorized(data, kmeans_result$cluster)
      
      # Calculate the average Silhouette score
      average_silhouette_score <- mean(silhouette_scores)
      
      return(average_silhouette_score)
    }
    
    ## Sammon's Stresss
    sammon_stress <- function(original_embeddings, projected_embeddings) {
      # Compute pairwise distances
      orig_dist <- as.vector(original_embeddings)
      proj_dist <- as.vector(dist(projected_embeddings))
      
      # Avoid division by zero
      mask <- orig_dist > 0
      orig_dist <- orig_dist[mask]
      proj_dist <- proj_dist[mask]
      
      # Sammon stress formula
      stress <- sum(((orig_dist - proj_dist)^2) / orig_dist) / sum(orig_dist)
      return(stress)
    }

   
    knn_retention <- function(high_dim_embeddings, low_dim_embeddings, k = 8) {
      high_dim_embeddings <- as.matrix(high_dim_embeddings)
      low_dim_embeddings <- as.matrix(low_dim_embeddings)
      
      n <- nrow(high_dim_embeddings)  # Total data points
      k <- min(k, n - 1)  # Ensure k is not greater than available points
      
      # Find k-nearest neighbors in high-dimensional space
      high_dim_nn <- FNN::knn.index(high_dim_embeddings, k = k)
      
      # Find k-nearest neighbors in low-dimensional space
      low_dim_nn <- FNN::knn.index(low_dim_embeddings, k = k)
      
      # Calculate total overlap
      total_overlap <- sum(sapply(1:n, function(i) {
        length(intersect(high_dim_nn[i, ], low_dim_nn[i, ]))
      }))
      
      # Calculate average retention
      avg_retention <- total_overlap / (n * k)
      return(avg_retention)
    }
    
    
    if(depth == 1){
      points2d_df = as.data.frame(points2d)
      new_tessdata = unique(tessdata[[1]][,(newcols +1):ncol(hvqoutput)])
      
      ## Trustworthiness
      trustworthiness <- calculate_trustworthiness(new_tessdata, points2d_df, k = 5)
      
      ## Continuity Score
      continuity_score <- calculate_continuity(new_tessdata, points2d_df, k = 5)
      
      ## RMSE Score
      embeddingss <- list(
        original = new_tessdata,  
        points_df = points2d_df
      )
      rmse_value<- calculate_normalized_rmse(embeddingss)

      ## Silhouette Score
      average_silhouette_score <- calculate_average_silhouette(points2d_df)
      
      ## Sammon's Stress
      # Calculate pairwise distances in the original space
      original_distances <- dist(embeddingss$original)
      # Calculate Sammon's Stress 
      stress_score <- sammon_stress(original_distances, embeddingss$points_df/projection.scale)
      
      ## KNN Retention
      avg_retention_score <- knn_retention(embeddingss[["original"]], embeddingss[["points_df"]], k = 8)
      
      ## Creating the Metrics df with the values.
      metrics_df<- data.frame(
        `L1_Metrics` = c("Structure Preservation Metrics", "Structure Preservation Metrics", "Structure Preservation Metrics", "Distance Preservation Metrics", "Interpretive Quality Metrics", "Interpretive Quality Metrics","Computational Efficiency Metrics"),
        `L2_Metrics` = c("Trustworthiness", "Continuity", "Sammon's Stress", "RMSE", "Silhouette Score", "KNN Retention Score","Execution Duration(sec)"),
        Value = c(trustworthiness, continuity_score, stress_score, rmse_value, average_silhouette_score, avg_retention_score, round(time_for_projection,4))
        # stringsAsFactors = FALSE
      )
    }else{
      
    }
    
   
    
    for (i in 1:nlevel) {
     
      # #sammon datapoints grouped according to the hierarchy
      intermediate.rawdata <- list()
      rn = row.names(points2d[[i]])
      vec = as.integer(rn) - sum(n_cells ^ (0:(i - 1))) + 1
      
      index = 1
      for (j in unique(tessdata[[i]][, "Segment.Parent"])) {
     
        if (any(dplyr::between(j - vec / n_cells, 0, 0.9999))) {
          intermediate.rawdata[[index]] <-
            cbind(points2d[[i]][rn[dplyr::between(j - vec / n_cells, 0, 0.9999)], 1],
                  points2d[[i]][rn[dplyr::between(j - vec /
                                                    n_cells, 0, 0.9999)], 2])
          index = index + 1
        }
      }
      rawdeldata[[i]] <- intermediate.rawdata
      
    }
    rm(intermediate.rawdata)

    #new_rawdeldata contains the transformed points of rawdeldata
    new_rawdeldata[[1]] <- rawdeldata[[1]]
    
    #deldat1 is the output of the deldir::deldir function and contains the tessellation information
    deldat1 <- deldat2 <- list()
    #deldir function of deldir package gives the tessellations output
    deldat2[[1]] <- deldir::deldir(new_rawdeldata[[1]][[1]][, 1],
                                   new_rawdeldata[[1]][[1]][, 2])
    deldat1[[1]] <- deldat2
    rm(deldat2)
    #plotting the tessellation

    #polygon_info stores parent tile vertices information
    #polygon_info is the modified tile.list output except for first level.
    pol_info[[1]] <- deldir::tile.list(deldat1[[1]][[1]])
    
    #loop throught the polygon and  add the row name as id variable
    # row names of info in rawdeldata is preserved from the hvqoutput and this row name ties the rawdeldata back to the hierarchy
    # in the hvqoutput dataset
    pol_info[[1]] <- mapply(function(info, id) {
      info$id = as.numeric(id)
      info <-
        append(info, as.list(gdata[row.names(gdata) == id, c('Segment.Level', 'Segment.Parent', 'Segment.Child')]))
      return(info)
    }, pol_info[[1]], as.list(row.names(new_rawdeldata[[1]][[1]])), SIMPLIFY = FALSE)
    
    polygon_info[[1]] <- pol_info
    rm(pol_info)
    par_tile_indices <- n_par_tile <- list()
    par_tile_indices[[1]] <-
      unique(tessdata[[1]][, "Segment.Parent"])
    n_par_tile[[1]] <-
      length(unique(tessdata[[1]][, "Segment.Parent"]))
    if (nlevel < 2) {
      polinfo <- polygon_info
      
      fin_out <- list()
      
      fin_out[[1]] <- deldat1
      fin_out[[2]] <- polinfo
      fin_out[[3]] <- hvq_k
      fin_out[[6]] <- hvq_k
      fin_out[[3]][['scale_summary']] <- scale_summary
      level = 1
      level_names <- list()
      level_names[[level]] <- fin_out[[2]][[level]] %>% purrr::map(~ {
        this <- .
        element <- this[[1]]
        return(element$Segment.Parent)
      }) %>% unlist()
      
      names(fin_out[[2]][[level]]) <- level_names[[level]]
      
      ####### Adding Code for Diagnosis Plot
      diag_Suggestion=NA
      fin_out[[4]] = NA
      fin_out[[5]] <- NA
      if(diagnose){
        # fin_out[[4]] = NA
        diag_list =  diagPlot(hvt.results = fin_out,
                              data = dataset,
                              level = depth,
                              quant.err = quant.err,
                              distance_metric=distance_metric,
                              error_metric=error_metric
        )
        diag_Suggestion =  diagSuggestion(hvt.results = fin_out,
                                          data = dataset,
                                          level = depth,
                                          quant.err = quant.err,
                                          distance_metric=distance_metric,
                                          error_metric=error_metric
        )
        fin_out[[4]] = diag_list
        
      }
      
      ####### Adding Code for MAD Plot  
      if(hvt_validation){
        ####### MAD Plot ################
        
        predictions_validation = list()
        predictions_validation <- scoreHVT(
          dataset = validation_data,
          hvt.results.model=fin_out,
          child.level = depth,
          line.width = c(0.6, 0.4, 0.2),
          color.vec = c("#141B41", "#6369D1", "#D8D2E1"),
          mad.threshold = quant.err,
          distance_metric = distance_metric,
          error_metric = error_metric
        )
        
        mad_plot=madPlot(predictions_validation)
        fin_out[[5]] <- list()
        fin_out[[5]][["mad_plot"]] <- mad_plot
      }
      ### Model Info
      Dataset <- paste0(data_structure[1] ," Rows & ", data_structure[2], " Columns")
      
      model_train_date=Sys.time()
      input_parameters = list(
        input_dataset = Dataset,
        n_cells = n_cells,
        depth = depth,
        quant.err = quant.err,
        normalize = normalize,
        distance_metric = distance_metric[1],
        error_metric = error_metric[1],
        quant_method = quant_method[1],
        diagnose = diagnose,
        hvt_validation = hvt_validation,
        train_validation_split_ratio = train_validation_split_ratio
      )
      
#browser()    
  
      if(depth == 1){
        
        validation_result=diag_Suggestion
        fin_out[["model_info"]]=list(model_train_date=model_train_date,
                                     type="hvt_model",
                                     input_parameters = input_parameters,
                                     validation_result = validation_result,
                                     distance_measures = metrics_df)
      }else{
        validation_result=diag_Suggestion
        fin_out[["model_info"]]=list(model_train_date=model_train_date,
                                     type="hvt_model",
                                     input_parameters = input_parameters,
                                     validation_result = validation_result)
      }
      

      
      
      # generating cell ID using getCellId
      fin_out[[3]]$summary <- getCellId(hvt.results=fin_out)
      fin_out[[3]]$summary <- fin_out[[3]]$summary %>% select(Segment.Level, Segment.Parent, Segment.Child, n, Cell.ID, Quant.Error, colnames(dataset))
      
      
  
      
      
      
      class(fin_out) <- "hvt.object"
      return(fin_out)
      
    } else { 
      for (i in 2:nlevel) {
        new_rawdeldata[[i]] <- list()
        par_tile_indices[[i]] <-
          unique(tessdata[[i]][, "Segment.Parent"])
        n_par_tile[[i]] <-
          length(unique(tessdata[[i]][, "Segment.Parent"]))
        par_map[[i - 1]] <- list()
        
        for (tileIndex in 1:n_par_tile[[(i - 1)]]) {
          #a chunk of hvqdata which contains the rows corresponding to a particular parent tile
          gi_par_tiles <-
            par_tile_indices[[i]][intersect(which((par_tile_indices[[i]] / n_cells) <= par_tile_indices[[(i - 1)]][tileIndex]),
                                            which((par_tile_indices[[(i - 1)]][tileIndex] - 1) < (par_tile_indices[[i]] / n_cells)))]
          gidata <-
            tessdata[[i]][which(tessdata[[i]][, "Segment.Parent"] %in% gi_par_tiles),]
          par_map[[i - 1]][[par_tile_indices[[(i - 1)]][tileIndex]]] <-
            gi_par_tiles
          
          #datapoints corresponding to a particular parent tile
          rawdeldati <-
            rawdeldata[[i]][intersect(which((par_tile_indices[[i]] / n_cells) <= par_tile_indices[[(i - 1)]][tileIndex]),
                                      which((par_tile_indices[[(i - 1)]][tileIndex] - 1) < (par_tile_indices[[i]] / n_cells)))]
          if (nrow(gidata) != 0) {
            #transformation of rawdeldata such that they are inside the parent tile and output is stored in new_rawdeldata
            transpoints <-
              DelaunayInfo(gidata, polygon_info[[(i - 1)]][[tileIndex]], rawdeldati, n_cells)
            if (all(transpoints != "-1")) {
              new_rawdeldata[[i]] <- append(new_rawdeldata[[i]], transpoints)
            } else{
              deldat1[i] <- 0
              polinfo <<- polygon_info
              return(deldat1)
            }
            
          }
        }

        deldat2 <- list()
        par_tile_polygon <- list()
        for (tileNo in 1:n_par_tile[[i]]) {
          #modulus operation for the last index in polygon_info
          if (((par_tile_indices[[i]][tileNo]) %% n_cells)) {
            last_index <- (par_tile_indices[[i]][tileNo] %% n_cells)
          } else{
            last_index <- n_cells
          }
          
          #divide to get the parent tile
          sec_index <-
            Hmisc::ceil(par_tile_indices[[i]][tileNo] / n_cells)
          
          deldat2[[tileNo]] <-
            deldir::deldir(new_rawdeldata[[i]][[tileNo]][, 1],
                           new_rawdeldata[[i]][[tileNo]][, 2],
                           rw = c(
                             range(polygon_info[[(i - 1)]][[which(par_tile_indices[[(i - 1)]] == sec_index)]][[last_index]]$x) - c(0.5,-0.5),
                             range(polygon_info[[(i - 1)]][[which(par_tile_indices[[(i - 1)]] == sec_index)]][[last_index]]$y) - c(0.5,-0.5)
                           ))
          
          #CORRECT ROW NAMES
          deldat2[[tileNo]]$rownames.orig <-
            as.numeric(row.names(new_rawdeldata[[i]][[tileNo]]))
          
          #constructing parent polygons
          par_tile_polygon[[tileNo]] <-
            matrix(
              c(polygon_info[[(i - 1)]][[which(par_tile_indices[[(i - 1)]] == sec_index)]][[last_index]]$x,
                polygon_info[[(i - 1)]][[which(par_tile_indices[[(i - 1)]] == sec_index)]][[last_index]]$y),
              ncol = 2,
              byrow = FALSE
            )
          #correct the tessellations
          cur_dirsgs <- deldat2[[tileNo]]$dirsgs
          cur_tile <-
            polygon_info[[(i - 1)]][[which(par_tile_indices[[(i - 1)]] == sec_index)]][[last_index]]
          cur_polygon <- par_tile_polygon[[tileNo]]
          verify_dirsgs <- Corrected_Tessellations(cur_dirsgs,
                                                   cur_tile,
                                                   cur_polygon)
          
          #polygon is sufficient to calculate next level tessellations inside this
          if (all(verify_dirsgs[, 1:8] != "-1")) {
            deldat2[[tileNo]]$dirsgs <- verify_dirsgs
          } else{
            # deldat1[[i]] <- 0
            # polinfo <<- polygon_info
            # return(deldat1)
            stop(paste0(
              "Tessellation failed at depth ", i, ": polygon is too small for further subdivision. ",
              "Try reducing 'depth', increasing 'n_cells', or using a larger dataset."
            ))
          }
          
          #delete lines with both the points identical
          new_dirsgs <- deldat2[[tileNo]]$dirsgs
          del_rows <- 0
          for (j in 1:nrow(new_dirsgs)) {
            if (round(new_dirsgs[j, "x1"], 6) == round(new_dirsgs[j, "x2"], 6) &&
                round(new_dirsgs[j, "y1"], 6) == round(new_dirsgs[j, "y2"], 6)) {
              del_rows <- c(del_rows, j)
            }
          }
          if (length(del_rows) > 1) {
            new_dirsgs <- new_dirsgs[-del_rows, ]
          }
          deldat2[[tileNo]]$dirsgs <- new_dirsgs
        }
        
        deldat1[[i]] <- deldat2
        
        polygon_info[[i]] <- list()
        #polygon information to correct the polygons
        for (parentIndex in 1:n_par_tile[[i]]) {
          polygon_info[[i]][[parentIndex]] <-
            suppressMessages(deldir::tile.list(deldat1[[i]][[parentIndex]]))
          # Add id corresponding to each segment from rownames.orig information
          polygon_info[[i]][[parentIndex]] <-
            mapply(function(info, id) {
              info$id = as.numeric(id)
              info <-
                append(info, as.list(gdata[row.names(gdata) == id, c('Segment.Level', 'Segment.Parent', 'Segment.Child')]))
              return(info)
            },
            polygon_info[[i]][[parentIndex]],
            as.list(deldat1[[i]][[parentIndex]]$rownames.orig),
            SIMPLIFY = FALSE)
        }
        
        #to delete the points which are outside the parent tile
        
        for (parentIndex in 1:length(polygon_info[[i]])) {
          for (tileIndex in 1:length(polygon_info[[i]][[parentIndex]])) {
            tile <- polygon_info[[i]][[parentIndex]][[tileIndex]]
            check_dirsgs <- deldat1[[i]][[parentIndex]]$dirsgs
            if (nrow(check_dirsgs) != 0 && length(tile$x) != 0) {
              polygon_info[[i]][[parentIndex]][[tileIndex]] <-
                Delete_Outpoints (tile, check_dirsgs)
            }
          }
        }

        for (parentIndex in 1:n_par_tile[[i]]) {
          #modulo operation to obtain the last index
          if (((par_tile_indices[[i]][parentIndex]) %% n_cells)) {
            last_index <- (par_tile_indices[[i]][parentIndex] %% n_cells)
          } else{
            last_index <- n_cells
          }
          #divide to get second index
          sec_index <-
            Hmisc::ceil(par_tile_indices[[i]] / n_cells)[parentIndex]
          
          polygon_info[[i]][[parentIndex]] <-
            Add_boundary_points(polygon_info[[i]][[parentIndex]],
                                polygon_info[[(i - 1)]][[which(par_tile_indices[[(i - 1)]] == sec_index)]][[last_index]],
                                new_rawdeldata[[i]][[parentIndex]])
        }
      
      }
      
      polinfo <- polygon_info
      
    
      fin_out <- list()
      
      fin_out[[1]] <- deldat1
      fin_out[[2]] <- polinfo
      fin_out[[3]] <- hvq_k
      fin_out[[6]] <- hvq_k
      fin_out[[3]][['scale_summary']] <- scale_summary
      level_names <- list()
      
      for (level in 1:nlevel) {
        level_names[[level]] <- fin_out[[2]][[level]] %>% purrr::map(~ {
          this <- .
          element <- this[[1]]
          return(element$Segment.Parent)
        }) %>% unlist()
        
        names(fin_out[[2]][[level]]) <- level_names[[level]]
      }
      

      emptyParents  <- depth - length(fin_out[[2]])
      for (i in 1:emptyParents) {
        fin_out[[2]][[length(fin_out[[2]]) + 1]] <- list()
      }

      
      
      
            
      # ###### Adding Code for Diagnosis Plot
      diag_Suggestion =NA
      fin_out[[4]] = NA
      fin_out[[5]] <- NA
      
      if(diagnose){
        fin_out[[4]] = NA
        diag_list = diagPlot(hvt.results = fin_out,
                             data = dataset,
                             level = depth,
                             quant.err = quant.err,
                             distance_metric=distance_metric,
                             error_metric=error_metric
        )
        
        diag_Suggestion = diagSuggestion(hvt.results = fin_out,
                                         data = dataset,
                                         level = depth,
                                         quant.err = quant.err,
                                         distance_metric=distance_metric,
                                         error_metric=error_metric
        )
        fin_out[[4]] = diag_list
        
      }
      

      ####### Adding Code for MAD Plot
      if(hvt_validation){
        # browser()
        
        ####### MAD Plot ################
        predictions_validation = list()
        predictions_validation <- scoreHVT(
          dataset = validation_data,
          hvt.results.model=fin_out,
          child.level = depth,
          line.width = c(0.6, 0.4, 0.2),
          color.vec = c("#141B41", "#6369D1", "#D8D2E1"),
          mad.threshold = quant.err,
          distance_metric = distance_metric,
          error_metric = error_metric
        )
        
        mad_plot=madPlot(predictions_validation)
        fin_out[[5]] <- list()
        fin_out[[5]][["mad_plot"]] <- mad_plot
      }
      

      ### Model Info
      Dataset <- paste0(data_structure[1] ," Rows & ", data_structure[2], " Columns")
      
      model_train_date=Sys.time()
      input_parameters = list(
        input_dataset = Dataset,
        n_cells = n_cells,
        depth = depth,
        quant.err = quant.err,
        normalize = normalize,
        distance_metric = distance_metric[1],
        error_metric = error_metric[1],
        quant_method = quant_method[1],
        diagnose = diagnose,
        hvt_validation = hvt_validation,
        train_validation_split_ratio = train_validation_split_ratio,
        dim_reduction_method = dim_reduction_method[1],
        tsne_theta = tsne_theta[1],
        tsne_eta = tsne_eta[1],
        tsne_perplexity = tsne_perplexity[1],
        tsne_verbose = tsne_verbose[1],
        tsne_max_iter = tsne_max_iter[1],
        umap_n_neighbors = umap_n_neighbors[1],
        umap_n_components = umap_n_components[1],
        umap_min_dist = umap_min_dist[1]
        
      )
      
      validation_result=diag_Suggestion
      fin_out[["model_info"]]=list(model_train_date=model_train_date,
                                   type="hvt_model",
                                   input_parameters=input_parameters,
                                   validation_result=validation_result
      )
      
      # generating cell ID using getCellId
      fin_out[[3]]$summary <- getCellId(hvt.results=fin_out)
      fin_out[[3]]$summary <- fin_out[[3]]$summary %>% select(Segment.Level, Segment.Parent, Segment.Child, n, Cell.ID, Quant.Error, colnames(dataset))
 
      
      class(fin_out) <- "hvt.object"
      return(fin_out)
    }
  }

#' @name getTransitionProbability
#' @title Creating Transition Probability Matrix
#' @description This is the main function to create transition probability matrix
#' The transition probability matrix quantifies the likelihood of transitioning from one state to another. 
#' States: The table includes the current states and the possible next states.
#' Probabilities: For each current state, it lists the probability of transitioning to each of the next possible states. 
#' @param df Data frame. The input data frame should contain two columns, 
#' cell ID from scoreHVT function and time stamp of that dataset.
#' @param cellid_column Character. Name of the column containing cell IDs.
#' @param time_column Character. Name of the column containing time stamps.
#' @param type Character. A character value indicating the type of transition probability table to create.
#' Accepted entries are "with_self_state" and "without_self_state".
#' @return Stores a data frames with transition probabilities.
#' @author PonAnuReka Seenivasan <ponanureka.s@@mu-sigma.com>, Vishwavani <vishwavani@@mu-sigma.com>
#' @keywords Timeseries_Analysis
#' @importFrom magrittr %>%
#' @examples
#' dataset <- data.frame(t = as.numeric(time(EuStockMarkets)),
#'                       DAX = EuStockMarkets[, "DAX"],
#'                       SMI = EuStockMarkets[, "SMI"],
#'                       CAC = EuStockMarkets[, "CAC"],
#'                       FTSE = EuStockMarkets[, "FTSE"])
#' hvt.results<- trainHVT(dataset[-1],n_cells = 60, depth = 1, quant.err = 0.1,
#'                        distance_metric = "L1_Norm", error_metric = "max",
#'                        normalize = TRUE,quant_method = "kmeans")
#' scoring <- scoreHVT(dataset, hvt.results)
#' cell_id <- scoring$scoredPredictedData$Cell.ID
#' time_stamp <- dataset$t
#' dataset <- data.frame(cell_id, time_stamp)
#' table <- getTransitionProbability(dataset, cellid_column = "cell_id",time_column = "time_stamp")
#' @export getTransitionProbability


getTransitionProbability <- function(df, cellid_column, time_column, type = "with_self_state") {
    Current_State <- Transition_Probability <- Cumulative_Probability <- Next_State <- Relative_Frequency <- NULL
  
  # Rename columns for consistency
  colnames(df)[colnames(df) == time_column] <- "Timestamp"
  colnames(df)[colnames(df) == cellid_column] <- "Cell.ID"
  
  
   if (type == "with_self_state" || type == "without_self_state") {
  
  # Get a sorted list of unique Cell.ID values
  cell_id_list <- unique(df$Cell.ID) %>% sort()
  
  prob_results <- lapply(cell_id_list, function(state) {
    # Find rows with current state
    row_numbers <- which(df$Cell.ID == state)
    
    # Check if the next state exists (not at the end of dataset)
    valid_rows <- row_numbers[row_numbers + 1 <= nrow(df)]
    
    # If no valid transitions exist, return NULL or empty dataframe
    if(length(valid_rows) == 0) {
      return(NULL)
    }
    
    # Get next states only for valid rows
    tplus1_states <- df[valid_rows + 1, "Cell.ID"]
    
    prob_table <- table(tplus1_states)
    total_count <- sum(prob_table)
    probabilities <- as.vector(prob_table) / total_count
    
    result_df <- data.frame(
      Current_State = state,
      Next_State = as.numeric(names(prob_table)),
      Relative_Frequency = as.vector(prob_table),
      Transition_Probability = round(probabilities, 4)
    )
    
    result_df <- result_df %>%
      group_by(Current_State) %>%
      mutate(Cumulative_Probability = round(cumsum(Transition_Probability), 4),
             Cumulative_Probability = if_else(row_number() == n(), 1, Cumulative_Probability)) %>% 
      as.data.frame()
    
    return(result_df)
  })
  
  # Remove NULL results and combine remaining dataframes
  prob_results <- do.call(rbind, prob_results[!sapply(prob_results, is.null)])
  

    trans_prob_matx <- prob_results %>%
      group_by(Current_State) %>%
      filter(!(Current_State == Next_State & n() > 1)) %>%
      mutate(
        Transition_Probability = round((Relative_Frequency / sum(Relative_Frequency)), 4),
        Cumulative_Probability = round(cumsum(Transition_Probability),4),
        Cumulative_Probability = if_else(row_number() == n(), 1, Cumulative_Probability)
      ) %>%
      ungroup()
    
    trans_prob_matx$Next_State <- as.numeric(trans_prob_matx$Next_State)

  } else {
    stop("Invalid input for `type` argument. Only `with_self_state` , `without_self_state`, or `mean_var_table` is allowed.")
  }
    
  
  if (type == "with_self_state") {
    return(prob_results)
  } else if (type == "without_self_state") {
    return(trans_prob_matx)}
 
}

#' @name HVTMSMoptimization
#' @title HVT-MSM Optimization Function
#' @description This function runs multiple iterations/experiments over the dataset across 
#' different cell counts, and computes the MAE. If a given cell configuration 
#' not results in problematic states, the baseline simulation is used; otherwise, 
#' the simulation proceeds with problematic state handling. This process helps identify 
#' the best-performing model (lowest MAE). In essence, it performs 
#' trainHVT → scoreHVT → transition probability estimation on the training 
#' data, followed by msm simulation on the ex-post (test) dataset.  
#' Note: This is not applicable for ex-ante analysis.
#' @param entire_dataset Dataframe. Train dataset for model training
#' @param expost_forecasting Dataframe. Test dataset for ex-post forecasting  
#' @param time_column Character. Name of the time column
#' @param ncell_range Numeric vector. Range of cells to run experiments (default 3:5)
#' @param k_range Numeric vector. Range of clusters to run experiments (default: 2:9)
#' @param nn_range Numeric vector. Range of nearest neighbors to run experiments (default: 2:7)
#' @param num_simulations Integer. Number of simulations (default: 10)
#' @param mae_metric Character. MAE calculation method(s): "mean", "median", "mode", or
#'  "all" (default: "median")
#' @param hvt_params List. Set of parameters for Model Training (refer trainHVT)
#' @param parallel Character. Whether to use parallel processing (default: TRUE)
#' @param verbose Character. Whether to print progress information (default: TRUE)
#' @param parallel_strategy Character. Parallel processing strategy: "multisession", 
#' "multicore", etc. (default: "multisession")
#' @param max_workers Maximum number of parallel workers 
#' (default: NULL for auto-detect) 
#' @return List containing optimization results for each MAE metric:
#' \item{[[successful_results]] }{All successful parameter combinations} 
#' \item{[[nclust_best_results]] }{Best combination for each cell}
#' \item{[[overall_best]] }{Overall best parameter combination}
#' \item{[[all_results]] }{All attempted combinations with status}
#' @author Vishwavani <vishwavani@@mu-sigma.com>, Nithya <nithya.sn@@mu-sigma.com>
#' @keywords Hyperparameter_Tuning
#' @include HVTMSM_support.R
#' @importFrom magrittr %>%
#' @export HVTMSMoptimization

HVTMSMoptimization <- function(entire_dataset,
                               expost_forecasting,
                               time_column,
                               ncell_range = 3:5,
                               k_range = 2:9,
                               nn_range = 2:7,
                               num_simulations = 10,
                               mae_metric,
                               hvt_params = list(
                                            depth = 1,
                                            quant.err = 0.2,
                                            normalize = TRUE,
                                            distance_metric = "L1_Norm",
                                            error_metric = "max",
                                            dim_reduction_method = "sammon"
                              ),
                               parallel = TRUE,
                               verbose = TRUE,
                               parallel_strategy = "multisession",
                               max_workers = NULL) {
  
  # Validate inputs
  if (!time_column %in% names(entire_dataset)) {
    stop("Time column '", time_column, "' not found in entire_dataset")
  }
  if (!time_column %in% names(expost_forecasting)) {
    stop("Time column '", time_column, "' not found in expost_forecasting")
  }
  
  if (hvt_params$depth >= 2) {
    stop("HVTMSMOptimization doesn't work for depth equal to or greater than 2")
  }
  
  # Normalize MAE metric
  if (length(mae_metric) == 1 && mae_metric == "all") {
    mae_metric <- c("mean", "median", "mode")
  }
  if (!all(mae_metric %in% c("median", "mean", "mode"))) {
    stop("mae_metric must be one or more of: 'mean', 'median', 'mode', or 'all'")
  }
  
  # Validate k_range and nn_range against ncell_range
  min_ncell <- min(ncell_range)
  max_k <- max(k_range)
  max_nn <- max(nn_range)
  
  # Check if k_range is compatible with ncell_range
  if (max_k >= min_ncell) {
    stop("Invalid parameter ranges: max(k_range) = ", max_k, 
         " must be less than min(ncell_range) = ", min_ncell, 
         ".\nk represents number of clusters and must always be less than the number of cells available.",
         "\nSuggestion: Set k_range with max value < ", min_ncell, 
         " or increase min(ncell_range) to > ", max_k)
  }
  
  # Check if nn_range is compatible with ncell_range and k_range
  min_k <- min(k_range)
  max_possible_nn <- min_ncell - min_k
  if (max_nn > max_possible_nn) {
    stop("Invalid parameter ranges: max(nn_range) = ", max_nn,
         " exceeds maximum possible neighbors = ", max_possible_nn,
         " (calculated as min(ncell_range) - min(k_range) = ", min_ncell, " - ", min_k, ").",
         "\nSuggestion: Set nn_range with max value <= ", max_possible_nn,
         " or increase min(ncell_range) or reduce min(k_range)")
  }
  
  # Setup parallel processing
  if (parallel) {
    if (!requireNamespace("furrr", quietly = TRUE) || !requireNamespace("future", quietly = TRUE)) {
      stop("Packages 'furrr' and 'future' are required for parallel processing. ",
           "Please install them with: install.packages(c('furrr', 'future'))")
    }
    if (is.null(max_workers)) {
      max_workers <- min(parallel::detectCores() - 1, 16)
    }
    if (verbose) cat("Setting up parallel processing with", max_workers, "workers\n")
    future::plan(parallel_strategy, workers = max_workers)
    on.exit(future::plan(future::sequential))
  }
  
  # Prepare data
  numeric_dataset <- entire_dataset %>% dplyr::select(-dplyr::all_of(time_column))
  n_ahead <- nrow(expost_forecasting)
  time_values <- expost_forecasting[[time_column]]
  raw_dataset_stats <- calculate_dataset_stats(numeric_dataset)
  
  if (verbose) cat("Starting optimization with", length(ncell_range), "cell configurations\n")
  
  # Pre-compute HVT models
  hvt_models <- precompute_hvt_models(
    entire_dataset = entire_dataset,
    ncell_range = ncell_range,
    hvt_params = hvt_params,
    #dependent_variables = dependent_variables,
    verbose = verbose,
    time_column = time_column
  )
  
  # Main optimization process
  if (verbose) cat("Running optimization...\n")
  
  if (parallel) {
    optimization_results <- furrr::future_map(ncell_range, function(nclust) {
      process_single_nclust_with_precomputed_model(
        nclust = nclust,
        hvt_model = hvt_models[[as.character(nclust)]],
        numeric_dataset = numeric_dataset,
        entire_dataset = entire_dataset,
        expost_forecasting = expost_forecasting,
        time_column = time_column,
        k_range = k_range,
        nn_range = nn_range,
        num_simulations = num_simulations,
        time_values = time_values,
        n_ahead = n_ahead,
        mae_metric = mae_metric,
        raw_dataset_stats = raw_dataset_stats,
        parallel = FALSE,
        verbose = verbose
      )
    }, .options = furrr::furrr_options(seed = TRUE))
  } else {
    optimization_results <- purrr::map(ncell_range, function(nclust) {
      process_single_nclust_with_precomputed_model(
        nclust = nclust,
        hvt_model = hvt_models[[as.character(nclust)]],
        numeric_dataset = numeric_dataset,
        entire_dataset = entire_dataset,
        expost_forecasting = expost_forecasting,
        time_column = time_column,
        k_range = k_range,
        nn_range = nn_range,
        num_simulations = num_simulations,
        time_values = time_values,
        n_ahead = n_ahead,
        mae_metric = mae_metric,
        raw_dataset_stats = raw_dataset_stats,
        parallel = parallel,
        verbose = verbose
      )
    })
  }
  
  # Collect and format final results
  final_output <- collect_final_results(optimization_results, mae_metric, verbose)
  
  if (verbose) cat("Optimization complete\n")
  
  return(final_output)
}

#' @name plotQuantErrorHistogram
#' @title Make the quantization error plots for training and scoring.
#' @description This is the function that produces histograms displaying the distribution of Quantization Error (QE) values 
#' for both train and test datasets, highlighting mean values with dashed lines for quick evaluation.
#' @param hvt.results List. A list of hvt.results obtained from the trainHVT function.
#' @param hvt.scoring List. A list of hvt.scoring obtained from the scoreHVT function.
#' @return Returns the ggplot object containing the quantization error distribution plots for 
#' the given HVT results of training and scoring
#' @author Shubhra Prakash <shubhra.prakash@@mu-sigma.com>
#' @seealso \code{\link{plotHVT}}
#' @keywords Diagnostics_or_Validation
#' @importFrom magrittr %>%
#' @import ggplot2 
#' @examples
#' data("EuStockMarkets")
#' dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
#'                      DAX = EuStockMarkets[, "DAX"],
#'                      SMI = EuStockMarkets[, "SMI"],
#'                      CAC = EuStockMarkets[, "CAC"],
#'                      FTSE = EuStockMarkets[, "FTSE"])
#' rownames(EuStockMarkets) <- dataset$date
#' #Split in train and test
#' train <- EuStockMarkets[1:1302, ]
#' test <- EuStockMarkets[1303:1860, ]
#' 
#' hvt.results<- trainHVT(train,n_cells = 60, depth = 1, quant.err = 0.1,
#'                       distance_metric = "L1_Norm", error_metric = "max",
#'                       normalize = TRUE, quant_method = "kmeans")
#' scoring <- scoreHVT(test, hvt.results)
#' plotQuantErrorHistogram(hvt.results, scoring) 
#' @export plotQuantErrorHistogram
#' 

plotQuantErrorHistogram <- function(hvt.results, hvt.scoring) {
  if (!requireNamespace("patchwork", quietly = TRUE)) {
    stop("Package 'patchwork' is required but not installed.")
  }
  val <- hvt.results[[3]][["max_QE"]] %>%
    unlist() %>%
    as.numeric()
  p1 <- ggplot2::ggplot() +
    ggplot2::aes(x = val) +
    ggplot2::geom_histogram(fill = "midnightblue", colour = "white", alpha = 0.75) +
    ggplot2::theme_minimal() +
    ggplot2::geom_vline(xintercept = mean(val[val > 0.00000001]), linetype = "dashed") +
    ggplot2::geom_vline(xintercept = mean(val), colour = "red", linetype = "dashed") +
    ggplot2::ggtitle("1.Max QE Distribution: Train ") +
    ggplot2::xlab("Max QE values for cells: Train Data")



  pred_val <- hvt.scoring[["scoredPredictedData"]]
  p2 <- ggplot2::ggplot() +
    ggplot2::aes(x = pred_val$Quant.Error) +
    ggplot2::geom_histogram(fill = "midnightblue", colour = "white", alpha = 0.75) +
    ggplot2::theme_minimal() +
    ggplot2::geom_vline(xintercept = mean(pred_val$Quant.Error), colour = "red", linetype = "dashed") +
    ggplot2::ggtitle("2.Max QE Distribution: Test ") +
    ggplot2::xlab("Max QE values for cells: Test Data")

  plot_output <- p1 / p2
  return(plot_output)
}

#' @keywords internal


diagSuggestion <- function(hvt.results,
                  data,
                  level,
                  quant.err,
                  distance_metric="L1_Norm",
                  error_metric="max",
                  ...){
  
  requireNamespace("reshape2")     
  requireNamespace("dplyr")     

  ##### Min Inter-Centroid distance plot
  newdfMapping <- hvt.results[[3]][["summary"]]
  
  x <- newdfMapping %>% 
    dplyr::filter((n > 0 & Segment.Level == level) | (Segment.Level < level & (Quant.Error < quant.err | n <= 3)))
  x <- x[x%>%stats::complete.cases(),]
  
  singleton_count=sum(x$Quant.Error< 0.0001)
  num_cells=length(x$Quant.Error)
  
  d  <- stats::dist(x,method = "manhattan")
  df <- reshape2::melt(as.matrix(d), varnames = c("row", "col"))
  df <- df[df$value!=0,]
  df$value <- df$value/ncol(x)
  df_cent <- df %>% dplyr::group_by(row) %>% dplyr::summarise(min_dist = min(value, na.rm = TRUE))
  
  
  mean_cent_train=mean(df_cent$min_dist)
  
  ##### Min Inter-Point distance plot
  d = stats::dist(data,method = "manhattan")
  df <- reshape2::melt(as.matrix(d), varnames = c("row", "col"))
  df=df[df$value!=0,]
  df$value=df$value/ncol(data)
  df_data = df %>% dplyr::group_by(row) %>% dplyr::summarise(min_dist = min(value, na.rm = TRUE))
  
  mean_dist_train=mean(df_data$min_dist)

  
  diag_list=list(
    mean_centroid_train = mean_cent_train,
    mean_distance_train = mean_dist_train
  )
  return(diag_list)
  
} 

#' @keywords internal


extract_cell_coordinates <- function(hvt.results) {
  hvt_res1 <- hvt.results[[2]][[1]]$`1`
  hvt_res2 <- hvt.results[[3]]$summary$Cell.ID
  b <- stats::na.omit(as.vector((1:length(hvt_res1))[hvt_res2]))
  coordinates <- do.call(rbind.data.frame, lapply(1:length(hvt_res1), function(x) hvt_res1[[x]]$pt))
  colnames(coordinates) <- c("x", "y")
  coordinates$Cell.ID <- b
  coordinates
}

calculate_polygon_centroid <- function(x_coords, y_coords) {
  n <- length(x_coords)
  if (n < 3) {
    # If not enough points, return mean
    return(list(x = mean(x_coords), y = mean(y_coords)))
  }
  # Close the polygon if not already closed
  if (x_coords[1] != x_coords[n] || y_coords[1] != y_coords[n]) {
    x_coords <- c(x_coords, x_coords[1])
    y_coords <- c(y_coords, y_coords[1])
    n <- n + 1
  }
  # Calculate signed area using shoelace formula
  signed_area <- 0
  cx <- 0
  cy <- 0
  for (i in 1:(n - 1)) {
    cross_product <- x_coords[i] * y_coords[i + 1] - x_coords[i + 1] * y_coords[i]
    signed_area <- signed_area + cross_product
    cx <- cx + (x_coords[i] + x_coords[i + 1]) * cross_product
    cy <- cy + (y_coords[i] + y_coords[i + 1]) * cross_product
  }
  signed_area <- signed_area / 2
  if (abs(signed_area) < 1e-10) {
    # Degenerate polygon, return mean
    return(list(x = mean(x_coords[1:(n-1)]), y = mean(y_coords[1:(n-1)])))
  }
  centroid_x <- cx / (6 * signed_area)
  centroid_y <- cy / (6 * signed_area)
  return(list(x = centroid_x, y = centroid_y))
}


get_cell_label_alignment <- function(position, renderer = c("ggplot", "plotly")) {
  renderer <- match.arg(renderer)
  
  if (renderer == "ggplot") {
    alignments <- list(
      center = list(hjust = 0.5,  vjust =  0.5),
      right  = list(hjust = -0.5, vjust =  0.5),
      left   = list(hjust = 1.5,  vjust =  0.5),
      bottom = list(hjust = 0.5,  vjust =  1.5),
      top    = list(hjust = 0.5,  vjust = -0.5)
    )
  } else {
    alignments <- list(
      center = list(xshift = 0,  yshift =  0, xanchor = "center", yanchor = "middle"),
      right  = list(xshift = 5,  yshift =  0, xanchor = "left",   yanchor = "middle"),
      left   = list(xshift = -5, yshift =  0, xanchor = "right",  yanchor = "middle"),
      bottom = list(xshift = 0,  yshift = -7, xanchor = "center", yanchor = "top"),
      top    = list(xshift = 0,  yshift =  7, xanchor = "center", yanchor = "bottom")
    )
  }
  rlang::`%||%`(alignments[[position]], alignments$bottom)
}

build_boundary_coords <- function(plotList) {
  lapply(plotList, function(x) {
    data.frame(
      "Segment.Level" = x[["Segment.Level"]],
      "Segment.Parent" = x[["Segment.Parent"]],
      "Segment.Child" = x[["Segment.Child"]],
      "x" = x$pt["x"],
      "y" = x$pt["y"],
      "bp.x" = I(x$x),
      "bp.y" = I(x$y)
    )
  }) %>% bind_rows(.)
}


subtract_predicted_actual <- function(data, actual_prefix = "act_", predicted_prefix = "pred_") {
  actual_cols <- grep(paste0("^", actual_prefix), names(data), value = TRUE)
  df_new <- data.frame(matrix(ncol = 1, nrow = nrow(data)))
  temp0  <- data.frame(matrix(nrow = nrow(data)))
  for (col in actual_cols) {
    predicted_col <- gsub(actual_prefix, predicted_prefix, col)
    if (predicted_col %in% names(data)) {
      temp0[[predicted_col]] <- abs(data[[col]] - data[[predicted_col]])
    }
  }
  temp0 <- temp0 %>% purrr::discard(~ all(is.na(.) | . == ""))
  df_new[, 1] <- rowMeans(temp0)
  return(df_new)
}

#' @include ScaleMat.R  Transform_Coordinates.R
#' @keywords internal

DelaunayInfo <-
function(gidata, poly_info, rawdeldati, n_cells){

  requireNamespace("deldir")       #deldir function
  requireNamespace("grDevices")    #chull function
  requireNamespace("splancs")      #csr function
  requireNamespace("sp")           #point.in.polygon function
  requireNamespace("conf.design")  #factorize function

  new_pt <-  matrix(0, ncol = 2, nrow = 1)
  #function to perform the transformation to get the new points which will be inside the parent polygon
  New_Points <- function(scale_translate_info){
    mat_scale <- matrix(c(scale_translate_info[1], 0, 0, scale_translate_info[2]), ncol = 2)
    mat_tran <- matrix(c(scale_translate_info[3], scale_translate_info[4]), ncol = 1)
    new_pt[1, ] <- t((mat_scale %*% matrix(c(rawdeldati[[ind1]][j, 1],
                                             rawdeldati[[ind1]][j, 2]), ncol = 1)) + mat_tran)
    return(new_pt)
  }

  #function to chop parent tile into smaller boxes
  Parent_Tile_Boxes <- function(output_polygon){

    xrange_parent_poly <- range(output_polygon[, 1])
    yrange_parent_poly <- range(output_polygon[, 2])
    #divide the range into segments
    xseq_parent <- seq(xrange_parent_poly[1], xrange_parent_poly[2], length.out = 15)
    yseq_parent <- seq(yrange_parent_poly[1], yrange_parent_poly[2], length.out = 15)

    Points_on_ylines <- Points_on_xlines <- y_points_inside_polygon <- x_points_inside_polygon <- list()

    for(i in 1: length(xseq_parent)){
      #points on line with constant 'y'
      Points_on_ylines[[i]] <- matrix(c(xseq_parent,
                                        rep(yseq_parent[i], length(xseq_parent))),
                                      ncol = 2, byrow = FALSE)
      #points on line with constant 'x'
      Points_on_xlines[[i]] <- matrix(c(rep(xseq_parent[i], length(yseq_parent)),
                                        yseq_parent), ncol = 2, byrow = FALSE)
    }

    for(i in 1: length(xseq_parent)){
      #points on line with constant 'x' and those that are inside the polygon
      x_points_inside_polygon[[i]] <- Points_on_xlines[[i]][c(which(sp::point.in.polygon(Points_on_xlines[[i]][, 1],
                                                                                     Points_on_xlines[[i]][, 2],
                                                                                     output_polygon[, 1],
                                                                                     output_polygon[, 2]) != 0)), ,drop = FALSE]
      #lines(x_points_inside_polygon[[i]][, 1], x_points_inside_polygon[[i]][, 2])
    }

    #constructing smaller polygons inside parent polygon
    smaller_polygons_within_parent <- list()
    #npoly <-0
    no_of_small_poly <- 1
    for(i in 1: (length(x_points_inside_polygon) - 1)){

      #if there are 2 or more points on a line
      if(length(x_points_inside_polygon[[i]][, 1]) > 1){

        for(j in 1: (length(x_points_inside_polygon[[i]][, 1]) - 1)){

          #if the consecutive y co-ordinates are present in the next line
          if(x_points_inside_polygon[[i]][j, 2] %in% x_points_inside_polygon[[(i + 1)]][, 2] &&
               x_points_inside_polygon[[i]][(j + 1), 2] %in% x_points_inside_polygon[[(i + 1)]][, 2]){

            smaller_polygons_within_parent[[no_of_small_poly]] <- matrix(c(x_points_inside_polygon[[i]][j, ],
                                                                           x_points_inside_polygon[[(i + 1)]][1, 1],
                                                                           x_points_inside_polygon[[i]][j, 2],
                                                                           x_points_inside_polygon[[(i + 1)]][1, 1],
                                                                           x_points_inside_polygon[[i]][(j + 1), 2],
                                                                           x_points_inside_polygon[[i]][j, 1],
                                                                           x_points_inside_polygon[[i]][(j + 1), 2]),
                                                                         ncol = 2, byrow = TRUE)
            no_of_small_poly <- no_of_small_poly + 1
          }
        }
      }
    }
    return(smaller_polygons_within_parent)
  }

  #function to chop sammon space into smaller boxes
  Sammon_Space_Boxes <- function(input_polygon, nbox){

    requireNamespace("grDevices")    #chull function
    requireNamespace("splancs")      #csr function
    requireNamespace("sp")           #point.in.polygon function
    requireNamespace("conf.design")  #factorize function

    xrange_sammon_poly <- range(input_polygon[, 1])
    yrange_sammon_poly <- range(input_polygon[, 2])
    #create same number of boxes as that in parent polygon
    nbox_factors <- conf.design::factorize(nbox)
    if(length(nbox_factors) == 1){
      nbox <- nbox - 1
      nbox_factors <- conf.design::factorize(nbox)
    }
    xseq_sammon <- seq(xrange_sammon_poly[1],
                       xrange_sammon_poly[2],
                       length.out = nbox_factors[length(nbox_factors)] + 1)
    yseq_sammon <- seq(yrange_sammon_poly[1],
                       yrange_sammon_poly[2],
                       length.out = nbox / nbox_factors[length(nbox_factors)] + 1)

    samPoints_on_ylines <- samPoints_on_xlines <-  list()

    for(i in 1: length(xseq_sammon)){
      #points on line with constant 'x'
      samPoints_on_xlines[[i]] <- matrix(c(rep(xseq_sammon[i], length(yseq_sammon)),
                                           yseq_sammon),
                                         ncol = 2, byrow = FALSE)
    }

    #constructing polygons inside sammon polygon
    smaller_polygons_within_sammon <- list()
    #npoly <-0
    no_of_small_samPoly <- 1
    for(l in 1: (length(samPoints_on_xlines) - 1)){
      #if there are 2 or more points on a line
      if(length(samPoints_on_xlines[[l]][, 1]) > 1){
        for(m in 1: (length(samPoints_on_xlines[[l]][, 1]) - 1)){
          #if the consecutive y co-ordinates are present in the next line
          if(samPoints_on_xlines[[l]][m, 2] %in% samPoints_on_xlines[[(l + 1)]][, 2] &&
               samPoints_on_xlines[[l]][(m + 1), 2] %in% samPoints_on_xlines[[(l + 1)]][, 2]){

            smaller_polygons_within_sammon[[no_of_small_samPoly]] <- matrix(c(samPoints_on_xlines[[l]][m, ],
                                                                              samPoints_on_xlines[[(l + 1)]][1, 1],
                                                                              samPoints_on_xlines[[l]][m, 2],
                                                                              samPoints_on_xlines[[(l + 1)]][1, 1],
                                                                              samPoints_on_xlines[[l]][(m + 1), 2],
                                                                              samPoints_on_xlines[[l]][m, 1],
                                                                              samPoints_on_xlines[[l]][(m + 1), 2]),
                                                                            ncol = 2, byrow = TRUE)
            no_of_small_samPoly <- no_of_small_samPoly + 1
          }
        }
      }
    }
    return(smaller_polygons_within_sammon)
  }

  #construct the polygons in parent tile
  par_tile_polygon <- list()
  #variable storing appropriate parent tile index
  lindex <- c()
  #for constructing sammon space polygon
  xrange <- yrange <- sammon_space_polygon <- list()
  #to store the chopped rectangles in both parent tile and sammon space
  boxes_in_sammon_space <- boxes_in_parent_tile <- list()
  #to store the scale-factors and translate matrix information
  transformation_matrices_info <- int_list <- list()
  #for the indices of the smaller rectangles containing sammon points in the sammon space polygon
  rect_indices_containing_points <- inter_indices <- list()
  #to store the transformed points
  new_rawdeldati <- list()
  parentTiles <- unique(gidata[, "Segment.Parent"])

  for(ind1 in 1: length(parentTiles)){
    if(((parentTiles[ind1]) %% n_cells)){
      lindex[ind1] <- (parentTiles[ind1] %% n_cells)
    }else{
      lindex[ind1] <- n_cells
    }
    par_tile_polygon[[ind1]] <- matrix(c(poly_info[[lindex[ind1]]]$x,
                                         poly_info[[lindex[ind1]]]$y),
                                       ncol = 2, byrow = FALSE)

    #construct the sammon space polygon using xrange and yrange of sammon
    xrange[[ind1]] <- range(rawdeldati[[ind1]][, 1])
    yrange[[ind1]] <- range(rawdeldati[[ind1]][, 2])
    sammon_space_polygon[[ind1]] <- matrix(c(xrange[[ind1]][1], yrange[[ind1]][1],
                                             xrange[[ind1]][2], yrange[[ind1]][1],
                                             xrange[[ind1]][2], yrange[[ind1]][2],
                                             xrange[[ind1]][1], yrange[[ind1]][2]),
                                           ncol = 2, byrow = TRUE)

    #represent each small box in the bigger polygon as a polygon

    boxes_in_parent_tile[[ind1]] <- Parent_Tile_Boxes(par_tile_polygon[[ind1]])
    nbox <- length(boxes_in_parent_tile[[ind1]])
    if(nbox == 0){
      return("-1")
    }
    boxes_in_sammon_space[[ind1]] <- Sammon_Space_Boxes(sammon_space_polygon[[ind1]], nbox)

    #transformation_matrices_info will have scale matrix and translate matrix values for each
    #rectangle-rectangle transformation from sammon space to parent polygon

    #obtain the scale matrix
    scaleMat <- ScaleMat(boxes_in_sammon_space[[ind1]][[1]],
                         boxes_in_parent_tile[[ind1]][[1]])
    for(j in 1: length(boxes_in_sammon_space[[ind1]])){
      #get the translate matrix values by giving scale matrix also as input
      int_list[[j]] <- Transform_Coordinates(boxes_in_sammon_space[[ind1]][[j]],
                                             boxes_in_parent_tile[[ind1]][[j]],
                                             scaleMat)
    }
    transformation_matrices_info[[ind1]] <- int_list

    int_list <- list()
    #find the indices of the smaller boxes in sammon space in which the input sammon points exist

    for(j in 1: length(rawdeldati[[ind1]][, 1])){
      for(k in 1: length(boxes_in_sammon_space[[ind1]])){
        if(sp::point.in.polygon(rawdeldati[[ind1]][j, 1],
                            rawdeldati[[ind1]][j, 2],
                            boxes_in_sammon_space[[ind1]][[k]][, 1],
                            boxes_in_sammon_space[[ind1]][[k]][, 2]) != 0){
          inter_indices[[j]] <- k
        }
      }
    }
    rect_indices_containing_points[[ind1]] <- inter_indices

    #using the indices obtained above and the scale and transzlate matrices to
    #get the transformed coordinates for the input sammon datapoints
    int_matrix <- matrix(0, ncol = 2, nrow = length(rawdeldati[[ind1]][, 1]))
    row.names(int_matrix) <- row.names(rawdeldati[[ind1]])

    for(j in 1: length(rawdeldati[[ind1]][, 1])){
      int_matrix[j, ] <- New_Points(transformation_matrices_info[[ind1]][[rect_indices_containing_points[[ind1]][[j]]]])
    }
    new_rawdeldati[[ind1]] <- int_matrix
  }
  return(new_rawdeldati)
}

#' @name scoreHVT
#' @title Score which cell each point in the test dataset belongs to.
#' @description This function scores each data point in the test dataset based on a trained hierarchical Voronoi tessellations model. 
#' @param dataset Data frame. A data frame which to be scored. Can have categorical columns if `analysis.plots` are required.
#' @param hvt.results.model List. A list obtained from the trainHVT function 
#' @param child.level Numeric. A number indicating the depth for which the heat map is to be plotted. 
#' @param mad.threshold Numeric. A numeric value indicating the permissible Mean Absolute Deviation.
#' @param line.width Vector. A vector indicating the line widths of the tessellation boundaries for each layer. 
#' @param color.vec Vector. A vector indicating the colors of the tessellation boundaries at each layer. 
#' @param normalize Logical. A logical value indicating if the dataset should be normalized. When set to TRUE,
#'  the data (testing dataset) is standardized by ‘mean’ and ‘sd’ of the training dataset referred from the trainHVT(). 
#'  When set to FALSE, the data is used as such without any changes.
#' @param distance_metric Character. The distance metric can be L1_Norm(Manhattan) or L2_Norm(Eucledian). L1_Norm is selected by default.
#' The distance metric is used to calculate the distance between an n dimensional point and centroid.
#'  The distance metric can be different from the one used during training.
#' @param error_metric Character. The error metric can be mean or max. max is selected by default. 
#' max will return the max of m values and mean will take mean of m values where
#' each value is a distance between a point and centroid of the cell.
#' @param yVar Character. A character or a vector representing the name of the dependent variable(s)
#' @param analysis.plots Logical. A logical value indicating that the scored plot should be plotted or not. If TRUE, 
#' the identifier column(character column) name should be supplied in `names.column` argument. The output will
#' be a 2D heatmap plotly which gives info on the cell id and the observations of a cell.
#' @param names.column Character. A character or a vector representing the name of the identifier column/character column.
#' @param cell_id Logical. A logical indicating whether the cell IDs should be displayed. Default is TRUE.
#' @param cell_id_position Character. A character indicating the position of the cell IDs. Accepted entries are 'top', 
#' 'bottom', 'left', 'right', and 'center'. Default is 'center'.
#' @param cell_id_size Numeric. A numeric value indicating the size of the cell IDs. Default is 4
#' @param centroids Logical. A logical indicating whether the centroid points should be displayed. Default value is TRUE
#' @param centroid.size  Numeric. A numeric value or vector indicating the size of centroids for each level. Default is 0.8
#' @returns Dataframe containing scored data, plots and summary
#' @author Shubhra Prakash <shubhra.prakash@@mu-sigma.com>, Sangeet Moy Das <sangeet.das@@mu-sigma.com> ,
#' Vishwavani <vishwavani@@mu-sigma.com>
#' @seealso \code{\link{trainHVT}} \cr \code{\link{plotHVT}}
#' @keywords Scoring
#' @importFrom magrittr %>%
#' @include utilities.R constants.R
#' @examples
#' data("EuStockMarkets")
#' dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
#'                      DAX = EuStockMarkets[, "DAX"],
#'                      SMI = EuStockMarkets[, "SMI"],
#'                      CAC = EuStockMarkets[, "CAC"],
#'                      FTSE = EuStockMarkets[, "FTSE"])
#' rownames(EuStockMarkets) <- dataset$date
#' # Split in train and test
#' train <- EuStockMarkets[1:1302, ]
#' test <- EuStockMarkets[1303:1860, ]
#' #model training
#' hvt.results<- trainHVT(train,n_cells = 60, depth = 1, quant.err = 0.1,
#'                       distance_metric = "L1_Norm", error_metric = "max",
#'                       normalize = TRUE,quant_method = "kmeans")
#' scoring <- scoreHVT(test, hvt.results)
#' data_scored <- scoring$scoredPredictedData
#' @export scoreHVT


scoreHVT <- function(dataset,
                       hvt.results.model,
                       child.level = 1,
                       mad.threshold = 0.2,
                       line.width = 0.6,
                       color.vec = c("navyblue", "slateblue", "lavender"),
                       normalize = TRUE,
                       distance_metric = "L1_Norm",
                       error_metric = "max",
                       yVar = NULL,
                       analysis.plots = FALSE,
                       names.column = NULL,
                       cell_id = TRUE,
                       cell_id_position = "bottom",
                       centroids = TRUE,
                       cell_id_size = 4,
                       centroid.size = 0.8) {
  
  suppressWarnings({
  set.seed(300)
  sum_n = NULL
  requireNamespace("dplyr")
  requireNamespace("purrr")
  requireNamespace("data.table")
  
  # Ensure centroid.size is a vector of appropriate length
  if (length(centroid.size) == 1) {
    centroid.size <- rep(centroid.size, child.level)
  }
  
  if (any(is.na(dataset))) {
    stop("Input dataframe contains missing (NA) values. Please handle missing data before using this function (e.g., via na.omit(), na.fill(), or imputation).")
  } 
  

  if (!("Cell.ID" %in% colnames(hvt.results.model[[3]]$summary))) {
    hvt.results.model[[3]]$summary <- getCellId(hvt.results = hvt.results.model)
  }
 
  if (analysis.plots == TRUE && is.null(names.column)) {
   stop("names.column is not defined")
  }
  
  
  hvt.results.model[[3]]$summary <- cbind(hvt.results.model[[3]]$summary, centroidRadius = unlist(hvt.results.model[[3]]$max_QE))

  numeric_columns <- sapply(as.data.frame(dataset), is.numeric)
  data <- dataset[, numeric_columns, drop = FALSE]

  summary_list <- hvt.results.model[[3]]
  train_colnames <- names(summary_list[["nodes.clust"]][[1]][[1]])

    if (!all(train_colnames %in% colnames(data))) {
    stop("Not all training columns are part of test dataset")
  }
  
  data_structure_score  <- dim(data)
  
  if (!all(is.na(summary_list$scale_summary)) && normalize == TRUE) {
    scaled_test_data <- scale(
      data[, train_colnames],
      center = summary_list$scale_summary$mean_data[train_colnames],
      scale = summary_list$scale_summary$std_data[train_colnames]
    )
  } else {
     scaled_test_data <- data[, train_colnames]
 
  }

  colnames(scaled_test_data) <- train_colnames
 
  level <- child.level

  
  if (!is.null(yVar)) {
    yVardf <- data[, yVar]
    if (length(yVar) != 1) {
      colnames(yVardf) <- paste0("Scored.", yVar)
    }
  }
   


  find_path <- function(data_vec, centroid_data) {
  
    if (distance_metric == "L1_Norm") {
      centroidDist <- which.min(colSums(abs(centroid_data - data_vec), na.rm = TRUE))
      Quant.Error <- (colSums(abs(centroid_data - data_vec), na.rm = TRUE))[centroidDist]
    } else {
      centroidDist <- which.min(sqrt(colSums((centroid_data - data_vec)^2, na.rm = TRUE)))
      Quant.Error <- sqrt(colSums((centroid_data - data_vec)^2, na.rm = TRUE))[centroidDist]
    }
    return(data.frame("Index" = centroidDist, "Quant.Error" = Quant.Error / length(train_colnames)))
  }

 
  newdfMapping <- summary_list$summary

  innermostCells2 <- newdfMapping %>%
    dplyr::filter((n > 0 & Segment.Level == level) | (Segment.Level < level & (Quant.Error < mad.threshold | n <= 3)))

  transposedCells <- innermostCells2 %>%
    select(all_of(train_colnames)) %>%
    t()

  cent_dist_df2 <- apply(data.frame(scaled_test_data), 1, find_path, transposedCells) %>%
    bind_rows()

  groupCols2 <- c(paste0("Segment.", c("Level", "Parent", "Child")), yVar)
  
  if ("Cell.ID" %in% names(innermostCells2)) { 
       groupCols2 <- c(groupCols2, "Cell.ID", "centroidRadius")
    } else{
      groupCols2 <- groupCols2
    }

  predict_test_data2 <-
    cbind(data.frame(scaled_test_data, "n" = 1), cent_dist_df2) %>%
    dplyr::left_join(
      innermostCells2 %>%
        select(all_of(groupCols2)) %>%
        cbind(Index = as.integer(row.names(.))),
      by = "Index"
    ) %>%
    select(-Index) %>%
    select(names(newdfMapping))

  # Considering margin of error
  predict_test_data3 <- predict_test_data2 %>% mutate(diff = centroidRadius - Quant.Error)
  predict_test_data3 <- predict_test_data3 %>% mutate(anomalyFlag = ifelse(Quant.Error < (mad.threshold), 0, 1))

  # Renaming fitted yVar with prefix Fitted
  if (!is.null(yVar)) {
    if (length(yVar) == 3) {
      for (i in 1:length(yVar)) {
        indexScored <- which(colnames(predict_test_data2) == yVar[i])
        colnames(predict_test_data2)[indexScored] <- paste0("Fitted.", yVar[i])
      }
    } else {
      indexScored <- which(colnames(predict_test_data2) == yVar)
      colnames(predict_test_data2)[indexScored] <- paste0("Fitted.", yVar)
    }
  }


  # Adding dep Variables back to scored data
  if (!is.null(yVar)) {
    predict_test_data2 <- merge(predict_test_data2, yVardf, by = 0) %>% select(-"Row.names")
    if (length(yVar) == 1) {
      indexFitted <- which(colnames(predict_test_data2) == "y")
      colnames(predict_test_data2)[indexFitted] <- paste0("Scored.", yVar)
    }
  }

  groupCols2 <- c(paste0("Segment.", c("Level", "Parent", "Child")), "anomalyFlag")
  # Calculating mean for scored data for each centroid
  groupCols3 <- c(paste0("Segment.", c("Level", "Parent", "Child")))
  # filter anomalous values for test dataset

  if (error_metric == "mean") {
    predictQE2 <- predict_test_data3 %>%
      group_by_at(groupCols2) %>%
      dplyr::summarise(
        n = sum(n),
        Quant.Error = mean(Quant.Error)
      )
    predictQE3 <- predict_test_data3 %>%
      group_by_at(groupCols3) %>%
      dplyr::summarise(
        n = sum(n),
        Quant.Error = mean(Quant.Error)
      )
  } else {
    predictQE2 <- predict_test_data3 %>% # with anamoly flags
      group_by_at(groupCols2) %>%
      dplyr::summarise(
        n = sum(n),
        Quant.Error = max(Quant.Error)
      )
    predictQE3 <- predict_test_data3 %>% # wo anamoly flags
      group_by_at(groupCols3) %>%
      dplyr::summarise(
        n = sum(n),
        Quant.Error = max(Quant.Error)
      )
  }

  # Calculating mean for scored data for each centroid
  groupCols2 <- c(paste0("Segment.", c("Level", "Parent", "Child")))
  newdfMapping <- newdfMapping %>% mutate(sumOriginal = Quant.Error * n)
  df_temp <- inner_join(predictQE2,
    newdfMapping %>% select(c(groupCols2, Quant.Error, sumOriginal, n)),
    by = groupCols2
  )
  df_temp2 <- inner_join(predictQE3,
    newdfMapping %>% select(c(groupCols2, Quant.Error, sumOriginal, n)),
    by = groupCols2
  )


  if (error_metric == "mean") {
    df_temp <- df_temp %>% mutate(Scored.Quant.Error = (sumOriginal + (Quant.Error.x * n.x)) / (n.x + n.y)) 
    df_temp2 <- df_temp2 %>% mutate(Scored.Quant.Error = (sumOriginal + (Quant.Error.x * n.x)) / (n.x + n.y))
  } else {
    df_temp <- df_temp %>% mutate(Scored.Quant.Error = pmax(Quant.Error.x, Quant.Error.y))
    df_temp2 <- df_temp2 %>% mutate(Scored.Quant.Error = pmax(Quant.Error.x, Quant.Error.y)) 
  }

  QECompareDf2 <- df_temp %>%
    mutate(
      Quant.Error.Diff = abs(Scored.Quant.Error - Quant.Error.y),
      `Quant.Error.Diff (%)` = abs(Scored.Quant.Error - Quant.Error.y) / Quant.Error.y * 100
    ) %>%
    dplyr::rename(Fitted.Quant.Error = Quant.Error.y, n = n.x) %>%
    select(-c("Quant.Error.x", "sumOriginal", "n.y"))

  plotList <- hvt.results.model[[2]] %>%
    unlist(., recursive = FALSE) %>%
    unlist(., recursive = FALSE)

  

  cellID_coordinates <- extract_cell_coordinates(hvt.results.model)
  
  

  ##################
 
  
  boundaryCoords2 <- build_boundary_coords(plotList)
  
  boundaryCoords2 <- merge(boundaryCoords2, cellID_coordinates, by = c("x" ,"y")) %>%
    right_join(.,
               QECompareDf2 %>% dplyr::filter(anomalyFlag == 1),
               by = paste0("Segment.", c("Level", "Parent", "Child"))
    )
  
  
  anomalyPlot <- plotHVT(
    hvt.results.model,
    child.level = child.level,
    cell_id = cell_id,
    cell_id_position = cell_id_position,
    centroids = centroids,
    cell_id_size = cell_id_size,
    centroid.size = centroid.size
  ) + ggplot2::ggtitle(paste(
    "Hierarchical Voronoi Tessellation for Level",
    child.level
  )) +
    ggplot2::theme(
      plot.title = ggplot2::element_text(
        size = 18,
        hjust = 0.5,
        margin = ggplot2::margin(0, 0, 20, 0)
      ),
      # legend.title = element_blank(),
      panel.grid.major = ggplot2::element_blank(),
      panel.grid.minor = ggplot2::element_blank(),
      panel.background = ggplot2::element_blank(),
      axis.text.x = ggplot2::element_blank(),
      axis.ticks.x = ggplot2::element_blank(),
      axis.title.x = ggplot2::element_blank(),
      axis.text.y = ggplot2::element_blank(),
      axis.ticks.y = ggplot2::element_blank(),
      axis.title.y = ggplot2::element_blank(),
      legend.position = "bottom"
    ) +
    ggplot2::scale_x_continuous(expand = c(0, 0)) +
    ggplot2::scale_y_continuous(expand = c(0, 0))
  
  
  colour_scheme <- HVT_COLOR_SCHEME
  
  if (nrow(boundaryCoords2) != 0) {
    hoverText <- paste(
      " Cell ID:",
      boundaryCoords2$Cell.ID,
      "<br>",
      "Segment.Level:",
      boundaryCoords2$Segment.Level,
      "<br>",
      "Segment.Parent:",
      boundaryCoords2$Segment.Parent,
      "<br>",
      "Segment.Child:",
      boundaryCoords2$Segment.Child,
      "<br>",
      "Number of observations:",
      boundaryCoords2$n,
      "<br>"
    )
  } else {
    hoverText <- NULL
  }
  # browser()
  anomalyPlot <- anomalyPlot + ggplot2::geom_polygon(
    data = boundaryCoords2,
    ggplot2::aes(
      x = bp.x,
      y = bp.y,
      group = interaction(Segment.Level, Segment.Parent, Segment.Child),
      fill = n,
      text = hoverText
    ),
    color = "red",
    size = 1
  ) +
    ggplot2::geom_point(data = boundaryCoords2 %>% distinct(x, y), ggplot2::aes(x = x, y = y), size = 1.5) +
    ggplot2::scale_fill_gradientn(colours = colour_scheme) +
    ggplot2::guides(colour = "none")
  
  plotlyPredict <- plotly::ggplotly(anomalyPlot, tooltip = "text")
  
  hoverText <- lapply(plotlyPredict$x$data, function(x) {
    if (!is.null(x$text)) {
      return(x$text)
    }
  }) %>% unlist()

  checkCell <- substr(hoverText, 1, 5) %in% " Cell"
  trace_vec <- seq_along(checkCell)[!checkCell]

  plotlyPredict <- plotlyPredict %>%
    plotly::layout(
      hoverlabel = list(bgcolor = "rgba(255,255,0,0.2)"),
      legend = list(
        title = list(text = "Level"),
        itemdoubleclick = FALSE,
        itemclick = "toggleothers",
        traceorder = "reversed"
      )
    ) %>%
    plotly::style(plotlyPredict, hoverinfo = "none", traces = trace_vec) %>%
    plotly::config(displayModeBar = FALSE)
  
  
  predict_test_data3 <- predict_test_data3 %>% mutate_if(is.numeric, round, digits = 4) 
  predict_test_data3 <- predict_test_data3 %>% dplyr::select(c("Segment.Level",  "Segment.Parent" ,"Segment.Child" , "n" ,            
                                                               "Cell.ID", "Quant.Error",  "centroidRadius" ,"diff" , "anomalyFlag", everything()))
  predict_test_dataRaw <- predict_test_data3
  predict_test_dataRaw[, train_colnames] <- data[, train_colnames]
################################################# Changes #################
  predicted_result <- hvt.results.model[[3]]$summary 
  current_predicted <- colnames(predicted_result)
  new_names <- paste0("pred_", current_predicted)
  colnames(predicted_result) <- new_names
  Cell.ID <- data.frame(predicted_result$pred_Cell.ID)
  predicted_result <- predicted_result %>% select(-c("pred_Segment.Level", "pred_Segment.Parent", "pred_Segment.Child", "pred_n", "pred_Cell.ID", "pred_Quant.Error", "pred_centroidRadius"))
  predicted_result <- cbind(Cell.ID, predicted_result)
  predicted_result <- data.table::setnames(predicted_result, "predicted_result.pred_Cell.ID", "Cell.ID")



  actuals <- predict_test_data3
  current_actual <- colnames(actuals)
  new_names <- paste0("act_", current_actual)
  colnames(actuals) <- new_names

  actuals$Row.No <- row.names(data)
  data_with_row <- data.frame(actuals$Row.No)
  data_with_cell <- data.frame(actuals$act_Cell.ID)
  actuals <- actuals %>% select(-c("act_Segment.Level", "act_Segment.Parent", "act_Segment.Child", "act_n", "act_Cell.ID", "act_Quant.Error", "act_centroidRadius", "act_diff", "act_anomalyFlag", "Row.No"))
  actuals_data <- cbind(data_with_row, actuals, data_with_cell)
  actuals_data <- data.table::setnames(actuals_data, "actuals.act_Cell.ID", "Cell.ID")
  actuals_data <- data.table::setnames(actuals_data, "actuals.Row.No", "Row.No")



  merged_df <- merge(actuals_data, predicted_result, by = "Cell.ID")
  merged_result <- merged_df %>% arrange(as.numeric(merged_df$Row.No))




  diff <- subtract_predicted_actual(merged_result)
  merged_result$diff <- diff$matrix.ncol...1..nrow...nrow.data..
  merged_result <- rename(merged_result, c("diff" = "diff"))


  desired_order <- c("Row.No", grep("^act_", colnames(merged_result), value = TRUE), "Cell.ID", grep("^pred_", colnames(merged_result), value = TRUE), "diff")
  df_reordered <- merged_result[, desired_order]
#################################################

  .generate_scoring_plots <- function(hvt.results.model, predict_test_data3,
                                      QECompareDf2, cellID_coordinates,
                                      plotList, names.column, child.level,
                                      mad.threshold, cell_id, cell_id_position,
                                      cell_id_size, centroid.size) {
    
    scored_data <- data.frame(Cell.ID = predict_test_data3$Cell.ID, names.column)
    scored_data <- scored_data[order(scored_data$Cell.ID), c("Cell.ID", "names.column")]
    reformed_data <- stats::aggregate(names.column ~ Cell.ID, scored_data, FUN = function(x) paste(x, collapse = ", "))
    
    boundaryCoords2_1 <- build_boundary_coords(plotList)
    
    boundaryCoords2_1 <- merge(boundaryCoords2_1, cellID_coordinates, by = c("x","y")) %>%
      right_join(., QECompareDf2, by = paste0("Segment.", c("Level", "Parent", "Child")))
    boundaryCoords2_1 <- merge(boundaryCoords2_1, reformed_data, by = "Cell.ID")
    
    scored_data <- data.frame(Cell.ID = predict_test_data3$Cell.ID, names.column)
    scored_data <- scored_data[order(scored_data$Cell.ID), c("Cell.ID", "names.column")]
    reformed_data <- stats::aggregate(names.column ~ Cell.ID, scored_data, FUN = function(x) paste(x, collapse = ", "))
    
    boundaryCoords2_1 <-
      lapply(plotList, function(x) {
        data.frame(
          "Segment.Level" = x[["Segment.Level"]],
          "Segment.Parent" = x[["Segment.Parent"]],
          "Segment.Child" = x[["Segment.Child"]],
          "x" = x$pt["x"],
          "y" = x$pt["y"],
          "bp.x" = I(x$x),
          "bp.y" = I(x$y)
        )
      }) %>%
      bind_rows(.) 
    
    boundaryCoords2_1 <- merge(boundaryCoords2_1, cellID_coordinates, by = c("x" ,"y")) %>%
      right_join(.,
                 QECompareDf2,
                 by = paste0("Segment.", c("Level", "Parent", "Child"))
      )
    
    boundaryCoords2_1 <- merge(boundaryCoords2_1, reformed_data, by = "Cell.ID")
    
    
    boundaryCoords2_1<- boundaryCoords2_1 %>%
      group_by(Cell.ID) %>%
      mutate(
        sum_n = ifelse(any(anomalyFlag == 0) & any(anomalyFlag == 1), sum(n[anomalyFlag == 0][1], n[anomalyFlag == 1][1]), NA),
        n = ifelse(!is.na(sum_n), sum_n, n)
      ) %>%
      select(-sum_n)
    
    
    # Call plotHVT with cell_id = FALSE since we'll add annotations ourselves for better control
    scoredPlot <- plotHVT(
      hvt.results.model,
      child.level = child.level,
      cell_id = FALSE,  # We'll add cell IDs via plotly annotations for better positioning
      cell_id_position = cell_id_position,
      centroids = FALSE,  # We'll add centroids ourselves using polygon centroids when needed
      cell_id_size = cell_id_size,
      centroid.size = centroid.size
    ) +
      ggplot2::theme(
        plot.title = ggplot2::element_text(
          size = 18,
          hjust = 0.5,
          margin = ggplot2::margin(0, 0, 20, 0)
        ),
        # legend.title = element_blank(),
        panel.grid.major = ggplot2::element_blank(),
        panel.grid.minor = ggplot2::element_blank(),
        panel.background = ggplot2::element_blank(),
        axis.text.x = ggplot2::element_blank(),
        axis.ticks.x = ggplot2::element_blank(),
        axis.title.x = ggplot2::element_blank(),
        axis.text.y = ggplot2::element_blank(),
        axis.ticks.y = ggplot2::element_blank(),
        axis.title.y = ggplot2::element_blank(),
        legend.position = "bottom"
      ) +
      ggplot2::scale_x_continuous(expand = c(0, 0)) +
      ggplot2:: scale_y_continuous(expand = c(0, 0))
    
    
    colour_scheme <-HVT_COLOR_SCHEME
    
    wrap_and_limit_text <- function(text, line_length = 100, max_chars = 500) {
      # First, limit the total text to max_chars
      if (nchar(text) > max_chars) {
        text <- substr(text, 1, max_chars - 3)
        text <- paste0(text, "...")
      }
      
      # Then wrap the text
      wrapped <- sapply(seq(1, nchar(text), line_length), function(i) {
        substr(text, i, min(i + line_length - 1, nchar(text)))
      })
      
      # Join the wrapped lines with <br> for HTML line breaks
      paste(wrapped, collapse = "<br>")
    }
    
    if (nrow(boundaryCoords2_1) != 0) {
      boundaryCoords2_1$hoverText <- apply(boundaryCoords2_1, 1, function(row) {
        full_names <- row["names.column"]
        formatted_names <- wrap_and_limit_text(full_names, line_length = 100, max_chars = 500)
        
        paste(
          "Cell ID:", row["Cell.ID"],
          "<br>",
          "Number of observations:", row["n"],
          "<br>",
          "Name of observations:<br>", formatted_names
        )
      })
    } else {
      boundaryCoords2_1$hoverText <- NULL
    }
    
    # Create the plot (polygons + centroids; labels will be added as Plotly annotations)
    scoredPlot <- scoredPlot +
      ggplot2::geom_polygon(
        data = boundaryCoords2_1,
        ggplot2::aes(
          x = bp.x,
          y = bp.y,
          group = interaction(Segment.Level, Segment.Parent, Segment.Child),
          fill = n,
          text = hoverText
        ),
        color = "black",
        size = 0.5
      ) +
      ggplot2::scale_fill_gradientn(colours = colour_scheme) +
      ggplot2::guides(colour = "none")
    
   
    hvt_list <- hvt.results.model
    
    # Build positionsDataframe from base tessellation (same as plotHVT does)
    depthPos <- c()
    clusterPos <- c()
    childPos <- c()
    x_pos <- c()
    y_pos <- c()
    
    for (depth in 1:child.level) {
      for (clusterNo in 1:length(hvt_list[[2]][[depth]])) {
        for (childNo in 1:length(hvt_list[[2]][[depth]][[clusterNo]])) {
          current_cluster <- hvt_list[[2]][[depth]][[clusterNo]][[childNo]]
          x <- as.numeric(current_cluster[["x"]])
          y <- as.numeric(current_cluster[["y"]])
          depthPos <- c(depthPos, rep(depth, length(x)))
          clusterPos <- c(clusterPos, rep(clusterNo, length(x)))
          childPos <- c(childPos, rep(childNo, length(x)))
          x_pos <- c(x_pos, x)
          y_pos <- c(y_pos, y)
        }
      }
    }
    
    positionsDataframe <- data.frame(
      depth = depthPos,
      cluster = clusterPos,
      child = childPos,
      x = x_pos,
      y = y_pos
    )
    
    # Calculate polygon centroids grouped by depth, cluster, child (same as plotHVT)
    polygon_centroids_list <- list()
    for (depth_val in 1:child.level) {
      depth_data <- positionsDataframe[positionsDataframe$depth == depth_val, ]
      if (nrow(depth_data) > 0) {
        polygon_centroids <- depth_data %>%
          dplyr::group_by(depth, cluster, child) %>%
          dplyr::summarise(
            x = calculate_polygon_centroid(x, y)$x,
            y = calculate_polygon_centroid(x, y)$y,
            .groups = 'drop'
          )
        polygon_centroids_list[[depth_val]] <- polygon_centroids
      }
    }
    polygon_centroids_df_base <- do.call(rbind, polygon_centroids_list)
    
    # Match polygon centroids with Cell.IDs for level 1 (same as plotHVT does)
    if (child.level >= 1 && nrow(polygon_centroids_df_base) > 0) {
      level1_poly_centroids <- polygon_centroids_df_base[polygon_centroids_df_base$depth == 1, ]
      cellID_mapping <- hvt_list[[3]][["summary"]] %>%
        dplyr::filter(Segment.Level == 1) %>%
        dplyr::select(Segment.Parent, Segment.Child, Cell.ID) %>%
        dplyr::mutate(
          Segment.Parent = as.character(Segment.Parent),
          Segment.Child = as.numeric(Segment.Child)
        )
      
      polygon_centroids_df <- level1_poly_centroids %>%
        dplyr::mutate(
          cluster = as.character(cluster),
          child = as.numeric(child)
        ) %>%
        dplyr::left_join(cellID_mapping, 
                         by = c("cluster" = "Segment.Parent", "child" = "Segment.Child")) %>%
        dplyr::filter(!is.na(Cell.ID)) %>%
        dplyr::select(Cell.ID, x, y)
    } else {
      polygon_centroids_df <- data.frame(Cell.ID = numeric(0), x = numeric(0), y = numeric(0))
    }
    
    # Add centroid points conditionally based on centroids parameter
    # Centroids always use the original point centroids (x, y from boundaryCoords2_1) - no change from original behavior
    # Store color for later use in plotly styling
    centroid_fill_color <- NULL
    centroid_stroke_color <- NULL
    centroid.color = "black"
    
    if (centroids == TRUE) {
      # Ensure centroid.color is a vector of appropriate length
      if (length(centroid.color) == 1) {
        centroid_color_vec <- rep(centroid.color, child.level)
      } else {
        centroid_color_vec <- centroid.color
      }
      
      # Get the color value (not vector) - use same color for fill and stroke
      centroid_fill_color <- centroid_color_vec[1]
      centroid_stroke_color <- centroid_color_vec[1]
      
      scoredPlot <- scoredPlot +
        ggplot2::geom_point(
          data = boundaryCoords2_1 %>% dplyr::distinct(Cell.ID, .keep_all = TRUE),
          ggplot2::aes(x = x, y = y, text = hoverText),
          size = centroid.size[1],
          pch = 21,  # Filled circle (matches plotHVT)
          fill = centroid_fill_color,
          color = centroid_stroke_color
        )
    }
    
    # Prepare annotation data for cell IDs
    if (cell_id == TRUE && cell_id_position == "center") {
      # Use polygon centroids for center positioning
      annotation_data <- polygon_centroids_df
    } else if (cell_id == TRUE) {
      # Use point centroids for other positions
      annotation_data <- boundaryCoords2_1 %>% 
        dplyr::distinct(Cell.ID, .keep_all = TRUE) %>%
        dplyr::select(Cell.ID, x, y)
    } else {
      annotation_data <- data.frame(Cell.ID = numeric(0), x = numeric(0), y = numeric(0))
    }
    
    plotlyscored <- plotly::ggplotly(scoredPlot, tooltip = "text")
    
    # Ensure centroids are filled circles in plotly
    if (centroids == TRUE && !is.null(centroid_fill_color)) {
      # Find the trace index for the centroid points (usually one of the last traces)
      n_traces <- length(plotlyscored$x$data)
      # Look for traces with markers (point traces) - centroids should be one of them
      for (i in seq_len(n_traces)) {
        trace <- plotlyscored$x$data[[i]]
        # Check if this is a marker trace (centroid points)
        if (!is.null(trace$mode) && grepl("markers", trace$mode)) {
          # Directly modify the marker properties to ensure filled circle
          if (is.null(plotlyscored$x$data[[i]]$marker)) {
            plotlyscored$x$data[[i]]$marker <- list()
          }
          plotlyscored$x$data[[i]]$marker$fillcolor <- centroid_fill_color
          if (is.null(plotlyscored$x$data[[i]]$marker$line)) {
            plotlyscored$x$data[[i]]$marker$line <- list()
          }
          plotlyscored$x$data[[i]]$marker$line$color <- centroid_stroke_color
          plotlyscored$x$data[[i]]$marker$line$width <- 0.5
        }
      }
    }
    
    # Add Cell.ID labels as Plotly annotations based on cell_id parameter
    if (cell_id == TRUE && nrow(annotation_data) > 0) {
      # Calculate position offsets based on cell_id_position
      alignment <- get_cell_label_alignment(cell_id_position, "plotly")
      xshift  <- alignment$xshift
      yshift  <- alignment$yshift
      xanchor <- alignment$xanchor
      yanchor <- alignment$yanchor
      

      plotly_font_size <- max(8, cell_id_size * 2.845)
      
      plotlyscored <- plotlyscored %>%
        plotly::add_annotations(
          x = annotation_data$x,
          y = annotation_data$y,
          text = as.character(annotation_data$Cell.ID),
          showarrow = FALSE,
          font = list(size = plotly_font_size, color = "black"),
          xshift = xshift,
          yshift = yshift,
          xanchor = xanchor,
          yanchor = yanchor,
          xref = "x",
          yref = "y"
        )
    }
    
    # Add layout modifications after annotations
    plotlyscored <- plotlyscored %>%
      plotly::layout(
        title = list(
          text = "Scored Heatmap with Cell-Level Observations",
          x = 0, 
          y = 0.99,
          xanchor = "left", font = list(size = 20)),
        hovermode = 'closest',
        hoverdistance = 100,
        hoverlabel = list(
          bgcolor = "rgba(255,255,0,0.2)",
          font = list(size = 10),
          align = "left",
          namelength = -1),
        legend = list(
          title = list(text = "Level"),
          itemdoubleclick = FALSE,
          itemclick = "toggleothers",
          traceorder = "reversed"
        )
      ) %>%
      plotly::config(displayModeBar = TRUE)
    
    names_data  <- boundaryCoords2_1 %>% dplyr::select("Cell.ID","names.column") %>% distinct(Cell.ID, .keep_all = TRUE)
    a           <- cellID_coordinates %>% arrange(Cell.ID)
    centroid_data <- merge(a, names_data, by = "Cell.ID") %>% as.data.frame()
    states_data <- boundaryCoords2_1 %>% dplyr::select("Cell.ID","names.column") %>% distinct(Cell.ID, .keep_all = TRUE)
    
    list(
      scoredPlotly     = plotlyscored,
      scoredPlot_test  = scoredPlot,
      centroidData     = centroid_data,
      states_data      = states_data
    )
  }
  
    
 #################################################
  #MODEL INFO Rewriting
  input_dataset <- hvt.results.model[["model_info"]][["input_parameters"]][["input_dataset"]]
  n_cells <-hvt.results.model[["model_info"]][["input_parameters"]][["n_cells"]]
  compression_percentage <- compression_percentage <- paste0(hvt.results.model[[3]][["compression_summary"]][["percentOfCellsBelowQuantizationErrorThreshold"]] * 100, "%")
  quantization_error <-  hvt.results.model[["model_info"]][["input_parameters"]][["quant.err"]]
  
  trained_model <- list(
    input_dataset = input_dataset, 
    no_of_cells = n_cells, 
    compression_percentage = compression_percentage,
    quantization_error = quantization_error)
  
  score_dataset <- paste0(data_structure_score[1] ," Rows & ", data_structure_score[2], " Columns")
  qe_range_min <-min(predict_test_data3$Quant.Error)
  qe_range_max <- max(predict_test_data3$Quant.Error)
  qe_range <- paste0( qe_range_min, " to " ,qe_range_max)
  no_of_anomaly_cells <- sum(QECompareDf2$anomalyFlag == 1)
  no_of_anomaly_data <- sum(QECompareDf2$n[QECompareDf2$anomalyFlag == 1])
  
  scored_model <- list(
    input_dataset = score_dataset,
    scored_qe_range = qe_range,
    mad.threshold = mad.threshold,
    no_of_anomaly_datapoints = no_of_anomaly_data,
    no_of_anomaly_cells = no_of_anomaly_cells
  )
  
  
  ####################################
  base_list <- list(
    scoredPredictedData  = predict_test_data3,
    actual_predictedTable = df_reordered,
    QECompareDf          = QECompareDf2,
    anomalyPlot          = plotlyPredict,
    cellID_coordinates   = cellID_coordinates,
    predictInput         = c("depth" = child.level, "quant.err" = mad.threshold),
    model_mad_plots      = list(),
    model_info           = list(type = "hvt_prediction",
                                trained_model_summary = trained_model,
                                scored_model_summary  = scored_model)
  )
  
  if (analysis.plots) {
    plot_extras  <- .generate_scoring_plots(hvt.results.model, predict_test_data3,
                                            QECompareDf2, cellID_coordinates,
                                            plotList, names.column, child.level,
                                            mad.threshold, cell_id, cell_id_position,
                                            cell_id_size, centroid.size)
    prediction_list <- c(base_list, plot_extras)
  } else {
    prediction_list <- base_list
  }
  
  model_mad_plots <- NA
 
  if (!all(is.na(hvt.results.model[[4]]))) {
    mtrain <- hvt.results.model[[4]]$mad_plot_train + ggplot2::ggtitle("Mean Absolute Deviation Plot: Calibration on Train Data")
  }
  if (!all(is.na(hvt.results.model[[5]]))) {
    mtest <- hvt.results.model[[5]][["mad_plot"]] + ggplot2::ggtitle("Mean Absolute Deviation Plot:Validation")
  }

  if (!all(is.na(hvt.results.model[[4]])) & !all(is.na(hvt.results.model[[5]]))) {
    model_mad_plots <- list(mtrain = mtrain, mtest = mtest)
  }


  prediction_list$model_mad_plots <- model_mad_plots
  
  class(prediction_list) <- "hvt.object"
  return(prediction_list)
})
  
}

#' @keywords internal

getCellId <-  function(hvt.results, seed = 123) {
  generic_col=c("Segment.Level","Segment.Parent","Segment.Child","n","Quant.Error")
  temp_summary=hvt.results[[3]][["summary"]] %>%
    dplyr::select(!all_of(generic_col)) %>% 
    dplyr::mutate(id=row_number())
  cent_val= temp_summary %>% subset(.,stats::complete.cases(.)) 
  set.seed(seed)
  sammon_1d_cord <- MASS::sammon(
    d = stats::dist(cent_val %>% dplyr::select(!all_of("id")),method = "manhattan"),
    niter = 10 ^ 5,
    trace = FALSE,
    k=1
  )$points
  temp_df=data.frame(sammon_1d_cord,id=cent_val$id)%>%dplyr::arrange(sammon_1d_cord) %>% dplyr::mutate(Cell.ID=row_number()) %>% dplyr::select(!all_of("sammon_1d_cord"))
  temp_summary = dplyr::left_join(temp_summary,temp_df,by="id") %>% dplyr::select(!all_of("id"))
  hvt.results[[3]][["summary"]]$Cell.ID=temp_summary$Cell.ID

    return(hvt.results[[3]][["summary"]])
  }

#' @importFrom magrittr %>%
#' @keywords internal


getCentroids <-
  function (x,
            kout,
            n_cells,
            function_to_calculate_distance_metric,
            function_to_calculate_error_metric,
            distance_metric = "L1_Norm",
            error_metric,
            quant_method=c("kmeans","kmedoids")) {
    set.seed(279)
    
    if (quant_method == "kmeans") {
      centl <- nout <- outl <- list()
      x <- data.frame(x) %>% stats::na.omit()
      col_dim <- ncol(x)
  ###############################################################    
      
      calculate_error <-
        x %>%
        group_by(kout$cluster) %>%
        tidyr::nest()
    
      centroid_data <- kout$centers %>% as.data.frame()
      calculate_error <- calculate_error %>% arrange(calculate_error$`kout$cluster`)
      
      
      cluster_distances <- purrr::map2(
        1:nrow(centroid_data),
        calculate_error$data,
        function(i, cluster_data) {
          centroid_row <- centroid_data[i, ]
          apply(cluster_data, 1, function(row) function_to_calculate_distance_metric(centroid_row, row))
        }
      )
     
      calculate_error_for_each_cluster <- unlist(lapply(cluster_distances, function(x) {
        if (error_metric == "mean") {
          return(mean(x)/col_dim)
        } else if (error_metric == "max") {
          return(max(x)/col_dim)
        } else {
          return(error_metric(x))
        }
      }))
      

      maxQE_each_cluster <- unlist(lapply(cluster_distances, function(x) {   return(max(x))}))
      meanQE_each_cluster <-  unlist(lapply(cluster_distances, function(x) {   return(mean(x))}))
      centl <- calculate_error_for_each_cluster
      outl <-
        c(1:n_cells) %>% purrr::map( ~ x[kout$cluster == .x, ])
      nout <- as.list(kout$size)
      #return centroids, datapoints and size of each cluster
      return(
        list(
          centers = centl,
          maxQE = maxQE_each_cluster,
          values = outl,
          nsize = nout,
          meanQE = meanQE_each_cluster,
          dataframe_clusters = centroid_data,
          cluster_distance = cluster_distances
        )
      )
     
      
       
      ############################## Medoid implementation ###############################################
      
    } else if (quant_method == "kmedoids") {
      if (distance_metric == "L1_Norm") {
        distance_metric = "manhattan"
      } else if (distance_metric == "L2_Norm") {
        distance_metric = "euclidean"
      }
      


      kmedoids_model <-
        cluster::pam(
          x = cluster::daisy(x, metric = distance_metric),
          k = n_cells,
          diss = TRUE,
          keep.data = FALSE
        )
      
      centl <- nout <- outl <- list()
      x <- data.frame(x)
      col_dim <- ncol(x)
      centroid_data <- x[kmedoids_model[["medoids"]], ] %>% as.data.frame()
      
      calculate_error <-
        x %>%
        dplyr::group_by(kmedoids_model[["clustering"]]) %>%
        tidyr::nest()
      calculate_error <- calculate_error %>% arrange(calculate_error$`kmedoids_model[["clustering"]]`)
      
      cluster_distances <- purrr::map2(
        1:nrow(centroid_data),
        calculate_error$data,
        function(i, cluster_data) {
          centroid_row <- centroid_data[i, ]
          apply(cluster_data, 1, function(row) function_to_calculate_distance_metric(centroid_row, row))
        }
      )
      
      calculate_error_for_each_cluster <- unlist(lapply(cluster_distances, function(x) {
        if (error_metric == "mean") {
          return(mean(x)/col_dim)
        } else if (error_metric == "max") {
          return(max(x)/col_dim)
        } else {
          return(error_metric(x))
        }
      }))
      maxQE_each_cluster  <- unlist(lapply(cluster_distances, function(x) { return(max(x)) }))
      meanQE_each_cluster <- unlist(lapply(cluster_distances, function(x) { return(mean(x)) }))
      centl <- calculate_error_for_each_cluster
      outl <-
        c(1:n_cells) %>% purrr::map(~ x[kmedoids_model[["clustering"]] == .x,])
      nout <-
        as.list(as.numeric(table(kmedoids_model[["clustering"]])))
      
      #sum_val = as.data.frame(x[kmedoids_model[["medoids"]],])
      #return centroids, datapoints and size of each cluster
      return(
        list(
          centers = centl,
          maxQE = maxQE_each_cluster,
          values = outl,
          nsize = nout,
          meanQE = meanQE_each_cluster,
          dataframe_clusters = centroid_data,
          cluster_distance = cluster_distances
        )
      )
    }
  }

#' @name clustHVT
#' @title Performing Hierarchical Clustering Analysis
#' @description This is the main function to perform hierarchical clustering
#' analysis which determines optimal number of clusters, perform AGNES clustering
#' and plot the 2D cluster hvt plot.
#' @param data Data frame. A data frame intended for performing hierarchical clustering analysis.
#' @param trainHVT_results List.  A list object which is obtained as a result of trainHVT function.
#' @param scoreHVT_results List. A list object which is obtained as a result of scoreHVT function.
#' @param clusters_k Character.  A parameter that specifies the number of clusters for the provided data. 
#' The options include “champion,” “challenger,” or any integer between 1 and 20. 
#' Selecting “champion” will use the highest number of clusters recommended by the ‘NbClust’ function,
#' while “challenger” will use the second-highest recommendation. If a numerical value from 1 to 20
#' is provided, that exact number will be used as the number of clusters.
#' @param indices Character. The indices used for determining the optimal number of clusters in NbClust function.
#'  By default it uses 20 different indices.
#' @param clustering_method Character. The method used for clustering in both NbClust and hclust function. Defaults to ‘ward.D2’.
#' @param type Character. The type of output required. Default is 'default'. Other option is 'plot' which
#'  will return only the clustered heatmap.
#' @param domains.column Character. A vector of cluster names for the clustered heatmap.
#' Used only when type is 'plot'.
#' @param highlight_labels Vector. Numeric vector specifying problematic states and their neighboring states to be highlighted in the dendrogram. 
#' This argument is intended for internal use only. The default value is NULL.
#' @param only_dendro Logical. A logical string specifies whether to generate only the dendrogram or to include the results as well.
#' This argument is intended for internal use only. The default value is FALSE.
#' @return A list object that contains the hierarchical clustering results.
#' \item{[[1]] }{Summary of k suggested by all indices with plots} 
#' \item{[[2]] }{A dendrogram plot with the selected number of clusters} 
#' \item{[[3]] }{A 2D Cluster HVT Plotly visualization that colors cells according to clusters derived from AGNES clustering results. 
#' It is interactive, allowing users to view cell contents by hovering over them}
#' @author Vishwavani <vishwavani@@mu-sigma.com>
#' @keywords Clustering_Analysis
#' @include clusterPlot.R
#' @importFrom utils data head tail
#' @importFrom stats as.dendrogram median 
#' @examples 
#'data("EuStockMarkets")
#'dataset <- data.frame(t = as.numeric(time(EuStockMarkets)),
#'                      DAX = EuStockMarkets[, "DAX"],
#'                      SMI = EuStockMarkets[, "SMI"],
#'                      CAC = EuStockMarkets[, "CAC"],
#'                      FTSE = EuStockMarkets[, "FTSE"])
#'rownames(EuStockMarkets) <- dataset$t
#'hvt.results<- trainHVT(dataset[-1],n_cells = 30, depth = 1, quant.err = 0.1,
#'                       distance_metric = "L1_Norm", error_metric = "max",
#'                       normalize = TRUE,quant_method = "kmeans")
#'scoring <- scoreHVT(dataset, hvt.results, analysis.plots = TRUE, names.column = dataset[,1])
#'centroid_data <- scoring$centroidData
#'hclust_data_1 <- centroid_data[,2:3]
#'clust.results <- clustHVT(data = hclust_data_1, 
#'                          trainHVT_results = hvt.results,
#'                          scoreHVT_results = scoring, 
#'                          clusters_k = 'champion', indices = 'hartigan')
#' @export clustHVT




clustHVT <- function(data, trainHVT_results, scoreHVT_results, clustering_method = 'ward.D2',
                     indices = c("kl", "ch", "hartigan", "cindex", "db", "silhouette", 
                                 "ratkowsky", "ball", "hubert", "dindex", "ptbiserial", 
                                 "gap", "frey", "mcclain", "gamma", "gplus", "tau", 
                                 "dunn", "sdindex", "sdbw"), 
                     clusters_k = "champion", type = "default", domains.column = NULL, highlight_labels = NULL, only_dendro = FALSE) {
  requireNamespace('NbClust')
  
  if (type == "plot"){
    
    if (is.null(data) || is.null(trainHVT_results) || is.null(domains.column)) {
      stop("For type 'plot',  the arguments `data`, `trainHVT_results` and `domains.column` are required.")
    }
    
    plot_a <- clusterPlot(dataset= data, hvt.results  = trainHVT_results, domains.column = domains.column )
    
    return(plot_a)
    
  } 
  else if (type == "default"){
    
  
  hclust_data <- data
  
  results_df <- c()
  
  
  # Function to run NbClust and extract results
  print_nbclust_results <- function(hclust_data, indices) {
    get_nbclust_result <- function(index) {
      res <- NbClust::NbClust(hclust_data, method = clustering_method, index = index)$Best.nc
      if (!is.null(res)) {
        return(data.frame(Index = index, 
                          Number_clusters = res["Number_clusters"], 
                          Value_Index = res["Value_Index"]))
      } else {
        return(data.frame(Index = index, 
                          Number_clusters = NA, 
                          Value_Index = NA))
      }
    }
    
    results_list <- lapply(indices, get_nbclust_result)
    results_df <<- do.call(rbind, results_list)
    
    cluster_summary <- table(results_df$Number_clusters)
    cat("** Among all indices:\n")
    lapply(seq_along(cluster_summary), function(i) {
      cat(sprintf("** * %d proposed %d as the best number of clusters\n", 
                  cluster_summary[i], as.numeric(names(cluster_summary)[i])))})
    
    
    best_clusters <- as.numeric(names(cluster_summary)[which.max(cluster_summary)])
    cat("*******************************************************************\n")
    cat("***** Conclusion *****\n")
    cat(sprintf("** * According to the majority rule, the best number of clusters is %d\n", best_clusters))
    cat("*******************************************************************\n")
    cat("                         Index     Number_clusters    Value_Index\n")
    cat("-------------------------------------------------------------------\n")
    apply(results_df, 1, function(row) {
      cat(sprintf("%24s  %15.4f  %15.4f\n", 
                  row["Index"], 
                  as.numeric(row["Number_clusters"]), 
                  as.numeric(row["Value_Index"])))
    })
    cat("*******************************************************************\n")
  }
  
  
  # Decide whether to run NbClust
  use_nbclust <- !is.null(indices) && length(indices) > 0 && clusters_k %in% c("champion","challenger")
  if (use_nbclust) {
    # If caller passed indices, use them; otherwise default set
    if (identical(indices, TRUE)) {
      indices <- c("kl", "ch", "hartigan", "cindex", "db", "silhouette", "ratkowsky", "ball","hubert","dindex",
                   "ptbiserial", "gap", "frey", "mcclain", "gamma", "gplus", "tau", "dunn", "sdindex", "sdbw")
    }
    print_nbclust_results(hclust_data, indices)
    cluster_summary <- table(results_df$Number_clusters)
    if (clusters_k == "champion") {
      no_of_clusters <- as.numeric(names(cluster_summary)[which.max(cluster_summary)])
    } else { # challenger
      sorted_summary <- sort(cluster_summary, decreasing = TRUE)
      no_of_clusters <- as.numeric(names(sorted_summary)[2])
    }
  } else {
    # Bypass NbClust and honor explicit numeric clusters_k
    if (is.numeric(clusters_k) && clusters_k >= 1 && clusters_k <= 20) {
      no_of_clusters <- as.integer(clusters_k)
    } else if (clusters_k %in% c("champion","challenger")) {
      stop("indices must be supplied to use 'champion' or 'challenger'")
    } else {
      stop("Invalid input for clusters_k. Use numeric 1..20, or provide indices with 'champion'/'challenger'.")
    }
  }
  # Ensure k is a single integer and within allowable range for this data
  no_of_clusters <- as.integer(no_of_clusters[1])
  n_items <- nrow(hclust_data)
  if (is.na(no_of_clusters) || n_items < 2) {
    stop("Clustering error: insufficient data for clustering")
  }
  # cutree requires 1 < k <= n_items; in practice k must be in [1, n_items]
  if (no_of_clusters < 1 || no_of_clusters > n_items) {
    stop(sprintf("Clustering error: k=%d out of range [1..%d]", no_of_clusters, n_items))
  }
  
  #########cell_id#####
  hvt_list <- trainHVT_results
  hvt_res1 <- hvt_list[[2]][[1]]$`1`
  hvt_res2 <- hvt_list[[3]]$summary$Cell.ID
  a <- seq_along(hvt_res1)
  b <- a[hvt_res2]
  b <- as.vector(b)
  hvt_res2 <- stats::na.omit(b)
  hclust_data_1 <- hclust_data
  rownames(hclust_data_1)<- hvt_res2


  # Perform hierarchical clustering
  hc <- stats::hclust(stats::dist(as.matrix(hclust_data_1)), method = clustering_method)
  clusters <- stats::cutree(hc, k = no_of_clusters)
  
  
  if(only_dendro){
    plot_dendrogram <- function(hc_1, no_of_clusters_1, highlight_labels = NULL) {
      
      function() {
        dend <- as.dendrogram(hc_1)
        
        if (!is.null(highlight_labels)) {
          labels_colors <- ifelse(labels(dend) %in% highlight_labels, "red", "black")
          dend <- dendextend::set(dend, "labels_col", labels_colors)
        }
        dend <- dendextend::set(dend, "labels_cex", 1)
        plot(dend, ylab = "Distance", xlab = "Clusters", leaflab = "perpendicular")
        dendextend::rect.dendrogram(dend, k = no_of_clusters_1, 
                                    border = grDevices::rainbow(no_of_clusters_1))
      }
    }
    
    
    a <- plot_dendrogram(hc_1 = hc,no_of_clusters_1 = no_of_clusters,highlight_labels = highlight_labels)
    
    output_list <- list(dendrogram = a)
    return(output_list)
  }else {
    plot_dendrogram <- function(hc_1, no_of_clusters_1, highlight_labels = NULL) {
      
      function() {
        dend <- as.dendrogram(hc_1)
        dend <- dendextend::set(dend, "labels_cex", 1)
        plot(dend, ylab = "Distance", xlab = "Clusters", leaflab = "perpendicular")
        dendextend::rect.dendrogram(dend, k = no_of_clusters_1, 
                                    border = grDevices::rainbow(no_of_clusters_1))
      }
    }
    
    
    
    a <- plot_dendrogram(hc_1 = hc,no_of_clusters_1 = no_of_clusters,highlight_labels = highlight_labels)
    
    
    # Prepare data for clusterPlotly
    centroid_src <- scoreHVT_results$centroidData
    if (is.null(centroid_src)) {
      ch <- scoreHVT_results$cellID_coordinates
      ch <- ch[!duplicated(ch$Cell.ID), ]
      ch <- ch[order(ch$Cell.ID), ]
      centroid_src <- data.frame(
        Cell.ID      = as.integer(ch$Cell.ID),
        names.column = as.character(ch$Cell.ID),
        x            = as.numeric(ch$x),
        y            = as.numeric(ch$y),
        stringsAsFactors = FALSE
      )
    }
    cluster_data <- centroid_src %>%
      dplyr::select("Cell.ID", "names.column") %>%
      mutate(clusters = clusters)
    
    cluster_data <- cluster_data[order(cluster_data$Cell.ID), ]
    #browser()  
    
    b <- clusterPlot(dataset= cluster_data, hvt.results  = trainHVT_results, domains.column = "clusters" )
    
    output_list <- list(
      hc = hc,
      clusters = clusters,
      cluster_data =cluster_data,
      dendrogram = a,
      clusterplot = b
    )
    
    return(output_list)
  }
  }
}

#' @name plotNovelCells
#' @title Plot the identified outlier cells in the voronoi tessellation map.
#' @description This is the main plotting function to construct hierarchical voronoi tessellations 
#' and highlight the outlier cells 
#' @param plot.cells Vector. A vector indicating the cells to be highlighted in the map
#' @param hvt.map List. A list containing the output of \code{trainHVT} function
#' which has the details of the tessellations to be plotted
#' @param line.width Numeric Vector. A vector indicating the line widths of the
#' tessellation boundaries for each level
#' @param color.vec Vector. A vector indicating the colors of the boundaries of
#' the tessellations at each level
#' @param pch Numeric. Symbol of the centroids of the tessellations
#' (parent levels) Default value is 21.
#' @param centroid.size Numeric. Size of centroids of first level
#' tessellations. Default value is 0.5
#' @param title String. Set a title for the plot. (default = NULL)
#' @param maxDepth Numeric. An integer indicating the number of levels. (default = NULL)
#' @returns Returns a ggplot object containing hierarchical voronoi tessellation plot
#' highlighting the outlier cells
#' @author Shantanu Vaidya <shantanu.vaidya@@mu-sigma.com>
#' @seealso \code{\link{trainHVT}} \cr \code{\link{plotHVT}}
#' @keywords Novelty_or_Outliers
#' @importFrom magrittr %>%
#' @import ggplot2
#' @examples
#' data("EuStockMarkets")
#' hvt.results <- trainHVT(EuStockMarkets, n_cells = 60, depth = 1, quant.err = 0.1, 
#'                        distance_metric = "L1_Norm", error_metric = "max",
#'                        normalize = TRUE,quant_method="kmeans")
#' #selected 55,58 are for demo purpose
#' plotNovelCells(c(55,58),hvt.results)
#' @export plotNovelCells


plotNovelCells <-
  function(plot.cells,
           hvt.map,
           line.width = c(0.6),
           color.vec = c("#141B41"),
           pch = 21,
           centroid.size = 0.5,
           title = NULL,
           maxDepth = 1) {

    hvt_list <- hvt.map
    outlier_cell_k <- plot.cells

    maxDepth <- min(maxDepth, max(hvt_list[[3]][["summary"]] %>% stats::na.omit() %>% dplyr::select("Segment.Level")))

    min_x <- 1e9
    min_y <- 1e9
    max_x <- -1e9
    max_y <- -1e9
    depthVal <- c()
    clusterVal <- c()
    childVal <- c()
    value <- c()
    x_pos <- c()
    y_pos <- c()
    x_cor <- c()
    y_cor <- c()
    depthPos <- c()
    clusterPos <- c()
    childPos <- c()
    levelCluster <- c()
    for (clusterNo in 1:length(hvt_list[[2]][[1]][[1]])) {
      bp_x <- hvt_list[[2]][[1]][[1]][[clusterNo]][["x"]]
      bp_y <- hvt_list[[2]][[1]][[1]][[clusterNo]][["y"]]


      if (min(bp_x) < min_x) {
        min_x <- min(bp_x)
      }
      if (max(bp_x) > max_x) {
        max_x <- max(bp_x)
      }
      if (min(bp_y) < min_y) {
        min_y <- min(bp_y)
      }
      if (max(bp_y) > max_y) {
        max_y <- max(bp_y)
      }
    }

    for (depth in 1:maxDepth) {
      for (clusterNo in 1:length(hvt_list[[2]][[depth]])) {
        for (childNo in 1:length(hvt_list[[2]][[depth]][[clusterNo]])) {
          current_cluster <- hvt_list[[2]][[depth]][[clusterNo]][[childNo]]

          x <- as.numeric(current_cluster[["x"]])
          y <- as.numeric(current_cluster[["y"]])
          x_cor <- c(x_cor, as.numeric(current_cluster[["pt"]][["x"]]))
          y_cor <- c(y_cor, as.numeric(current_cluster[["pt"]][["y"]]))
          depthVal <- c(depthVal, depth)
          clusterVal <- c(clusterVal, clusterNo)
          childVal <- c(childVal, childNo)
          depthPos <- c(depthPos, rep(depth, length(x)))
          clusterPos <- c(clusterPos, rep(clusterNo, length(x)))
          childPos <- c(childPos, rep(childNo, length(x)))
          x_pos <- c(x_pos, x)
          y_pos <- c(y_pos, y)
          levelCluster <- c(levelCluster, depth)
        }
      }
    }

    # outlier_cell_k = c(53)


    valuesDataframe <- data.frame(
      depth = depthVal,
      cluster = clusterVal,
      child = childVal
    )
    valuesDataframe <- valuesDataframe %>%
      dplyr::mutate(outlier_cell = ifelse((child %in% outlier_cell_k),
        "Outlier_Cell", ""
      ))

    positionsDataframe <- data.frame(
      depth = depthPos,
      cluster = clusterPos,
      child = childPos,
      x = x_pos,
      y = y_pos
    )


    centroidDataframe <-
      data.frame(x = x_cor, y = y_cor, lev = levelCluster)

    datapoly <-
      merge(valuesDataframe,
        positionsDataframe,
        by = c("depth", "cluster", "child")
      )

    datapoly_outlier_cell <- datapoly[which(datapoly$outlier_cell == "Outlier_Cell"), ]
    datapoly_non_outlier_cell <- datapoly[which(datapoly$outlier_cell != "Outlier_Cell"), ]


    p <- ggplot2::ggplot()

    for (i in maxDepth:1) {
      p <-
        p + ggplot2::geom_polygon(
          data = datapoly_outlier_cell[which(datapoly_outlier_cell$depth == i), ],
          ggplot2::aes(
            x = x,
            y = y,
            color = factor(depth),
            size = factor(depth),
            group = interaction(depth, cluster, child),
          ),
          fill = "red"
        ) +
        ggplot2::geom_polygon(
          data = datapoly_non_outlier_cell[which(datapoly_non_outlier_cell$depth == i), ],
          ggplot2::aes(
            x = x,
            y = y,
            color = factor(depth),
            size = factor(depth),
            group = interaction(depth, cluster, child),
          ),
          fill = NA
        ) +
        ggplot2::scale_colour_manual(values = color.vec) +
        ggplot2::scale_size_manual(values = line.width, guide = "none") +
        ggplot2::labs(color = "Level") +
        ggplot2::ggtitle(title)
    }

    for (depth in 1:maxDepth) {
      p <- p + ggplot2::geom_point(
        data = centroidDataframe[centroidDataframe["lev"] == depth, ],
        ggplot2::aes(x = x, y = y),
        size = (centroid.size / (2^(depth - 1))),
        pch = 21,
        fill = color.vec[depth],
        color = color.vec[depth]
      ) +
        ggplot2::geom_point(
          data = centroidDataframe[centroidDataframe["lev"] == depth, ],
          ggplot2::aes(x = x, y = y),
          size = (centroid.size / (2^(depth - 1))),
          pch = 21,
          fill = color.vec[depth],
          color = color.vec[depth]
        )
    }

    p <- p +
      ggplot2::scale_color_manual(
        name = "Outlier Cell",
        values = color.vec
      ) +
      ggplot2::theme_bw() + ggplot2::theme(
        plot.background = ggplot2::element_blank(),
        plot.title = element_text(
          size = 20,
          hjust = 0.5,
          margin = margin(0, 0, 20, 0)
        ),
        panel.grid = ggplot2::element_blank(),
        panel.border = ggplot2::element_blank(),
        axis.ticks = element_blank(),
        axis.text = element_blank(),
        axis.title = element_blank(),
        panel.background = element_blank()
      ) + ggplot2::theme(plot.title = element_text(hjust = 0.5)) +
      ggplot2::scale_x_continuous(expand = c(0, 0)) +
      ggplot2::scale_y_continuous(expand = c(0, 0)) +
      ggplot2::geom_label(
        label = centroidDataframe$outlier_cell,
        nudge_x = 0.45, nudge_y = 0.1,
        check_overlap = TRUE,
        label.padding = unit(0.55, "lines"),
        label.size = 0.4,
        color = "white",
        fill = "#038225"
      ) +
      theme(legend.position = "none")


    return(suppressMessages(p))
  }

#' @keywords internal

Corrected_Tessellations <-
function(current_dirsgs, current_tile, current_polygon){

  requireNamespace("splancs")      #csr function
  requireNamespace("sp")           #point.in.polygon function
  requireNamespace("polyclip")


  initial_dirsgs <- current_dirsgs
  #vert_outside1 and vert_outside2 check for the points which are outside the parent polygon
  vert_outside1 <- which(sp::point.in.polygon(current_dirsgs[, "x1"],
                                          current_dirsgs[, "y1"],
                                          current_tile$x,
                                          current_tile$y) == 0)

  vert_outside2 <- which(sp::point.in.polygon(current_dirsgs[, "x2"],
                                          current_dirsgs[, "y2"],
                                          current_tile$x,
                                          current_tile$y) == 0)

  #if the same index is present in both vert_outside1 and vert_outside2, it means the line
  #joining these points has both the end-points outside the parent polygon
  vert_outside <- intersect(vert_outside1, vert_outside2)

  #if the line joining the points p1 and p2 is outside the parent polygon and
  #p1 is boundary point and p2 is not, then make p2 the boundary point or vice versa
  if (length(vert_outside) > 0 ) {
    for (j in 1: length(vert_outside)) {
      #obtain the index of the line for which the point is outside
      x_coord_index1 <- which((current_dirsgs[vert_outside[j], "x1"] == current_dirsgs[, c("x1", "x2")]) == TRUE)
      y_coord_index1 <- which((current_dirsgs[vert_outside[j], "y1"] == current_dirsgs[, c("y1", "y2")]) == TRUE)
      x_coord_index2 <- which((current_dirsgs[vert_outside[j], "x2"] == current_dirsgs[, c("x1", "x2")]) == TRUE)
      y_coord_index2 <- which((current_dirsgs[vert_outside[j], "y2"] == current_dirsgs[, c("y1", "y2")]) == TRUE)
      x_coord_index <- c(x_coord_index1, x_coord_index2)
      y_coord_index <- c(y_coord_index1, y_coord_index2)
      xy_index <- intersect(x_coord_index, y_coord_index)
      ind_len <- length(xy_index)
      #change the boundary points of those indices to TRUE
      for (i in 1: ind_len) {
        if (xy_index[i] > nrow(current_dirsgs)) {
          current_dirsgs[(xy_index[i] - nrow(current_dirsgs)), "bp2"] <- TRUE
        } else {
          current_dirsgs[xy_index[i], "bp1"] <- TRUE
        }
      }
    }
  }

  #boundary points in the first index
  firstpt <- current_dirsgs[which(current_dirsgs[, "bp1"] == TRUE), ]
  #boundary points in the second index
  secondpt <- current_dirsgs[which(current_dirsgs[, "bp2"] == TRUE), ]

  #use all these points together and remove the common rows
  pts <- rbind(firstpt, secondpt)
  if(any(duplicated(pts) == TRUE)){
    ptsout <- pts[-which(duplicated(pts) == TRUE), ]
  }else{
    ptsout <- pts
  }

  #2 matrices, 'm' to store slope and 'c' to store constants of the lines in the polygon
  m_parent <-  c_parent <- matrix(ncol = 1, nrow = nrow(current_polygon))
  #mat1 is the current polygon
  mat1 <-  current_polygon
  #closing the polygon by making the first and last points same
  mat1 <- rbind(mat1, mat1[1, ])

  # Polygon offset to check if points lie within this margin
  A<-list(list(x=mat1[,1], y=mat1[,2]))
  C <- polyclip::polylineoffset(A, 0.0001, jointype="square", endtype="opensquare")
  offset_polygon<-cbind(C[[1]][["x"]],C[[1]][["y"]])

  #slopes and intercepts for lines in the parent polygon
  for (i in 1: (nrow(mat1) - 1)) {
    m_parent[i] <- ((mat1[(i + 1), 2] - mat1[i, 2]) /
                      (mat1[(i + 1), 1] - mat1[i, 1]))
    c_parent[i] <- mat1[i, 2] - (m_parent[i] * mat1[i, 1])
  }

  #polygon is too small to calculate child level tessellations
  if(any(is.nan(m_parent) == TRUE)){
    #flog.warn("Projection is not scaled enough")
    return("-1")
  }

  #slopes and intercepts for lines with endpoints outside the polgyon
  if(nrow (ptsout) != 0){
    m_ptsout <- c_ptsout <- matrix(ncol = 1, nrow = nrow(ptsout))
    #for all lines in the child tessellations for which the points are outside
    # i<-3
    for (i in 1: nrow(ptsout)){
      m_ptsout[i] <- ((ptsout[i, "y2"] - ptsout[i, "y1"]) /
                        (ptsout[i, "x2"] - ptsout[i, "x1"]))
      c_ptsout[i] <- ptsout[i, "y1"] - (m_ptsout[i] * ptsout[i, "x1"])

      #for all lines in the child tessellation for which the points are outside,,
      #calculate the point of intersection of each line
      #with all the boundary lines of the parent polygon
      x <- y <- c()
      for (j in 1: (nrow(mat1) - 1)) {
        if ((abs(m_parent[j]) == Inf) || (m_parent[j] == m_ptsout[i])) {
          x[j] <- mat1[j, 1]
          if((abs(m_ptsout[i]) == Inf) || (m_parent[j] == m_ptsout[i])){
            #both are vertical lines or both are parallel
            y[j] <- 500
          }else{
            y[j] <- (m_ptsout[i] * x[j]) + c_ptsout[i]
          }
        } else {
          if (abs(m_ptsout[i]) == Inf) {
            x[j] <- ptsout[i, "x1"]
            y[j] <- (m_parent[j] * x[j]) + c_parent[j]
          } else {
            x[j] <- (c_ptsout[i] - c_parent[j]) / (m_parent[j] - m_ptsout[i])
            y[j] <- (m_ptsout[i] * x[j]) + c_ptsout[i]
          }
        }
      }

      #check which points among all the points of intersection lies on the boundary of parent polygon
      pt_intersect <- matrix(NA, ncol = 2, nrow = (nrow(mat1) - 1))
      xy.poly <- splancs::as.points(mat1)
      for(j in 1: (nrow(mat1) - 1)){
        # j<-2
        xy <- splancs::as.points(x[j], y[j])
        if(
          # (nrow(splancs::pip(xy, xy.poly, out = FALSE, bound = TRUE)) != 0) ||  obsolete code
             (sp::point.in.polygon(x[j], y[j], offset_polygon[, 1], offset_polygon[, 2]) !=0) ||
             (sp::point.in.polygon(as.character(x[j]), as.character(y[j]), offset_polygon[, 1], offset_polygon[, 2]) != 0)){
          pt_intersect[j, ] <- xy
        }
      }
      #discard the points which are not on the boundary.
      rows_del <- NULL
      for(k in 1: (nrow(mat1) - 1)){
        if(length(all(is.na(pt_intersect[k,])))){
          rows_del[k] <- all(is.na(pt_intersect[k,]))
        }
      }

      if(any(rows_del == TRUE)){
        pt_intersect <- pt_intersect[-which(rows_del == TRUE), , drop = FALSE]
      }

      #calculate the distance between the boundary point of the second level tessellation with the
      #points of intersection which are on the boundary of the parent tile.
      if(nrow(pt_intersect) != 0){
        distmat <- matrix(0, ncol = 2, nrow = (nrow(pt_intersect) + 1))
        distmat[2: nrow(distmat), ] <- pt_intersect
        #choose the point which is a boundary point
        if(ptsout[i, "bp1"] == TRUE){
          distmat[1, ] <- as.matrix(ptsout[i, c("x1", "y1")])
          desired_pt <- distmat[which(stats::dist(distmat) == min(as.matrix(stats::dist(distmat))[2: nrow(distmat),1])) + 1,]   # +1 to get the correct index from the output of dist function
          ptsout[i, c("x1", "y1")] <- desired_pt
        }
        if(ptsout[i, "bp2"] == TRUE){
          distmat[1, ] <- as.matrix(ptsout[i, c("x2", "y2")])
          desired_pt <- distmat[which(stats::dist(distmat) == min(as.matrix(stats::dist(distmat))[2: nrow(distmat),1])) + 1,]   # +1 to get the correct index from the output of dist function
          ptsout[i, c("x2", "y2")] <- desired_pt
        }
      }
    }
  }

  #assign the new points to the output of the tessellation
  current_dirsgs[rownames(ptsout), ] <- ptsout

  #return the result
  return (current_dirsgs)
}

#' @name reconcileTransitionProbability
#' @title Reconciliation of Transition Probability
#' @description This is the main function for creating reconciliation plots and tables which 
#' helps in comparing the transition probabilities calculated manually and from markovchain function
#' @param df Data frame. The input data frame should contain two columns, cell ID from scoreHVT function and timestamp of that dataset.
#' @param cellid_column Character. Name of the column containing cell IDs.
#' @param time_column Character. Name of the column containing timestamps
#' @param  hmap_type Character. ('self_state', 'without_self_state', or 'All')
#' @return A list of plotly heatmap objects and tables representing the transition probability heatmaps.
#' @author PonAnuReka Seenivasan <ponanureka.s@@mu-sigma.com>, Vishwavani <vishwavani@@mu-sigma.com>
#' @keywords Timeseries_Analysis
#' @importFrom magrittr %>%
#' @import markovchain methods
#' @examples
#' dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
#'                       DAX = EuStockMarkets[, "DAX"],
#'                       SMI = EuStockMarkets[, "SMI"],
#'                       CAC = EuStockMarkets[, "CAC"],
#'                       FTSE = EuStockMarkets[, "FTSE"])
#' hvt.results<- trainHVT(dataset,n_cells = 60, depth = 1, quant.err = 0.1,
#'                        distance_metric = "L1_Norm", error_metric = "max",
#'                        normalize = TRUE,quant_method = "kmeans")
#' scoring <- scoreHVT(dataset, hvt.results)
#' cell_id <- scoring$scoredPredictedData$Cell.ID
#' time_stamp <- dataset$date
#' dataset <- data.frame(cell_id, time_stamp)
#' reconcileTransitionProbability(dataset, hmap_type = "All", 
#' cellid_column = "cell_id", time_column = "time_stamp")
#' @export reconcileTransitionProbability


reconcileTransitionProbability <- function(df, hmap_type = NULL, cellid_column, time_column) {
 
  StateFrom <- StateTo <- Current_State <- Next_State_manual <- Next_State_markov <- Probability_manual_calculation <- Probability_markov_function<- NULL
  
  # Rename columns for consistency
  colnames(df)[colnames(df) == time_column] <- "Timestamp"
  colnames(df)[colnames(df) == cellid_column] <- "Cell.ID"
  
  # Set default value for hmap_type if it is NULL
  if (is.null(hmap_type)) hmap_type <- "self_state"
  
  # Heatmap 1: Transition probability with self-state
  raw_data <- df %>% dplyr::select("Cell.ID", "Timestamp")
  transition_values <- table(raw_data$Cell.ID[-nrow(raw_data)], raw_data$Cell.ID[-1])
  mat <- unclass(transition_values)
  normalized_value <- mat / rowSums(mat)
  melted_matrix <- reshape2::melt(normalized_value)
  a_df <- melted_matrix %>% as.data.frame()
  colnames(a_df) <- c("StateFrom", "StateTo", "Probabilty")
  a_df$StateTo <- as.integer(a_df$StateTo)
  a_df$StateFrom <- as.integer(a_df$StateFrom)
  a_df$Probabilty <- round(a_df$Probabilty, digits = 4)
  a_df <- a_df %>% arrange( StateFrom, StateTo)
  
  hmap1 <- plotly::plot_ly(
    data = a_df,
    x = ~StateFrom,
    y = ~StateTo,
    z = ~Probabilty,
    type = "heatmap",
    colorscale = "colz",
    name = "Probability",
    hovertemplate = "Cell t: %{x}<br>Cell t+1: %{y}<br>Probability: %{z}"
  ) %>%
    plotly::layout(
      xaxis = list(title = "Cell ID From"),
      yaxis = list(title = "Cell ID To"),
      autosize = TRUE
    )
  
  # Heatmap 2: Transition probability without self-state
  raw_data1 <- df
  transition_values1 <- table(raw_data1$Cell.ID[-nrow(raw_data1)], raw_data1$Cell.ID[-1])
  mat1 <- unclass(transition_values1)
  normalized_value1 <- mat1 / rowSums(mat1)

  #nullifying self state probabilities 
  normalized_value1[cbind(1:nrow(normalized_value1), 1:ncol(normalized_value1))] <- 0
  
  melted_matrix1 <- reshape2::melt(normalized_value1)
  a_df1 <- melted_matrix1 %>% as.data.frame()
  colnames(a_df1) <- c("StateFrom", "StateTo", "Probabilty")
  a_df1$StateTo <- as.integer(a_df1$StateTo)
  a_df1$StateFrom <- as.integer(a_df1$StateFrom)
  a_df1$Probabilty <- round(a_df1$Probabilty,4)
  a_df1 <- a_df1 %>% arrange( StateFrom, StateTo)
    
  hmap2 <- plotly::plot_ly(
    data = a_df1,
    x = ~StateFrom,
    y = ~StateTo,
    z = ~Probabilty,
    type = "heatmap",
    colorscale = "colz",
    hovertemplate = "Cell t: %{x}<br>Cell t+1: %{y}<br>Probability: %{z}"
  ) %>%
    plotly::layout(
      xaxis = list(title = "Cell ID From"),
      yaxis = list(title = "Cell ID To"),
      autosize = TRUE
    )
  
  # Heatmap 3: Markov Chain state transition Probability
  mc_data  <- df$Cell.ID[-c(1, nrow(df))]
  mc <- markovchain::markovchainFit(data = mc_data)
  trans_matrix <- mc$estimate
  trans_plot <-methods::as(trans_matrix, "matrix")
  #trans_plot <- as.matrix.default(trans_matrix)
  melted_matrix_mc <- reshape2::melt(trans_plot)
  a_df_mc <- melted_matrix_mc %>% as.data.frame()
  colnames(a_df_mc) <- c("StateFrom", "StateTo", "Probability")
  a_df_mc$Probability <- round(a_df_mc$Probability,4)
  a_df_mc <- a_df_mc %>% arrange( StateFrom, StateTo)

  hmap3 <- plotly::plot_ly(
    data = a_df_mc,
    x = ~StateFrom,
    y = ~StateTo,
    z = ~Probability,
    type = "heatmap",
    colors = grDevices::colorRampPalette(c("white", "blue"))(100),
    hovertemplate = "Cell t: %{x}<br>Cell t+1: %{y}<br>Probability: %{z}"
  ) %>%
    plotly::layout(
      xaxis = list(title = "Cell ID From"),
      yaxis = list(title = "Cell ID To"),
      autosize = TRUE
    )

  
  #without self - state
  trans_matrix_no_self <- trans_plot
  diag(trans_matrix_no_self) <- 0  # Set self-transitions to 0
  
  # Create the heatmap data frame
  melted_matrix_mc1 <- reshape2::melt(trans_matrix_no_self)
  a_df_mc1 <- melted_matrix_mc1 %>% as.data.frame()
  colnames(a_df_mc1) <- c("StateFrom", "StateTo", "Probability")
  a_df_mc1$Probability <- round(a_df_mc1$Probability,4)
  a_df_mc1 <- a_df_mc1 %>% arrange( StateFrom, StateTo) 
  
  # Create the heatmap without self-state
  hmap4 <- plotly::plot_ly(
    data = a_df_mc1,
    x = ~StateFrom,
    y = ~StateTo,
    z = ~Probability,
    type = "heatmap",
    colors = grDevices::colorRampPalette(c("white", "blue"))(100),
    hovertemplate = "Cell t: %{x}<br>Cell t+1: %{y}<br>Probability: %{z}"
  ) %>%
    plotly::layout(
      xaxis = list(title = "Cell ID From"),
      yaxis = list(title = "Cell ID To"),
      autosize = TRUE
    )
 
   a_df <- a_df[a_df$Probabilty !=0,]
   a_df <- a_df %>% arrange( StateFrom, StateTo)
  colnames(a_df) <- c("Current_State", "Next_State_manual", "Probability_manual_calculation")
  
  
   a_df_mc <- a_df_mc[a_df_mc$Probability != 0,]
   a_df_mc <- a_df_mc %>% arrange( StateFrom, StateTo) 
  colnames(a_df_mc) <- c("Current_State", "Next_State_markov", "Probability_markov_function")
  
  
  a_df_mc <- a_df_mc[,-1]
  self_state_table <- cbind (a_df, a_df_mc)
  self_state_table$diff <- (self_state_table$Probability_manual_calculation - self_state_table$Probability_markov_function)
  self_state_table <- self_state_table %>% dplyr::select(Current_State,Next_State_manual, Next_State_markov, Probability_manual_calculation,
                                                          Probability_markov_function, diff)
  
  

  a_df1 <- a_df1[a_df1$Probabilty !=0,]
  a_df1 <- a_df1 %>% arrange( StateFrom, StateTo)
  colnames(a_df1) <- c("Current_State", "Next_State_manual", "Probability_manual_calculation")
  
  
  a_df_mc1 <- a_df_mc1[a_df_mc1$Probability != 0,]
  a_df_mc1 <- a_df_mc1 %>% arrange( StateFrom, StateTo)
  colnames(a_df_mc1) <- c("Current_State", "Next_State_markov", "Probability_markov_function")
  
  
  a_df_mc1 <- a_df_mc1[,-1]
  non_self_state_table <- cbind (a_df1, a_df_mc1)
  non_self_state_table$diff <- (non_self_state_table$Probability_manual_calculation - non_self_state_table$Probability_markov_function)
  non_self_state_table <- non_self_state_table %>% dplyr::select(Current_State,Next_State_manual, Next_State_markov, Probability_manual_calculation,
                                                         Probability_markov_function, diff)
  
  

  ############
  self_state_plots <-  plotly::subplot(hmap1, hmap3, shareY = TRUE ) %>%
    plotly::layout(annotations = list(
      list(x = 0.15, y = 1.05, text = "Using Manual calculation", showarrow = F, xref='paper', yref='paper', width = 150),
      list(x = 0.9, y = 1.05, text = "Using Markovchain method", showarrow = F, xref='paper', yref='paper', width = 165))) %>%
    plotly::layout(xaxis = list(title = "Cell ID From"), 
                    xaxis2 = list(title = "Cell ID From"))
  
    
  
  
  ###############
  non_self_state_plots <- plotly::subplot(hmap2, hmap4,shareY = TRUE) %>%
    plotly::layout(annotations = list(
      list(x = 0.15, y = 1.05,text = "Using Manual calculation", showarrow = F, xref='paper', yref='paper',width = 150),
      list(x = 0.9, y = 1.05,text = "Using Markovchain method", showarrow = F, xref='paper', yref='paper',width = 165)))%>% 
  plotly::layout(xaxis = list(title = "Cell ID From"), 
                 xaxis2 = list(title = "Cell ID From"))
  
  
  # Determine which heatmaps to return based on the hmap_type parameter
  if (hmap_type == "without_self_state") {
    return(list(non_self_state_plots, non_self_state_table))
  } else if (hmap_type == "self_state") {
    return(list(self_state_plots, self_state_table))
  } else if (hmap_type == "All") {
    return(list(self_state_plots,non_self_state_table, non_self_state_plots, non_self_state_table))
  } else {
    stop("Invalid plot_type parameter. Use 'without_self_state', 'self_state', or 'both'.")
  }
}

#' @keywords internal

diagPlot <- function(hvt.results,
                  data,
                  level,
                  quant.err,
                  distance_metric="L1_Norm",
                  error_metric="max",
                  ...){

  requireNamespace("reshape2")     
  requireNamespace("ggplot2")     
  requireNamespace("dplyr")     

  ##### Min Inter-Centroid distance plot
  newdfMapping <- hvt.results[[3]][["summary"]]
  
  x <- newdfMapping %>% 
    dplyr::filter((n > 0 & Segment.Level == level) | (Segment.Level < level & (Quant.Error < quant.err | n <= 3)))
  x <- x[x%>%stats::complete.cases(),]
  
  singleton_count=sum(x$Quant.Error< 0.0001)
  num_cells=length(x$Quant.Error)
  
  # x <- hvt.results[[3]][["summary"]]
  d  <- stats::dist(x,method = "manhattan")
  df <- reshape2::melt(as.matrix(d), varnames = c("row", "col"))
  df <- df[df$value!=0,]
  df$value <- df$value/ncol(x)
  df_cent <- df %>% dplyr::group_by(row) %>% dplyr::summarise(min_dist = min(value, na.rm = TRUE))
  
  
  p_cent <- ggplot2::ggplot(df_cent,aes(x = min_dist)) +
    ggplot2::geom_histogram(ggplot2::aes(y = stat(density)),fill = "midnightblue",colour="white",alpha=0.75) +
    
    ggplot2::stat_function(ggplot2::aes(color = "Normal"), fun = stats::dnorm,args = list(mean = mean(df_cent$min_dist),sd = stats::sd(df_cent$min_dist))) +
    ggplot2::stat_density(ggplot2::aes(color = "Density"), geom = "line", linetype = "dashed")  +
    ggplot2::scale_colour_manual("", values = c("dodgerblue", "#EE7600")) +
    ggplot2::scale_linetype_manual("", values = c("Normal" = 2, "Density" = 1)) +
    
    ggplot2::theme_minimal()+
    ggplot2::geom_vline(xintercept = mean(df_cent$min_dist), colour="#EE7600",linetype="dashed")+
    ggplot2::ggtitle("Minimum Intra-Centroid Distance Plot")+
    ggplot2::xlab("Distance between Centroids")+
    ggplot2::ylab("Density")+
    ggplot2::annotate("text",x = mean(df_cent$min_dist),
             y = Inf,vjust=1,hjust=-0.5,
             label = mean(df_cent$min_dist)%>%round(2),fontface = "bold")
  
  
  ##### Min Inter-Point distance plot

  d = stats::dist(data,method = "manhattan")
  
  df <- reshape2::melt(as.matrix(d), varnames = c("row", "col"))
  df=df[df$value!=0,]
  df$value=df$value/ncol(data)
  df_data = df %>% dplyr::group_by(row) %>% dplyr::summarise(min_dist = min(value, na.rm = TRUE))

  p_datapoint = ggplot2::ggplot(df_data,aes(x = min_dist)) +
    
    ggplot2::geom_histogram(ggplot2::aes(y = stat(density)),fill = "midnightblue",colour="white",alpha=0.75) +
    
    ggplot2::stat_function(ggplot2::aes(color = "Normal"), fun = stats::dnorm,args = list(mean = mean(df_data$min_dist),sd = stats::sd(df_data$min_dist))) +
    ggplot2::stat_density(ggplot2::aes(color = "Density"), geom = "line", linetype = "dashed")  +
    ggplot2::scale_colour_manual("", values = c("dodgerblue", "#EE7600")) +
    ggplot2::scale_linetype_manual("", values = c("Normal" = 2, "Density" = 1)) +
    
    ggplot2::theme_minimal()+
    ggplot2::geom_vline(xintercept = mean(df_data$min_dist), colour="#EE7600",linetype="dashed")+
    ggtitle("Minimum Intra-DataPoint Distance Plot")+
    ggplot2::xlab("Distance between DataPoints")+
    ggplot2::ylab("Density")+
    ggplot2::annotate("text",x = mean(df_data$min_dist),
             y = Inf,vjust=1,hjust=-0.5,
             label = mean(df_data$min_dist)%>%round(2),fontface = "bold") 
  

  ####### Number of datapoints
  
  n_cells=x%>% dplyr::select("n") %>% unlist()
  p_obs=ggplot2::ggplot() +
    ggplot2::aes(x = n_cells) +
    ggplot2::geom_histogram(fill = "midnightblue") +
    ggplot2::theme_minimal()+
    ggplot2::geom_vline(xintercept = round(mean(n_cells),0), colour="red",linetype="dashed")+
    ggplot2::ggtitle("Distribution of number of observation in cells")+
    ggplot2::xlab("Number of Observation in Cells")+
    ggplot2::ylab("Number of Occurrences")+
    ggplot2::annotate("text",x =round(mean(n_cells),0),
             y = Inf,vjust=1,hjust=-0.5,
             label = round(mean(n_cells),0),fontface = "bold") 
  
  # MAD Calibration
  ####### MAD Plot ################
  predictions_train = list()
  predictions_train <- scoreHVT(
    dataset = data,
    hvt.results.model=hvt.results,
    child.level = level,
    line.width = c(0.6, 0.4, 0.2),
    color.vec = c("#141B41", "#6369D1", "#D8D2E1"),
    mad.threshold = quant.err,
    distance_metric = distance_metric,
    error_metric = error_metric
  )
  
  mad_plot_train=madPlot(predictions_train)
  
  df_sing=data.frame(
    "Centroid_Type"=c("Singletons","Non-Singletons"),
    n=c(singleton_count,(num_cells-singleton_count))
  )%>% 
    dplyr::mutate(perc = `n` / sum(`n`)) %>% 
    dplyr::arrange(perc) %>%
    dplyr::mutate(labels = scales::percent(perc))
  
  
  sing_pie=ggplot2::ggplot(df_sing, aes(x = "", y = perc, fill = Centroid_Type)) +
    ggplot2::geom_col(color = "white") +
    ggplot2::geom_label(aes(label = labels),
               position = position_stack(vjust = 0.5),
               show.legend = FALSE,colour="white",fontface = "bold") +
    ggplot2::guides(fill = guide_legend(title = "Centroid_Type")) +
    # scale_fill_viridis_d() +
    ggplot2::coord_polar(theta = "y") + 
    ggplot2::scale_fill_manual(values=c("#2C83B5","#EE6A42"))+
    ggplot2::theme_void()+ggplot2::labs(title = "Singletons Pie Chart: HVT Model | Train Data",
                      caption = paste("Total number of centroids:",num_cells,",Singletons:",singleton_count,",Non-Singletons:",(num_cells-singleton_count)))
  sing_pie

  
  plot_list=list(
    cent_plot = p_cent,
    datapoint_plot = p_datapoint,
    number_plot = p_obs,
    mad_plot_train=mad_plot_train,
    singleton_piechart=sing_pie,
    singleton_count=list(singleton_count=singleton_count,
                         num_cells=num_cells,
                         singleton_count_percentage=(singleton_count/num_cells)*100)
   

  )
  return(plot_list)
  
} 

#' @name plotAnimatedFlowmap
#' @title Generating flow maps and animations based on transition probabilities
#' @description This is the main function for generating flow maps and animations based on transition probabilities
#' including self states and excluding self states.
#' Flow maps are a type of data visualization used to represent the transition probability of different states. 
#' Animations are the gifs used to represent the movement of data through the cells. 
#' @param hvt_model_output List. Output from a trainHVT function.
#' @param transition_probability_df List. Output from getTransitionProbability function
#' @param df Data frame. The input dataframe should contain two columns, 
#' cell ID from scoreHVT function and time stamp of that dataset.
#' @param animation Character. Type of animation ('state_based', 'time_based', 'All' or  NULL)
#' @param flow_map Character. Type of flow map ('self_state', 'without_self_state', 'All' or NULL)
#' @param fps_time Numeric. A numeric value for the frames per second of the time transition gif.
#' (Must be a numeric value and a factor of 100). Default value is 1.
#' @param fps_state Numeric. A numeric value for the frames per second of the state transition gif.
#' (Must be a numeric value and a factor of 100). Default value is 1.
#' @param time_duration Numeric. A numeric value for the duration of the time transition gif.
#' Default value is 2.
#' @param state_duration Numeric. A numeric value for the duration of the state transition gif.
#' Default value is 2.
#' @param cellid_column Character. Name of the column containing cell IDs.
#' @param time_column Character. Name of the column containing time stamps
#' @return A list of flow maps and animation gifs.
#' @author PonAnuReka Seenivasan <ponanureka.s@@mu-sigma.com>, Vishwavani <vishwavani@@mu-sigma.com>
#' @seealso \code{\link{trainHVT}} \cr \code{\link{scoreHVT}} \cr \code{\link{getTransitionProbability}}
#' @keywords Timeseries_Analysis
#' @importFrom magrittr %>%
#' @include utilities.R
#' @examples
#' dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
#'                       DAX = EuStockMarkets[, "DAX"],
#'                       SMI = EuStockMarkets[, "SMI"],
#'                       CAC = EuStockMarkets[, "CAC"],
#'                       FTSE = EuStockMarkets[, "FTSE"])
#' hvt.results<- trainHVT(dataset,n_cells = 10, depth = 1, quant.err = 0.1,
#'                        distance_metric = "L1_Norm", error_metric = "max",
#'                        normalize = TRUE,quant_method = "kmeans")
#' scoring <- scoreHVT(dataset, hvt.results)
#' cell_id <- scoring$scoredPredictedData$Cell.ID
#' time_stamp <- dataset$date
#' dataset <- data.frame(cell_id, time_stamp)
#' table <- getTransitionProbability(dataset, cellid_column = "cell_id",time_column = "time_stamp")
#' plots <- plotAnimatedFlowmap(hvt_model_output = hvt.results, transition_probability_df = table,
#' df = dataset, animation = 'All', flow_map = 'All',fps_time = 1,fps_state =  1,time_duration = 2,
#' state_duration = 2,cellid_column = "cell_id", time_column = "time_stamp")
#' @export plotAnimatedFlowmap

plotAnimatedFlowmap <- function(hvt_model_output, transition_probability_df, df, 
                                animation = "All", flow_map = "All", 
                                fps_time = 1, fps_state = 1, time_duration = 2, 
                                state_duration = 2, cellid_column, time_column) {
  
  
  ##for cran warnings
  CircleSize  <- Current_State <- Next_State <- Transition_Probability <- Cumulative_Probability <- Relative_Frequency <- latency <- NULL
  
  valid_animation <- c("All", "state_based", "time_based", NULL)
  valid_flowmap <- c("All", "self_state", "without_self_state", NULL)
  
  if (!is.null(animation) && !(animation %in% valid_animation)) {
    stop("Invalid animation argument. Must be one of: 'All', 'state_based', 'time_based', or NULL")
  }
  
  if (!is.null(flow_map) && !(flow_map %in% valid_flowmap)) {
    stop("Invalid flow_map argument. Must be one of: 'All', 'self_state', 'without_self_state', or NULL") 
  }
  
  
  ##for cran warnings, initializing empty vectors for these variables.
  segment_len <- grp <- colour<-order_map<- Frequency <-Timestamp<-y2 <-x2 <-y1.y<- x1.y<- y1 <-x1<-label<-NULL
  
  
  
  # Rename columns for consistency
  colnames(df)[colnames(df) == time_column] <- "Timestamp"
  colnames(df)[colnames(df) == cellid_column] <- "Cell.ID"
  
  ###########centroid - x and y coordinates################
  cellID_coordinates <- extract_cell_coordinates(hvt_model_output)
  current_state_data <- dplyr::arrange(cellID_coordinates, Cell.ID)
  colnames(current_state_data) <- c("x1", "y1", "Cell.ID")
  
  ################################################################
  #####Function to get highest state and probability excluding self-state
  get_second_highest <- function(df) {
    df$Next_State <- as.integer(df$Next_State)
    # Remove self transitions first
    df <- df[df$Next_State != df$Current_State[1], ]
    
    if(nrow(df) > 0) {
      # Get the highest probability from remaining transitions
      max_probability_row <- df[which.max(df$Transition_Probability), ]
      return(data.frame(
        Next_State = max_probability_row$Next_State,
        Probability = max_probability_row$Transition_Probability
      ))
    }
    return(data.frame(Next_State = 0, Probability = 0))
  }
  
  grouped_df <- split(transition_probability_df, transition_probability_df$Current_State)
  second_highest_states_list <- lapply(grouped_df, get_second_highest)
  second_state_df <- do.call(rbind, second_highest_states_list)
  ##########################################################
  
  
  
  ######################################################################
  # Function to get the highest probability state and probability
  get_highest_probability <- function(df) {
    df$Next_State <- as.integer(df$Next_State)
    # Get max probability
    max_prob <- max(df$Transition_Probability)
    # Filter rows with max probability
    max_prob_rows <- df[df$Transition_Probability == max_prob, ]
    
    # If current state exists in max probability rows, select it
    current_state <- df$Current_State[1]
    if(current_state %in% max_prob_rows$Next_State) {
      selected_row <- max_prob_rows[max_prob_rows$Next_State == current_state, ]
    } else {
      selected_row <- max_prob_rows[1, ]
    }
    
    return(data.frame(Next_State = selected_row$Next_State, 
                      Probability = selected_row$Transition_Probability))
  }
  
  # Apply the function to each data frame in 'transition_probability_df'
  grouped_df <- split(transition_probability_df, transition_probability_df$Current_State)
  highest_probability_states_list <- lapply(grouped_df, get_highest_probability)
  
  # Combine the results into a single data frame
  first_state_df <- do.call(rbind, highest_probability_states_list)
  ###################################################################################
  
  ############## Dataframe for second_highest state (without self state)
  merged_df1 <- cbind(current_state_data, second_state_df)
  merged_df1 <- merged_df1 %>%
    left_join(dplyr::select(merged_df1, Cell.ID, x1, y1), by = c("Next_State" = "Cell.ID")) %>%
    dplyr::mutate(x2 = x1.y, y2 = y1.y) %>%
    dplyr::select(-x1.y, -y1.y)
  colnames(merged_df1) <- c("x1", "y1", "Cell.ID", "Next_State", "Probability", "x2", "y2")
  
  # Dataframe for highest state  (with self state)
  merged_df2 <- cbind(current_state_data, first_state_df)
  merged_df2 <- merged_df2 %>%
    left_join(dplyr::select(merged_df2, Cell.ID, x1, y1), by = c("Next_State" = "Cell.ID")) %>%
    dplyr::mutate(x2 = x1.y, y2 = y1.y) %>%
    dplyr::select(-x1.y, -y1.y)
  colnames(merged_df2) <- c("x1", "y1", "Cell.ID", "Next_State", "Probability", "x2", "y2")
  merged_df2$Probability <- round(merged_df2$Probability, digits = 3)
  
  # Self-state Plot
  prob1 <- merged_df2$Probability 
  cellID_coordinates$prob1 <- prob1
  
  # Initialize plot variables
  self_state_plot <- NULL
  arrow_flow_map <- NULL
  time_animation <- NULL
  state_animation <- NULL
  
  if (!is.null(flow_map)) {
    if(flow_map %in% c('self_state', 'All')) {
      
      # Get min and max probabilities
      min_prob <- min(merged_df2$Probability, na.rm = TRUE)
      max_prob <- max(merged_df2$Probability, na.rm = TRUE)
      
      # Check if all probability values are the same
      if(length(unique(merged_df2$Probability)) == 1) {
        # If all values are identical, use a single circle size for all points
        merged_df2$CircleSize <- 3  # Set a moderate circle size
        
        # Create a simple legend with just the single value
        single_value <- unique(merged_df2$Probability)
        breaks <- c(single_value - 0.001, single_value + 0.001)
        legend_labels <- format(round(single_value, 3), nsmall = 3)
        legend_size <- 5  # A single, moderate legend size
      } else {
        # Normal case with varying probabilities
        # Create breaks based on 30% quantiles
        probs_seq <- seq(0, 1, by = 0.3)
        custom_breaks <- stats::quantile(merged_df2$Probability, probs = probs_seq)
        
        # Fix the duplicate breaks issue
        if(any(duplicated(custom_breaks))) {
          # Find the unique probability values
          unique_probs <- sort(unique(merged_df2$Probability))
          
          # If there are few unique values, use those directly
          if(length(unique_probs) <= length(probs_seq)) {
            # Add small epsilon to create slightly different breaks
            epsilon <- (max_prob - min_prob) / 1000
            if(epsilon == 0) epsilon <- 0.001 # Ensure non-zero epsilon
            
            # Create breaks that ensure each unique probability value falls in a different bin
            custom_breaks <- c(min_prob - epsilon)
            for(i in 1:(length(unique_probs) - 1)) {
              custom_breaks <- c(custom_breaks, (unique_probs[i] + unique_probs[i+1]) / 2)
            }
            custom_breaks <- c(custom_breaks, max_prob + epsilon)
          } else {
            # If there are many unique values but still have duplicates in the quantiles
            # Use evenly spaced breaks between min and max
            custom_breaks <- seq(min_prob, max_prob, length.out = length(probs_seq))
          }
        }
        
        # Make sure first break is slightly lower than minimum
        custom_breaks[1] <- min_prob - 0.001
        custom_breaks[length(custom_breaks)] <- max_prob + 0.001
        
        # Assign circle sizes
        merged_df2$CircleSize <- as.numeric(cut(merged_df2$Probability, 
                                                breaks = custom_breaks, 
                                                labels = seq(1, length(custom_breaks) - 1),
                                                include.lowest = TRUE))
        
        # Handle NA values
        merged_df2$CircleSize <- ifelse(is.na(merged_df2$CircleSize), 
                                        max(merged_df2$CircleSize, na.rm = TRUE), 
                                        merged_df2$CircleSize)
        
        # Create legend sizes dynamically based on number of unique categories
        n_categories <- length(unique(merged_df2$CircleSize))
        legend_size <- 2.6 * seq_len(n_categories)
        
        # Create breaks for legend
        breaks <- as.numeric(custom_breaks)
        breaks[1] <- breaks[1] + 0.001
        
        # Generate legend labels - fix to avoid duplicate values
        last_label_max <- min(max_prob, 1.000)
        
        generate_legend_labels <- function(custom_breaks) {
          # Ensure custom_breaks is sorted (just in case)
          custom_breaks <- sort(custom_breaks)
          
          # Initialize an empty vector for labels
          legend_labels <- c()
          
          # Loop through the breakpoints to create labels
          for (i in seq_along(custom_breaks)[-length(custom_breaks)]) {
            lower_bound <- round(custom_breaks[i], 3)
            upper_bound <- round(custom_breaks[i + 1] - 0.001, 3)
            
            # Ensure lower bound is not equal to upper bound
            if (lower_bound == upper_bound) {
              legend_labels <- c(legend_labels, paste0("\u2264", format(lower_bound, nsmall = 3)))
            } else {
              legend_labels <- c(legend_labels, paste0(format(lower_bound, nsmall = 3), " - ", format(upper_bound, nsmall = 3)))
            }
          }
          
          # Append the last range correctly
          last_label_max <- min(max(custom_breaks), 1.000)
          legend_labels[length(legend_labels)] <- paste0(format(round(custom_breaks[length(custom_breaks) - 1], 3), nsmall = 3), " - ", format(last_label_max, nsmall = 3))
          
          return(legend_labels)
        }
        legend_labels <- generate_legend_labels(custom_breaks)
        
        
      }
      
      self_state_plot <- ggplot2::ggplot() +
        ggplot2::geom_point(data = cellID_coordinates, aes(x = x, y = y, color = prob1), size = 0.9) +
        ggplot2::geom_point(data = merged_df2,
                            aes(x = x1, y = y1, size = CircleSize),
                            shape = 1,
                            color = "blue") +
        ggplot2::geom_text(data = cellID_coordinates, aes(x = x, y = y, label = Cell.ID), vjust = -1, size = 3) +
        scale_color_gradient(low = "black", high = "black",
                             name = "Probability",
                             breaks = if(length(unique(merged_df2$Probability)) == 1) unique(merged_df2$Probability) else breaks[-length(breaks)],
                             labels = legend_labels) +
        scale_size(range = c(2, 10)) +
        labs(title = "State Transitions: Circle size based on Transition Probability",
             subtitle = "considering self state transitions",
             x = "x-coordinates",
             y = "y-coordinates") +
        guides(color = guide_legend(title = "Transition\nProbability",
                                    override.aes = list(shape = 21, size = legend_size, color = "blue")),
               fill = guide_legend(title = "Probability", override.aes = list(color = "blue", size = legend_size)),
               size = "none") +
        theme_minimal()
    }
    
    
    
    if(flow_map %in% c('without_self_state', 'All')) {
      
      # Initial transition matrix with removal of self-transitions
      trans_prob_matx <- transition_probability_df
      
      # Keep track of states that have only single transition AND it's to themselves
      single_transition_states <- trans_prob_matx %>%
        group_by(Current_State) %>%
        filter(n() == 1 & Current_State == Next_State) %>%
        ungroup()
      
      # Remove self-transitions for states with multiple transitions 
      trans_prob_matx <- trans_prob_matx %>%
        group_by(Current_State) %>%
        filter(!(n() == 1 & Current_State == Next_State)) %>%
        filter(Current_State != Next_State) %>%
        ungroup()
      
      # Calculate probabilities
      trans_prob_matx <- trans_prob_matx %>%
        group_by(Current_State) %>%
        mutate(
          Transition_Probability = round((Relative_Frequency / sum(Relative_Frequency)), 4),
          Cumulative_Probability = cumsum(Transition_Probability),
          Cumulative_Probability = if_else(row_number() == n(), 1, Cumulative_Probability)
        ) %>% 
        ungroup()
      
      # Combine with single self-transition states
      trans_prob_matx <- rbind(trans_prob_matx, single_transition_states)
      filtered_trans_prob <- trans_prob_matx %>%
        group_by(Current_State) %>%
        arrange(desc(Transition_Probability), Next_State) %>%
        slice_head(n = 1) %>%
        ungroup()
      
      third_df <- filtered_trans_prob %>% dplyr::select(c(Next_State, Transition_Probability))
      
      # Dataframe for second_highest state
      merged_df3 <- cbind(current_state_data, third_df)
      merged_df3 <- merged_df3 %>%
        left_join(dplyr::select(merged_df3, Cell.ID, x1, y1), by = c("Next_State" = "Cell.ID")) %>%
        dplyr::mutate(x2 = x1.y, y2 = y1.y) %>%
        dplyr::select(-x1.y, -y1.y)
      colnames(merged_df3) <- c("x1", "y1", "Cell.ID", "Next_State", "Probability", "x2", "y2")
      merged_df3$Probability <- round(merged_df3$Probability, digits = 1)
      
      
      # NEW CODE FOR PLOT LEGEND
      segment_length <- function(p) {
        case_when(
          p <= 0.3 ~ 0.2,
          p <= 0.6 ~ 0.4,
          p <= 0.8 ~ 0.7,
          TRUE ~ 0.9
        )
      }
      
      # Calculate segment lengths
      merged_df3$segment_len <- segment_length(merged_df3$Probability)
      
      # Calculate relative ranges to position legend consistently
      y_range <- range(merged_df3$y1)
      x_range <- range(merged_df3$x1)
      plot_width <- diff(x_range)
      plot_height <- diff(y_range)
      
      # Set fixed positions relative to plot dimensions
      y1_l <- max(y_range) - 0.1*plot_height  
      y2_l <- max(y_range) - 0.15*plot_height
      
      x1_l <- max(x_range) + 0.15*plot_width
      x2_l <- max(x_range) + 0.2*plot_width 
      x3_l <- max(x_range) + 0.35*plot_width
      
      y1_l <- mean(y_range)  # Center vertically
      x1_l <- max(x_range) + 0.05*plot_width  # Slightly right of plot
      
      # Dynamically scale arrow lengths based on plot dimensions 
      arrow_scale <- min(plot_width, plot_height) / 20  # Scale factor
      
      # Check actual probability ranges in data
      prob_ranges <- case_when(
        merged_df3$Probability <= 0.3 ~ "0 to 0.3",
        merged_df3$Probability > 0.3 & merged_df3$Probability <= 0.6 ~ "0.4 to 0.6",
        merged_df3$Probability > 0.6 & merged_df3$Probability <= 0.8 ~ "0.7 to 0.8", 
        merged_df3$Probability > 0.8 ~ "0.9 to 1"
      )
      
      # Define proper order for ranges (ascending by lower bound)
      range_order <- c("0 to 0.3", "0.4 to 0.6", "0.7 to 0.8", "0.9 to 1")
      # Get unique ranges and sort them according to the defined order (ascending)
      unique_ranges <- range_order[range_order %in% unique(prob_ranges)]
      
      create_legend <- function(x1_l, x2_l, x3_l, y1_l, y2_l) {
        labels <- unique_ranges  # Use actual ranges found in data
        arrow_lengths <- seq(1, length(labels)*2 , by=1) * arrow_scale
        spacing <- plot_height / 20
        
        legend_elements <- list(
          annotate("text", x = x1_l + 2*arrow_scale, y = y1_l + spacing, 
                   label = "Transition\nProbability", size = 4)
        )
        
        for(i in seq_along(labels)) {
          y_pos <- y1_l - i * spacing
          
          legend_elements[[length(legend_elements) + 1]] <- 
            annotate("segment", 
                     x = x1_l, 
                     xend = x1_l + arrow_lengths[i],
                     y = y_pos, 
                     yend = y_pos,
                     color = "blue",
                     arrow = arrow(length = unit(1.5, "mm"), type = "open"))
          
          legend_elements[[length(legend_elements) + 1]] <- 
            annotate("text", 
                     x = x1_l + arrow_lengths[i] + arrow_scale/2,
                     y = y_pos, 
                     label = labels[i], 
                     size = 3,
                     hjust = 0)
        }
        return(legend_elements)
      }
      # Generate legend
      annotate_list <- create_legend(
        x1_l = x1_l,
        x2_l = x2_l,
        x3_l = x3_l,
        y1_l = y1_l,
        y2_l = y2_l
      )
      
      # Create the plot
      arrow_flow_map <- ggplot(merged_df3, aes(x = x1, y = y1)) +
        geom_point(color = "black", size = 0.9) +
        geom_text(aes(label = Cell.ID), vjust = -1, size = 3) +
        geom_segment(aes(xend = x1 + (x2 - x1) * segment_len,
                         yend = y1 + (y2 - y1) * segment_len),
                     arrow = arrow(length = unit(merged_df3$segment_len * 3, "mm")),
                     color = "blue") +
        labs(title = "State Transitions: Arrow size based on Transition Probability",
             subtitle = "without considering self state transitions",
             x = "x-coordinates",
             y = "y-coordinates") +
        theme_minimal()+
        scale_x_continuous(limits = c(min(merged_df3$x1), max(merged_df3$x1) + plot_width/3))
      
      # Add legend annotations
      for(annotation in annotate_list) {
        arrow_flow_map <- arrow_flow_map + annotation
      }
    }
  }  
  
  if (!is.null(animation)) {
    if(animation %in% c('time_based', 'All')) {
      sampled_df <- df %>% dplyr::select(Cell.ID, Timestamp)
      anime_data <- merge(sampled_df, cellID_coordinates, by = "Cell.ID", all.x = TRUE) %>%
        dplyr::arrange(Timestamp) %>%
        dplyr::select(-prob1) %>%
        dplyr::group_by(grp = cumsum(Cell.ID != lag(Cell.ID, default = first(Cell.ID)))) %>% 
        dplyr::mutate(colour = 2 - row_number() %% 2) %>%
        dplyr::ungroup() %>%
        dplyr::select(-grp) 
      
      
      if (!inherits(df$Timestamp, c("POSIXct", "POSIXt", "numeric"))) {
        stop("Accepted Timestamp data types: POSIXct, POSIXt and numeric")
      }
      
      if (inherits(anime_data$Timestamp, c("POSIXct", "POSIXt"))) {
        # Calculate time differences in days
        time_diffs <- as.numeric(difftime(lead(anime_data$Timestamp), 
                                          anime_data$Timestamp, 
                                          units = "days"))
        
        # Determine frequency
        med_diff <- stats::median(time_diffs, na.rm = TRUE)
        if (med_diff >= 365) {
          freq_unit <- "years"
          scale_factor <- 1/365
        } else if (med_diff >= 28) {
          freq_unit <- "months"
          scale_factor <- 1/30.44
        } else if (med_diff >= 1) {
          freq_unit <- "days"
          scale_factor <- 1
        } else if (med_diff >= 1/24) {
          freq_unit <- "hours"
          scale_factor <- 24
        } else {
          freq_unit <- "mins"
          scale_factor <- 24*60
        }
        
        anime_data <- anime_data %>%
          arrange(Timestamp) %>%
          mutate(
            latency = as.numeric(difftime(lead(Timestamp), Timestamp, units = "days")) * scale_factor,
            freq_unit = freq_unit
          ) %>%
          ungroup()
        anime_data$latency <- round(anime_data$latency,0)
        
        
        format_string <- switch(freq_unit,
                                "years" = "%Y",
                                "months" = "%Y-%m",
                                "days" = "%Y-%m-%d",
                                "hours" = "%Y-%m-%d %H:00",
                                "mins" = "%Y-%m-%d %H:%M"
        )
        
        subtitle_text <- paste0(
          "\n\ntime(t): {format(frame_time, format_string)}\n",
          "Latency: {round(anime_data$latency[findInterval(as.numeric(frame_time), 
                                     as.numeric(anime_data$Timestamp))], 2)} ",
          freq_unit
        )
      } else if (is.numeric(anime_data$Timestamp)) {
        anime_data <- anime_data %>%
          arrange(Timestamp) %>%
          mutate(latency = lead(Timestamp) - Timestamp) %>%
          mutate(latency = formatC(latency, format = "f", digits = 5)) %>%
          ungroup()
        
        subtitle_text <- paste0(
          "\n\ntime(t): {round(frame_time, 3)} seconds\n",
          "Latency: {anime_data$latency[frame]} seconds"
        )
      }
      
      dot_anim <- ggplot2::ggplot(anime_data, aes(x = x, y = y)) +
        ggplot2::geom_point(data = cellID_coordinates, aes(x = x, y = y), color = "black", size = 1) +
        ggplot2::geom_text(data = cellID_coordinates, aes(x = x, y = y, label = Cell.ID), vjust = -1, size = 3) +
        ggplot2::geom_point(aes(x = x, y = y, color = ifelse(colour == 1, "Active state at t", " ")), alpha = 0.7, size = 5) +
        scale_color_manual(values = c("Active state at t" = "red", " " = "white")) +
        theme_minimal() +
        labs(x = "x-coordinates", 
             y = "y-coordinates", 
             color = "Time Transition",
             title = "Animation showing state transitions considering self state transitions",
             subtitle = subtitle_text) +
        theme(plot.subtitle = element_text(margin = margin(t = 20))) +
        gganimate::transition_time(Timestamp) +
        gganimate::shadow_wake(wake_length = 0.05, alpha = FALSE, wrap = FALSE)
      
      time_animation <- gganimate::animate(dot_anim, 
                                           fps = fps_time, 
                                           duration = time_duration, 
                                           renderer = gganimate::gifski_renderer())
    } 
    
    
    if(animation %in% c('state_based', 'All')) {
      ### Animation based on next state
      df_a <- df %>%group_by(Cell.ID) %>%dplyr::mutate(Frequency = with(rle(Cell.ID), rep(lengths, lengths)))
      state_data <- df_a %>%group_by(grp = cumsum(c(TRUE, diff(Cell.ID) != 0))) %>% slice(n())
      order <- unique(state_data$Cell.ID)
      anime_df <- merged_df1[order, ]
      anime_df$order_map <- 1:nrow(anime_df)
      anime_df$label <- rep("Successive states", nrow(anime_df))
      
      
      #NEWLY ADDED FOR ARROW HEAD SIZE
      x <- anime_df$x1
      y <- anime_df$y1
      xend <- anime_df$x1 + (anime_df$x2 - anime_df$x1) * 0.5
      yend <- anime_df$y1 + (anime_df$y2 - anime_df$y1) * 0.5
      segment_lengths <- sqrt((anime_df$x2 - anime_df$x1)^2 + (anime_df$y2 - anime_df$y1)^2)
      
      # Scale arrow head length based on segment length
      arrow_head_length <- median(case_when(
        segment_lengths <= 5 ~ 1,
        segment_lengths <= 10 ~ 2,
        segment_lengths <= 20 ~ 3,
        TRUE ~ 4
      ))
      
      arrow_anim <- ggplot2::ggplot(anime_df, aes(x = x1, y = y1)) +
        geom_segment(data = anime_df, mapping = aes(x = x1, y = y1, xend = x1 + (x2 - x1) * 0.5, yend = y1 + (y2 - y1) * 0.5, color = label),
                     arrow = arrow(length = unit(arrow_head_length, "mm")), show.legend = TRUE) +
        ggplot2::geom_point(data = cellID_coordinates, aes(x = x, y = y), size = 1, show.legend = FALSE) +
        geom_text(data = cellID_coordinates, aes(x = x, y = y, label = Cell.ID), vjust = -1, size = 3 ) +
        scale_color_manual(values = c("Successive states" = "blue")) + 
        labs(x = "x-coordinates", y = "y-coordinates", color = "State Transition") + theme_minimal()
      
      animation1 <- arrow_anim + gganimate::transition_states(order_map, wrap = FALSE) + gganimate::shadow_mark() +
        labs(title = "Animation showing state transitions",
             subtitle = "without considering self-state transitions")
      
      state_animation <- gganimate::animate(animation1, fps = fps_state, duration = state_duration, renderer = gganimate::gifski_renderer())
    } 
  }
  
  plots <- list()
  
  if (!is.null(flow_map)) {
    if(flow_map %in% c('self_state', 'All')) {
      plots$self_state <- self_state_plot
    }
    if(flow_map %in% c('without_self_state', 'All')) {
      plots$without_self_state <- arrow_flow_map
    }
  }
  
  if (!is.null(animation)) {
    if(animation %in% c('time_based', 'All')) {
      plots$time_based <- time_animation
    }
    if(animation %in% c('state_based', 'All')) {
      plots$state_based <- state_animation
    } 
  }
  
  return(plots)
}

#' getOptimalCentroids
#'
#' Get Optimal Centroids
#'
#' The raw data is first scaled and this scaled data is supplied as input to
#' the vector quantization algorithm. Vector quantization technique uses a
#' parameter called quantization error. This parameter acts as a threshold and
#' determines the number of levels in the hierarchy. It means that, if there
#' are 'n' number of levels in the hierarchy, then all the clusters formed till
#' this level will have quantization error equal or greater than the threshold
#' quantization error. The user can define the number of clusters in the first
#' level of hierarchy and then each cluster in first level is sub-divided into
#' the same number of clusters as there are in the first level. This process
#' continues and each group is divided into smaller clusters as long as the
#' threshold quantization error is met. The output of this technique will be
#' hierarchically arranged vector quantized data.
#'
#' @param x Data Frame. A dataframe of multivariate data. Each row corresponds to an
#' observation, and each column corresponds to a variable. Missing values are
#' not accepted.
#' @param n_cells Numeric. Indicating the number of nodes per hierarchy.
#' @param quant.err Numeric. The quantization error for the algorithm.
#' @param algorithm String. The type of algorithm used for quantization.
#' Available algorithms are Hartigan and Wong, "Lloyd", "Forgy", "MacQueen".
#' (default is "Hartigan-Wong")
#' @param seed Numeric. Random Seed.
#' @param distance_metric Character. The distance metric to calculate inter point distance. It can be 'L1_Norm" or "L2_Norm". L1_Norm is selected by default.
#' @param function_to_calculate_distance_metric Function. The function is to find 'L1_Norm" or "L2_Norm" distances. L1_Norm is selected by default.
#' @param function_to_calculate_error_metric Character. The error metric can be "mean" or "max". mean is selected by default
#' @param quant_method Character. The quant_method can be "kmeans" or "kmedoids". kmeans is selected by default
#' @return \item{values}{ List. A list showing observations assigned to a cluster.}
#' \item{maxQE}{ List. A list corresponding to maximum QE values for each cell. }
#' \item{meanQE}{ List. A list corresponding to mean QE values for each cell. }
#' \item{centers}{ List. A list of quantization error for all levels and nodes. } 
#' \item{nsize}{ List. A list corresponding to number of observations in respective groups. } 
#' @author Shubhra Prakash <shubhra.prakash@@mu-sigma.com>, Sangeet Moy Das <sangeet.das@@mu-sigma.com>
#' @include getCentroids_for_opti.R
#' @keywords internal

getOptimalCentroids <-
  function (x, 
            iter.max, 
            algorithm, 
            n_cells,
            seed = 100, 
            function_to_calculate_distance_metric, 
            function_to_calculate_error_metric=c("mean","max"), 
            quant.err,
            distance_metric = "L1_Norm",
            quant_method=c("kmeans","kmedoids")
  ){
    
    print(paste0("Data dimensions before clustering:", nrow(x), "and", ncol(x)))
    
    if(quant_method == "kmeans"){
      
      # Start with splitting data into three clusters
      nclust_iter <- 3
      outkinit <- list(centers = numeric(),maxQE = numeric(),meanQE = numeric(), values = logical() , nsize = numeric())
      quantok <- rep(T, n_cells)
      
      # In trainHVT before calling getOptimalCentroids

      # Check if 3 <= No of clusters AND No of cells in a cluster > 3 AND flag to check QE for all clusters
      while(nclust_iter <= n_cells & nrow(x) > nclust_iter & (sum(quantok,na.rm = TRUE) > 0)) {
        resplt <- list()
        #outkinit will have centroids and datapoints and size of the cluster
        set.seed(seed)
        kout<-stats::kmeans(x, nclust_iter, iter.max=10^5, algorithm=algorithm)
        result <- getCentroids_for_opti(x,kout, nclust_iter,function_to_calculate_distance_metric,function_to_calculate_error_metric)
        
        outkinit[[1]] <- result[[1]]
        outkinit[[2]] <- result[[2]]
        outkinit[[3]] <- result[[3]]
        #flag to check for quantization error
        resplt <- unlist(outkinit$centers) > quant.err
        quantok <- unlist(resplt)
        nclust_iter <- nclust_iter + 1      
      }
      clusts <- nclust_iter-1
      outkinit[[4]] <- c(1:clusts) %>% purrr::map(~x[kout$cluster==.x,])
      outkinit[[5]] <- as.list(kout$size)
      outkinit[["centroid_val"]] <- kout$centers
      
      dummy_iter = n_cells - nclust_iter + 1
      
      if(dummy_iter !=0){
        outkinit[["centers"]] <- c(outkinit[["centers"]],as.list(rep(NA,dummy_iter)))
        outkinit[["maxQE"]] <- c(outkinit[["maxQE"]],as.list(rep(NA,dummy_iter)))
        outkinit[["meanQE"]] <- c(outkinit[["meanQE"]],as.list(rep(NA,dummy_iter)))
        outkinit[["values"]] <- c(outkinit[["values"]],as.list(rep(NA,dummy_iter)))
        outkinit[["nsize"]] <- c(outkinit[["nsize"]],as.list(rep(0,dummy_iter)))
        temprow <- matrix(c(rep.int(NA,length(outkinit[["centroid_val"]][1,]))),nrow=length(nclust_iter:n_cells),ncol=length(outkinit[["centroid_val"]][1,]))
        outkinit[["centroid_val"]] <- rbind(outkinit[["centroid_val"]],temprow)
      }
      # return centroids, datapoints and size of each cluster
      return(outkinit)
    }else if(quant_method == "kmedoids"){
      # Start with splitting data into three clusters
      nclust_iter <- 3
      outkinit <- list(centers = numeric(),maxQE = numeric(),meanQE = numeric(), values = logical() , nsize = numeric())
      quantok <- rep(T, n_cells)
      if(distance_metric == "L1_Norm") {
        distance_metric = "manhattan"
      }else if(distance_metric == "L2_Norm") {
        distance_metric = "euclidean"
      }
      kout=list()
      while(nclust_iter <= n_cells & nrow(x) > nclust_iter & (sum(quantok,na.rm = TRUE) > 0)) {
        resplt <- list()
        #outkinit will have centroids and datapoints and size of the cluster
        set.seed(seed)
        
        kmedoids_model <-
          cluster::pam(
            x = cluster::daisy(x, metric = distance_metric),
            k = nclust_iter,
            diss = TRUE,
            keep.data = FALSE
          )
        kout$cluster=kmedoids_model[["clustering"]]
        result<- getCentroids_for_opti(x,kout, nclust_iter,function_to_calculate_distance_metric,function_to_calculate_error_metric)
        outkinit[[1]] <-result[[1]]
        outkinit[[2]]<- result[[2]]
        outkinit[[3]]<- result[[3]]
        
        
        #flag to check for quantization error
        resplt <- unlist(outkinit$centers) > quant.err
        quantok <- unlist(resplt)
        nclust_iter <- nclust_iter + 1      
      }
      kout$size=as.numeric(table(kmedoids_model[["clustering"]]))
      kout$centers=as.matrix(x[kmedoids_model[["medoids"]],])
      clusts <- nclust_iter-1
      outkinit[[4]] <- c(1:clusts) %>% purrr::map(~x[kout$cluster==.x,])
      outkinit[[5]] <- as.list(kout$size)
      outkinit[["centroid_val"]] <- kout$centers
      
      dummy_iter = n_cells - nclust_iter + 1
      
      if(dummy_iter !=0){
        outkinit[["centers"]] <- c(outkinit[["centers"]],as.list(rep(NA,dummy_iter)))
        outkinit[["maxQE"]] <- c(outkinit[["maxQE"]],as.list(rep(NA,dummy_iter)))
        outkinit[["meanQE"]] <- c(outkinit[["meanQE"]],as.list(rep(NA,dummy_iter)))
        outkinit[["values"]] <- c(outkinit[["values"]],as.list(rep(NA,dummy_iter)))
        outkinit[["nsize"]] <- c(outkinit[["nsize"]],as.list(rep(0,dummy_iter)))
        temprow <- matrix(c(rep.int(NA,length(outkinit[["centroid_val"]][1,]))),nrow=length(nclust_iter:n_cells),ncol=length(outkinit[["centroid_val"]][1,]))
        outkinit[["centroid_val"]] <- rbind(outkinit[["centroid_val"]],temprow)
        
      }
      # return centroids, datapoints and size of each cluster
      return(outkinit)
    }
    
  }

#' @importFrom utils globalVariables

utils::globalVariables(c(".","k","Segment.Level","Segment.Child","Quant.Error",
                         "noOfCellsBelowQuantizationError","noOfCells","index",
                         "Segment.Parent","cluster","child","value", "hvq_k","..count..",":=",
                         "min_dist", "perc", "Centroid_Type", "polygon_x", "polygon_y",
                         "quantError", "kout$cluster", 'kmedoids_model[["clustering"]]',
                         "cellID", "Mean_Absolute_Deviation", "Percentage_of_Anomalous_Points",
                         "hvt.results", "Index", "centroidRadius", "sumOriginal", "Quant.Error.x",
                         "n.x", "n.y", "Quant.Error.y", "Scored.Quant.Error", "anomalyFlag", "Row.No", "x_coord", "y_coord",
                         "bp.x", "bp.y", "x", "y", "density","Cell.ID","Cell.Number","MapA.Cell.ID","MapA.Segment.Child","MapA.Segment.Level"
                         ,"MapA.Segment.Parent","MapC.Cell.ID","MapC.Cell.Path","MapC.Segment.Child","MapC.Segment.Level","MapC.Segment.Parent",
                         "Row.Number","identified_outlier_cells", "Cell.ID.x", "Cell.ID.y", "Layer1.Cell.ID", "Layer2.Cell.ID", "Segment.Child.x",
                        "Segment.Child.y", "Segment.Level.x", "Segment.Level.y", "Segment.Parent.x", "Segment.Parent.y", "identified_Novelty_cells",
                        "mapC_cell_path", "results_df","hc","clusters","cluster_data","centroid_data",
                        "Current_State","Next_State","is_self_only","nearest_neighbor",
                        "k_numeric","nn_numeric","mae_numeric","k_sort","nn_sort", "id",
                        "Number of Cells","Number of nearest neighbors","mae","status",
                        "temp_cluster_data","hoverText","lev","nclust","nn","formatted_time",
                        "is_overall_best","is_overall_worst","hover_text", "label"))

#' @importFrom magrittr %>%
#' @import dplyr
#' @import tidyr
#' @keywords internal

getCentroids_for_opti <- function (x, kout, n_cells,function_to_calculate_distance_metric,function_to_calculate_error_metric){
  requireNamespace("dplyr")   
  requireNamespace("tidyr")   
  requireNamespace("magrittr")   
  calculate_error<- centl <- list()
  
  calculate_error <-
    x %>%
    group_by(kout$cluster) %>%
    tidyr::nest()
  
  centroid_data <- kout[["centers"]] %>% as.data.frame()
  calculate_error <- calculate_error %>% arrange(calculate_error$`kout$cluster`)
  
  cluster_distances <- purrr::map2(
    1:nrow(centroid_data),
    calculate_error$data,
    function(i, cluster_data) {
      centroid_row <- centroid_data[i, ]
      apply(cluster_data, 1, function(row) function_to_calculate_distance_metric(centroid_row, row))
    }
  )
  
  
  calculate_error_for_each_cluster <- unlist(lapply(cluster_distances, function_to_calculate_error_metric)) 
  maxQE <- unlist(lapply(cluster_distances, function(x) {   return(max(x))}))
  meanQE <-  unlist(lapply(cluster_distances, function(x) {   return(mean(x))}))
  centl <- calculate_error_for_each_cluster
  
  
  return(list(centl,maxQE,meanQE))
  
}

#' @name displayTable
#' @title function for displaying table
#' @description This is the main function for displaying data in table format
#' @param data Data frame. The dataframe to be displayed in table format.
#' @param scroll Logical. A value to have a scroll or not in the table. 
#' Default is TRUE. 
#' @param limit Numeric. A value to indicate how many rows to display.
#' Default is 20.
#' @return A table with proper formatting for html notebook
#' @author Vishwavani <vishwavani@@mu-sigma.com>
#' @importFrom dplyr mutate 
#' @keywords Table_Formatting
#' @examples
#' data <- datasets::EuStockMarkets
#' dataset <- as.data.frame(data)
#' displayTable(dataset)
#' @export displayTable

displayTable <- function(data, scroll = TRUE,limit= 20) {
  data <- data %>% 
    as.data.frame() %>%
    dplyr::mutate_if(is.numeric, round, 4)
  
  # Function to calculate scroll height based on the number of rows
  scrolLimit <- function(noOfRows) {
    if (noOfRows < 10) {
      swe <- paste(as.character(noOfRows * 50), "px")
    } else {
      swe <- "400px"
    }
    return(swe)
  }
  
  # Check if scrolling is needed (this variable is defined but not used)
  scroll <- nrow(data) > 10 || ncol(data) > 10
  
  # Limit the number of rows displayed
    limit <- if (nrow(data) > 20) limit else nrow(data)
    data <- head(data, limit)
  
    table_3 <- knitr::kable(data, "html", escape = FALSE, align = "c") %>%
      kableExtra::kable_styling(bootstrap_options = c("striped", "hover", "responsive"),font_size = 10.5) 
    if (scroll){
      table_3 <- table_3 %>% kableExtra::scroll_box(width = "100%", height = scrolLimit(nrow(data)))
    }
    return(table_3)
  
}

#' @name removeNovelty
#' @title Remove identified novelty cell(s) 
#' @description This function is used to remove the identified novelty cells.
#' @param outlier_cells Vector. A vector with the cell number of the identified novelty
#' @param hvt_results List. A list having the results of the compressed map i.e. output of \code{trainHVT} function
#' @return A list of two items
#' \item{[[1]] }{Dataframe of novelty cell(s)}
#' \item{[[2]] }{Dataframe without the novelty cell(s) from the dataset used in model training}
#' @author Shantanu Vaidya <shantanu.vaidya@@mu-sigma.com>
#' @seealso \code{\link{trainHVT}} \cr \code{\link{scoreLayeredHVT}}
#' @importFrom magrittr %>%
#' @importFrom plyr rbind.fill
#' @examples
#' data("EuStockMarkets")
#' hvt.results <- trainHVT(EuStockMarkets, n_cells = 60, depth = 1, quant.err = 0.1, 
#'                        distance_metric = "L1_Norm", error_metric = "max",
#'                        normalize = TRUE,quant_method="kmeans")
#' identified_Novelty_cells <<- c(2, 10)
#' output_list <- removeNovelty(identified_Novelty_cells, hvt.results) 
#' data_with_novelty <- output_list[[1]]
#' data_without_novelty <- output_list[[2]]                      
#' @keywords Novelty_or_Outliers
#' @export removeNovelty


removeNovelty <- function(outlier_cells, hvt_results) {
  remove_outlier_cells <- as.vector(outlier_cells)

  if (length(remove_outlier_cells) == 0) {
    message("No outlier cells were removed from the dataset.")
    hvt_results_cells <- hvt_results[[3]]$clusters
    scaled_data <- do.call(rbind, lapply(hvt_results_cells, function(x) {
      temp_df <- as.data.frame(x)
      temp_df$Row.Number <- row.names(x)
      temp_df
    }))
    row.names(scaled_data) <- scaled_data$Row.Number
    scaled_data <- select(scaled_data, -c(Row.Number))

    return(list(hvt_mapB = NA, dataset_without_outliers = scaled_data))
  } else {
    hvt_results_cells <- hvt_results[[3]]$clusters
    scaled_data <- do.call(rbind, lapply(seq_along(hvt_results_cells), function(i) {
      temp_df <- as.data.frame(hvt_results_cells[[i]])
      temp_df$Cell.Number <- i
      temp_df$Row.Number <- row.names(hvt_results_cells[[i]])
      temp_df
    }))
    row.names(scaled_data) <- scaled_data$Row.Number
    Cell.ID.mapping <- hvt_results[[3]]$summary[, c("Segment.Child", "Cell.ID")]
    scaled_data <- merge(x = scaled_data, y = Cell.ID.mapping, by.x = c("Cell.Number"), by.y = c("Segment.Child"))
    outlier_rows_df <- scaled_data %>% dplyr::filter(Cell.Number %in% identified_Novelty_cells)
    removed_outlier_rows <- as.vector(outlier_rows_df$Row.Number)
    outliers_data_scaled <- select(outlier_rows_df, -c(Row.Number))
    outliers_data_scaled <- outliers_data_scaled[, c(ncol(outliers_data_scaled), 1:(ncol(outliers_data_scaled) - 1))]
    `%notin%` <- Negate(`%in%`)
    dataset_without_outliers_scaled <- scaled_data %>% dplyr::filter(Cell.Number %notin% remove_outlier_cells)
    row.names(dataset_without_outliers_scaled) <- dataset_without_outliers_scaled$Row.Number
    dataset_without_outliers_scaled <- select(dataset_without_outliers_scaled, -c(Row.Number, Cell.Number, Cell.ID))
    dataset_without_outliers_scaled <- dataset_without_outliers_scaled[order(as.numeric(row.names(dataset_without_outliers_scaled))), ]
    message(paste0(c("The following cell(s) have been removed as novelties from the dataset: ", remove_outlier_cells), collapse = " "))

    return(list(hvt_mapB = outliers_data_scaled, dataset_without_outliers = dataset_without_outliers_scaled))
  }
}

#' @name mcmc_plots
#' @title Consolidated MCMC Plots Function
#' @description Creates comprehensive plots for HVT-MSM simulation results with minimal code duplication.
#' @param simulation_results Data frame with simulation results
#' @param centroid_data Data frame with centroid information
#' @param centroid_2d_points Data frame with 2D coordinates
#' @param actual_data Data frame with actual values
#' @param state_time_data Data frame with state-time information
#' @param forecast_type Type of forecast ("ex-post" or "ex-ante")
#' @param n_ahead_ante Number of ahead steps for ex-ante
#' @param type Plot type identifier
#' @param raw_dataset Raw dataset for scaling
#' @param show_simulation Whether to show simulation lines
#' @param mae_metric MAE metric to use
#' @param time_column Name of time column
#' @param trainHVT_results HVT training results
#' @return List containing all plots and data frames
#' @author Vishwavani <vishwavani@mu-sigma.com>
#' @keywords internal
#' @importFrom magrittr %>%
#' @importFrom stats na.omit

mcmc_plots <- function(simulation_results, centroid_data, centroid_2d_points, actual_data, 
                      state_time_data, forecast_type, n_ahead_ante, type, raw_dataset, 
                      show_simulation, mae_metric = mae_metric, time_column, trainHVT_results) {
  
  # Global variables for CRAN warnings
  time <- simulation <- median <- sd <- studentized_residuals <- value <- Cell.ID <- NULL

  # Standardize timestamp function
  standardize_timestamp <- function(data, time_col) {
    sample_time <- na.omit(data[[time_col]])[1]
    if(inherits(sample_time, "POSIXct")) return(data)
    
    formats <- c("%m/%d/%Y", "%m-%d-%Y", "%Y/%m/%d", "%Y-%m-%d", 
                 "%Y-%m-%d %H:%M:%S", "%d/%m/%Y")
    
    for(fmt in formats) {
      tryCatch({
        data[[time_col]] <- as.POSIXct(data[[time_col]], format = fmt)
        if(!all(is.na(data[[time_col]]))) break
      }, error = function(e) {})
    }
    return(data)
  }
  
  # Analyze timestamp characteristics
  analyze_timestamps <- function(data, time_col) {
    data <- standardize_timestamp(data, time_col)
    sorted_times <- sort(data[[time_col]])
    
    changes <- list(
      year = length(unique(format(sorted_times, "%Y"))),
      month = length(unique(format(sorted_times, "%Y-%m"))),
      day = length(unique(format(sorted_times, "%Y-%m-%d"))),
      hour = length(unique(format(sorted_times, "%Y-%m-%d %H"))),
      minute = length(unique(format(sorted_times, "%Y-%m-%d %H:%M"))),
      second = length(unique(format(sorted_times, "%Y-%m-%d %H:%M:%S")))
    )
    
    date_format <- if(changes$minute == changes$second) {
      if(changes$hour == changes$minute) {
        if(changes$day == changes$hour) {
          if(changes$month == changes$day) {
            if(changes$year == changes$month) "%Y" else "%Y-%m"
          } else "%Y-%m-%d"
        } else "%Y-%m-%d %H"
      } else "%Y-%m-%d %H:%M"
    } else "%Y-%m-%d %H:%M:%S"
    
    time_diffs <- diff(sorted_times)
    median_diff <- median(time_diffs)
    days_diff <- as.numeric(median_diff, units="days")
    
    interval_type <- dplyr::case_when(
      between(days_diff, 57, 62) ~ "bimonthly",
      between(days_diff, 27, 31) ~ "monthly",
      between(days_diff, 88, 92) ~ "quarterly",
      between(days_diff, 364, 366) ~ "yearly",
      between(days_diff, 6, 8) ~ "weekly",
      between(days_diff, 13, 15) ~ "biweekly",
      days_diff < 1 ~ "intraday",
      days_diff == 1 ~ "daily",
      TRUE ~ "irregular"
    )
    
    total_span <- as.numeric(difftime(max(sorted_times), min(sorted_times), units = "days"))
    
    return(list(
      interval_type = interval_type,
      date_format = date_format,
      total_span = total_span,
      standardized_data = data
    ))
  }
  
  # Get regular timestamp breaks
  get_regular_breaks <- function(data_time, interval_type) {
    data_time <- sort(data_time)
    first_timestamp <- data_time[1]
    max_timestamp <- max(data_time)
    
    if(interval_type == "bimonthly") {
      breaks <- seq(from = first_timestamp, to = max_timestamp, by = "2 months")
      if(length(breaks) > 10) {
        step <- ceiling(length(breaks) / 10)
        indices <- c(1, seq(from = 1 + step, to = length(breaks), by = step))
        breaks <- breaks[indices]
      }
      return(breaks)
    }
    
    interval <- switch(interval_type,
                       "daily" = "1 day", "weekly" = "1 week", "biweekly" = "2 weeks",
                       "monthly" = "1 month", "quarterly" = "3 months", "yearly" = "1 year",
                       "intraday" = "6 hours", "irregular" = "1 month")
    
    breaks <- seq(from = first_timestamp, to = max_timestamp, by = interval)
    
    if(length(breaks) > 20) {
      step <- ceiling(length(breaks) / 20)
      indices <- c(1, seq(from = 1 + step, to = length(breaks), by = step))
      breaks <- breaks[indices]
    }
    
    return(breaks)
  }
  
  # Get axis breaks
  get_axis_breaks <- function(time_analysis) {
    break_interval <- switch(time_analysis$interval_type,
                             "intraday" = if(grepl("%H:%M:%S", time_analysis$date_format)) "1 minute" 
                             else if(grepl("%H:%M", time_analysis$date_format)) "15 minutes" 
                             else "1 hour",
                             "daily" = "1 day", "weekly" = "1 week", "biweekly" = "2 weeks",
                             "monthly" = "1 month", "quarterly" = "3 months", "yearly" = "1 year",
                             "bimonthly" = "2 months",
                             {
                               months_span <- time_analysis$total_span / 30.44
                               if(months_span <= 1) "1 week"
                               else if(months_span <= 3) "2 weeks"
                               else if(months_span <= 6) "1 month"
                               else if(months_span <= 24) "1 month"
                               else if(months_span <= 60) "2 months"
                               else "1 year"
                             })
    
    return(list(break_interval = break_interval, date_format = time_analysis$date_format))
  }
  
  # Common theme
  theme_plot <- ggplot2::theme(
    plot.title = ggplot2::element_text(size = 16, face = "bold"),
    plot.subtitle = ggplot2::element_text(size = 12),
    axis.title = ggplot2::element_text(size = 12),
    axis.text = ggplot2::element_text(size = 10),
    legend.title = ggplot2::element_text(size = 14),
    legend.text = ggplot2::element_text(size = 12),
    legend.position = "right",
    legend.key.width = grid::unit(0.5, "cm")
  )
  
  # Integer breaks function
  integer_breaks <- function(n = 5, ...) {
    function(x) {
      breaks <- floor(pretty(x, n, ...))
      names(breaks) <- attr(breaks, "labels")
      breaks
    }
  }
  
  # Format date to always show 1st of month for tooltips
  format_date_first <- function(date_val) {
    if(inherits(date_val, "POSIXct") || inherits(date_val, "Date")) {
      date_str <- format(date_val, "%Y-%m-%d")
      # Extract year-month and set day to 01
      year_month <- substr(date_str, 1, 7)
      return(paste0(year_month, "-01"))
    }
    return(as.character(date_val))
  }
  
  # Create base plot with common elements
  create_base_plot <- function(plot_data, summary_data, show_simulation, mae_metric, 
                              variable_name = NULL, is_states = FALSE) {
    
    # Color scheme
    color_values <- c("Simulations" = "darkgray", "Median" = "red", "Mean" = "darkgreen", 
                     "Mode" = "#0901FF")
    if(!is_states) color_values["Actual"] <- "black"
    
    plot <- ggplot2::ggplot() +
      # Simulation lines
      {if(show_simulation) ggplot2::geom_line(data = plot_data,
                                             ggplot2::aes(x = time, y = value, group = simulation, 
                                                         colour = "Simulations",
                                                         text = paste("Time:", format_date_first(time), "<br>Value:", value, "<br>Simulation:", simulation)), 
                                             alpha = 0.4, size = 0.4)} +
      # Mode line and points
      ggplot2::geom_line(data = summary_data, ggplot2::aes(x = time, y = mode, colour = "Mode"), 
                        size = ifelse(mae_metric == "mode", 1.0, 0.4)) +
      ggplot2::geom_point(data = summary_data, ggplot2::aes(x = time, y = mode, colour = "Mode",
                                                          text = paste("Time:", format_date_first(time), "<br>Mode", 
                                                                      ifelse(is.null(variable_name), "", paste(variable_name, ":")), mode)), 
                         size = ifelse(mae_metric == "mode", 1.5, 1.0)) +
      # Mean line and points
      ggplot2::geom_line(data = summary_data, ggplot2::aes(x = time, y = mean, colour = "Mean"), 
                        size = ifelse(mae_metric == "mean", 1.0, 0.4)) +
      ggplot2::geom_point(data = summary_data, ggplot2::aes(x = time, y = mean, colour = "Mean",
                                                           text = paste("Time:", format_date_first(time), "<br>Mean", 
                                                                       ifelse(is.null(variable_name), "", paste(variable_name, ":")), mean)), 
                         size = ifelse(mae_metric == "mean", 1.5, 1.0)) +
      # Median line and points
      ggplot2::geom_line(data = summary_data, ggplot2::aes(x = time, y = median, colour = "Median"), 
                        size = ifelse(mae_metric == "median", 1.0, 0.4)) +
      ggplot2::geom_point(data = summary_data, ggplot2::aes(x = time, y = median, colour = "Median",
                                                           text = paste("Time:", format_date_first(time), "<br>Median", 
                                                                       ifelse(is.null(variable_name), "", paste(variable_name, ":")), median)), 
                         size = ifelse(mae_metric == "median", 1.5, 1.0)) +
      ggplot2::scale_colour_manual(values = color_values) +
      ggplot2::theme_minimal() +
      theme_plot
    
    return(plot)
  }
  
  # Add actual data layer
  add_actual_layer <- function(plot, actual_data, variable_name, time_col = "time") {
    actual_col_name <- paste0("actual_", variable_name)
    # Create formatted time column for tooltips
    actual_data$formatted_time <- sapply(actual_data[[time_col]], format_date_first)
    plot + 
      ggplot2::geom_line(data = actual_data, 
                        ggplot2::aes(x = !!rlang::sym(time_col), y = !!rlang::sym(actual_col_name), 
                                    colour = "Actual"), size = 1.0) +
      ggplot2::geom_point(data = actual_data, 
                         ggplot2::aes(x = !!rlang::sym(time_col), y = !!rlang::sym(actual_col_name), 
                                     colour = "Actual",
                                     text = paste("Time:", formatted_time, "<br>Actual", variable_name, ":", 
                                                 round(!!rlang::sym(actual_col_name), 4))), 
                         size = 1.5)
  }
  
  # Create residual plot
  create_residual_plot <- function(residuals_df, variable_name = NULL) {
    if(all(residuals_df$residuals == 0)) {
      residuals_df$studentized_residuals <- 0
    }
    
    # Determine the correct time column name
    time_col <- if("time" %in% names(residuals_df)) "time" else "t"
    # Create formatted time column for tooltips
    residuals_df$formatted_time <- sapply(residuals_df[[time_col]], format_date_first)
    
    ggplot2::ggplot() +
      ggplot2::geom_line(data = residuals_df, 
                        ggplot2::aes(x = !!rlang::sym(time_col), y = studentized_residuals, color = "Studentized\nResiduals"), 
                        size = 0.8) +
      ggplot2::geom_point(data = residuals_df, 
                         ggplot2::aes(x = !!rlang::sym(time_col), y = studentized_residuals, color = "Studentized\nResiduals",
                                     text = paste("Time:", formatted_time, "<br>Residuals", 
                                                 ifelse(is.null(variable_name), "", paste(variable_name, ":")), 
                                                 round(studentized_residuals, 4))), 
                         size = 1.0) +
      ggplot2::geom_hline(yintercept = 0, col = "black", linetype = "dashed") +
      ggplot2::geom_hline(yintercept = -1, col = "blue", linetype = "dashed") +
      ggplot2::geom_hline(yintercept = 1, col = "blue", linetype = "dashed") +
      ggplot2::scale_color_manual(name = NULL, values = c("Studentized\nResiduals" = "black")) +
      ggplot2::theme_minimal() +
      ggplot2::theme(legend.position = "right", legend.key.width = grid::unit(0.5, "cm")) +
      theme_plot
  }
  
  # Apply time scaling to plot
  apply_time_scaling <- function(plot, summary_data) {
    if(inherits(summary_data$time, "POSIXct")) {
      time_analysis <- analyze_timestamps(summary_data, "time")
      axis_settings <- get_axis_breaks(time_analysis)
      all_breaks <- get_regular_breaks(summary_data$time, time_analysis$interval_type)
      
      plot + 
        ggplot2::scale_x_datetime(
          breaks = all_breaks,
          minor_breaks = summary_data$time,
          date_labels = axis_settings$date_format,
          expand = ggplot2::expansion(mult = 0.02)
        ) +
        ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
    } else {
      plot + 
        ggplot2::scale_x_continuous(expand = ggplot2::expansion(mult = 0.02)) +
        ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1))
    }
  }
  
  # Apply plotly layout
  apply_plotly_layout <- function(plot, plot_type, variable_name, height = 400) {
    plotly::ggplotly(plot, tooltip = "text") %>%
      plotly::layout(
        margin = list(r = 150, b = 50, t = 50),
        height = height,
        yaxis = list(title = list(text = ifelse(is.null(variable_name), "States", 
                                              paste0(variable_name, " (transformed units)")))),
        xaxis = list(title = list(text = "Timestamps")),
        legend = list(title = list(text = ""))
      )
  }
  
  # Create HTML layout
  create_html_layout <- function(plot1, plot2, type, variable_name, mae_metric, mae) {
    htmltools::tagList(
      htmltools::div(
        style = "display: grid; grid-template-columns: 1fr; max-width: 100%; font-family: Arial, sans-serif; margin-bottom: 0;",
        htmltools::div(
          style = "grid-column: 1; width: 100%; text-align: left; font-weight: bold; font-size: 18px; padding-left: 60px; font-family: Arial, sans-serif;",
          paste0(ifelse(is.null(type), "DefaultType", type), 
                 ": Ex-post Actual ", 
                 ifelse(is.null(variable_name), "Variable", variable_name), 
                 " vs Predicted ", 
                 ifelse(is.null(variable_name), "Variable", variable_name))
        ),
        htmltools::div(style = "grid-column: 1; width: 100%;", plot1),
        htmltools::div(
          style = "grid-column: 1; width: 100%; text-align: left; font-weight: bold; font-size: 18px; padding-left: 60px; font-family: Arial, sans-serif;",
          paste0(ifelse(is.null(type), "DefaultType", type), 
                 ": Ex-post Studentized Residuals for the ", 
                 ifelse(is.null(mae_metric), "DefaultMetric", mae_metric), " forecast")
        ),
        htmltools::div(
          style = "grid-column: 1; width: 100%; text-align: left; font-size: 14px; padding-left: 60px; font-style: italic;",
          paste("MAE:", ifelse(is.null(mae), "N/A", mae))
        ),
        htmltools::div(style = "grid-column: 1; width: 100%; margin-bottom: 0;", plot2)
      )
    )
  }
  
  # Get column names and prepare centroid data
  name_columns <- colnames(centroid_data)
  centroid_data <- centroid_data %>% dplyr::mutate(dplyr::across(where(is.numeric), ~ round(., 4)))
  
  # Calculate mean and sd of raw dataset
  raw_dataset_wo_time <- raw_dataset %>% dplyr::select(-all_of(time_column))
  mean_raw <- raw_dataset_wo_time %>% dplyr::summarise(dplyr::across(dplyr::everything(), ~ round(mean(.), 4)))
  sd_raw <- raw_dataset_wo_time %>% dplyr::summarise(dplyr::across(dplyr::everything(), ~ round(stats::sd(.), 4)))
  
  # Join dataframes
  centroid_dataframe <- cbind(centroid_data, Cell.ID = centroid_2d_points$Cell.ID)
  
  # Generate predicted dataframes
  generate_predicted_df <- function(coord) {
    centroid_map <- centroid_dataframe %>% dplyr::select(Cell.ID, all_of(coord))
    
    replace_with_value <- function(column) {
      centroid_map[[coord]][match(column, centroid_map$Cell.ID)]
    }
    
    sim_df <- simulation_results[,-1] %>%
      dplyr::mutate_all(replace_with_value) %>%
      dplyr::rename_with(~ paste0(., "_", coord), starts_with("Sim_"))
    
    simulation_results %>%
      dplyr::select(time) %>%
      dplyr::bind_cols(sim_df)
  }
  
  predicted_dfs <- lapply(name_columns, generate_predicted_df)
  names(predicted_dfs) <- name_columns
  
  # Scale predicted centroids
  if(trainHVT_results[["model_info"]][["input_parameters"]][["normalize"]]){
    scaled_dfs <- lapply(names(predicted_dfs), function(name) {
      scale_df <- predicted_dfs[[name]] %>%
        dplyr::select(-time) %>%
        dplyr::mutate(dplyr::across(dplyr::everything(), ~ (. * sd_raw[1, name]) + mean_raw[1, name])) 
      
      scale_df <- round(scale_df, 4)
      predicted_df <- cbind(time = predicted_dfs[[name]]$time, scale_df)
      return(predicted_df)
    })
  } else {
    scaled_dfs <- predicted_dfs
  }
  names(scaled_dfs) <- names(predicted_dfs)
  
  if(is.numeric(state_time_data[[time_column]])){  
    if (forecast_type == "ex-post") {
      
      test_dataset <- actual_data
      test_data <- state_time_data[state_time_data[[time_column]] %in% simulation_results$time, ]
      
      # Prepare actual data
      if (!is.null(test_dataset)) {
        actual_raw_dfs <- lapply(name_columns, function(col_name) {
          df <- test_dataset[, c(time_column, col_name)]
          names(df) <- c("time", paste0("actual_", col_name))
          return(df)
        })
        names(actual_raw_dfs) <- name_columns
    
      # Generate all plots for actual vs predicted and residuals
      all_plots <- lapply(name_columns, function(variable_name) {
        
        # Prepare data
        plot_data <- scaled_dfs[[variable_name]] %>%
          dplyr::select(time, starts_with("Sim_")) %>%
          tidyr::pivot_longer(cols = starts_with("Sim_"), names_to = "simulation", values_to = "value")
        
        summary_data <- scaled_dfs[[variable_name]] %>%
          dplyr::select(time, mean, median, mode)
        
        predicted_metric <- summary_data[[mae_metric]]
        actual_col_name <- paste0("actual_", variable_name)
        
        # Calculate residuals
        residuals_df <- data.frame(
          t = test_dataset[[time_column]][seq_along(predicted_metric)],
          x = actual_raw_dfs[[variable_name]][[actual_col_name]][seq_along(predicted_metric)],
          predicted_metric
        )
        residuals_df$residuals <- residuals_df$x - residuals_df$predicted_metric
        residuals_sd <- stats::sd(residuals_df$residuals)
        residuals_df$studentized_residuals <- residuals_df$residuals / residuals_sd
        residuals_df$mape_component <- abs((residuals_df$x - residuals_df$predicted_metric) / residuals_df$x) * 100
        options(scipen = 999)
        # mape <- round(mean(residuals_df$mape_component), 4)  # Commented out as not used
        mae <- round(mean(abs(residuals_df$residuals)), 4)
        
        # Create plots
        p1 <- create_base_plot(plot_data, summary_data, show_simulation, mae_metric, variable_name)
        p1 <- add_actual_layer(p1, actual_raw_dfs[[variable_name]], variable_name)
        p1 <- apply_time_scaling(p1, summary_data)
        
        p2 <- create_residual_plot(residuals_df, variable_name)
        p2 <- apply_time_scaling(p2, summary_data)
        
        # Convert to plotly
        p1_plotly <- apply_plotly_layout(p1, type, variable_name, 400)
        p2_plotly <- apply_plotly_layout(p2, type, NULL, 250)
        
        # Create combined layout
        x_plots <- create_html_layout(p1_plotly, p2_plotly, type, variable_name, mae_metric, mae)
        
        list(centroids_plot = x_plots, mae = mae)
      })
      } else {
        actual_raw_dfs <- NULL
        all_plots <- NULL
      }
    
      
      # States plot
      plot_data <- simulation_results %>%
        dplyr::select(time, starts_with("Sim_")) %>%
        tidyr::pivot_longer(cols = starts_with("Sim_"), names_to = "simulation", values_to = "value")
      
      summary_data <- simulation_results %>%
        dplyr::select(time, mean, median, mode)
      
      # Format dates for tooltips
      test_data$formatted_time <- sapply(test_data$t, format_date_first)
      
      pa <- create_base_plot(plot_data, summary_data, show_simulation, mae_metric, NULL, TRUE) +
        ggplot2::geom_line(data = test_data, ggplot2::aes(x = t, y = Cell.ID, color = "Actual States"), size = 1) +
        ggplot2::geom_point(data = test_data, ggplot2::aes(x = t, y = Cell.ID, color = "Actual States", 
                                                          text = paste("Time:", formatted_time, "<br>Actual States:", Cell.ID)), 
                           size = 1.5) +
        ggplot2::scale_colour_manual(values = c("Simulations" = "darkgray", "Median" = "red",
                                               "Mean" = "darkgreen", "Actual States" = "black", "Mode" = "#0901FF")) +
        ggplot2::scale_y_continuous(limits = c(1, NA), breaks = integer_breaks()) +
        ggplot2::theme_minimal() +
        ggplot2::theme(legend.position = "right")
      
      # States residual plot
      predicted_metric <- summary_data[[mae_metric]]
      residuals_df <- data.frame(test_data[seq_along(predicted_metric),], predicted_metric)
      residuals_df$residuals <- (residuals_df$Cell.ID - residuals_df$predicted_metric)
      residuals_sd <- stats::sd(residuals_df$residuals)
      residuals_df$studentized_residuals <- residuals_df$residuals / residuals_sd
      residuals_df$mape_component <- abs((residuals_df$Cell.ID - residuals_df$predicted_metric) / residuals_df$Cell.ID) * 100
      options(scipen = 999)
      mape <- round(mean(residuals_df$mape_component), 4)
      mae <- round(mean(abs(residuals_df$residuals)), 4)
      
      pb <- create_residual_plot(residuals_df, NULL)
      
      # Convert to plotly
      pa_plotly <- apply_plotly_layout(pa, type, NULL, 400)
      pb_plotly <- apply_plotly_layout(pb, type, NULL, 250)
      
      # Create states layout
      x_plots <- create_html_layout(pa_plotly, pb_plotly, type, "States", mae_metric, mae)
      states_plots <- list(x_plots, mae = mae)
      
      return(list(all_plots, states_plots))
      
    } else {
      # Ex-ante forecasts for numeric time
      all_plots <- lapply(name_columns, function(variable_name) {
        plot_data <- scaled_dfs[[variable_name]] %>%
          dplyr::select(time, starts_with("Sim_")) %>%
          tidyr::pivot_longer(cols = starts_with("Sim_"), names_to = "simulation", values_to = "value")
        
        summary_data <- scaled_dfs[[variable_name]] %>%
          dplyr::select(time, mean, median, mode)
        
        p1 <- create_base_plot(plot_data, summary_data, show_simulation, mae_metric, variable_name)
        p1 <- apply_time_scaling(p1, summary_data)
        p1 <- p1 + ggplot2::labs(x = "Timestamps", y = paste0(variable_name, " (transformed units)"),
                                 title = paste0(type, ": Ex-ante Predicted ", variable_name), color = " ")
        
        plotly::ggplotly(p1, tooltip = "text")
      })
      
      # States plot for ex-ante
      plot_data <- simulation_results %>%
        dplyr::select(time, starts_with("Sim_")) %>%
        tidyr::pivot_longer(cols = starts_with("Sim_"), names_to = "simulation", values_to = "value")
      summary_data <- simulation_results %>%
        dplyr::select(time, mean, median, mode)
      
      p2 <- create_base_plot(plot_data, summary_data, show_simulation, mae_metric, NULL, TRUE) +
        ggplot2::scale_y_continuous(limits = c(1, NA), breaks = integer_breaks()) +
        ggplot2::theme_minimal() +
        ggplot2::theme(legend.position = "right") +
        ggplot2::labs(x = "Timestamps", y = "States", title = paste0(type, ": Ex-ante Predicted States"), color = " ")
      
      p2 <- apply_time_scaling(p2, summary_data)
      p2 <- plotly::ggplotly(p2, tooltip = "text")
      
      # Helper functions for data frames
      create_variable_dfs <- function(scaled_dfs, name_columns) {
        variable_dfs <- lapply(name_columns, function(variable_name) {
          summary_data <- scaled_dfs[[variable_name]] %>%
            dplyr::select(time, mean, median, mode)
          return(summary_data)
        })
        names(variable_dfs) <- name_columns
        return(variable_dfs)
      }
      
      create_state_df <- function(simulation_results) {
        state_df <- simulation_results %>%
          dplyr::select(time, mean, median, mode)
        return(state_df)
      }
      
      return(list(
        centroids_plot = all_plots, 
        states_plots = p2, 
        variable_dfs = create_variable_dfs(scaled_dfs, name_columns), 
        state_df = create_state_df(simulation_results)
      ))
    }
  }
  
  if(!is.numeric(state_time_data[[time_column]])){
    if (forecast_type == "ex-post") {
      
      test_dataset <- actual_data
      test_data <- state_time_data[state_time_data[[time_column]] %in% simulation_results$time, ]
      colnames(test_data)[colnames(test_data) == time_column] <- 'time'
      
      actual_raw_dfs <- lapply(name_columns, function(col_name) {
        test_dataset %>%
          dplyr::select(all_of(time_column), all_of(col_name)) %>%
          dplyr::rename(time = !!time_column, !!paste0("actual_", col_name) := all_of(col_name))
      })
      names(actual_raw_dfs) <- name_columns
      
      # Generate all plots for actual vs predicted and residuals
      all_plots <- lapply(name_columns, function(variable_name) {
        
        plot_data <- scaled_dfs[[variable_name]] %>%
          dplyr::select(time, starts_with("Sim_")) %>%
          tidyr::pivot_longer(cols = starts_with("Sim_"), names_to = "simulation", values_to = "value")
        
        summary_data <- scaled_dfs[[variable_name]] %>%
          dplyr::select(time, mean, median, mode)
        
        predicted_metric <- summary_data[[mae_metric]]
        actual_col_name <- paste0("actual_", variable_name)
        
        residuals_df <- data.frame(
          t = test_dataset[[time_column]][seq_along(predicted_metric)],
          x = actual_raw_dfs[[variable_name]][[actual_col_name]][seq_along(predicted_metric)],
          predicted_metric
        )
        residuals_df$residuals <- residuals_df$x - residuals_df$predicted_metric
        residuals_sd <- stats::sd(residuals_df$residuals)
        residuals_df$studentized_residuals <- residuals_df$residuals / residuals_sd
        residuals_df$mape_component <- abs((residuals_df$x - residuals_df$predicted_metric) / residuals_df$x) * 100
        options(scipen = 999)
        # mape <- round(mean(residuals_df$mape_component), 4)  # Commented out as not used
        mae <- round(mean(abs(residuals_df$residuals)), 4)
        
        # Create plots
        p1 <- create_base_plot(plot_data, summary_data, show_simulation, mae_metric, variable_name)
        p1 <- add_actual_layer(p1, actual_raw_dfs[[variable_name]], variable_name)
        p1 <- apply_time_scaling(p1, summary_data)
        
        p2 <- create_residual_plot(residuals_df, variable_name)
        p2 <- apply_time_scaling(p2, summary_data)
        
        # Convert to plotly
        p1_plotly <- apply_plotly_layout(p1, type, variable_name, 400)
        p2_plotly <- apply_plotly_layout(p2, type, NULL, 250)
        
        # Create combined layout
        x_plots <- create_html_layout(p1_plotly, p2_plotly, type, variable_name, mae_metric, mae)
        
        list(centroids_plot = x_plots, mae = mae)
      })
      
      # States plot
      plot_data <- simulation_results %>%
        dplyr::select(time, starts_with("Sim_")) %>%
        tidyr::pivot_longer(cols = starts_with("Sim_"), names_to = "simulation", values_to = "value")
      
      summary_data <- simulation_results %>%
        dplyr::select(time, mean, median, mode)
      
      # Analyze timestamps
      time_analysis <- analyze_timestamps(summary_data, "time")
      summary_data <- time_analysis$standardized_data
      if(!is.null(plot_data)) {
        plot_data <- standardize_timestamp(plot_data, "time")
      }
      
      # Format dates for tooltips
      test_data$formatted_time <- sapply(test_data$time, format_date_first)
      
      pa <- create_base_plot(plot_data, summary_data, show_simulation, mae_metric, NULL, TRUE) +
        ggplot2::geom_line(data = test_data, ggplot2::aes(x = time, y = Cell.ID, color = "Actual States"), size = 1.0) +
        ggplot2::geom_point(data = test_data, ggplot2::aes(x = time, y = Cell.ID, color = "Actual States", 
                                                          text = paste("Time:", formatted_time, "<br>Actual States:", Cell.ID)), 
                           size = 1.5, show.legend = TRUE) +
        ggplot2::scale_colour_manual(values = c("Simulations" = "darkgray", "Median" = "red",
                                               "Mean" = "darkgreen", "Actual States" = "black", "Mode" = "#0901FF")) +
        ggplot2::scale_y_continuous(limits = c(1, NA), breaks = integer_breaks()) +
        ggplot2::theme_minimal() +
        ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1)) +
        theme_plot
      
      pa <- apply_time_scaling(pa, summary_data)
      
      # States residual plot
      predicted_metric <- summary_data[[mae_metric]]
      residuals_df <- data.frame(test_data[seq_along(predicted_metric),], predicted_metric)
      residuals_df$residuals <- (residuals_df$Cell.ID - residuals_df$predicted_metric)
      residuals_sd <- stats::sd(residuals_df$residuals)
      residuals_df$studentized_residuals <- residuals_df$residuals / residuals_sd
      residuals_df$mape_component <- abs((residuals_df$Cell.ID - residuals_df$predicted_metric) / residuals_df$Cell.ID) * 100
      options(scipen = 999)
      mape <- round(mean(residuals_df$mape_component), 4)
      mae <- round(mean(abs(residuals_df$residuals)), 4)
      
      pb <- create_residual_plot(residuals_df, NULL)
      pb <- apply_time_scaling(pb, summary_data)
      
      # Convert to plotly
      pa_plotly <- apply_plotly_layout(pa, type, NULL, 400)
      pb_plotly <- apply_plotly_layout(pb, type, NULL, 250)
      
      # Create states layout
      x_plots <- create_html_layout(pa_plotly, pb_plotly, type, "States", mae_metric, mae)
      states_plots <- list(x_plots, mae = mae)
      
      return(list(all_plots, states_plots))
      
    } else {
      # Ex-ante forecasts for non-numeric time
      all_plots <- lapply(name_columns, function(variable_name) {
        plot_data <- scaled_dfs[[variable_name]] %>%
          dplyr::select(time, starts_with("Sim_")) %>%
          tidyr::pivot_longer(cols = starts_with("Sim_"), names_to = "simulation", values_to = "value")
        
        summary_data <- scaled_dfs[[variable_name]] %>%
          dplyr::select(time, mean, median, mode)
        
        # Analyze timestamps
        time_analysis <- analyze_timestamps(summary_data, "time")
        summary_data <- time_analysis$standardized_data
        if(!is.null(plot_data)) {
          plot_data <- standardize_timestamp(plot_data, "time")
        }
        
        p1 <- create_base_plot(plot_data, summary_data, show_simulation, mae_metric, variable_name)
        p1 <- apply_time_scaling(p1, summary_data)
        p1 <- p1 + ggplot2::labs(x = "Timestamps", y = paste0(variable_name, " (transformed units)"),
                                 title = paste0(type, ": Ex-ante Predicted ", variable_name), color = " ")
        
        plotly::ggplotly(p1, tooltip = "text")
      })
      
      # States plot for ex-ante
      plot_data <- simulation_results %>%
        dplyr::select(time, starts_with("Sim_")) %>%
        tidyr::pivot_longer(cols = starts_with("Sim_"), names_to = "simulation", values_to = "value")
      summary_data <- simulation_results %>%
        dplyr::select(time, mean, median, mode)
      
      # Analyze timestamps
      time_analysis <- analyze_timestamps(summary_data, "time")
      summary_data <- time_analysis$standardized_data
      if(!is.null(plot_data)) {
        plot_data <- standardize_timestamp(plot_data, "time")
      }
      
      p2 <- create_base_plot(plot_data, summary_data, show_simulation, mae_metric, NULL, TRUE) +
        ggplot2::scale_y_continuous(limits = c(1, NA), breaks = integer_breaks()) +
        ggplot2::theme_minimal() +
        ggplot2::theme(axis.text.x = ggplot2::element_text(angle = 45, hjust = 1)) +
        ggplot2::labs(x = "Timestamps", y = "States", title = paste0(type, ": Ex-ante Predicted States"), color = " ") +
        theme_plot
      
      p2 <- apply_time_scaling(p2, summary_data)
      
      p2 <- plotly::ggplotly(p2, tooltip = "text")
      
      # Helper functions for data frames
      create_variable_dfs <- function(scaled_dfs, name_columns) {
        variable_dfs <- lapply(name_columns, function(variable_name) {
          summary_data <- scaled_dfs[[variable_name]] %>%
            dplyr::select(time, mean, median, mode)
          return(summary_data)
        })
        names(variable_dfs) <- name_columns
        return(variable_dfs)
      }
      
      create_state_df <- function(simulation_results) {
        state_df <- simulation_results %>%
          dplyr::select(time, mean, median, mode)
        return(state_df)
      }
      
      return(list(
        centroids_plot = all_plots, 
        states_plots = p2, 
        variable_dfs = create_variable_dfs(scaled_dfs, name_columns), 
        state_df = create_state_df(simulation_results)
      ))
    }
  }
}

#' @name scoreLayeredHVT
#' @title Score which cell and what layer each data point in the test dataset belongs to
#' @description
#' This function that scores the cell and corresponding layer for each data point in a test dataset using three 
#' hierarchical vector quantization (HVT) models (Map A, Map B, Map C) and returns a data frame containing the scored layer output. 
#' The function incorporates the scored results from each map and merges them to provide a comprehensive result.
#' @param data Data Frame. A data frame containing test dataset. 
#' The data frame should have all the variable(features) used for training. 
#' @param hvt_mapA A list of hvt.results.model obtained from trainHVT function while performing
#'  `trainHVT()` on train data
#' @param hvt_mapB A list of hvt.results.model obtained from trainHVT function while performing 
#' `trainHVT()` on data with novelty(s)
#' @param hvt_mapC A list of hvt.results.model obtained from trainHVT function while performing
#'  `trainHVT()` on data without novelty(s)
#' @param child.level Numeric. A number indicating the level for which the heat map is to be plotted.
#' @param mad.threshold Numeric. A number indicating the permissible Mean Absolute Deviation
#' @param normalize Logical. A logical value indicating if the dataset should be normalized. 
#' When set to TRUE, the data (testing dataset) is standardized by 'mean' and 'sd' of the training dataset 
#' referred from the trainHVT(). When set to FALSE, the data is used as such without any changes.
#' (Default value is TRUE).
#' @param seed Numeric. Random Seed.
#' @param distance_metric Character. The distance metric can be L1_Norm(Manhattan) or L2_Norm(Eucledian). L1_Norm is selected by default.
#' The distance metric is used to calculate the distance between an n dimensional point and centroid.
#' The distance metric can be different from the one used during training.
#' @param error_metric Character. The error metric can be mean or max. max is selected by default. 
#' max will return the max of m values and mean will take mean of m values where
#' each value is a distance between a point and centroid of the cell.
#' @param yVar Character. A character or a vector representing the name of the dependent variable(s)
#' @return Dataframe containing scored layer output
#' @author Shubhra Prakash <shubhra.prakash@@mu-sigma.com>, Sangeet Moy Das <sangeet.das@@mu-sigma.com>, Shantanu Vaidya <shantanu.vaidya@@mu-sigma.com>,Somya Shambhawi <somya.shambhawi@@mu-sigma.com>
#' @seealso \code{\link{trainHVT}} \cr \code{\link{plotHVT}}
#' @importFrom magrittr %>%
#' @examples
#' data("EuStockMarkets")
#' dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
#'                      DAX = EuStockMarkets[, "DAX"],
#'                      SMI = EuStockMarkets[, "SMI"],
#'                      CAC = EuStockMarkets[, "CAC"],
#'                      FTSE = EuStockMarkets[, "FTSE"])
#'rownames(EuStockMarkets) <- dataset$date
#'
#  #Split in train and test
#' train <- EuStockMarkets[1:500, ]
#' test <- EuStockMarkets[501:800, ]
#' 
#' ###MAP-A
#' hvt_mapA <- trainHVT(train, n_cells = 70, depth = 1, quant.err = 0.1,
#'                     distance_metric = "L1_Norm", error_metric = "max",
#'                     normalize = TRUE,quant_method = "kmeans")
#'                     
#' identified_Novelty_cells <- c(7,55) #example
#' output_list <- removeNovelty(identified_Novelty_cells, hvt_mapA)
#' data_with_novelty <- output_list[[1]] 
#' data_with_novelty <- data_with_novelty[, -c(1,2)]
#' 
#' ### MAP-B
#' hvt_mapB <- trainHVT(data_with_novelty,n_cells = 10, depth = 1, quant.err = 0.1,
#'                     distance_metric = "L1_Norm", error_metric = "max",
#'                     normalize = TRUE,quant_method = "kmeans")
#' data_without_novelty <- output_list[[2]]
#' 
#' ### MAP-C
#' hvt_mapC <- trainHVT(data_without_novelty,n_cells = 30,
#'                     depth = 1, quant.err = 0.1, distance_metric = "L1_Norm",
#'                     error_metric = "max", quant_method = "kmeans",
#'                     normalize = TRUE)
#'                     
#' ##SCORE LAYERED
#' data_scored <- scoreLayeredHVT(test, hvt_mapA, hvt_mapB, hvt_mapC)
#' @keywords Scoring
#' @export scoreLayeredHVT

scoreLayeredHVT <- function(data,
                            hvt_mapA,
                            hvt_mapB,
                            hvt_mapC,
                            mad.threshold = 0.2,
                            normalize = TRUE,
                            seed = 300,
                            distance_metric = "L1_Norm",
                            error_metric = "max",
                            child.level = 1,
                            yVar = NULL) {

  set.seed(seed)
  requireNamespace("dplyr")
  requireNamespace("purrr")

  # Predictions from Map A

  mapA_predictions <- scoreHVT(data,
    hvt_mapA,
    child.level = 1,
    mad.threshold = 0.2,
    normalize = normalize,
    distance_metric = distance_metric,
    error_metric = error_metric,
    yVar = NULL,
  )

  mapA_scoredPredictedData <- as.data.frame(mapA_predictions$scoredPredictedData)
  mapA_scoredPredictedData$row_number <- row.names(mapA_scoredPredictedData)

  # Predictions  from map B
  mapB_predictions <- scoreHVT(data,
    hvt_mapB,
    child.level = 1,
    mad.threshold = 0.2,
    normalize = normalize,
    distance_metric = distance_metric,
    error_metric = error_metric,
    yVar = NULL,
  )

  mapB_scoredPredictedData <- as.data.frame(mapB_predictions$scoredPredictedData)
  mapB_scoredPredictedData$row_number <- row.names(mapB_scoredPredictedData)

  # Predictions from map C

  mapC_predictions <- scoreHVT(data,
    hvt_mapC,
    child.level = 1,
    mad.threshold = 0.2,
    normalize = normalize,
    distance_metric = distance_metric,
    error_metric = error_metric,
    yVar = NULL,
  )

  mapC_scoredPredictedData <- as.data.frame(mapC_predictions$scoredPredictedData)
  mapC_scoredPredictedData$row_number <- row.names(mapC_scoredPredictedData)
  mapC_scoredPredictedData$mapC_cell_path <- paste0(mapC_scoredPredictedData$Segment.Level, "-", mapC_scoredPredictedData$Segment.Parent, "-", mapC_scoredPredictedData$Segment.Child)

  # Merged map A, map B, map C predictions value.

  predictions_set <- merge(merge(mapA_scoredPredictedData, mapB_scoredPredictedData, by = "row_number"), mapC_scoredPredictedData, by = "row_number")

  predictions_table <- predictions_set %>% select(
    row_number, Segment.Level.x, Segment.Parent.x, Segment.Child.x, Cell.ID.x,
    Segment.Level.y, Segment.Parent.y, Segment.Child.y, Cell.ID.y, mapC_cell_path, Segment.Level, Segment.Parent, Segment.Child, Cell.ID
  )


  colnames(predictions_table) <- c(
    "Row.Number", "MapA.Segment.Level", "MapA.Segment.Parent", "MapA.Segment.Child", "MapA.Cell.ID",
    "MapB.Segment.Level", "MapB.Segment.Parent", "MapB.Segment.Child", "MapB.Cell.ID", "Map.Cell.Path", "MapC.Segment.Level", "MapC.Segment.Parent", "MapC.Segment.Child", "MapC.Cell.ID"
  )



  predictions_table$MapB.Cell.ID_original <- ifelse(
    (predictions_table$MapA.Segment.Child %in% identified_Novelty_cells),
    paste0(predictions_table$MapB.Cell.ID), 
    NA 
  )

  predictions_table$MapC.Cell.ID <- ifelse(
    (is.na(predictions_table$MapB.Cell.ID_original)),
    paste0(predictions_table$MapC.Cell.ID), 
    NA 
  )

  predictions_table$Row.Number <- as.integer(predictions_table$Row.Number)
  predictions_table <- predictions_table[order(predictions_table$Row.Number, decreasing = FALSE), ]

  scoredPredictionsData_CellID <- predictions_table %>% select(c("Row.Number", "MapA.Cell.ID", "MapB.Cell.ID", "MapB.Cell.ID_original", "MapC.Cell.ID"))
  scoredPredictionsData_CellID <- lapply(scoredPredictionsData_CellID, as.numeric)

  scoredPredictionsData_CellID <- as.data.frame(scoredPredictionsData_CellID)
  scoredPredictionsData_CellID$MapB.Cell.ID_original <- unlist(scoredPredictionsData_CellID$MapB.Cell.ID_original)

  scoredPredictionsData_CellID <- scoredPredictionsData_CellID %>% dplyr::select("Row.Number", "MapA.Cell.ID", "MapB.Cell.ID_original", "MapC.Cell.ID")

  scoredPredictionsData_CellID$MapA.Cell.ID <- ifelse(is.na(scoredPredictionsData_CellID$MapA.Cell.ID), NA, paste0("A", scoredPredictionsData_CellID$MapA.Cell.ID))
  scoredPredictionsData_CellID$MapB.Cell.ID_original <- ifelse(is.na(scoredPredictionsData_CellID$MapB.Cell.ID_original), NA, paste0("B", scoredPredictionsData_CellID$MapB.Cell.ID_original))
  scoredPredictionsData_CellID$MapC.Cell.ID <- ifelse(is.na(scoredPredictionsData_CellID$MapC.Cell.ID), NA, paste0("C", scoredPredictionsData_CellID$MapC.Cell.ID))

  scoredPredictionsData_CellID$Combined.Cell.ID <- coalesce(scoredPredictionsData_CellID$MapB.Cell.ID_original, scoredPredictionsData_CellID$MapC.Cell.ID)
  scoredPredictionsData_CellID <- scoredPredictionsData_CellID %>% dplyr::select("Row.Number", "MapA.Cell.ID", "Combined.Cell.ID")
  colnames(scoredPredictionsData_CellID) <- c("Row.Number", "Layer1.Cell.ID", "Layer2.Cell.ID")

  ##################### Changes ###############
  new_predict <- scoredPredictionsData_CellID
  scaled_test_data <- mapA_predictions[["scoredPredictedData"]]
  current_actual <- colnames(scaled_test_data)
  new_names <- paste0("act_", current_actual)
  colnames(scaled_test_data) <- new_names
  merge <- cbind(new_predict, scaled_test_data) %>% select(-c("act_Segment.Level", "act_Segment.Parent", "act_Segment.Child", "act_n", "act_centroidRadius", "act_diff", "act_anomalyFlag", "act_Cell.ID"))

  data <- merge
  combined <- data %>%
    mutate(Cell.ID = gsub("[C]", "", Layer2.Cell.ID)) %>%
    merge(hvt_mapC[[3]]$summary, by = "Cell.ID", all.x = TRUE) %>%
    arrange(as.numeric(Row.Number))

  data1 <- combined
  combined <- data1 %>%
    mutate(Cell.ID = gsub("[B]", "", Layer2.Cell.ID)) %>%
    merge(hvt_mapB[[3]]$summary, by = "Cell.ID", all.x = TRUE) %>%
    arrange(as.numeric(Row.Number)) %>%
    select(-c(Cell.ID))


  combined <- combined %>% select(-c("act_Quant.Error", "Segment.Level.x", "Segment.Parent.x", "Segment.Child.x", "n.x", "Quant.Error.x", "Segment.Level.y", "Segment.Parent.y", "Segment.Child.y", "n.y", "Quant.Error.y"))


  df_data <- combined %>%
    select(matches("^act_|Row.Number|Layer1.Cell.ID|Layer2.Cell.ID"))

  df <- combined %>%
    select(-c(Row.Number, Layer1.Cell.ID, Layer2.Cell.ID)) %>%
    select(-starts_with("act_"))

  create_predictions <- function(df) {
    pred_cols <- unique(sub("\\..*", "", names(df))) # Extract unique column names
    pred_values <- vector("list", length(pred_cols))

    for (i in seq_along(pred_cols)) {
      col_x <- paste0(pred_cols[i], ".x")
      col_y <- paste0(pred_cols[i], ".y")
      pred_values[[i]] <- coalesce(df[[col_x]], df[[col_y]])
    }

    df1 <- as.data.frame(pred_values)
    names(df1) <- paste0("pred_", pred_cols) 
    return(df1)
  }

  df1 <- create_predictions(df)
  combined_data <- cbind(df_data, df1)



  diff <- subtract_predicted_actual(combined_data) 
  combined_data$diff <- diff$matrix.ncol...1..nrow...nrow.data..
  desired_order <- c("Row.Number", grep("^act_", colnames(combined_data), value = TRUE), "Layer1.Cell.ID", "Layer2.Cell.ID", grep("^pred_", colnames(combined_data), value = TRUE), "diff")
  df_reordered <- combined_data[, desired_order]

  ####################

  prediction_list <- list(
    predictLayer_Output = scoredPredictionsData_CellID,
    actual_predictedTable = df_reordered
  )

  class(prediction_list) <- "hvt.object"
  return(prediction_list)
}

#' @keywords internal
Add_boundary_points <-
  function (tile.child, tile.parent, data.points) {
    desired_index <- c()
    dist_to_tile_endpoints <- list()
    #for all the vertices in the parent tile
    for (m in 1: length(tile.parent$x)) {
      #for all the centroids of the child tile.
      for (n in 1: length(data.points[, 1])) {
        #find out the distance of each vertex from the all the centroids of the child tile
        dist_to_tile_endpoints[[n]] <- sum(((tile.parent$x[m] - data.points[n, 1]) ^ 2),
                                           ((tile.parent$y[m] - data.points[n, 2]) ^ 2))
      }
      # index of the child tile to which the vertex of the parent tile belongs is the one to which
      # the vertex of the parent tile is closest
      desired_index <- which(dist_to_tile_endpoints == min(as.numeric(dist_to_tile_endpoints)))
      #if more than one centroid of the child is close to a given vertex then it should be
      #added to all those child tiles.
      for(i in 1: length(desired_index)){
        tile.x <- c(tile.child[[desired_index[i]]]$x, tile.parent$x[m])
        tile.y <- c(tile.child[[desired_index[i]]]$y, tile.parent$y[m])
        tile.bp <- c(tile.child[[desired_index[i]]]$bp, tile.parent$bp[m])
        #use only those points which form a convex hull
        var1 <- grDevices::chull(tile.x,tile.y)
        tile.child[[desired_index[i]]]$x <- tile.x[var1]
        tile.child[[desired_index[i]]]$y <- tile.y[var1]
        tile.child[[desired_index[i]]]$bp <- tile.bp[var1]
      }
    }
    return (tile.child)
  }

#' @name OptimizationResults
#' @title Table display of all combinations performed in `HVTMSMoptimization`
#' @description This function helps to tabulate the `HVTMSMoptimization` all iterations
#' performed (successful and failed combinations)
#' @param msm_results Dataframe. The `all_results` object of `HVTMSMoptimization` 
#' output based on a specific mae_metric
#' @param show_top_global Numeric. To tabulate top n lowest MAE results
#' @return Table of all iterations will be displayed
#' @author Vishwavani <vishwavani@@mu-sigma.com>, Nithya <nithya.sn@@mu-sigma.com>
#' @keywords Hyperparameter_Tuning
#' @include HVTMSMoptimization.R
#' @importFrom magrittr %>%
#' @export OptimizationResults


OptimizationResults <- function(msm_results,show_top_global = NULL) {
  
  color_cell_min = "#d4f6d4"
  color_global_min = "#ffebee"
  
  # Check for DT package (in Suggests)
  if (!requireNamespace("DT", quietly = TRUE)) {
    stop("Package 'DT' is required for this function. Install it with: install.packages('DT')",
         call. = FALSE)
  }
  
  flag <- NULL
  # Input validation and data extraction
  if (is.data.frame(msm_results)) {
    df <- msm_results
  } else if (!is.null(msm_results[["all_results"]])) {
    df <- msm_results[["all_results"]]
  } else {
    return(DT::datatable(data.frame(Message = "No results to display"), rownames = FALSE))
  }
  
  # Data type conversions
  df <- df %>%
    mutate(
      `Number of Cells` = suppressWarnings(as.integer(`Number of Cells`)),
      k = suppressWarnings(as.integer(k)),
      `Number of nearest neighbors` = suppressWarnings(as.integer(`Number of nearest neighbors`)),
      mae = suppressWarnings(as.numeric(mae))
    )
  
  # Calculate best results per cell count and globally
  df <- df %>%
    group_by(`Number of Cells`) %>%
    mutate(is_cell_min = mae == min(mae, na.rm = TRUE)) %>%
    ungroup()
  
  global_min <- suppressWarnings(min(df$mae, na.rm = TRUE))
  if (!is.finite(global_min)) global_min <- NA_real_
  
  # Apply top-N global filter if specified
  if (!is.null(show_top_global)) {
    df <- df %>%
      arrange(mae) %>%
      slice(seq_len(min(show_top_global, n())))
  }
  
  # Create highlighting flags
  if (!is.null(show_top_global)) {
    # Rank top-N globally and color-code top 3 as Gold/Silver/Bronze
    df <- df %>%
      mutate(
        .rank_global = dplyr::row_number(),
        flag = dplyr::case_when(
          .rank_global == 1 ~ "rank_1",
          .rank_global == 2 ~ "rank_2",
          .rank_global == 3 ~ "rank_3",
          TRUE              ~ "top_n"
        )
      )
  } else {
    df <- df %>%
      mutate(
        flag = dplyr::case_when(
          !is.na(global_min) & mae == global_min ~ "overall_min",
          is_cell_min                           ~ "cell_min",
          TRUE                                  ~ "none"
        )
      )
  }
  
  # Sort results
  if (!is.null(show_top_global)) {
    df <- df %>% arrange(mae, `Number of Cells`, k, `Number of nearest neighbors`)
  } else {
    df <- df %>% arrange(`Number of Cells`, k, `Number of nearest neighbors`, mae)
  }
  
  # Select final columns for display - ONLY status, not status_category and display_status
  final_df <- df %>%
    select(`Number of Cells`, k, `Number of nearest neighbors`, mae, status, flag)
  # Round MAE values for display consistency
  final_df$mae <- round(final_df$mae, 4)
  
  # Column indices for DT options (0-based)
  col_flag      <- which(names(final_df) == "flag") - 1
  col_mae       <- which(names(final_df) == "mae") - 1
  col_cells     <- which(names(final_df) == "Number of Cells") - 1
  col_k         <- which(names(final_df) == "k") - 1
  col_neighbors <- which(names(final_df) == "Number of nearest neighbors") - 1
  
  # Dynamic length menu calculation
  calculate_dynamic_length_menu <- function(total_entries, base_step = 100) {
    max_option <- ceiling(total_entries / base_step) * base_step
    max_option <- max(max_option, 100)
    options <- seq(base_step, by = base_step, length.out = max_option / base_step)
    options <- c(25, options)
    return(options)
  }
  
  dynamic_length_menu <- calculate_dynamic_length_menu(nrow(final_df))
  
  # Create user-friendly headers
  display_names <- names(final_df)
  display_names[display_names == "Number of nearest neighbors"] <- "NN (Nearest Neighbors)"
 # display_names[display_names == "status"] <- "Optimization Status"
  
  # Create and configure DT table
  dt_table <- DT::datatable(
    final_df,
    colnames = display_names,
    options = list(
      pageLength = if (!is.null(show_top_global)) min(show_top_global, 25) else 25,
      scrollX = TRUE,
      lengthMenu = dynamic_length_menu,
      columnDefs = list(list(visible = FALSE, targets = col_flag)),
      order = if (!is.null(show_top_global)) {
        list(list(col_mae, "asc"), list(col_cells, "asc"), list(col_k, "asc"), list(col_neighbors, "asc"))
      } else {
        list(list(col_cells, "asc"), list(col_k, "asc"), list(col_neighbors, "asc"), list(col_mae, "asc"))
      }
    ),
    class = "nowrap display",
    rownames = FALSE
  ) %>%
    DT::formatRound(columns = "mae", digits = 4) %>%
    {
      # Choose styling depending on whether we are showing Top-N globally
      if (!is.null(show_top_global)) {
        # Gold / Silver / Bronze for ranks 1/2/3 respectively.
        gold   <- "#d4f6d4"  # 1st
        silver <- "#C0C0C0"  # 2nd
        bronze <- "#E5E4E2"  # 3rd
        DT::formatStyle(
          ., "flag", target = "row",
          backgroundColor = DT::styleEqual(
            c("rank_1", "rank_2", "rank_3"),
            c(gold, silver, bronze)
          ),
          fontWeight = DT::styleEqual(
            c("rank_1", "rank_2", "rank_3"),
            c("bold", "bold", "bold")
          ),
          color = DT::styleEqual(
            c("rank_1", "rank_2", "rank_3"),
            c("#2e7d32", "#434343","#656565")
          )
        )
      } else {
        DT::formatStyle(
          ., "flag", target = "row",
          backgroundColor = DT::styleEqual(
            c("cell_min", "overall_min", "top_n"),
            c(color_cell_min, color_cell_min, color_cell_min)
          ),
          fontWeight = DT::styleEqual(
            c("cell_min", "overall_min", "top_n"),
            c("bold", "bold", "bold")
          ),
          color = DT::styleEqual(
            c("cell_min", "overall_min", "top_n"),
            c("#2e7d32", "#2e7d32", "#2e7d32")
          )
        )
      }
    }
  
  # Additional emphasis on global minimum MAE
  if (!is.na(global_min)) {
    dt_table <- dt_table %>% 
      DT::formatStyle("mae", fontWeight = DT::styleEqual(global_min, "bold"))
  }
  
  return(dt_table)
}

---
title: "HVT: Collection of functions used to build hierarchical topology preserving maps"
author: "Zubin Dowlaty, Shubhra Prakash, Sangeet Moy Das, Shantanu Vaidya, Praditi Shah, Srinivasan Sudarsanam, Somya Shambhawi, Chepuri Gopi Krishna, Siddharth Shorya,  PonAnureka Seenivasan, Vishwavani, Bidesh Ghosh, Alimpan Dey"
date: "Created Date: 2018-05-17  <br> Modified Date: `r Sys.Date()`"
fig.height: 4
fig.width: 15
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth : 3
vignette: >
  %\VignetteIndexEntry{HVT: Collection of functions used to build hierarchical topology preserving maps}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---


```{css, echo=FALSE}
/* CSS for floating TOC on the left side */
#TOC {
    /* float: left; */
    position: fixed;
    margin-left: -22vw;
    width: 18vw;
    height: fit-content;
    overflow-y: auto;
    padding-top: 20px;
    padding-bottom: 20px;
    background-color: #f9f9f9;
    border-right: 1px solid #ddd;
    margin-top: -14.5em;
}

.main-container {
  margin-left: 220px; /* Adjust this value to match the width of the TOC + some margin */
}
body{
max-width:1200px;
width: 53%;
}
p {
text-align: justify;
}

li {
  padding-bottom: 5px;
}

.caption {
  text-align: center;
}
```

```{r setup, warning = FALSE, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  out.width = "auto",
  out.height = "480px",
  fig.width = 7,
  fig.height = 5,
  fig.align = "center",
  fig.retina = 1,
  dpi = 150,
  tangle = TRUE,
  eval=TRUE
)
```

# 1. Abstract

The HVT package is a collection of R functions to facilitate building <a href="https://link.springer.com/chapter/10.1007/1-84628-118-0_7" target="_blank">topology preserving maps</a> for rich multivariate data analysis, see `Figure 1` as an example of a 3D torus map generated from the package. Tending towards a big data preponderance, a large number of rows. A collection of R functions for this typical workflow is organized below:

1.  **Data Compression**: Vector quantization (VQ), HVQ (hierarchical vector quantization) using means or medians. This step compresses the rows (long data frame) using a compression objective.

2.  **Data Projection**: Dimension projection of the compressed cells to 1D,2D or Interactive surface plot with the Sammons Non-linear Algorithm. This step creates topology preserving map (also called <a href="https://en.wikipedia.org/wiki/Embedding" target="_blank">embeddings</a>) coordinates into the desired output dimension. 

3.  **Tessellation**: Create cells required for object visualization using the Voronoi Tessellation method, package includes heatmap plots for hierarchical Voronoi tessellations (HVT). This step enables data insights, visualization, and interaction with the topology preserving map useful for semi-supervised tasks.

4.  **Scoring**: Scoring new data sets or test data and recording their assignment using the map objects from the above steps, in a sequence of maps if required.

5. **Temporal Analysis and Visualization**: Collection of functions that analyzes time series data for its underlying patterns, calculation of transitioning probabilities and the visualizations for the flow of data over time.

6. **Dynamic Forecasting**: Simulate future states of dynamic systems using Monte Carlo simulations of Markov Chain (MSM), enabling ex-ante predictions for time-series data.



The HVT package allows creation of visually stunning tessellations, showcasing the power of topology preserving maps. below is an image depicting a captivating tessellation of a torus.

```{r predictlayer_flow,echo=FALSE,warning=FALSE,fig.show='hold',message=FALSE,fig.cap='Figure 1:  Heatmap Visualization of a Torus with 500 Cells'}
knitr::include_graphics('./pngs/torus2.png')
```

# 2. Vignettes

Following are the links to the vignettes for the HVT package:

```{r,eval=TRUE, echo=FALSE}
library(knitr)
library(kableExtra)
# Create a data frame with hyperlinks
vignette_data <- data.frame(
  Version_Number = c("v18.05.17", "v18.05.17", "v23.05.16", "v23.10.26", "v24.05.16","v24.08.14", "v25.01.01"),
  Vignette_Title = c(
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/HVT_vignette.html' target='_blank'>HVT Vignette</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/HVT_model_diagnostics_vignette.html' target='_blank'>HVT Model Diagnostics Vignette</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/Scoring_Cells_with_Layers_using_scoreLayeredHVT.html' target='_blank'>HVT Scoring Cells with Layers using scoreLayeredHVT</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/HVT_Temporal_Analysis.html' target='_blank'>Temporal Analysis and Visualization: Leveraging Time Series Capabilities in HVT</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/LLM_Embeddings_in_HVT.html' target='_blank'>Visualizing LLM Embeddings using HVT</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/Implementation_of_tsne_umap_in_trainHVT.html' target='_blank'>Implementation of t-SNE and UMAP in trainHVT function</a>",
  "<a href='https://nbviewer.org/github/Mu-Sigma/HVT/blob/master/vignettes/Dynamic_Forecasting_macroeconomic_data.html' target='_blank'>Dynamic Forecasting of Macroeconomic Time Series Dataset using HVT</a>"
),
  Description = c(
    "Contains descriptions of the functions used for vector quantization and construction of Hierarchical Voronoi Tessellations for data analysis.",
    "Contains descriptions of functions used to perform model diagnostics and validation for the HVT model.",
    "Contains descriptions of the functions used for scoring cells with layers based on a sequence of maps using scoreLayeredHVT.",
    "Contains descriptions of the functions used for analyzing time series data and its flow maps.",
    "Contains the implementation and analysis of hierarchical clustering using the `clustHVT` function to evaluate and visualize token embeddings generated by OpenAI.",
    "Contains enhancements to the `trainHVT` function with advanced dimensionality reduction techniques such as  t-SNE and UMAP, and includes a table of evaluation metrics to improve analysis, visualization, and interpretability.",
    "Contains enhancements to the HVT package for dynamic forecasting using Monte Carlo Simulations of Markov Chain (MSM) on macroeconomic time series dataset."
  ),
  stringsAsFactors = FALSE
)
 
# Generate the table
kable(vignette_data, format = "html", 
      col.names = c("Version", "Vignette Title", "Description"), 
      align = "l", escape = FALSE) %>% 
  kable_styling(full_width = TRUE) %>% column_spec(1, width = "15%") 
```

# 3. Version History


## HVT (v25.2.2) | What's New? {-}

27th January, 2025

In this version of the HVT package, the following new features and vignette have been introduced:

**Features**

1. **Dynamic Forecasting of a Time Series Dataset**: This update introduces a new function called `msm` Monte Carlo Simulations of Markov Chain for dynamic forecasting of states in time series dataset. It supports both ex-post and ex-ante forecasting, offering valuable insights into future trends while resolving state transition challenges through clustering and nearest-neighbor methods to enhance simulation accuracy.

2. **Z score Plots**: This update introduces a new function called `plotZscore` that generates Z-score plots corresponding to the HVT cells for the given data, offering a visual representation of data distribution and highlighting potential outliers.


**Vignette** 

1. **Dynamic Forecasting of Macroeconomic Time Series Dataset using HVT**: This vignette illustrates the practical use of the new `msm` function on a macroeconomic dataset with 10 variables. It covers all steps, including data preparation, model training, scoring, and forecasting, while addressing challenges related to state transitions and evaluating performance using Mean Absolute Error (MAE).


## HVT (v24.9.1) 

4th September, 2024

In this version of the HVT package, the following new features and vignettes have been introduced:

**Features**

1. **Implementation of t-SNE and UMAP in `trainHVT`**: This update incorporates dimensionality reduction methods like t-SNE and UMAP in the `trainHVT` function, complementing the existing Sammon's projection. It also enables the visualization of these techniques across all hierarchical levels within the HVT framework.

2. **Implementation of dimensionality reduction evaluation metrics**: This update introduces highly effective dimensionality reduction evaluation metrics as part of the output list of the `trainHVT` function. These metrics are organized into two levels: Level 1 (L1) and Level 2 (L2). The L1 metrics address key areas of dimensionality reduction which are mentioned below, by ensuring comprehensive evaluation and performance.

- Structure Preservation Metrics 
- Distance Preservation Metrics
- Human Centered Metrics
- Interpretive Quality Metrics
- Computational Efficiency Metrics


3. **Introduction of `clustHVT` function**: In this update, we introduced a new function called `clustHVT` specifically designed for Hierarchical clustering analysis. The function performs clustering of cells exclusively when the hierarchy level is set to 1, determining the optimal number of clusters by evaluating various indices. Based on user input, it conducts hierarchical clustering using AGNES with the default ward.D2 method. The output includes a dendrogram and an interactive 2D clustered HVT map that reveals cell context upon hovering. This function is not applicable when the hierarchy level is greater than 1.



**Vignettes** 

1. **Implementation of t-SNE and UMAP in `trainHVT` function**: This vignette showcases the integration of t-SNE and UMAP in the `trainHVT` function, offering a comprehensive guide on how to apply and visualize these dimensionality reduction techniques. It also covers the dimensionality reduction evaluation metrics and provides insights into their interpretation.

2. **Visualizing LLM Embeddings using HVT (Hierarchical Voronoi Tessellation)**: This vignette will outline the process of analyzing OpenAI-generated token embeddings using the HVT package, covering data compression, visualization, and hierarchical clustering, as well as comparing domain name assignments for clusters. It examines HVT's effectiveness in preserving contextual relationships between embeddings. Additionally, it provides a brief overview of the newly added `clustHVT` function and its parameters.


## HVT (v24.5.2) 

2nd May, 2024

In this version of the HVT package, the following new features have been introduced:

1. **Updated Nomenclature:** To make the function names more consistent and understandable/intuitive, we have renamed the functions throughout the package. Given below are a few instances.

* `HVT` to `trainHVT`
* `predictHVT` to `scoreHVT`
* `predictLayerHVT` to `scoreLayeredHVT`

2. **Restructured Functions:** The functions have been rearranged and grouped into new sections which are highlighted on the index page of the package’s PDF documentation. Given below are a few instances.

* `trainHVT` function now resides within the `Training_or_Compression` section.
* `plotHVT` function now resides within the `Tessellation_and_Heatmap` section.
* `scoreHVT` function now resides within the `Scoring` section.

3. **Enhancements:** The pre-existed functions, `hvtHmap` and `exploded_hmap`, have been combined and incorporated into the `plotHVT` function. Additionally, `plotHVT` now includes the ability to perform 1D plotting.

4. **Temporal Analysis** 
- The new update focuses on the integration of time series capabilities into the HVT package by extending its foundational operations to time series data which is emphasized in this vignette.
- The new functionalities are introduced to analyze underlying patterns and trends within the data, providing insights into its evolution over time and also offering the capability to analyze the movement of the data by calculating its transitioning probability and create elegant plots and GIFs.

Below are the new functions and their brief descriptions:

- `plotStateTransition`: Provides the time series flowmap plot.
- `getTransitionProbability`: Provides a list of transition probabilities.
- `reconcileTransitionProbability`: Provides plots and tables for comparing transition probabilities calculated manually and from markovchain function.
- `plotAnimatedFlowmap`: Creates flowmaps and animations for both self state and without self state scenarios.

## HVT (v23.11.02) {-}

17th November, 2023


This version of the HVT package offers functionality to score cells with layers based on a sequence of maps created using `scoreLayeredHVT`. Given below are the steps to create the successive set of maps.

1. **Map A** - The output of `trainHVT` function which is trained on parent data.

2. **Map B** - The output of `trainHVT` function which is trained on the 'data with novelty' created from `removeNovelty` function.

3. **Map C** - The output of `trainHVT` function which is trained on the 'data without novelty' created from `removeNovelty` function.

The `scoreLayeredHVT` function uses these three maps to score the test datapoints.

Let us try to understand the steps with the help of the diagram below

```{r mlayer_flow,echo=FALSE,warning=FALSE,fig.show='hold',message=FALSE,fig.cap='Figure 2: Data Segregation for scoring based on a sequence of maps using scoreLayeredHVT()'}
knitr::include_graphics('./pngs/scoreLayeredHVT_function.png')
```

## HVT (v22.12.06) {-}

06th December, 2022

This version of the HVT package offers features for both training an HVT model and eliminating outlier cells from the trained model.

1. **Training or Compression:** The initial step entails training the parent data using the `trainHVT` function, specifying the desired compression percentage and quantization error.

2. **Remove novelty cells:** Following the training process, outlier cells can be identified manually from the 2D hvt plot. These outlier cells can then be inputted into the `removeNovelty` function, which subsequently produces two datasets in its output: one containing 'data with novelty' and the other containing 'data without novelty'.



# 4. Installation of HVT (v25.2.2)

**CRAN Installation**

``` r
install.packages("HVT")
```

**Git Hub Installation**

``` r
library(devtools)
##Increase the timeout duration for the initial installation process.
options(timeout = 1200)
devtools::install_github(repo = "Mu-Sigma/HVT")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/removeNovelty.R
\name{removeNovelty}
\alias{removeNovelty}
\title{Remove identified novelty cell(s)}
\usage{
removeNovelty(outlier_cells, hvt_results)
}
\arguments{
\item{outlier_cells}{Vector. A vector with the cell number of the identified novelty}

\item{hvt_results}{List. A list having the results of the compressed map i.e. output of \code{trainHVT} function}
}
\value{
A list of two items
\item{[[1]] }{Dataframe of novelty cell(s)}
\item{[[2]] }{Dataframe without the novelty cell(s) from the dataset used in model training}
}
\description{
This function is used to remove the identified novelty cells.
}
\examples{
data("EuStockMarkets")
hvt.results <- trainHVT(EuStockMarkets, n_cells = 60, depth = 1, quant.err = 0.1, 
                       distance_metric = "L1_Norm", error_metric = "max",
                       normalize = TRUE,quant_method="kmeans")
identified_Novelty_cells <<- c(2, 10)
output_list <- removeNovelty(identified_Novelty_cells, hvt.results) 
data_with_novelty <- output_list[[1]]
data_without_novelty <- output_list[[2]]                      
}
\seealso{
\code{\link{trainHVT}} \cr \code{\link{scoreLayeredHVT}}
}
\author{
Shantanu Vaidya <shantanu.vaidya@mu-sigma.com>
}
\keyword{Novelty_or_Outliers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scoreLayeredHVT.R
\name{scoreLayeredHVT}
\alias{scoreLayeredHVT}
\title{Score which cell and what layer each data point in the test dataset belongs to}
\usage{
scoreLayeredHVT(
  data,
  hvt_mapA,
  hvt_mapB,
  hvt_mapC,
  mad.threshold = 0.2,
  normalize = TRUE,
  seed = 300,
  distance_metric = "L1_Norm",
  error_metric = "max",
  child.level = 1,
  yVar = NULL
)
}
\arguments{
\item{data}{Data Frame. A data frame containing test dataset. 
The data frame should have all the variable(features) used for training.}

\item{hvt_mapA}{A list of hvt.results.model obtained from trainHVT function while performing
`trainHVT()` on train data}

\item{hvt_mapB}{A list of hvt.results.model obtained from trainHVT function while performing 
`trainHVT()` on data with novelty(s)}

\item{hvt_mapC}{A list of hvt.results.model obtained from trainHVT function while performing
`trainHVT()` on data without novelty(s)}

\item{mad.threshold}{Numeric. A number indicating the permissible Mean Absolute Deviation}

\item{normalize}{Logical. A logical value indicating if the dataset should be normalized. 
When set to TRUE, the data (testing dataset) is standardized by 'mean' and 'sd' of the training dataset 
referred from the trainHVT(). When set to FALSE, the data is used as such without any changes.
(Default value is TRUE).}

\item{seed}{Numeric. Random Seed.}

\item{distance_metric}{Character. The distance metric can be L1_Norm(Manhattan) or L2_Norm(Eucledian). L1_Norm is selected by default.
The distance metric is used to calculate the distance between an n dimensional point and centroid.
The distance metric can be different from the one used during training.}

\item{error_metric}{Character. The error metric can be mean or max. max is selected by default. 
max will return the max of m values and mean will take mean of m values where
each value is a distance between a point and centroid of the cell.}

\item{child.level}{Numeric. A number indicating the level for which the heat map is to be plotted.}

\item{yVar}{Character. A character or a vector representing the name of the dependent variable(s)}
}
\value{
Dataframe containing scored layer output
}
\description{
This function that scores the cell and corresponding layer for each data point in a test dataset using three 
hierarchical vector quantization (HVT) models (Map A, Map B, Map C) and returns a data frame containing the scored layer output. 
The function incorporates the scored results from each map and merges them to provide a comprehensive result.
}
\examples{
data("EuStockMarkets")
dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
                     DAX = EuStockMarkets[, "DAX"],
                     SMI = EuStockMarkets[, "SMI"],
                     CAC = EuStockMarkets[, "CAC"],
                     FTSE = EuStockMarkets[, "FTSE"])
rownames(EuStockMarkets) <- dataset$date

train <- EuStockMarkets[1:500, ]
test <- EuStockMarkets[501:800, ]

###MAP-A
hvt_mapA <- trainHVT(train, n_cells = 70, depth = 1, quant.err = 0.1,
                    distance_metric = "L1_Norm", error_metric = "max",
                    normalize = TRUE,quant_method = "kmeans")
                    
identified_Novelty_cells <- c(7,55) #example
output_list <- removeNovelty(identified_Novelty_cells, hvt_mapA)
data_with_novelty <- output_list[[1]] 
data_with_novelty <- data_with_novelty[, -c(1,2)]

### MAP-B
hvt_mapB <- trainHVT(data_with_novelty,n_cells = 10, depth = 1, quant.err = 0.1,
                    distance_metric = "L1_Norm", error_metric = "max",
                    normalize = TRUE,quant_method = "kmeans")
data_without_novelty <- output_list[[2]]

### MAP-C
hvt_mapC <- trainHVT(data_without_novelty,n_cells = 30,
                    depth = 1, quant.err = 0.1, distance_metric = "L1_Norm",
                    error_metric = "max", quant_method = "kmeans",
                    normalize = TRUE)
                    
##SCORE LAYERED
data_scored <- scoreLayeredHVT(test, hvt_mapA, hvt_mapB, hvt_mapC)
}
\seealso{
\code{\link{trainHVT}} \cr \code{\link{plotHVT}}
}
\author{
Shubhra Prakash <shubhra.prakash@mu-sigma.com>, Sangeet Moy Das <sangeet.das@mu-sigma.com>, Shantanu Vaidya <shantanu.vaidya@mu-sigma.com>,Somya Shambhawi <somya.shambhawi@mu-sigma.com>
}
\keyword{Scoring}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotNovelCells.R
\name{plotNovelCells}
\alias{plotNovelCells}
\title{Plot the identified outlier cells in the voronoi tessellation map.}
\usage{
plotNovelCells(
  plot.cells,
  hvt.map,
  line.width = c(0.6),
  color.vec = c("#141B41"),
  pch = 21,
  centroid.size = 0.5,
  title = NULL,
  maxDepth = 1
)
}
\arguments{
\item{plot.cells}{Vector. A vector indicating the cells to be highlighted in the map}

\item{hvt.map}{List. A list containing the output of \code{trainHVT} function
which has the details of the tessellations to be plotted}

\item{line.width}{Numeric Vector. A vector indicating the line widths of the
tessellation boundaries for each level}

\item{color.vec}{Vector. A vector indicating the colors of the boundaries of
the tessellations at each level}

\item{pch}{Numeric. Symbol of the centroids of the tessellations
(parent levels) Default value is 21.}

\item{centroid.size}{Numeric. Size of centroids of first level
tessellations. Default value is 0.5}

\item{title}{String. Set a title for the plot. (default = NULL)}

\item{maxDepth}{Numeric. An integer indicating the number of levels. (default = NULL)}
}
\value{
Returns a ggplot object containing hierarchical voronoi tessellation plot
highlighting the outlier cells
}
\description{
This is the main plotting function to construct hierarchical voronoi tessellations 
and highlight the outlier cells
}
\examples{
data("EuStockMarkets")
hvt.results <- trainHVT(EuStockMarkets, n_cells = 60, depth = 1, quant.err = 0.1, 
                       distance_metric = "L1_Norm", error_metric = "max",
                       normalize = TRUE,quant_method="kmeans")
#selected 55,58 are for demo purpose
plotNovelCells(c(55,58),hvt.results)
}
\seealso{
\code{\link{trainHVT}} \cr \code{\link{plotHVT}}
}
\author{
Shantanu Vaidya <shantanu.vaidya@mu-sigma.com>
}
\keyword{Novelty_or_Outliers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scoreHVT.R
\name{scoreHVT}
\alias{scoreHVT}
\title{Score which cell each point in the test dataset belongs to.}
\usage{
scoreHVT(
  dataset,
  hvt.results.model,
  child.level = 1,
  mad.threshold = 0.2,
  line.width = 0.6,
  color.vec = c("navyblue", "slateblue", "lavender"),
  normalize = TRUE,
  distance_metric = "L1_Norm",
  error_metric = "max",
  yVar = NULL,
  analysis.plots = FALSE,
  names.column = NULL,
  cell_id = TRUE,
  cell_id_position = "bottom",
  centroids = TRUE,
  cell_id_size = 4,
  centroid.size = 0.8
)
}
\arguments{
\item{dataset}{Data frame. A data frame which to be scored. Can have categorical columns if `analysis.plots` are required.}

\item{hvt.results.model}{List. A list obtained from the trainHVT function}

\item{child.level}{Numeric. A number indicating the depth for which the heat map is to be plotted.}

\item{mad.threshold}{Numeric. A numeric value indicating the permissible Mean Absolute Deviation.}

\item{line.width}{Vector. A vector indicating the line widths of the tessellation boundaries for each layer.}

\item{color.vec}{Vector. A vector indicating the colors of the tessellation boundaries at each layer.}

\item{normalize}{Logical. A logical value indicating if the dataset should be normalized. When set to TRUE,
the data (testing dataset) is standardized by ‘mean’ and ‘sd’ of the training dataset referred from the trainHVT(). 
When set to FALSE, the data is used as such without any changes.}

\item{distance_metric}{Character. The distance metric can be L1_Norm(Manhattan) or L2_Norm(Eucledian). L1_Norm is selected by default.
The distance metric is used to calculate the distance between an n dimensional point and centroid.
 The distance metric can be different from the one used during training.}

\item{error_metric}{Character. The error metric can be mean or max. max is selected by default. 
max will return the max of m values and mean will take mean of m values where
each value is a distance between a point and centroid of the cell.}

\item{yVar}{Character. A character or a vector representing the name of the dependent variable(s)}

\item{analysis.plots}{Logical. A logical value indicating that the scored plot should be plotted or not. If TRUE, 
the identifier column(character column) name should be supplied in `names.column` argument. The output will
be a 2D heatmap plotly which gives info on the cell id and the observations of a cell.}

\item{names.column}{Character. A character or a vector representing the name of the identifier column/character column.}

\item{cell_id}{Logical. A logical indicating whether the cell IDs should be displayed. Default is TRUE.}

\item{cell_id_position}{Character. A character indicating the position of the cell IDs. Accepted entries are 'top', 
'bottom', 'left', 'right', and 'center'. Default is 'center'.}

\item{centroids}{Logical. A logical indicating whether the centroid points should be displayed. Default value is TRUE}

\item{cell_id_size}{Numeric. A numeric value indicating the size of the cell IDs. Default is 4}

\item{centroid.size}{Numeric. A numeric value or vector indicating the size of centroids for each level. Default is 0.8}
}
\value{
Dataframe containing scored data, plots and summary
}
\description{
This function scores each data point in the test dataset based on a trained hierarchical Voronoi tessellations model.
}
\examples{
data("EuStockMarkets")
dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
                     DAX = EuStockMarkets[, "DAX"],
                     SMI = EuStockMarkets[, "SMI"],
                     CAC = EuStockMarkets[, "CAC"],
                     FTSE = EuStockMarkets[, "FTSE"])
rownames(EuStockMarkets) <- dataset$date
# Split in train and test
train <- EuStockMarkets[1:1302, ]
test <- EuStockMarkets[1303:1860, ]
#model training
hvt.results<- trainHVT(train,n_cells = 60, depth = 1, quant.err = 0.1,
                      distance_metric = "L1_Norm", error_metric = "max",
                      normalize = TRUE,quant_method = "kmeans")
scoring <- scoreHVT(test, hvt.results)
data_scored <- scoring$scoredPredictedData
}
\seealso{
\code{\link{trainHVT}} \cr \code{\link{plotHVT}}
}
\author{
Shubhra Prakash <shubhra.prakash@mu-sigma.com>, Sangeet Moy Das <sangeet.das@mu-sigma.com> ,
Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{Scoring}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mcmc_plots.R
\name{mcmc_plots}
\alias{mcmc_plots}
\title{Consolidated MCMC Plots Function}
\usage{
mcmc_plots(
  simulation_results,
  centroid_data,
  centroid_2d_points,
  actual_data,
  state_time_data,
  forecast_type,
  n_ahead_ante,
  type,
  raw_dataset,
  show_simulation,
  mae_metric = mae_metric,
  time_column,
  trainHVT_results
)
}
\arguments{
\item{simulation_results}{Data frame with simulation results}

\item{centroid_data}{Data frame with centroid information}

\item{centroid_2d_points}{Data frame with 2D coordinates}

\item{actual_data}{Data frame with actual values}

\item{state_time_data}{Data frame with state-time information}

\item{forecast_type}{Type of forecast ("ex-post" or "ex-ante")}

\item{n_ahead_ante}{Number of ahead steps for ex-ante}

\item{type}{Plot type identifier}

\item{raw_dataset}{Raw dataset for scaling}

\item{show_simulation}{Whether to show simulation lines}

\item{mae_metric}{MAE metric to use}

\item{time_column}{Name of time column}

\item{trainHVT_results}{HVT training results}
}
\value{
List containing all plots and data frames
}
\description{
Creates comprehensive plots for HVT-MSM simulation results with minimal code duplication.
}
\author{
Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotStateTransition.R
\name{plotStateTransition}
\alias{plotStateTransition}
\title{Creating State Transition Plot}
\usage{
plotStateTransition(
  df,
  sample_size = NULL,
  line_plot = NULL,
  cellid_column,
  time_column,
  v_intercept = NULL,
  time_periods = NULL
)
}
\arguments{
\item{df}{Data frame. The Input data frame should contain two columns. 
Cell ID from scoreHVT function and time stamp of that dataset.}

\item{sample_size}{Numeric. An integer indicating the fraction of the data frame to visualize in the plot.
Default value is 0.2}

\item{line_plot}{Logical. A logical value indicating to create a line plot. Default value is NULL.}

\item{cellid_column}{Character. Name of the column containing cell IDs.}

\item{time_column}{Character. Name of the column containing time stamps.}

\item{v_intercept}{Numeric. A numeric value indicating the time stamp to draw a vertical line on the plot.}

\item{time_periods}{List. A list of vectors, each containing start and end times for highlighting time periods.}
}
\value{
A plotly object representing the state transition plot for the given data frame.
}
\description{
This is the main function to create a state transition plot from a data frame.
A state transition plot is a type of data visualization used to represent 
the changes or transitions in states over time for a given system. 
State refers to a particular condition or status of a cell at a specific point in time. 
Transition refers to the change of state for a cell from one condition to another over time.
}
\examples{
dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
DAX = EuStockMarkets[, "DAX"],
SMI = EuStockMarkets[, "SMI"],
CAC = EuStockMarkets[, "CAC"],
FTSE = EuStockMarkets[, "FTSE"])

hvt.results<- trainHVT(dataset,n_cells = 60, depth = 1, quant.err = 0.1,
                       distance_metric = "L1_Norm", error_metric = "max",
                       normalize = TRUE,quant_method = "kmeans")
scoring <- scoreHVT(dataset, hvt.results)
cell_id <- scoring$scoredPredictedData$Cell.ID
time_stamp <- dataset$date
dataset <- data.frame(cell_id, time_stamp)
plotStateTransition(dataset, sample_size = 1, cellid_column = "cell_id",time_column = "time_stamp")
}
\author{
PonAnuReka Seenivasan <ponanureka.s@mu-sigma.com>, Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{Timeseries_Analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary}
\alias{summary}
\title{Table for displaying summary}
\usage{
summary(data, limit = 20, scroll = TRUE)
}
\arguments{
\item{data}{List. A listed object from trainHVT or scoreHVT}

\item{limit}{Numeric. A value to indicate how many rows to display.}

\item{scroll}{Logical. A value to indicate whether to display scroll bar or not. Default value is TRUE.}
}
\value{
A consolidated table of summary for training, scoring and forecasting
}
\description{
This is the main function for displaying summary from model training and scoring
}
\examples{
data <- datasets::EuStockMarkets
dataset <- as.data.frame(data)
#model training
hvt.results <- trainHVT(dataset, n_cells = 60, depth = 1, quant.err = 0.1,
                      distance_metric = "L1_Norm", error_metric = "max",
                      normalize = TRUE, quant_method = "kmeans", dim_reduction_method = 'sammon')
summary(data =  hvt.results)
}
\author{
Vishwavani <vishwavani@mu-sigma.com>, Alimpan Dey <alimpan.dey@mu-sigma.com>
}
\keyword{Table_Formatting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/edaPlots.R
\name{edaPlots}
\alias{edaPlots}
\title{plots for data analysis}
\usage{
edaPlots(
  df,
  time_column,
  output_type = "summary",
  n_cols = -1,
  grey_bars = NULL
)
}
\arguments{
\item{df}{Dataframe. A data frame object.}

\item{time_column}{Character. The name of the time column in the data frame.
Can be given only when the data is time series}

\item{output_type}{Character. The name of the output to be displayed. Options are 'summary',
'histogram', 'boxplot', 'timeseries' & 'correlation'. Default value is summary.}

\item{n_cols}{Numeric. A value to indicate how many columns to be included in the output.}

\item{grey_bars}{List. A list of timestamps where each list contains two elements: start and end period,
which will be highlighted in gray in the time series plot. Default value is NULL.}
}
\value{
Five objects which include time series plots, data distribution plots, 
box plots, correlation plot and a descriptive statistics table.
}
\description{
This is the main function that provides exploratory data analysis plots
}
\examples{
dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
DAX = as.numeric(EuStockMarkets[, "DAX"]),
SMI = as.numeric(EuStockMarkets[, "SMI"]),
CAC = as.numeric(EuStockMarkets[, "CAC"]),
FTSE = as.numeric(EuStockMarkets[, "FTSE"]))
edaPlots(dataset)
edaPlots(dataset, time_column = 'date', output_type = 'timeseries', n_cols = 4)
}
\author{
Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{EDA}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotHVT.R
\name{plotHVT}
\alias{plotHVT}
\title{Plot the hierarchical tessellations.}
\usage{
plotHVT(
  hvt.results,
  line.width = 0.5,
  color.vec = "black",
  centroid.size = 0.6,
  centroid.color = "black",
  child.level = 1,
  hmap.cols,
  separation_width = 7,
  layer_opacity = c(0.5, 0.75, 0.99),
  dim_size = 1000,
  plot.type = "2Dhvt",
  quant.error.hmap = NULL,
  cell_id = FALSE,
  cell_id_position = "bottom",
  cell_id_size = 2.6,
  centroids = TRUE
)
}
\arguments{
\item{hvt.results}{(1D/2DProj/2Dhvt/2Dheatmap/surface_plot) List. A list containing the output of \code{trainHVT} function
which has the details of the tessellations to be plotted.}

\item{line.width}{(2Dhvt/2Dheatmap) Numeric Vector. A vector indicating the line widths of the
tessellation boundaries for each level.}

\item{color.vec}{(2Dhvt/2Dheatmap) Vector. A vector indicating the colors of the boundaries of
the tessellations at each level.}

\item{centroid.size}{(2Dhvt/2Dheatmap) Numeric Vector. A vector indicating the size of centroids
for each level.}

\item{centroid.color}{(2Dhvt/2Dheatmap) Numeric Vector. A vector indicating the color of centroids
for each level.}

\item{child.level}{(2Dheatmap/surface_plot) Numeric. Indicating the level for which the plot should
be displayed}

\item{hmap.cols}{(2Dheatmap/surface_plot) Numeric or Character. The column number or column name from
the dataset indicating the variables for which the heat map is to be plotted.}

\item{separation_width}{(surface_plot) Numeric. An integer indicating the width between hierarchical levels in surface plot}

\item{layer_opacity}{(surface_plot) Numeric. A vector indicating the opacity of each hierarchical levels in surface plot}

\item{dim_size}{(surface_plot) Numeric. An integer controls the resolution or granularity of the 3D surface grid}

\item{plot.type}{Character. An option to indicate which type of plot should be generated. Accepted entries are 
'1D','2Dproj','2Dhvt','2Dheatmap'and 'surface_plot'. Default value is '2Dhvt'.}

\item{quant.error.hmap}{(2Dheatmap) Numeric. A number representing the quantization error threshold to be highlighted in the heatmap. 
When a value is provided, it will emphasize cells with quantization errors equal or less than the specified threshold,
indicating that these cells cannot be further subdivided in the next depth layer. The default value is NULL,
meaning all cells will be colored in the heatmap across various depths.}

\item{cell_id}{(2Dhvt/2Dheatmap) Logical. A logical indicating whether the cell IDs should be displayed. Default is FALSE}

\item{cell_id_position}{(2Dhvt/2Dheatmap) Character. A character indicating the position of the cell IDs. Accepted entries are 'top' , 
'bottom', 'left', 'right', and 'center'. Default is 'bottom'}

\item{cell_id_size}{(2Dhvt/2Dheatmap) Numeric. A numeric vector indicating the size of the cell IDs. Default is 2.6}

\item{centroids}{(2Dhvt/2Dheatmap) Logical. A logical indicating whether the centroid points should be displayed. Default value is TRUE.}
}
\value{
plot object containing the visualizations of reduced dimension(1D/2D) for the given dataset.
}
\description{
This is the main plotting function to construct hierarchical voronoi tessellations in 1D,2D or
Interactive surface plot.
}
\examples{
data("EuStockMarkets")
hvt.results <- trainHVT(EuStockMarkets, n_cells = 60, depth = 1, quant.err = 0.1, 
                       distance_metric = "L1_Norm", error_metric = "max",
                       normalize = TRUE,quant_method="kmeans")
                       
#change the 'plot.type' argument to '2Dproj' or '2DHVT' to visualize respective plots.                      
plotHVT(hvt.results, plot.type='1D')

#change the 'plot.type' argument to 'surface_plot' to visualize the Interactive surface plot                   
plotHVT(hvt.results,child.level = 1, 
hmap.cols = "DAX", plot.type = '2Dheatmap')
}
\seealso{
\code{\link{trainHVT}}
}
\author{
Shubhra Prakash <shubhra.prakash@mu-sigma.com>, Sangeet Moy Das <sangeet.das@mu-sigma.com>, Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{Tessellation_and_Heatmap}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/displayTable.R
\name{displayTable}
\alias{displayTable}
\title{function for displaying table}
\usage{
displayTable(data, scroll = TRUE, limit = 20)
}
\arguments{
\item{data}{Data frame. The dataframe to be displayed in table format.}

\item{scroll}{Logical. A value to have a scroll or not in the table. 
Default is TRUE.}

\item{limit}{Numeric. A value to indicate how many rows to display.
Default is 20.}
}
\value{
A table with proper formatting for html notebook
}
\description{
This is the main function for displaying data in table format
}
\examples{
data <- datasets::EuStockMarkets
dataset <- as.data.frame(data)
displayTable(dataset)
}
\author{
Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{Table_Formatting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getOptimalCentroids.R
\name{getOptimalCentroids}
\alias{getOptimalCentroids}
\title{getOptimalCentroids}
\usage{
getOptimalCentroids(
  x,
  iter.max,
  algorithm,
  n_cells,
  seed = 100,
  function_to_calculate_distance_metric,
  function_to_calculate_error_metric = c("mean", "max"),
  quant.err,
  distance_metric = "L1_Norm",
  quant_method = c("kmeans", "kmedoids")
)
}
\arguments{
\item{x}{Data Frame. A dataframe of multivariate data. Each row corresponds to an
observation, and each column corresponds to a variable. Missing values are
not accepted.}

\item{algorithm}{String. The type of algorithm used for quantization.
Available algorithms are Hartigan and Wong, "Lloyd", "Forgy", "MacQueen".
(default is "Hartigan-Wong")}

\item{n_cells}{Numeric. Indicating the number of nodes per hierarchy.}

\item{seed}{Numeric. Random Seed.}

\item{function_to_calculate_distance_metric}{Function. The function is to find 'L1_Norm" or "L2_Norm" distances. L1_Norm is selected by default.}

\item{function_to_calculate_error_metric}{Character. The error metric can be "mean" or "max". mean is selected by default}

\item{quant.err}{Numeric. The quantization error for the algorithm.}

\item{distance_metric}{Character. The distance metric to calculate inter point distance. It can be 'L1_Norm" or "L2_Norm". L1_Norm is selected by default.}

\item{quant_method}{Character. The quant_method can be "kmeans" or "kmedoids". kmeans is selected by default}
}
\value{
\item{values}{ List. A list showing observations assigned to a cluster.}
\item{maxQE}{ List. A list corresponding to maximum QE values for each cell. }
\item{meanQE}{ List. A list corresponding to mean QE values for each cell. }
\item{centers}{ List. A list of quantization error for all levels and nodes. } 
\item{nsize}{ List. A list corresponding to number of observations in respective groups. }
}
\description{
Get Optimal Centroids
}
\details{
The raw data is first scaled and this scaled data is supplied as input to
the vector quantization algorithm. Vector quantization technique uses a
parameter called quantization error. This parameter acts as a threshold and
determines the number of levels in the hierarchy. It means that, if there
are 'n' number of levels in the hierarchy, then all the clusters formed till
this level will have quantization error equal or greater than the threshold
quantization error. The user can define the number of clusters in the first
level of hierarchy and then each cluster in first level is sub-divided into
the same number of clusters as there are in the first level. This process
continues and each group is divided into smaller clusters as long as the
threshold quantization error is met. The output of this technique will be
hierarchically arranged vector quantized data.
}
\author{
Shubhra Prakash <shubhra.prakash@mu-sigma.com>, Sangeet Moy Das <sangeet.das@mu-sigma.com>
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotZscore.R
\name{plotZscore}
\alias{plotZscore}
\title{Plots of z scores}
\usage{
plotZscore(
  data,
  cell_range = NULL,
  segment_size = 2,
  reference_lines = c(-1.65, 1.65)
)
}
\arguments{
\item{data}{Data frame. A data frame of cell id and features.}

\item{cell_range}{Vector. A numeric vector of cell id range for which the plot should be displayed. 
Default is NULL, which plots all the cells.}

\item{segment_size}{Integer.  A numeric value to indicate the size of the bars in the plot. 
Default is 2.}

\item{reference_lines}{Vector. A numeric vector of confidence interval values for the 
reference lines in the plot. Default is c(-1.65, 1.65).}
}
\value{
A grid of plots of z score against cell id of teh given features.
}
\description{
This is the main function to plot the z scores against cell ids.
}
\examples{
data("EuStockMarkets")
dataset <- data.frame(t = as.numeric(time(EuStockMarkets)),
                     DAX = EuStockMarkets[, "DAX"],
                     SMI = EuStockMarkets[, "SMI"],
                     CAC = EuStockMarkets[, "CAC"],
                     FTSE = EuStockMarkets[, "FTSE"])
rownames(EuStockMarkets) <- dataset$t
hvt.results<- trainHVT(dataset[-1],n_cells = 60, depth = 1, quant.err = 0.1,
                      distance_metric = "L1_Norm", error_metric = "max",
                      normalize = TRUE,quant_method = "kmeans")
col_names <- c("Cell.ID","DAX","SMI","CAC","FTSE")
data <- dplyr::arrange(dplyr::select(hvt.results[[3]][["summary"]],col_names),Cell.ID)
data <- round(data, 2)
plotZscore(data)
}
\author{
Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{EDA}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotQuantErrorHistogram.R
\name{plotQuantErrorHistogram}
\alias{plotQuantErrorHistogram}
\title{Make the quantization error plots for training and scoring.}
\usage{
plotQuantErrorHistogram(hvt.results, hvt.scoring)
}
\arguments{
\item{hvt.results}{List. A list of hvt.results obtained from the trainHVT function.}

\item{hvt.scoring}{List. A list of hvt.scoring obtained from the scoreHVT function.}
}
\value{
Returns the ggplot object containing the quantization error distribution plots for 
the given HVT results of training and scoring
}
\description{
This is the function that produces histograms displaying the distribution of Quantization Error (QE) values 
for both train and test datasets, highlighting mean values with dashed lines for quick evaluation.
}
\examples{
data("EuStockMarkets")
dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
                     DAX = EuStockMarkets[, "DAX"],
                     SMI = EuStockMarkets[, "SMI"],
                     CAC = EuStockMarkets[, "CAC"],
                     FTSE = EuStockMarkets[, "FTSE"])
rownames(EuStockMarkets) <- dataset$date
#Split in train and test
train <- EuStockMarkets[1:1302, ]
test <- EuStockMarkets[1303:1860, ]

hvt.results<- trainHVT(train,n_cells = 60, depth = 1, quant.err = 0.1,
                      distance_metric = "L1_Norm", error_metric = "max",
                      normalize = TRUE, quant_method = "kmeans")
scoring <- scoreHVT(test, hvt.results)
plotQuantErrorHistogram(hvt.results, scoring) 
}
\seealso{
\code{\link{plotHVT}}
}
\author{
Shubhra Prakash <shubhra.prakash@mu-sigma.com>
}
\keyword{Diagnostics_or_Validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clustHVT.R
\name{clustHVT}
\alias{clustHVT}
\title{Performing Hierarchical Clustering Analysis}
\usage{
clustHVT(
  data,
  trainHVT_results,
  scoreHVT_results,
  clustering_method = "ward.D2",
  indices = c("kl", "ch", "hartigan", "cindex", "db", "silhouette", "ratkowsky", "ball",
    "hubert", "dindex", "ptbiserial", "gap", "frey", "mcclain", "gamma", "gplus", "tau",
    "dunn", "sdindex", "sdbw"),
  clusters_k = "champion",
  type = "default",
  domains.column = NULL,
  highlight_labels = NULL,
  only_dendro = FALSE
)
}
\arguments{
\item{data}{Data frame. A data frame intended for performing hierarchical clustering analysis.}

\item{trainHVT_results}{List.  A list object which is obtained as a result of trainHVT function.}

\item{scoreHVT_results}{List. A list object which is obtained as a result of scoreHVT function.}

\item{clustering_method}{Character. The method used for clustering in both NbClust and hclust function. Defaults to ‘ward.D2’.}

\item{indices}{Character. The indices used for determining the optimal number of clusters in NbClust function.
By default it uses 20 different indices.}

\item{clusters_k}{Character.  A parameter that specifies the number of clusters for the provided data. 
The options include “champion,” “challenger,” or any integer between 1 and 20. 
Selecting “champion” will use the highest number of clusters recommended by the ‘NbClust’ function,
while “challenger” will use the second-highest recommendation. If a numerical value from 1 to 20
is provided, that exact number will be used as the number of clusters.}

\item{type}{Character. The type of output required. Default is 'default'. Other option is 'plot' which
will return only the clustered heatmap.}

\item{domains.column}{Character. A vector of cluster names for the clustered heatmap.
Used only when type is 'plot'.}

\item{highlight_labels}{Vector. Numeric vector specifying problematic states and their neighboring states to be highlighted in the dendrogram. 
This argument is intended for internal use only. The default value is NULL.}

\item{only_dendro}{Logical. A logical string specifies whether to generate only the dendrogram or to include the results as well.
This argument is intended for internal use only. The default value is FALSE.}
}
\value{
A list object that contains the hierarchical clustering results.
\item{[[1]] }{Summary of k suggested by all indices with plots} 
\item{[[2]] }{A dendrogram plot with the selected number of clusters} 
\item{[[3]] }{A 2D Cluster HVT Plotly visualization that colors cells according to clusters derived from AGNES clustering results. 
It is interactive, allowing users to view cell contents by hovering over them}
}
\description{
This is the main function to perform hierarchical clustering
analysis which determines optimal number of clusters, perform AGNES clustering
and plot the 2D cluster hvt plot.
}
\examples{
data("EuStockMarkets")
dataset <- data.frame(t = as.numeric(time(EuStockMarkets)),
                     DAX = EuStockMarkets[, "DAX"],
                     SMI = EuStockMarkets[, "SMI"],
                     CAC = EuStockMarkets[, "CAC"],
                     FTSE = EuStockMarkets[, "FTSE"])
rownames(EuStockMarkets) <- dataset$t
hvt.results<- trainHVT(dataset[-1],n_cells = 30, depth = 1, quant.err = 0.1,
                      distance_metric = "L1_Norm", error_metric = "max",
                      normalize = TRUE,quant_method = "kmeans")
scoring <- scoreHVT(dataset, hvt.results, analysis.plots = TRUE, names.column = dataset[,1])
centroid_data <- scoring$centroidData
hclust_data_1 <- centroid_data[,2:3]
clust.results <- clustHVT(data = hclust_data_1, 
                         trainHVT_results = hvt.results,
                         scoreHVT_results = scoring, 
                         clusters_k = 'champion', indices = 'hartigan')
}
\author{
Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{Clustering_Analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotMsmKN.R
\name{plotMsmKN}
\alias{plotMsmKN}
\title{plotting MAE across Number of cells}
\usage{
plotMsmKN(optimization_results)
}
\arguments{
\item{optimization_results}{List. The output of `HVTMSMoptimization` function 
specific for a mae_metric}
}
\value{
List containing visualization results for the selected MAE metric:
\item{[[plot]] }{A Plotly object of cell-level lowest MAE combination} 
\item{[[table]] }{A table of plotted data}
}
\description{
This function helps to visualize the `HVTMSMoptimization` function's
output. It creates a plotly object where the lowest MAE combination achieved for each cell 
will be plotted
}
\author{
Vishwavani <vishwavani@mu-sigma.com>, Nithya <nithya.sn@mu-sigma.com>
}
\keyword{Hyperparameter_Tuning}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/HVTMSMoptimization.R
\name{HVTMSMoptimization}
\alias{HVTMSMoptimization}
\title{HVT-MSM Optimization Function}
\usage{
HVTMSMoptimization(
  entire_dataset,
  expost_forecasting,
  time_column,
  ncell_range = 3:5,
  k_range = 2:9,
  nn_range = 2:7,
  num_simulations = 10,
  mae_metric,
  hvt_params = list(depth = 1, quant.err = 0.2, normalize = TRUE, distance_metric =
    "L1_Norm", error_metric = "max", dim_reduction_method = "sammon"),
  parallel = TRUE,
  verbose = TRUE,
  parallel_strategy = "multisession",
  max_workers = NULL
)
}
\arguments{
\item{entire_dataset}{Dataframe. Train dataset for model training}

\item{expost_forecasting}{Dataframe. Test dataset for ex-post forecasting}

\item{time_column}{Character. Name of the time column}

\item{ncell_range}{Numeric vector. Range of cells to run experiments (default 3:5)}

\item{k_range}{Numeric vector. Range of clusters to run experiments (default: 2:9)}

\item{nn_range}{Numeric vector. Range of nearest neighbors to run experiments (default: 2:7)}

\item{num_simulations}{Integer. Number of simulations (default: 10)}

\item{mae_metric}{Character. MAE calculation method(s): "mean", "median", "mode", or
"all" (default: "median")}

\item{hvt_params}{List. Set of parameters for Model Training (refer trainHVT)}

\item{parallel}{Character. Whether to use parallel processing (default: TRUE)}

\item{verbose}{Character. Whether to print progress information (default: TRUE)}

\item{parallel_strategy}{Character. Parallel processing strategy: "multisession", 
"multicore", etc. (default: "multisession")}

\item{max_workers}{Maximum number of parallel workers 
(default: NULL for auto-detect)}
}
\value{
List containing optimization results for each MAE metric:
\item{[[successful_results]] }{All successful parameter combinations} 
\item{[[nclust_best_results]] }{Best combination for each cell}
\item{[[overall_best]] }{Overall best parameter combination}
\item{[[all_results]] }{All attempted combinations with status}
}
\description{
This function runs multiple iterations/experiments over the dataset across 
different cell counts, and computes the MAE. If a given cell configuration 
not results in problematic states, the baseline simulation is used; otherwise, 
the simulation proceeds with problematic state handling. This process helps identify 
the best-performing model (lowest MAE). In essence, it performs 
trainHVT → scoreHVT → transition probability estimation on the training 
data, followed by msm simulation on the ex-post (test) dataset.  
Note: This is not applicable for ex-ante analysis.
}
\author{
Vishwavani <vishwavani@mu-sigma.com>, Nithya <nithya.sn@mu-sigma.com>
}
\keyword{Hyperparameter_Tuning}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OptimizationResults.R
\name{OptimizationResults}
\alias{OptimizationResults}
\title{Table display of all combinations performed in `HVTMSMoptimization`}
\usage{
OptimizationResults(msm_results, show_top_global = NULL)
}
\arguments{
\item{msm_results}{Dataframe. The `all_results` object of `HVTMSMoptimization` 
output based on a specific mae_metric}

\item{show_top_global}{Numeric. To tabulate top n lowest MAE results}
}
\value{
Table of all iterations will be displayed
}
\description{
This function helps to tabulate the `HVTMSMoptimization` all iterations
performed (successful and failed combinations)
}
\author{
Vishwavani <vishwavani@mu-sigma.com>, Nithya <nithya.sn@mu-sigma.com>
}
\keyword{Hyperparameter_Tuning}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trainHVT.R
\name{trainHVT}
\alias{trainHVT}
\title{Constructing Hierarchical Voronoi Tessellations}
\usage{
trainHVT(
  dataset,
  min_compression_perc = NA,
  n_cells = NA,
  depth = 1,
  quant.err = 0.2,
  normalize = FALSE,
  distance_metric = "L1_Norm",
  error_metric = "max",
  quant_method = "kmeans",
  scale_summary = NA,
  diagnose = FALSE,
  hvt_validation = FALSE,
  train_validation_split_ratio = 0.8,
  dim_reduction_method = "sammon",
  tsne_theta = 0.2,
  tsne_eta = 200,
  tsne_perplexity = 30,
  tsne_verbose = TRUE,
  tsne_max_iter = 500,
  umap_n_neighbors = 60,
  umap_n_components = 2,
  umap_min_dist = 0.1
)
}
\arguments{
\item{dataset}{Data frame. A data frame, with numeric columns (features) will be used for training the model.}

\item{min_compression_perc}{Numeric. An integer, indicating the minimum compression percentage to be achieved for the dataset. 
It indicates the desired level of reduction in dataset size compared to its original size.}

\item{n_cells}{Numeric. An integer, indicating the number of cells per hierarchy (level).}

\item{depth}{Numeric. An integer, indicating the number of levels. A depth of 1 means no hierarchy (single level), 
while higher values indicate multiple levels (hierarchy).}

\item{quant.err}{Numeric. A number indicating the quantization error threshold.
A cell will only breakdown into further cells if the quantization error of the cell is 
above the defined quantization error threshold.}

\item{normalize}{Logical. A logical value indicating if the dataset should be normalized. When set to TRUE, 
scales the values of all features to have a mean of 0 and a standard deviation of 1 (Z-score).}

\item{distance_metric}{Character. The distance metric can be L1_Norm(Manhattan) or L2_Norm(Eucledian). L1_Norm is selected by default.
The distance metric is used to calculate the distance between an n dimensional point and centroid.}

\item{error_metric}{Character. The error metric can be mean or max. max is selected by default. 
max will return the max of m values and mean will take mean of m values where
each value is a distance between a point and centroid of the cell.}

\item{quant_method}{Character. The quantization method can be kmeans or kmedoids. Kmeans uses means (centroids) as cluster centers
while Kmedoids uses actual data points (medoids) as cluster centers. kmeans is selected by default.}

\item{scale_summary}{List. A list with user-defined mean and standard deviation values for all the features in the dataset. 
Pass the scale summary when normalize is set to FALSE.}

\item{diagnose}{Logical. A logical value indicating whether user wants to perform diagnostics on the model. 
Default value is FALSE.}

\item{hvt_validation}{Logical. A logical value indicating whether user wants to holdout a validation set and find 
mean absolute deviation of the validation points from the centroid. Default value is FALSE.}

\item{train_validation_split_ratio}{Numeric. A numeric value indicating train validation split ratio. 
This argument is only used when hvt_validation has been set to TRUE. Default value for the argument is 0.8.}

\item{dim_reduction_method}{Character.The dim_reduction_method can be one of "tsne", "umap", "sammon".}

\item{tsne_theta}{Numeric.The tsne_theta is only used when dim_reduction_method is set to "tsne".
Default value is 0.5 and common values are between 0.2 and 0.5.}

\item{tsne_eta}{Numeric.The tsne_eta are used only when dim_reduction method is set to "tsne".
Default value is 200.}

\item{tsne_perplexity}{Numeric.The tsne_perplexity is only used when dim_reduction_method is set to "tsne".
Default value is 30 and common values are between between 30 and 50.}

\item{tsne_verbose}{Logical. A logical value which indicates the t-SNE algorithm to print 
detailed information about its progress to the console.}

\item{tsne_max_iter}{Numeric.The tsne_max_iter is used only when dim_reduction_method is set to "tsne".
Default value is 1000.More iterations can improve results but increase computation time.}

\item{umap_n_neighbors}{Integer.The umap_n_neighbors is used only when dim_reduction_method is set to "umap".
Default value is 15.Controls the balance between local and global structure in data.}

\item{umap_n_components}{Integer.The umap_n_components is used only when dim_reduction_method is set to "umap".
Default value is 2.Indicates the number of dimensions for embedding.}

\item{umap_min_dist}{Numeric.The umap_map_dist is used only when dim_reduction_method is set to "umap".
Default value is 0.1.Controls how tightly UMAP packs points together.}
}
\value{
A Nested list that contains the hierarchical tessellation information. This
list has to be given as input argument to plot the tessellations.
\item{[[1]] }{A list containing information related to plotting tessellations. 
This information will include coordinates, boundaries, and other details necessary for visualizing the tessellations} 
\item{[[2]] }{A list containing information related to Sammon’s projection coordinates of the data points
in the reduced-dimensional space.}
\item{[[3]] }{A list containing detailed information about the hierarchical vector quantized data along with 
a summary section containing no of points, Quantization Error and the centroids for each cell.}
\item{[[4]] }{A list that contains all the diagnostics information of the model when diagnose is set to TRUE. 
Otherwise NA.}
\item{[[5]] }{A list that contains all the information required to generates a Mean Absolute Deviation (MAD) plot, 
if hvt_validation is set to TRUE. Otherwise NA}
\item{[[6]]}{A list containing detailed information about the hierarchical vector quantized data along with a
 summary section containing no of points, Quantization Error and the centroids for each cell which is the output of `hvq`}
\item{[[7]]}{model info: A list that contains model-generated timestamp, input parameters passed to the model ,
the validation results and the dimensionality reduction evaluation metrics table.}
}
\description{
This is the main function to construct hierarchical voronoi tessellations.
This is done using hierarchical vector quantization(hvq). The data is represented in 2D
coordinates and the tessellations are plotted using these coordinates as
centroids. For subsequent levels, transformation is performed on the 2D
coordinates to get all the points within its parent tile. Tessellations are
plotted using these transformed points as centroids.
}
\examples{
data("EuStockMarkets")
hvt.results <- trainHVT(EuStockMarkets, n_cells = 60, depth = 1, quant.err = 0.1, 
                       distance_metric = "L1_Norm", error_metric = "max",
                       normalize = TRUE,quant_method="kmeans")
}
\seealso{
\code{\link{plotHVT}}
}
\author{
Shubhra Prakash <shubhra.prakash@mu-sigma.com>, Sangeet Moy Das <sangeet.das@mu-sigma.com>, Shantanu Vaidya <shantanu.vaidya@mu-sigma.com>,Bidesh Ghosh <bidesh.gosh@mu-sigma.com>,Alimpan Dey <alimpan.dey@mu-sigma.com>
}
\keyword{Training_or_Compression}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotExAnteRawSeries.R
\name{plotExAnteRawSeries}
\alias{plotExAnteRawSeries}
\title{Ex-ante raw series forecasting}
\usage{
plotExAnteRawSeries(
  ex_ante_results,
  original_dataset,
  transformed_dataset = NULL,
  time_column,
  mae_metric = "median"
)
}
\arguments{
\item{ex_ante_results}{List. Output from `msm()` function with `forecast_type = "ex-ante"`}

\item{original_dataset}{Dataframe. The dataset imported with all features in its original scale of measures 
without normalization or log difference}

\item{transformed_dataset}{Dataframe. The dataset that is transformed for analysis including scaling and log difference}

\item{time_column}{Character. Name of the time column in the dataset.}

\item{mae_metric}{Character. Metric to highlight in plots ("mean", "median", or "mode"). Default is "median".}
}
\value{
A list containing:
\item{reverted_forecasts}{List of data frames, one per variable, with columns: time, mean, median, mode (reverted values)}
\item{plots}{List of plotly objects, one per variable, showing historical data and reverted forecasts}
}
\description{
Transforms ex-ante forecasts generated on a log-difference-12 scale back to the original raw data scale
}
\examples{
\dontrun{
# After running msm() for ex-ante forecasting
# Note: trainHVT_results and scoreHVT_results are needed for msm(), 
# but NOT for plotExAnteRawSeries() which uses the forecast values directly
ex_ante <- msm(state_time_data = temporal_data,
               forecast_type = "ex-ante",
               transition_probability_matrix = prob_trans_matx,
               initial_state = tail(temporal_data$Cell.ID, 1),
               n_ahead_ante = ex_ante_period,
               num_simulations = 500,
               scoreHVT_results = scoring,
               trainHVT_results = hvt.results,
               raw_dataset = entire_dataset,
               time_column = "t")

# Revert to raw scale - only needs ex_ante_results and original_dataset
raw_forecasts <- plotExAnteRawSeries(
  ex_ante_results = ex_ante,
  original_dataset = entire_dataset_original,  # Pre-transformation raw data
  transformed_dataset = entire_dataset,        # Post-transformation data 
  time_column = "t",
  mae_metric = "median"
)

# Access reverted forecasts
raw_forecasts$reverted_forecasts$CPI_Food

# Access plots
raw_forecasts$plots$CPI_Food
}
}
\author{
Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{Timeseries_Analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/msm.R
\name{msm}
\alias{msm}
\title{Performing Monte Carlo Simulations of Markov Chain}
\usage{
msm(
  state_time_data,
  forecast_type = "ex-post",
  initial_state,
  n_ahead_ante = 10,
  transition_probability_matrix,
  num_simulations = 500,
  trainHVT_results,
  scoreHVT_results,
  actual_data = NULL,
  raw_dataset,
  k = NULL,
  handle_problematic_states = TRUE,
  n_nearest_neighbor = NULL,
  show_simulation = TRUE,
  mae_metric = "median",
  time_column,
  precomputed_problematic_states = NULL,
  plot_mode = c("all", "mae-only")
)
}
\arguments{
\item{state_time_data}{Dataframe. A dataframe containing state transitions over time(cell id and timestamp)}

\item{forecast_type}{Character. A character to indicate the type of forecasting.
Accepted values are "ex-post" or "ex-ante".}

\item{initial_state}{Numeric. An integer indicatiog the state at t0.}

\item{n_ahead_ante}{Numeric. A vector of n ahead points to be predicted further in ex-ante analyzes.}

\item{transition_probability_matrix}{Dataframe. A dataframe of transition probabilities/ output of 
`getTransitionProbability` function}

\item{num_simulations}{Integer. A number indicating the total number of simulations to run.
Default is 100.}

\item{trainHVT_results}{List.`trainHVT` function output}

\item{scoreHVT_results}{List. `scoreHVT` function output}

\item{actual_data}{Dataframe. A dataframe for ex-post prediction period with the actual raw data values}

\item{raw_dataset}{DataFrame. A dataframe of input raw dataset from the mean and standard deviation will 
be calculated to scale up the predicted values}

\item{k}{Integer. A number of optimal clusters when handling problematic states. 
Default is 5.}

\item{handle_problematic_states}{Logical. To indicate whether to handle problematic states or not.
Default is FALSE.}

\item{n_nearest_neighbor}{Integer. A number of nearest neighbors to consider when handling problematic states.
Default is 1.}

\item{show_simulation}{Logical. To indicate whether to show the simulation lines in plots or not. Default is TRUE.}

\item{mae_metric}{Character. A character to indicate which metric to calculate Mean Absolute Error.
Accepted entries are "mean", "median", or "mode". Default is "median".}

\item{time_column}{Character. The name of the column containing time data. Used for aligning and plotting the results.}

\item{precomputed_problematic_states}{Vector. An internal parameter used in HVTMSMoptimization call 
to pass the problematic states detected}

\item{plot_mode}{Character. A character to indicate what should be included in output list 
"mae-only" values (used in HVTMSMoptimization) and "all" (used in direct call, include plots and mae)}
}
\value{
A list object that contains the forecasting plots and MAE values.
\item{[[1]]}{Simulation plots and MAE values for state and centroids plot} 
\item{[[2]]}{Summary Table, Dendrogram plot and Clustered Heatmap when handle_problematic_states is TRUE}
}
\description{
This is the main function to perform Monte Carlo simulations of Markov Chain on
the dynamic forecasting of HVT States of a time series dataset. It includes both ex-post and ex-ante analysis 
offering valuable insights into future trends while resolving state transition 
challenges through clustering and nearest-neighbor methods to enhance simulation accuracy.
}
\examples{
dataset <- data.frame(t = as.numeric(time(EuStockMarkets)),
DAX = EuStockMarkets[, "DAX"],
SMI = EuStockMarkets[, "SMI"],
CAC = EuStockMarkets[, "CAC"],
FTSE = EuStockMarkets[, "FTSE"])
hvt.results<- trainHVT(dataset[,-1],n_cells = 60, depth = 1, quant.err = 0.1,
                      distance_metric = "L1_Norm", error_metric = "max",
                      normalize = TRUE,quant_method = "kmeans")
scoring <- scoreHVT(dataset, hvt.results)
cell_id <- scoring$scoredPredictedData$Cell.ID
time_stamp <- dataset$t
temporal_data <- data.frame(cell_id, time_stamp)
table <- getTransitionProbability(temporal_data, 
cellid_column = "cell_id",time_column = "time_stamp")
colnames(temporal_data) <- c("Cell.ID","t")
ex_post_forecasting <- dataset[1830:1860,]
ex_post <- msm(state_time_data = temporal_data,
              forecast_type = "ex-post",
              transition_probability_matrix = table,
              initial_state = 2,
              num_simulations = 2,
              scoreHVT_results = scoring,
              trainHVT_results = hvt.results,
              actual_data = ex_post_forecasting,
              raw_dataset = dataset,
              mae_metric = "median",
              handle_problematic_states = FALSE,
             show_simulation = FALSE,
             time_column = 't')
}
\author{
Vishwavani <vishwavani@mu-sigma.com>, Nithya <nithya.sn@mu-sigma.com>
}
\keyword{Timeseries_Analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconcileTransitionProbability.R
\name{reconcileTransitionProbability}
\alias{reconcileTransitionProbability}
\title{Reconciliation of Transition Probability}
\usage{
reconcileTransitionProbability(
  df,
  hmap_type = NULL,
  cellid_column,
  time_column
)
}
\arguments{
\item{df}{Data frame. The input data frame should contain two columns, cell ID from scoreHVT function and timestamp of that dataset.}

\item{hmap_type}{Character. ('self_state', 'without_self_state', or 'All')}

\item{cellid_column}{Character. Name of the column containing cell IDs.}

\item{time_column}{Character. Name of the column containing timestamps}
}
\value{
A list of plotly heatmap objects and tables representing the transition probability heatmaps.
}
\description{
This is the main function for creating reconciliation plots and tables which 
helps in comparing the transition probabilities calculated manually and from markovchain function
}
\examples{
dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
                      DAX = EuStockMarkets[, "DAX"],
                      SMI = EuStockMarkets[, "SMI"],
                      CAC = EuStockMarkets[, "CAC"],
                      FTSE = EuStockMarkets[, "FTSE"])
hvt.results<- trainHVT(dataset,n_cells = 60, depth = 1, quant.err = 0.1,
                       distance_metric = "L1_Norm", error_metric = "max",
                       normalize = TRUE,quant_method = "kmeans")
scoring <- scoreHVT(dataset, hvt.results)
cell_id <- scoring$scoredPredictedData$Cell.ID
time_stamp <- dataset$date
dataset <- data.frame(cell_id, time_stamp)
reconcileTransitionProbability(dataset, hmap_type = "All", 
cellid_column = "cell_id", time_column = "time_stamp")
}
\author{
PonAnuReka Seenivasan <ponanureka.s@mu-sigma.com>, Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{Timeseries_Analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getTransitionProbability.R
\name{getTransitionProbability}
\alias{getTransitionProbability}
\title{Creating Transition Probability Matrix}
\usage{
getTransitionProbability(
  df,
  cellid_column,
  time_column,
  type = "with_self_state"
)
}
\arguments{
\item{df}{Data frame. The input data frame should contain two columns, 
cell ID from scoreHVT function and time stamp of that dataset.}

\item{cellid_column}{Character. Name of the column containing cell IDs.}

\item{time_column}{Character. Name of the column containing time stamps.}

\item{type}{Character. A character value indicating the type of transition probability table to create.
Accepted entries are "with_self_state" and "without_self_state".}
}
\value{
Stores a data frames with transition probabilities.
}
\description{
This is the main function to create transition probability matrix
The transition probability matrix quantifies the likelihood of transitioning from one state to another. 
States: The table includes the current states and the possible next states.
Probabilities: For each current state, it lists the probability of transitioning to each of the next possible states.
}
\examples{
dataset <- data.frame(t = as.numeric(time(EuStockMarkets)),
                      DAX = EuStockMarkets[, "DAX"],
                      SMI = EuStockMarkets[, "SMI"],
                      CAC = EuStockMarkets[, "CAC"],
                      FTSE = EuStockMarkets[, "FTSE"])
hvt.results<- trainHVT(dataset[-1],n_cells = 60, depth = 1, quant.err = 0.1,
                       distance_metric = "L1_Norm", error_metric = "max",
                       normalize = TRUE,quant_method = "kmeans")
scoring <- scoreHVT(dataset, hvt.results)
cell_id <- scoring$scoredPredictedData$Cell.ID
time_stamp <- dataset$t
dataset <- data.frame(cell_id, time_stamp)
table <- getTransitionProbability(dataset, cellid_column = "cell_id",time_column = "time_stamp")
}
\author{
PonAnuReka Seenivasan <ponanureka.s@mu-sigma.com>, Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{Timeseries_Analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotAnimatedFlowmap.R
\name{plotAnimatedFlowmap}
\alias{plotAnimatedFlowmap}
\title{Generating flow maps and animations based on transition probabilities}
\usage{
plotAnimatedFlowmap(
  hvt_model_output,
  transition_probability_df,
  df,
  animation = "All",
  flow_map = "All",
  fps_time = 1,
  fps_state = 1,
  time_duration = 2,
  state_duration = 2,
  cellid_column,
  time_column
)
}
\arguments{
\item{hvt_model_output}{List. Output from a trainHVT function.}

\item{transition_probability_df}{List. Output from getTransitionProbability function}

\item{df}{Data frame. The input dataframe should contain two columns, 
cell ID from scoreHVT function and time stamp of that dataset.}

\item{animation}{Character. Type of animation ('state_based', 'time_based', 'All' or  NULL)}

\item{flow_map}{Character. Type of flow map ('self_state', 'without_self_state', 'All' or NULL)}

\item{fps_time}{Numeric. A numeric value for the frames per second of the time transition gif.
(Must be a numeric value and a factor of 100). Default value is 1.}

\item{fps_state}{Numeric. A numeric value for the frames per second of the state transition gif.
(Must be a numeric value and a factor of 100). Default value is 1.}

\item{time_duration}{Numeric. A numeric value for the duration of the time transition gif.
Default value is 2.}

\item{state_duration}{Numeric. A numeric value for the duration of the state transition gif.
Default value is 2.}

\item{cellid_column}{Character. Name of the column containing cell IDs.}

\item{time_column}{Character. Name of the column containing time stamps}
}
\value{
A list of flow maps and animation gifs.
}
\description{
This is the main function for generating flow maps and animations based on transition probabilities
including self states and excluding self states.
Flow maps are a type of data visualization used to represent the transition probability of different states. 
Animations are the gifs used to represent the movement of data through the cells.
}
\examples{
dataset <- data.frame(date = as.numeric(time(EuStockMarkets)),
                      DAX = EuStockMarkets[, "DAX"],
                      SMI = EuStockMarkets[, "SMI"],
                      CAC = EuStockMarkets[, "CAC"],
                      FTSE = EuStockMarkets[, "FTSE"])
hvt.results<- trainHVT(dataset,n_cells = 10, depth = 1, quant.err = 0.1,
                       distance_metric = "L1_Norm", error_metric = "max",
                       normalize = TRUE,quant_method = "kmeans")
scoring <- scoreHVT(dataset, hvt.results)
cell_id <- scoring$scoredPredictedData$Cell.ID
time_stamp <- dataset$date
dataset <- data.frame(cell_id, time_stamp)
table <- getTransitionProbability(dataset, cellid_column = "cell_id",time_column = "time_stamp")
plots <- plotAnimatedFlowmap(hvt_model_output = hvt.results, transition_probability_df = table,
df = dataset, animation = 'All', flow_map = 'All',fps_time = 1,fps_state =  1,time_duration = 2,
state_duration = 2,cellid_column = "cell_id", time_column = "time_stamp")
}
\seealso{
\code{\link{trainHVT}} \cr \code{\link{scoreHVT}} \cr \code{\link{getTransitionProbability}}
}
\author{
PonAnuReka Seenivasan <ponanureka.s@mu-sigma.com>, Vishwavani <vishwavani@mu-sigma.com>
}
\keyword{Timeseries_Analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotModelDiagnostics.R
\name{plotModelDiagnostics}
\alias{plotModelDiagnostics}
\title{Make the diagnostic plots for hierarchical voronoi tessellations}
\usage{
plotModelDiagnostics(model_obj)
}
\arguments{
\item{model_obj}{List. A list obtained from the trainHVT
function or scoreHVT function}
}
\value{
For trainHVT, Minimum Intra-DataPoint Distance Plot, Minimum Intra-Centroid Distance Plot
Mean Absolute Deviation Plot, Distribution of Number of Observations in Cells, for Training Data and
Mean Absolute Deviation Plot for Validation Data are plotted.
For scoreHVT Mean Absolute Deviation Plot for Training Data and Validation Data are plotted
}
\description{
This is the main function that generates diagnostic plots for 
hierarchical voronoi tessellations models and scoring.
}
\examples{
data("EuStockMarkets")
hvt.results <- trainHVT(EuStockMarkets, n_cells = 10, depth = 1, quant.err = 0.1, 
                       distance_metric = "L1_Norm", error_metric = "max",
                       normalize = TRUE, quant_method="kmeans", diagnose = TRUE, 
                       hvt_validation = TRUE)
plotModelDiagnostics(hvt.results)
}
\seealso{
\code{\link{plotHVT}}
}
\author{
Shubhra Prakash <shubhra.prakash@mu-sigma.com>
}
\keyword{Diagnostics_or_Validation}
